@@ -34,15 +34,20 @@ namespace dxvk {
           Rc<DxvkAdapter>  Adapter,
           UINT             Ordinal,
           UINT             DisplayIndex)
-    : m_parent          (pParent)
-    , m_adapter         (Adapter)
-    , m_ordinal         (Ordinal)
-    , m_displayIndex    (DisplayIndex)
-    , m_modeCacheFormat (D3D9Format::Unknown)
-    , m_d3d9Formats     (Adapter, m_parent->GetOptions()) {
+  : m_parent          (pParent),
+    m_adapter         (Adapter),
+    m_ordinal         (Ordinal),
+    m_displayIndex    (DisplayIndex),
+    m_modeCacheFormat (D3D9Format::Unknown),
+    m_d3d9Formats     (Adapter, m_parent->GetOptions()) {
     m_adapter->logAdapterInfo();
   }
 
+  template <size_t N>
+  static void copyToStringArray(char (&dst)[N], const char* src) {
+    dxvk::str::strlcpy(dst, src, N);
+  }
+
 
   HRESULT D3D9Adapter::GetAdapterIdentifier(
           DWORD                   Flags,
@@ -69,9 +74,9 @@ namespace dxvk {
     const char*  desc  = options.customDeviceDesc.empty() ? props.deviceName : options.customDeviceDesc.c_str();
     const char* driver = GetDriverDLL(DxvkGpuVendor(vendorId));
 
-    std::strncpy(pIdentifier->Description, desc,              countof(pIdentifier->Description));
-    std::strncpy(pIdentifier->DeviceName,  device.DeviceName, countof(pIdentifier->DeviceName)); // The GDI device name. Not the actual device name.
-    std::strncpy(pIdentifier->Driver,      driver,            countof(pIdentifier->Driver));     // This is the driver's dll.
+    copyToStringArray(pIdentifier->Description, desc);
+    copyToStringArray(pIdentifier->DeviceName,  device.DeviceName); // The GDI device name. Not the actual device name.
+    copyToStringArray(pIdentifier->Driver,      driver);            // This is the driver's dll.
 
     pIdentifier->DeviceIdentifier       = guid;
     pIdentifier->DeviceId               = deviceId;
@@ -90,10 +95,10 @@ namespace dxvk {
           D3D9Format AdapterFormat,
           D3D9Format BackBufferFormat,
           BOOL       bWindowed) {
-    if (!IsSupportedAdapterFormat(AdapterFormat, bWindowed) && !bWindowed)
+    if (!IsSupportedAdapterFormat(AdapterFormat))
       return D3DERR_NOTAVAILABLE;
 
-    if (!IsSupportedBackBufferFormat(BackBufferFormat, bWindowed))
+    if (!IsSupportedBackBufferFormat(AdapterFormat, BackBufferFormat, bWindowed))
       return D3DERR_NOTAVAILABLE;
 
     return D3D_OK;
@@ -106,7 +111,7 @@ namespace dxvk {
           DWORD           Usage,
           D3DRESOURCETYPE RType,
           D3D9Format      CheckFormat) {
-    if (!IsSupportedAdapterFormat(AdapterFormat, false))
+    if (!IsSupportedAdapterFormat(AdapterFormat))
       return D3DERR_NOTAVAILABLE;
 
     const bool dmap = Usage & D3DUSAGE_DMAP;
@@ -138,10 +143,10 @@ namespace dxvk {
     if (CheckFormat == D3D9Format::ATOC && surface)
       return D3D_OK;
 
-    if (m_adapter->features().core.features.depthBounds) {
-      if (CheckFormat == D3D9Format::NVDB && surface)
-        return D3D_OK;
-    }
+    if (CheckFormat == D3D9Format::NVDB && surface)
+      return m_adapter->features().core.features.depthBounds
+        ? D3D_OK
+        : D3DERR_NOTAVAILABLE;
 
     // I really don't want to support this...
     if (dmap)
@@ -158,7 +163,6 @@ namespace dxvk {
       return D3D_OK;
 
     // Let's actually ask Vulkan now that we got some quirks out the way!
-
     return CheckDeviceVkFormat(mapping.FormatColor, Usage, RType);
   }
 
@@ -179,7 +183,8 @@ namespace dxvk {
     if (MultiSampleType != D3DMULTISAMPLE_NONE
      && (SurfaceFormat == D3D9Format::D32_LOCKABLE
       || SurfaceFormat == D3D9Format::D32F_LOCKABLE
-      || SurfaceFormat == D3D9Format::D16_LOCKABLE))
+      || SurfaceFormat == D3D9Format::D16_LOCKABLE
+      || SurfaceFormat == D3D9Format::INTZ))
       return D3DERR_NOTAVAILABLE;
 
     uint32_t sampleCount = std::max<uint32_t>(MultiSampleType, 1u);
@@ -200,7 +205,7 @@ namespace dxvk {
 
     if (pQualityLevels != nullptr) {
       if (MultiSampleType == D3DMULTISAMPLE_NONMASKABLE)
-        *pQualityLevels = (32 - bit::lzcnt(availableFlags));
+        *pQualityLevels = 32 - bit::lzcnt(availableFlags);
       else
         *pQualityLevels = 1;
     }
@@ -217,6 +222,9 @@ namespace dxvk {
     if (!IsDepthFormat(DepthStencilFormat))
       return D3DERR_NOTAVAILABLE;
 
+    if (RenderTargetFormat == dxvk::D3D9Format::NULL_FORMAT)
+      return D3D_OK;
+
     auto mapping = ConvertFormatUnfixed(RenderTargetFormat);
     if (mapping.FormatColor == VK_FORMAT_UNDEFINED)
       return D3DERR_NOTAVAILABLE;
@@ -229,7 +237,8 @@ namespace dxvk {
           D3DDEVTYPE DeviceType,
           D3D9Format SourceFormat,
           D3D9Format TargetFormat) {
-    bool sourceSupported = IsSupportedBackBufferFormat(SourceFormat, FALSE);
+    bool sourceSupported = SourceFormat != D3D9Format::Unknown
+                        && IsSupportedBackBufferFormat(SourceFormat);
     bool targetSupported = TargetFormat == D3D9Format::X1R5G5B5
                         || TargetFormat == D3D9Format::A1R5G5B5
                         || TargetFormat == D3D9Format::R5G6B5
@@ -321,7 +330,7 @@ namespace dxvk {
                                     | D3DPMISCCAPS_CULLCCW
                                     | D3DPMISCCAPS_COLORWRITEENABLE
                                     | D3DPMISCCAPS_CLIPPLANESCALEDPOINTS
-                                 /* | D3DPMISCCAPS_CLIPTLVERTS */
+                                    | D3DPMISCCAPS_CLIPTLVERTS
                                     | D3DPMISCCAPS_TSSARGTEMP
                                     | D3DPMISCCAPS_BLENDOP
                                  /* | D3DPMISCCAPS_NULLREFERENCE */
@@ -618,7 +627,7 @@ namespace dxvk {
   }
 
 
-  UINT D3D9Adapter::GetAdapterModeCountEx(CONST D3DDISPLAYMODEFILTER* pFilter) {
+  UINT D3D9Adapter::GetAdapterModeCountEx(const D3DDISPLAYMODEFILTER* pFilter) {
     if (pFilter == nullptr)
       return 0;
 
@@ -754,7 +763,7 @@ namespace dxvk {
     m_modeCacheFormat = Format;
 
     // Skip unsupported formats
-    if (!IsSupportedAdapterFormat(Format, false))
+    if (!IsSupportedAdapterFormat(Format))
       return;
 
     auto& options = m_parent->GetOptions();
@@ -794,7 +803,7 @@ namespace dxvk {
     // Sort display modes by width, height and refresh rate,
     // in that order. Some games rely on correct ordering.
     std::sort(m_modes.begin(), m_modes.end(),
-      [](const D3DDISPLAYMODEEX & a, const D3DDISPLAYMODEEX & b) {
+      [](const D3DDISPLAYMODEEX& a, const D3DDISPLAYMODEEX& b) {
         if (a.Width < b.Width)   return true;
         if (a.Width > b.Width)   return false;
         
@@ -805,4 +814,4 @@ namespace dxvk {
     });
   }
 
-}
\ No newline at end of file
+}
