@@ -91,74 +91,83 @@ namespace dxvk {
           ? fogCtx.vFog
           : spvModule.opFAbs(floatType, z);
     }
+    uint32_t fogFactor;
+    if (!fogCtx.IsPixel && fogCtx.IsFixedFunction && fogCtx.IsPositionT) {
+      fogFactor = fogCtx.HasSpecular
+        ? spvModule.opCompositeExtract(floatType, fogCtx.Specular, 1, &wIndex)
+        : spvModule.constf32(1.0f);
+    } else {
+      uint32_t applyFogFactor = spvModule.allocateId();
 
-    uint32_t applyFogFactor = spvModule.allocateId();
+      std::array<SpirvPhiLabel, 4> fogVariables;
 
-    std::array<SpirvPhiLabel, 4> fogVariables;
+      std::array<SpirvSwitchCaseLabel, 4> fogCaseLabels = { {
+        { uint32_t(D3DFOG_NONE),      spvModule.allocateId() },
+        { uint32_t(D3DFOG_EXP),       spvModule.allocateId() },
+        { uint32_t(D3DFOG_EXP2),      spvModule.allocateId() },
+        { uint32_t(D3DFOG_LINEAR),    spvModule.allocateId() },
+      } };
 
-    std::array<SpirvSwitchCaseLabel, 4> fogCaseLabels = { {
-      { uint32_t(D3DFOG_NONE),      spvModule.allocateId() },
-      { uint32_t(D3DFOG_EXP),       spvModule.allocateId() },
-      { uint32_t(D3DFOG_EXP2),      spvModule.allocateId() },
-      { uint32_t(D3DFOG_LINEAR),    spvModule.allocateId() },
-    } };
+      spvModule.opSelectionMerge(applyFogFactor, spv::SelectionControlMaskNone);
+      spvModule.opSwitch(fogMode,
+        fogCaseLabels[D3DFOG_NONE].labelId,
+        fogCaseLabels.size(),
+        fogCaseLabels.data());
 
-
-    spvModule.opSelectionMerge(applyFogFactor, spv::SelectionControlMaskNone);
-    spvModule.opSwitch(fogMode,
-      fogCaseLabels[D3DFOG_NONE].labelId,
-      fogCaseLabels.size(),
-      fogCaseLabels.data());
-
-    for (uint32_t i = 0; i < fogCaseLabels.size(); i++) {
-      spvModule.opLabel(fogCaseLabels[i].labelId);
+      for (uint32_t i = 0; i < fogCaseLabels.size(); i++) {
+        spvModule.opLabel(fogCaseLabels[i].labelId);
         
-      fogVariables[i].labelId = fogCaseLabels[i].labelId;
-      fogVariables[i].varId   = [&] {
-        auto mode = D3DFOGMODE(fogCaseLabels[i].literal);
-        switch (mode) {
-          default:
-          // vFog
-          case D3DFOG_NONE: {
-            return fogCtx.IsPixel
-              ? fogCtx.vFog
-              : spvModule.constf32(1.0f);
+        fogVariables[i].labelId = fogCaseLabels[i].labelId;
+        fogVariables[i].varId   = [&] {
+          auto mode = D3DFOGMODE(fogCaseLabels[i].literal);
+          switch (mode) {
+            default:
+            // vFog
+            case D3DFOG_NONE: {
+              if (fogCtx.IsPixel)
+                return fogCtx.vFog;
+
+              if (fogCtx.IsFixedFunction && fogCtx.HasSpecular)
+                return spvModule.opCompositeExtract(floatType, fogCtx.Specular, 1, &wIndex);
+
+              return spvModule.constf32(1.0f);
+            }
+
+            // (end - d) / (end - start)
+            case D3DFOG_LINEAR: {
+              uint32_t fogFactor = spvModule.opFSub(floatType, fogEnd, depth);
+              fogFactor = spvModule.opFMul(floatType, fogFactor, fogScale);
+              fogFactor = spvModule.opNClamp(floatType, fogFactor, spvModule.constf32(0.0f), spvModule.constf32(1.0f));
+              return fogFactor;
+            }
+
+            // 1 / (e^[d * density])^2
+            case D3DFOG_EXP2:
+            // 1 / (e^[d * density])
+            case D3DFOG_EXP: {
+              uint32_t fogFactor = spvModule.opFMul(floatType, depth, fogDensity);
+
+              if (mode == D3DFOG_EXP2)
+                fogFactor = spvModule.opFMul(floatType, fogFactor, fogFactor);
+
+              // Provides the rcp.
+              fogFactor = spvModule.opFNegate(floatType, fogFactor);
+              fogFactor = spvModule.opExp(floatType, fogFactor);
+              return fogFactor;
+            }
           }
-
-          // (end - d) / (end - start)
-          case D3DFOG_LINEAR: {
-            uint32_t fogFactor = spvModule.opFSub(floatType, fogEnd, depth);
-            fogFactor = spvModule.opFMul(floatType, fogFactor, fogScale);
-            fogFactor = spvModule.opNClamp(floatType, fogFactor, spvModule.constf32(0.0f), spvModule.constf32(1.0f));
-            return fogFactor;
-          }
-
-          // 1 / (e^[d * density])^2
-          case D3DFOG_EXP2:
-          // 1 / (e^[d * density])
-          case D3DFOG_EXP: {
-            uint32_t fogFactor = spvModule.opFMul(floatType, depth, fogDensity);
-
-            if (mode == D3DFOG_EXP2)
-              fogFactor = spvModule.opFMul(floatType, fogFactor, fogFactor);
-
-            // Provides the rcp.
-            fogFactor = spvModule.opFNegate(floatType, fogFactor);
-            fogFactor = spvModule.opExp(floatType, fogFactor);
-            return fogFactor;
-          }
-        }
-      }();
+        }();
         
-      spvModule.opBranch(applyFogFactor);
+        spvModule.opBranch(applyFogFactor);
+      }
+
+      spvModule.opLabel(applyFogFactor);
+
+      fogFactor = spvModule.opPhi(floatType,
+        fogVariables.size(),
+        fogVariables.data());
     }
 
-    spvModule.opLabel(applyFogFactor);
-
-    uint32_t fogFactor = spvModule.opPhi(floatType,
-      fogVariables.size(),
-      fogVariables.data());
-
     uint32_t fogRetValue = 0;
 
     // Return the new color if we are doing this in PS
@@ -190,7 +199,7 @@ namespace dxvk {
   }
 
 
-  uint32_t SetupRenderStateBlock(SpirvModule& spvModule) {
+  uint32_t SetupRenderStateBlock(SpirvModule& spvModule, uint32_t count) {
     uint32_t floatType = spvModule.defFloatType(32);
     uint32_t vec3Type  = spvModule.defVectorType(floatType, 3);
 
@@ -208,44 +217,44 @@ namespace dxvk {
       floatType,
       floatType,
     }};
-    
-    uint32_t rsStruct = spvModule.defStructTypeUnique(rsMembers.size(), rsMembers.data());
+
+    uint32_t rsStruct = spvModule.defStructTypeUnique(count, rsMembers.data());
     uint32_t rsBlock = spvModule.newVar(
       spvModule.defPointerType(rsStruct, spv::StorageClassPushConstant),
       spv::StorageClassPushConstant);
     
+    spvModule.setDebugName         (rsBlock, "render_state");
+
     spvModule.setDebugName         (rsStruct, "render_state_t");
     spvModule.decorate             (rsStruct, spv::DecorationBlock);
-    spvModule.setDebugMemberName   (rsStruct, 0, "fog_color");
-    spvModule.memberDecorateOffset (rsStruct, 0, offsetof(D3D9RenderStateInfo, fogColor));
-    spvModule.setDebugMemberName   (rsStruct, 1, "fog_scale");
-    spvModule.memberDecorateOffset (rsStruct, 1, offsetof(D3D9RenderStateInfo, fogScale));
-    spvModule.setDebugMemberName   (rsStruct, 2, "fog_end");
-    spvModule.memberDecorateOffset (rsStruct, 2, offsetof(D3D9RenderStateInfo, fogEnd));
-    spvModule.setDebugMemberName   (rsStruct, 3, "fog_density");
-    spvModule.memberDecorateOffset (rsStruct, 3, offsetof(D3D9RenderStateInfo, fogDensity));
-    spvModule.setDebugMemberName   (rsStruct, 4, "alpha_ref");
-    spvModule.memberDecorateOffset (rsStruct, 4, offsetof(D3D9RenderStateInfo, alphaRef));
-    spvModule.setDebugMemberName   (rsStruct, 5, "point_size");
-    spvModule.memberDecorateOffset (rsStruct, 5, offsetof(D3D9RenderStateInfo, pointSize));
-    spvModule.setDebugMemberName   (rsStruct, 6, "point_size_min");
-    spvModule.memberDecorateOffset (rsStruct, 6, offsetof(D3D9RenderStateInfo, pointSizeMin));
-    spvModule.setDebugMemberName   (rsStruct, 7, "point_size_max");
-    spvModule.memberDecorateOffset (rsStruct, 7, offsetof(D3D9RenderStateInfo, pointSizeMax));
-    spvModule.setDebugMemberName   (rsStruct, 8, "point_scale_a");
-    spvModule.memberDecorateOffset (rsStruct, 8, offsetof(D3D9RenderStateInfo, pointScaleA));
-    spvModule.setDebugMemberName   (rsStruct, 9, "point_scale_b");
-    spvModule.memberDecorateOffset (rsStruct, 9, offsetof(D3D9RenderStateInfo, pointScaleB));
-    spvModule.setDebugMemberName   (rsStruct, 10, "point_scale_c");
-    spvModule.memberDecorateOffset (rsStruct, 10, offsetof(D3D9RenderStateInfo, pointScaleC));
-    
-    spvModule.setDebugName         (rsBlock, "render_state");
+
+    uint32_t memberIdx = 0;
+    auto SetMemberName = [&](const char* name, uint32_t offset) {
+      if (memberIdx >= count)
+        return;
+
+      spvModule.setDebugMemberName   (rsStruct, memberIdx, name);
+      spvModule.memberDecorateOffset (rsStruct, memberIdx, offset);
+      memberIdx++;
+    };
+
+    SetMemberName("fog_color",      offsetof(D3D9RenderStateInfo, fogColor));
+    SetMemberName("fog_scale",      offsetof(D3D9RenderStateInfo, fogScale));
+    SetMemberName("fog_end",        offsetof(D3D9RenderStateInfo, fogEnd));
+    SetMemberName("fog_density",    offsetof(D3D9RenderStateInfo, fogDensity));
+    SetMemberName("alpha_ref",      offsetof(D3D9RenderStateInfo, alphaRef));
+    SetMemberName("point_size",     offsetof(D3D9RenderStateInfo, pointSize));
+    SetMemberName("point_size_min", offsetof(D3D9RenderStateInfo, pointSizeMin));
+    SetMemberName("point_size_max", offsetof(D3D9RenderStateInfo, pointSizeMax));
+    SetMemberName("point_scale_a",  offsetof(D3D9RenderStateInfo, pointScaleA));
+    SetMemberName("point_scale_b",  offsetof(D3D9RenderStateInfo, pointScaleB));
+    SetMemberName("point_scale_c",  offsetof(D3D9RenderStateInfo, pointScaleC));
 
     return rsBlock;
   }
 
 
-  D3D9PointSizeInfoVS GetPointSizeInfoVS(SpirvModule& spvModule, uint32_t vPos, uint32_t vtx, uint32_t perVertPointSize, uint32_t rsBlock) {
+  D3D9PointSizeInfoVS GetPointSizeInfoVS(SpirvModule& spvModule, uint32_t vPos, uint32_t vtx, uint32_t perVertPointSize, uint32_t rsBlock, bool isFixedFunction) {
     uint32_t floatType  = spvModule.defFloatType(32);
     uint32_t floatPtr   = spvModule.defPointerType(floatType, spv::StorageClassPushConstant);
     uint32_t vec3Type   = spvModule.defVectorType(floatType, 3);
@@ -258,43 +267,44 @@ namespace dxvk {
       return spvModule.opLoad(floatType, spvModule.opAccessChain(floatPtr, rsBlock, 1, &index));
     };
 
-    uint32_t pointMode = spvModule.specConst32(uint32Type, 0);
-    spvModule.setDebugName(pointMode, "point_mode");
-    spvModule.decorateSpecId(pointMode, getSpecId(D3D9SpecConstantId::PointMode));
+    uint32_t value = perVertPointSize != 0 ? perVertPointSize : LoadFloat(D3D9RenderStateItem::PointSize);
 
-    uint32_t scaleBit  = spvModule.opBitFieldUExtract(uint32Type, pointMode, spvModule.consti32(0), spvModule.consti32(1));
-    uint32_t isScale   = spvModule.opIEqual(boolType, scaleBit, spvModule.constu32(1));
+    if (isFixedFunction) {
+      uint32_t pointMode = spvModule.specConst32(uint32Type, 0);
+      spvModule.setDebugName(pointMode, "point_mode");
+      spvModule.decorateSpecId(pointMode, getSpecId(D3D9SpecConstantId::PointMode));
 
-    uint32_t regularValue = perVertPointSize != 0 ? perVertPointSize : LoadFloat(D3D9RenderStateItem::PointSize);
+      uint32_t scaleBit  = spvModule.opBitFieldUExtract(uint32Type, pointMode, spvModule.consti32(0), spvModule.consti32(1));
+      uint32_t isScale   = spvModule.opIEqual(boolType, scaleBit, spvModule.constu32(1));
 
-    uint32_t scaleC = LoadFloat(D3D9RenderStateItem::PointScaleC);
-    uint32_t scaleB = LoadFloat(D3D9RenderStateItem::PointScaleB);
-    uint32_t scaleA = LoadFloat(D3D9RenderStateItem::PointScaleA);
+      uint32_t scaleC = LoadFloat(D3D9RenderStateItem::PointScaleC);
+      uint32_t scaleB = LoadFloat(D3D9RenderStateItem::PointScaleB);
+      uint32_t scaleA = LoadFloat(D3D9RenderStateItem::PointScaleA);
 
-    std::array<uint32_t, 4> indices = { 0, 1, 2, 3 };
+      std::array<uint32_t, 4> indices = { 0, 1, 2, 3 };
 
-    uint32_t vtx3;
-    if (vPos != 0) {
-      vPos = spvModule.opLoad(vec4Type, vPos);
+      uint32_t vtx3;
+      if (vPos != 0) {
+        vPos = spvModule.opLoad(vec4Type, vPos);
 
-      uint32_t rhw            = spvModule.opCompositeExtract(floatType, vPos, 1, &indices[3]);
-               rhw            = spvModule.opFDiv(floatType, spvModule.constf32(1.0f), rhw);
-      uint32_t pos3           = spvModule.opVectorShuffle(vec3Type, vPos, vPos, 3, indices.data());
-               vtx3           = spvModule.opVectorTimesScalar(vec3Type, pos3, rhw);
+        uint32_t rhw  = spvModule.opCompositeExtract(floatType, vPos, 1, &indices[3]);
+                 rhw  = spvModule.opFDiv(floatType, spvModule.constf32(1.0f), rhw);
+        uint32_t pos3 = spvModule.opVectorShuffle(vec3Type, vPos, vPos, 3, indices.data());
+                 vtx3 = spvModule.opVectorTimesScalar(vec3Type, pos3, rhw);
+      } else {
+                 vtx3 = spvModule.opVectorShuffle(vec3Type, vtx, vtx, 3, indices.data());
+      }
+
+      uint32_t DeSqr      = spvModule.opDot (floatType, vtx3, vtx3);
+      uint32_t De         = spvModule.opSqrt(floatType, DeSqr);
+      uint32_t scaleValue = spvModule.opFMul(floatType, scaleC, DeSqr);
+               scaleValue = spvModule.opFFma(floatType, scaleB, De, scaleValue);
+               scaleValue = spvModule.opFAdd(floatType, scaleA, scaleValue);
+               scaleValue = spvModule.opSqrt(floatType, scaleValue);
+               scaleValue = spvModule.opFDiv(floatType, value, scaleValue);
+
+      value = spvModule.opSelect(floatType, isScale, scaleValue, value);
     }
-    else {
-               vtx3           = spvModule.opVectorShuffle(vec3Type, vtx, vtx, 3, indices.data());
-    }
-
-    uint32_t DeSqr          = spvModule.opDot (floatType, vtx3, vtx3);
-    uint32_t De             = spvModule.opSqrt(floatType, DeSqr);
-    uint32_t scaleValue     = spvModule.opFMul(floatType, scaleC, DeSqr);
-             scaleValue     = spvModule.opFFma(floatType, scaleB, De, scaleValue);
-             scaleValue     = spvModule.opFAdd(floatType, scaleA, scaleValue);
-             scaleValue     = spvModule.opSqrt(floatType, scaleValue);
-             scaleValue     = spvModule.opFDiv(floatType, regularValue, scaleValue);
-
-    uint32_t value = spvModule.opSelect(floatType, isScale, scaleValue, regularValue);
 
     uint32_t min   = LoadFloat(D3D9RenderStateItem::PointSizeMin);
     uint32_t max   = LoadFloat(D3D9RenderStateItem::PointSizeMax);
@@ -586,6 +596,9 @@ namespace dxvk {
 
     void alphaTestPS();
 
+    uint32_t emitMatrixTimesVector(uint32_t rowCount, uint32_t colCount, uint32_t matrix, uint32_t vector);
+    uint32_t emitVectorTimesMatrix(uint32_t rowCount, uint32_t colCount, uint32_t vector, uint32_t matrix);
+
     bool isVS() { return m_programType == DxsoProgramType::VertexShader; }
     bool isPS() { return !isVS(); }
 
@@ -594,9 +607,13 @@ namespace dxvk {
     SpirvModule           m_module;
     std::vector
       <DxvkResourceSlot>  m_resourceSlots;
-    DxvkInterfaceSlots    m_interfaceSlots;
     std::vector<uint32_t> m_entryPointInterfaces;
 
+    uint32_t              m_inputMask = 0u;
+    uint32_t              m_outputMask = 0u;
+    uint32_t              m_pushConstOffset = 0u;
+    uint32_t              m_pushConstSize = 0u;
+
     DxsoProgramType       m_programType;
     D3D9FFShaderKeyVS     m_vsKey;
     D3D9FFShaderKeyFS     m_fsKey;
@@ -696,19 +713,17 @@ namespace dxvk {
       m_entryPointInterfaces.size(),
       m_entryPointInterfaces.data());
 
-    DxvkShaderOptions shaderOptions = { };
-
-    DxvkShaderConstData constData = { };
-
     // Create the shader module object
-    return new DxvkShader(
-      isVS() ? VK_SHADER_STAGE_VERTEX_BIT : VK_SHADER_STAGE_FRAGMENT_BIT,
-      m_resourceSlots.size(),
-      m_resourceSlots.data(),
-      m_interfaceSlots,
-      m_module.compile(),
-      shaderOptions,
-      std::move(constData));
+    DxvkShaderCreateInfo info;
+    info.stage = isVS() ? VK_SHADER_STAGE_VERTEX_BIT : VK_SHADER_STAGE_FRAGMENT_BIT;
+    info.resourceSlotCount = m_resourceSlots.size();
+    info.resourceSlots = m_resourceSlots.data();
+    info.inputMask = m_inputMask;
+    info.outputMask = m_outputMask;
+    info.pushConstOffset = m_pushConstOffset;
+    info.pushConstSize = m_pushConstOffset;
+
+    return new DxvkShader(info, m_module.compile());
   }
 
 
@@ -718,8 +733,8 @@ namespace dxvk {
       ? m_isgn : m_osgn;
 
     uint32_t& slots = input
-      ? m_interfaceSlots.inputSlots
-      : m_interfaceSlots.outputSlots;
+      ? m_inputMask
+      : m_outputMask;
 
     uint32_t i = sgn.elemCount++;
 
@@ -793,8 +808,8 @@ namespace dxvk {
     if (!m_vsKey.Data.Contents.HasPositionT) {
       if (m_vsKey.Data.Contents.VertexBlendMode == D3D9FF_VertexBlendMode_Normal) {
         uint32_t blendWeightRemaining = m_module.constf32(1);
-        uint32_t vtxSum               = m_module.constvec4f32(0, 0, 0, 0);
-        uint32_t nrmSum               = m_module.constvec3f32(0, 0, 0);
+        uint32_t vtxSum               = 0;
+        uint32_t nrmSum               = 0;
 
         for (uint32_t i = 0; i <= m_vsKey.Data.Contents.VertexBlendCount; i++) {
           std::array<uint32_t, 2> arrayIndices;
@@ -821,7 +836,7 @@ namespace dxvk {
           }
           nrmMtx = m_module.opCompositeConstruct(m_mat3Type, mtxIndices.size(), mtxIndices.data());
 
-          uint32_t vtxResult = m_module.opVectorTimesMatrix(m_vec4Type, vtx, worldview);
+          uint32_t vtxResult = emitVectorTimesMatrix(4, 4, vtx, worldview);
           uint32_t nrmResult = m_module.opVectorTimesMatrix(m_vec3Type, normal, nrmMtx);
 
           uint32_t weight;
@@ -832,18 +847,26 @@ namespace dxvk {
           else
             weight = blendWeightRemaining;
 
-          vtxResult = m_module.opVectorTimesScalar(m_vec4Type, vtxResult, weight);
-          nrmResult = m_module.opVectorTimesScalar(m_vec3Type, nrmResult, weight);
+          std::array<uint32_t, 4> weightIds = { weight, weight, weight, weight };
+          uint32_t weightVec4 = m_module.opCompositeConstruct(m_vec4Type, 4, weightIds.data());
+          uint32_t weightVec3 = m_module.opCompositeConstruct(m_vec3Type, 3, weightIds.data());
 
-          vtxSum = m_module.opFAdd(m_vec4Type, vtxSum, vtxResult);
-          nrmSum = m_module.opFAdd(m_vec3Type, nrmSum, nrmResult);
+          vtxSum = vtxSum
+            ? m_module.opFFma(m_vec4Type, vtxResult, weightVec4, vtxSum)
+            : m_module.opFMul(m_vec4Type, vtxResult, weightVec4);
+
+          nrmSum = nrmSum
+            ? m_module.opFFma(m_vec3Type, nrmResult, weightVec3, nrmSum)
+            : m_module.opFMul(m_vec3Type, nrmResult, weightVec3);
+
+          m_module.decorate(vtxSum, spv::DecorationNoContraction);
         }
 
         vtx    = vtxSum;
         normal = nrmSum;
       }
       else {
-        vtx = m_module.opVectorTimesMatrix(m_vec4Type, vtx, m_vs.constants.worldview);
+        vtx = emitVectorTimesMatrix(4, 4, vtx, m_vs.constants.worldview);
 
         uint32_t nrmMtx = m_vs.constants.normal;
 
@@ -871,7 +894,7 @@ namespace dxvk {
         normal = m_module.opSelect(m_vec3Type, isZeroNormal3, m_module.constvec3f32(0.0f, 0.0f, 0.0f), normal);
       }
       
-      gl_Position = m_module.opVectorTimesMatrix(m_vec4Type, vtx, m_vs.constants.proj);
+      gl_Position = emitVectorTimesMatrix(4, 4, vtx, m_vs.constants.proj);
     } else {
       gl_Position = m_module.opFMul(m_vec4Type, gl_Position, m_vs.constants.invExtent);
       gl_Position = m_module.opFAdd(m_vec4Type, gl_Position, m_vs.constants.invOffset);
@@ -1156,9 +1179,13 @@ namespace dxvk {
     fogCtx.HasFogInput = m_vsKey.Data.Contents.HasFog;
     fogCtx.vFog        = m_vs.in.FOG;
     fogCtx.oColor      = 0;
+    fogCtx.IsFixedFunction = true;
+    fogCtx.IsPositionT = m_vsKey.Data.Contents.HasPositionT;
+    fogCtx.HasSpecular = m_vsKey.Data.Contents.HasColor1;
+    fogCtx.Specular    = m_vs.in.COLOR[1];
     m_module.opStore(m_vs.out.FOG, DoFixedFunctionFog(m_module, fogCtx));
 
-    auto pointInfo = GetPointSizeInfoVS(m_module, 0, vtx, m_vs.in.POINTSIZE, m_rsBlock);
+    auto pointInfo = GetPointSizeInfoVS(m_module, 0, vtx, m_vs.in.POINTSIZE, m_rsBlock, true);
 
     uint32_t pointSize = m_module.opFClamp(m_floatType, pointInfo.defaultValue, pointInfo.min, pointInfo.max);
     m_module.opStore(m_vs.out.POINTSIZE, pointSize);
@@ -1169,16 +1196,20 @@ namespace dxvk {
 
 
   void D3D9FFShaderCompiler::setupRenderStateInfo() {
-    m_rsBlock = SetupRenderStateBlock(m_module);
+    uint32_t count;
 
     if (m_programType == DxsoProgramType::PixelShader) {
-      m_interfaceSlots.pushConstOffset = 0;
-      m_interfaceSlots.pushConstSize   = offsetof(D3D9RenderStateInfo, pointSize);
+      m_pushConstOffset = 0;
+      m_pushConstSize   = offsetof(D3D9RenderStateInfo, pointSize);
+      count = 5;
     }
     else {
-      m_interfaceSlots.pushConstOffset = offsetof(D3D9RenderStateInfo, pointSize);
-      m_interfaceSlots.pushConstSize   = sizeof(float) * 6;
+      m_pushConstOffset = offsetof(D3D9RenderStateInfo, pointSize);
+      m_pushConstSize   = sizeof(float) * 6;
+      count = 11;
     }
+
+    m_rsBlock = SetupRenderStateBlock(m_module, count);
   }
 
 
@@ -1486,7 +1517,7 @@ namespace dxvk {
       m_isgn.elemCount++;
     }
 
-    if (m_vsKey.Data.Contents.HasPointSize)
+    if (m_vsKey.Data.Contents.HasFog)
       m_vs.in.FOG = declareIO(true, DxsoSemantic{ DxsoUsage::Fog,   0 });
     else
       m_isgn.elemCount++;
@@ -1542,7 +1573,7 @@ namespace dxvk {
 
       bool processedTexture = false;
 
-      auto DoBumpmapCoords = [&](uint32_t baseCoords) {
+      auto DoBumpmapCoords = [&](uint32_t typeId, uint32_t baseCoords) {
         uint32_t stage = i - 1;
 
         uint32_t coords = baseCoords;
@@ -1561,7 +1592,7 @@ namespace dxvk {
           uint32_t dot    = m_module.opDot(m_floatType, bm, t);
 
           uint32_t result = m_module.opFAdd(m_floatType, tc_m_n, dot);
-          coords  = m_module.opCompositeInsert(m_vec4Type, result, coords, 1, &i);
+          coords  = m_module.opCompositeInsert(typeId, result, coords, 1, &i);
         }
 
         return coords;
@@ -1607,10 +1638,10 @@ namespace dxvk {
             m_fsKey.Stages[i - 1].Contents.ColorOp == D3DTOP_BUMPENVMAPLUMINANCE)) {
             if (shouldProject) {
               uint32_t projRcp = m_module.opFDiv(m_floatType, m_module.constf32(1.0), projValue);
-              texcoord = m_module.opVectorTimesScalar(m_vec4Type, texcoord, projRcp);
+              texcoord = m_module.opVectorTimesScalar(texcoord_t, texcoord, projRcp);
             }
 
-            texcoord = DoBumpmapCoords(texcoord);
+            texcoord = DoBumpmapCoords(texcoord_t, texcoord);
 
             shouldProject = false;
           }
@@ -1792,7 +1823,8 @@ namespace dxvk {
             break;
 
           case D3DTOP_BLENDTEXTUREALPHAPM:
-            Logger::warn("D3DTOP_BLENDTEXTUREALPHAPM: not implemented");
+            dst = m_module.opFFma(m_vec4Type, arg[2], Complement(AlphaReplicate(GetTexture())), arg[1]);
+            dst = Saturate(dst);
             break;
 
           case D3DTOP_BLENDCURRENTALPHA:
@@ -1942,6 +1974,10 @@ namespace dxvk {
     fogCtx.vPos        = m_ps.in.POS;
     fogCtx.vFog        = m_ps.in.FOG;
     fogCtx.oColor      = current;
+    fogCtx.IsFixedFunction = true;
+    fogCtx.IsPositionT = false;
+    fogCtx.HasSpecular = false;
+    fogCtx.Specular    = 0;
     current = DoFixedFunctionFog(m_module, fogCtx);
 
     m_module.opStore(m_ps.out.COLOR, current);
@@ -2069,7 +2105,7 @@ namespace dxvk {
       m_module.setDebugName(sampler.varId, name.c_str());
 
       const uint32_t bindingId = computeResourceSlotId(DxsoProgramType::PixelShader,
-        DxsoBindingType::ColorImage, i);
+        DxsoBindingType::Image, i);
 
       sampler.bound = m_module.specConstBool(true);
       m_module.decorateSpecId(sampler.bound, bindingId);
@@ -2112,7 +2148,7 @@ namespace dxvk {
 
 
   void D3D9FFShaderCompiler::emitVsClipping(uint32_t vtx) {
-    uint32_t worldPos = m_module.opMatrixTimesVector(m_vec4Type, m_vs.constants.inverseView, vtx);
+    uint32_t worldPos = emitMatrixTimesVector(4, 4, m_vs.constants.inverseView, vtx);
 
     uint32_t clipPlaneCountId = m_module.constu32(caps::MaxClipPlanes);
     
@@ -2188,12 +2224,7 @@ namespace dxvk {
     uint32_t floatPtr = m_module.defPointerType(m_floatType, spv::StorageClassPushConstant);
 
     // Declare spec constants for render states
-    uint32_t alphaTestId = m_module.specConstBool(false);
-    uint32_t alphaFuncId = m_module.specConst32(m_module.defIntType(32, 0), uint32_t(VK_COMPARE_OP_ALWAYS));
-
-    m_module.setDebugName(alphaTestId, "alpha_test");
-    m_module.decorateSpecId(alphaTestId, getSpecId(D3D9SpecConstantId::AlphaTestEnable));
-
+    uint32_t alphaFuncId = m_module.specConst32(m_module.defIntType(32, 0), 0);
     m_module.setDebugName(alphaFuncId, "alpha_func");
     m_module.decorateSpecId(alphaFuncId, getSpecId(D3D9SpecConstantId::AlphaCompareOp));
 
@@ -2218,8 +2249,9 @@ namespace dxvk {
     uint32_t atestSkipLabel = m_module.allocateId();
 
     // if (alpha_test) { ... }
+    uint32_t isNotAlways = m_module.opINotEqual(boolType, alphaFuncId, m_module.constu32(VK_COMPARE_OP_ALWAYS));
     m_module.opSelectionMerge(atestSkipLabel, spv::SelectionControlMaskNone);
-    m_module.opBranchConditional(alphaTestId, atestBeginLabel, atestSkipLabel);
+    m_module.opBranchConditional(isNotAlways, atestBeginLabel, atestSkipLabel);
     m_module.opLabel(atestBeginLabel);
 
     // Load alpha component
@@ -2271,8 +2303,6 @@ namespace dxvk {
       atestVariables.data());
     uint32_t atestDiscard = m_module.opLogicalNot(boolType, atestResult);
 
-    atestResult = m_module.opLogicalNot(boolType, atestResult);
-
     // if (do_discard) { ... }
     m_module.opSelectionMerge(atestKeepLabel, spv::SelectionControlMaskNone);
     m_module.opBranchConditional(atestDiscard, atestDiscardLabel, atestKeepLabel);
@@ -2289,6 +2319,38 @@ namespace dxvk {
   }
 
 
+  uint32_t D3D9FFShaderCompiler::emitMatrixTimesVector(uint32_t rowCount, uint32_t colCount, uint32_t matrix, uint32_t vector) {
+    uint32_t f32Type = m_module.defFloatType(32);
+    uint32_t vecType = m_module.defVectorType(f32Type, rowCount);
+    uint32_t accum = 0;
+
+    for (uint32_t i = 0; i < colCount; i++) {
+      std::array<uint32_t, 4> indices = { i, i, i, i };
+
+      uint32_t a = m_module.opVectorShuffle(vecType, vector, vector, rowCount, indices.data());
+      uint32_t b = m_module.opCompositeExtract(vecType, matrix, 1, &i);
+
+      accum = accum
+        ? m_module.opFFma(vecType, a, b, accum)
+        : m_module.opFMul(vecType, a, b);
+
+      m_module.decorate(accum, spv::DecorationNoContraction);
+    }
+
+    return accum;
+  }
+
+
+  uint32_t D3D9FFShaderCompiler::emitVectorTimesMatrix(uint32_t rowCount, uint32_t colCount, uint32_t vector, uint32_t matrix) {
+    uint32_t f32Type = m_module.defFloatType(32);
+    uint32_t vecType = m_module.defVectorType(f32Type, colCount);
+    uint32_t matType = m_module.defMatrixType(vecType, rowCount);
+
+    matrix = m_module.opTranspose(matType, matrix);
+    return emitMatrixTimesVector(colCount, rowCount, matrix, vector);
+  }
+
+
   D3D9FFShader::D3D9FFShader(
           D3D9DeviceEx*         pDevice,
     const D3D9FFShaderKeyVS&    Key) {
@@ -2340,7 +2402,7 @@ namespace dxvk {
 
     if (dumpPath.size() != 0) {
       std::ofstream dumpStream(
-        str::format(dumpPath, "/", Name, ".spv"),
+        str::tows(str::format(dumpPath, "/", Name, ".spv").c_str()).c_str(),
         std::ios_base::binary | std::ios_base::trunc);
       
       m_shader->dump(dumpStream);
@@ -2387,7 +2449,7 @@ namespace dxvk {
 
     std::hash<uint32_t> uint32hash;
 
-    for (uint32_t i = 0; i < countof(key.Data.Primitive); i++)
+    for (uint32_t i = 0; i < std::size(key.Data.Primitive); i++)
       state.add(uint32hash(key.Data.Primitive[i]));
 
     return state;
@@ -2400,7 +2462,7 @@ namespace dxvk {
     std::hash<uint32_t> uint32hash;
 
     for (uint32_t i = 0; i < caps::TextureStageCount; i++) {
-      for (uint32_t j = 0; j < countof(key.Stages[i].Primitive); j++)
+      for (uint32_t j = 0; j < std::size(key.Stages[i].Primitive); j++)
         state.add(uint32hash(key.Stages[i].Primitive[j]));
     }
 
@@ -2460,4 +2522,4 @@ namespace dxvk {
 
   DxsoIsgn g_ffIsgn = CreateFixedFunctionIsgn();
 
-}
\ No newline at end of file
+}
