@@ -4,22 +4,28 @@
 
 #include "d3d9_hud.h"
 
-#include <d3d9_presenter_frag.h>
-#include <d3d9_presenter_vert.h>
-
 namespace dxvk {
 
 
   struct D3D9WindowData {
     bool unicode;
     bool filter;
+    bool activateProcessed;
     WNDPROC proc;
+    D3D9SwapChainEx* swapchain;
   };
 
 
-  static std::recursive_mutex g_windowProcMapMutex;
+  static dxvk::recursive_mutex g_windowProcMapMutex;
   static std::unordered_map<HWND, D3D9WindowData> g_windowProcMap;
 
+  static void SetActivateProcessed(HWND window, bool processed)
+  {
+      std::lock_guard lock(g_windowProcMapMutex);
+      auto it = g_windowProcMap.find(window);
+      if (it != g_windowProcMap.end())
+        it->second.activateProcessed = processed;
+  }
 
   template <typename T, typename J, typename ... Args>
   auto CallCharsetFunction(T unicode, J ascii, bool isUnicode, Args... args) {
@@ -82,7 +88,7 @@ namespace dxvk {
   }
 
 
-  void HookWindowProc(HWND window) {
+  void HookWindowProc(HWND window, D3D9SwapChainEx* swapchain) {
     std::lock_guard lock(g_windowProcMapMutex);
 
     ResetWindowProc(window);
@@ -90,10 +96,12 @@ namespace dxvk {
     D3D9WindowData windowData;
     windowData.unicode = IsWindowUnicode(window);
     windowData.filter  = false;
+    windowData.activateProcessed = false;
     windowData.proc = reinterpret_cast<WNDPROC>(
       CallCharsetFunction(
       SetWindowLongPtrW, SetWindowLongPtrA, windowData.unicode,
         window, GWLP_WNDPROC, reinterpret_cast<LONG_PTR>(D3D9WindowProc)));
+    windowData.swapchain = swapchain;
 
     g_windowProcMap[window] = std::move(windowData);
   }
@@ -124,6 +132,38 @@ namespace dxvk {
 
     if (message == WM_DESTROY)
       ResetWindowProc(window);
+    else if (message == WM_ACTIVATEAPP) {
+      D3DDEVICE_CREATION_PARAMETERS create_parms;
+      windowData.swapchain->GetDevice()->GetCreationParameters(&create_parms);
+
+      if (!(create_parms.BehaviorFlags & D3DCREATE_NOWINDOWCHANGES)) {
+        D3D9WindowMessageFilter filter(window);
+        if (wparam && !windowData.activateProcessed) {
+          // Heroes of Might and Magic V needs this to resume drawing after a focus loss
+          D3DPRESENT_PARAMETERS params;
+          RECT rect;
+
+          GetMonitorRect(GetDefaultMonitor(), &rect);
+          windowData.swapchain->GetPresentParameters(&params);
+          SetWindowPos(window, nullptr, rect.left, rect.top, params.BackBufferWidth, params.BackBufferHeight,
+                       SWP_NOACTIVATE | SWP_NOZORDER);
+          SetActivateProcessed(window, true);
+        }
+        else if (!wparam) {
+          if (IsWindowVisible(window))
+            ShowWindow(window, SW_MINIMIZE);
+          SetActivateProcessed(window, false);
+        }
+      }
+    }
+    else if (message == WM_SIZE)
+    {
+      D3DDEVICE_CREATION_PARAMETERS create_parms;
+      windowData.swapchain->GetDevice()->GetCreationParameters(&create_parms);
+
+      if (!(create_parms.BehaviorFlags & D3DCREATE_NOWINDOWCHANGES) && !IsIconic(window))
+        PostMessageW(window, WM_ACTIVATEAPP, 1, GetCurrentThreadId());
+    }
 
     return CallCharsetFunction(
       CallWindowProcW, CallWindowProcA, unicode,
@@ -153,21 +193,19 @@ namespace dxvk {
     , m_context          (m_device->createContext())
     , m_frameLatencyCap  (pDevice->GetOptions()->maxFrameLatency)
     , m_frameLatencySignal(new sync::Fence(m_frameId))
-    , m_dialog            (pDevice->GetOptions()->enableDialogMode) {
+    , m_dialog           (pDevice->GetOptions()->enableDialogMode) {
     this->NormalizePresentParameters(pPresentParams);
     m_presentParams = *pPresentParams;
     m_window = m_presentParams.hDeviceWindow;
 
     UpdatePresentRegion(nullptr, nullptr);
-    if (!pDevice->GetOptions()->deferSurfaceCreation)
+    if (m_window && !pDevice->GetOptions()->deferSurfaceCreation)
       CreatePresenter();
 
     CreateBackBuffers(m_presentParams.BackBufferCount);
+    CreateBlitter();
     CreateHud();
 
-    InitRenderState();
-    InitSamplers();
-    InitShaders();
     InitRamp();
 
     // Apply initial window mode and fullscreen state
@@ -177,6 +215,11 @@ namespace dxvk {
 
 
   D3D9SwapChainEx::~D3D9SwapChainEx() {
+    // Avoids hanging when in this state, see comment
+    // in DxvkDevice::~DxvkDevice.
+    if (this_thread::isInModuleDetachment())
+      return;
+
     DestroyBackBuffers();
 
     ResetWindowProc(m_window);
@@ -295,7 +338,7 @@ namespace dxvk {
     D3D9CommonTexture* dstTexInfo = dst->GetCommonTexture();
     D3D9CommonTexture* srcTexInfo = m_backBuffers.back()->GetCommonTexture();
 
-    if (unlikely(dstTexInfo->Desc()->Pool != D3DPOOL_SYSTEMMEM))
+    if (unlikely(dstTexInfo->Desc()->Pool != D3DPOOL_SYSTEMMEM && dstTexInfo->Desc()->Pool != D3DPOOL_SCRATCH))
       return D3DERR_INVALIDCALL;
 
     Rc<DxvkBuffer> dstBuffer = dstTexInfo->GetBuffer(dst->GetSubresource());
@@ -438,13 +481,13 @@ namespace dxvk {
       cSubresources = srcSubresourceLayers,
       cLevelExtent  = srcExtent
     ] (DxvkContext* ctx) {
-      ctx->copyImageToBuffer(
-        cBuffer, 0, VkExtent2D { 0u, 0u },
+      ctx->copyImageToBuffer(cBuffer, 0, 4, 0,
         cImage, cSubresources, VkOffset3D { 0, 0, 0 },
         cLevelExtent);
     });
-    
-    dstTexInfo->SetDirty(dst->GetSubresource(), true);
+
+    dstTexInfo->SetNeedsReadback(dst->GetSubresource(), true);
+    m_parent->TrackTextureMappingBufferSequenceNumber(dstTexInfo, dst->GetSubresource());
 
     return D3D_OK;
   }
@@ -577,7 +620,7 @@ namespace dxvk {
   }
 
 
-  void    D3D9SwapChainEx::Reset(
+  HRESULT D3D9SwapChainEx::Reset(
           D3DPRESENT_PARAMETERS* pPresentParams,
           D3DDISPLAYMODEEX*      pFullscreenDisplayMode) {
     D3D9DeviceLock lock = m_parent->LockDevice();
@@ -593,29 +636,19 @@ namespace dxvk {
     if (pPresentParams->Windowed) {
       if (changeFullscreen)
         this->LeaveFullscreenMode();
-
-      // Adjust window position and size
-      RECT newRect = { 0, 0, 0, 0 };
-      RECT oldRect = { 0, 0, 0, 0 };
-      
-      ::GetWindowRect(m_window, &oldRect);
-      ::SetRect(&newRect, 0, 0, pPresentParams->BackBufferWidth, pPresentParams->BackBufferHeight);
-      ::AdjustWindowRectEx(&newRect,
-        ::GetWindowLongW(m_window, GWL_STYLE), FALSE,
-        ::GetWindowLongW(m_window, GWL_EXSTYLE));
-      ::SetRect(&newRect, 0, 0, newRect.right - newRect.left, newRect.bottom - newRect.top);
-      ::OffsetRect(&newRect, oldRect.left, oldRect.top);    
-      ::MoveWindow(m_window, newRect.left, newRect.top,
-        newRect.right - newRect.left, newRect.bottom - newRect.top, TRUE);
     }
     else {
-      if (changeFullscreen)
-        this->EnterFullscreenMode(pPresentParams, pFullscreenDisplayMode);
+      if (changeFullscreen) {
+        if (FAILED(this->EnterFullscreenMode(pPresentParams, pFullscreenDisplayMode)))
+          return D3DERR_INVALIDCALL;
+      }
 
       D3D9WindowMessageFilter filter(m_window);
 
-      if (!changeFullscreen)
-        ChangeDisplayMode(pPresentParams, pFullscreenDisplayMode);
+      if (!changeFullscreen) {
+        if (FAILED(ChangeDisplayMode(pPresentParams, pFullscreenDisplayMode)))
+          return D3DERR_INVALIDCALL;
+      }
 
       // Move the window so that it covers the entire output    
       RECT rect;
@@ -632,14 +665,40 @@ namespace dxvk {
       SetGammaRamp(0, &m_ramp);
 
     CreateBackBuffers(m_presentParams.BackBufferCount);
+
+    return D3D_OK;
   }
 
 
   HRESULT D3D9SwapChainEx::WaitForVBlank() {
-    Logger::warn("D3D9SwapChainEx::WaitForVBlank: Stub");
+    static bool s_errorShown = false;
+
+    if (!std::exchange(s_errorShown, true))
+      Logger::warn("D3D9SwapChainEx::WaitForVBlank: Stub");
+
     return D3D_OK;
   }
 
+  static bool validateGammaRamp(const WORD (&ramp)[256]) {
+    if (ramp[0] >= ramp[std::size(ramp) - 1]) {
+      Logger::err("validateGammaRamp: ramp inverted or flat");
+      return false;
+    }
+
+    for (size_t i = 1; i < std::size(ramp); i++) {
+      if (ramp[i] < ramp[i - 1]) {
+        Logger::err("validateGammaRamp: ramp not monotonically increasing");
+        return false;
+      }
+      if (ramp[i] - ramp[i - 1] >= UINT16_MAX / 2) {
+        Logger::err("validateGammaRamp: huuuge jump");
+        return false;
+      }
+    }
+
+    return true;
+  }
+
 
   void    D3D9SwapChainEx::SetGammaRamp(
             DWORD         Flags,
@@ -649,29 +708,34 @@ namespace dxvk {
     if (unlikely(pRamp == nullptr))
       return;
 
+    if (unlikely(!validateGammaRamp(pRamp->red)
+              && !validateGammaRamp(pRamp->blue)
+              && !validateGammaRamp(pRamp->green)))
+      return;
+
     m_ramp = *pRamp;
 
     bool isIdentity = true;
 
-    std::array<D3D9_VK_GAMMA_CP, NumControlPoints> cp;
+    std::array<DxvkGammaCp, NumControlPoints> cp;
       
     for (uint32_t i = 0; i < NumControlPoints; i++) {
       uint16_t identity = MapGammaControlPoint(float(i) / float(NumControlPoints - 1));
 
-      cp[i].R = pRamp->red[i];
-      cp[i].G = pRamp->green[i];
-      cp[i].B = pRamp->blue[i];
-      cp[i].A = 0;
+      cp[i].r = pRamp->red[i];
+      cp[i].g = pRamp->green[i];
+      cp[i].b = pRamp->blue[i];
+      cp[i].a = 0;
 
-      isIdentity &= cp[i].R == identity
-                 && cp[i].G == identity
-                 && cp[i].B == identity;
+      isIdentity &= cp[i].r == identity
+                 && cp[i].g == identity
+                 && cp[i].b == identity;
     }
 
-    if (isIdentity || m_presentParams.Windowed)
-      DestroyGammaTexture();
+    if (!isIdentity && !m_presentParams.Windowed)
+      m_blitter->setGammaRamp(NumControlPoints, cp.data());
     else
-      CreateGammaTexture(NumControlPoints, cp.data());
+      m_blitter->setGammaRamp(0, nullptr);
   }
 
 
@@ -754,127 +818,55 @@ namespace dxvk {
 
     // Retrieve the image and image view to present
     auto swapImage = m_backBuffers[0]->GetCommonTexture()->GetImage();
-    auto swapImageView = m_resolveImageView;
+    auto swapImageView = m_backBuffers[0]->GetImageView(false);
 
-    if (swapImageView == nullptr)
-      swapImageView = m_backBuffers[0]->GetImageView(false);
-
-    // Wait for the sync event so that we respect the maximum frame latency
-    uint64_t frameId = ++m_frameId;
-    m_frameLatencySignal->wait(frameId - GetActualFrameLatency());
+    // Bump our frame id.
+    ++m_frameId;
 
     for (uint32_t i = 0; i < SyncInterval || i < 1; i++) {
       SynchronizePresent();
 
-      m_context->beginRecording(
-        m_device->createCommandList());
-      
-      // Resolve back buffer if it is multisampled. We
-      // only have to do it only for the first frame.
-      if (m_resolveImage != nullptr && i == 0) {
-        VkImageSubresourceLayers resolveSubresource;
-        resolveSubresource.aspectMask      = VK_IMAGE_ASPECT_COLOR_BIT;
-        resolveSubresource.mipLevel        = 0;
-        resolveSubresource.baseArrayLayer  = 0;
-        resolveSubresource.layerCount      = 1;
-
-        VkImageResolve resolveRegion;
-        resolveRegion.srcSubresource = resolveSubresource;
-        resolveRegion.srcOffset      = VkOffset3D { 0, 0, 0 };
-        resolveRegion.dstSubresource = resolveSubresource;
-        resolveRegion.dstOffset      = VkOffset3D { 0, 0, 0 };
-        resolveRegion.extent         = swapImage->info().extent;
-        
-        m_context->resolveImage(
-          m_resolveImage, swapImage,
-          resolveRegion, VK_FORMAT_UNDEFINED);
-      }
-      
       // Presentation semaphores and WSI swap chain image
       vk::PresenterInfo info = m_presenter->info();
-      vk::PresenterSync sync = m_presenter->getSyncSemaphores();
+      vk::PresenterSync sync;
 
       uint32_t imageIndex = 0;
 
-      VkResult status = m_presenter->acquireNextImage(
-        sync.acquire, VK_NULL_HANDLE, imageIndex);
+      VkResult status = m_presenter->acquireNextImage(sync, imageIndex);
 
       while (status != VK_SUCCESS && status != VK_SUBOPTIMAL_KHR) {
         RecreateSwapChain(m_vsync);
         
         info = m_presenter->info();
-        sync = m_presenter->getSyncSemaphores();
-
-        status = m_presenter->acquireNextImage(
-          sync.acquire, VK_NULL_HANDLE, imageIndex);
+        status = m_presenter->acquireNextImage(sync, imageIndex);
       }
 
-      // Use an appropriate texture filter depending on whether
-      // the back buffer size matches the swap image size
-      m_context->bindShader(VK_SHADER_STAGE_VERTEX_BIT,   m_vertShader);
-      m_context->bindShader(VK_SHADER_STAGE_FRAGMENT_BIT, m_fragShader);
+      m_context->beginRecording(
+        m_device->createCommandList());
 
-      DxvkRenderTargets renderTargets;
-      renderTargets.color[0].view   = m_imageViews.at(imageIndex);
-      renderTargets.color[0].layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
-      m_context->bindRenderTargets(renderTargets);
+      VkRect2D srcRect = {
+        {  int32_t(m_srcRect.left),                    int32_t(m_srcRect.top)                    },
+        { uint32_t(m_srcRect.right - m_srcRect.left), uint32_t(m_srcRect.bottom - m_srcRect.top) } };
 
-      VkViewport viewport;
-      viewport.x        = float(m_dstRect.left);
-      viewport.y        = float(m_dstRect.top);
-      viewport.width    = float(m_dstRect.right  - m_dstRect.left);
-      viewport.height   = float(m_dstRect.bottom - m_dstRect.top);
-      viewport.minDepth = 0.0f;
-      viewport.maxDepth = 1.0f;
+      VkRect2D dstRect = {
+        {  int32_t(m_dstRect.left),                    int32_t(m_dstRect.top)                    },
+        { uint32_t(m_dstRect.right - m_dstRect.left), uint32_t(m_dstRect.bottom - m_dstRect.top) } };
 
-      VkRect2D scissor;
-      scissor.offset.x      = m_dstRect.left;
-      scissor.offset.y      = m_dstRect.top;
-      scissor.extent.width  = m_dstRect.right  - m_dstRect.left;
-      scissor.extent.height = m_dstRect.bottom - m_dstRect.top;
-
-      m_context->setViewports(1, &viewport, &scissor);
-
-      // Use an appropriate texture filter depending on whether
-      // the back buffer size matches the swap image size
-      bool fitSize = m_dstRect.right  - m_dstRect.left == m_srcRect.right  - m_srcRect.left
-                  && m_dstRect.bottom - m_dstRect.top  == m_srcRect.bottom - m_srcRect.top;
-
-      D3D9PresentInfo presentInfoConsts;
-      presentInfoConsts.scale[0]  = float(m_srcRect.right  - m_srcRect.left) / float(swapImage->info().extent.width);
-      presentInfoConsts.scale[1]  = float(m_srcRect.bottom - m_srcRect.top)  / float(swapImage->info().extent.height);
-
-      presentInfoConsts.offset[0] = float(m_srcRect.left) / float(swapImage->info().extent.width);
-      presentInfoConsts.offset[1] = float(m_srcRect.top)  / float(swapImage->info().extent.height);
-
-      m_context->pushConstants(0, sizeof(D3D9PresentInfo), &presentInfoConsts);
-
-      m_context->setRasterizerState(m_rsState);
-      m_context->setMultisampleState(m_msState);
-      m_context->setDepthStencilState(m_dsState);
-      m_context->setLogicOpState(m_loState);
-      m_context->setBlendMode(0, m_blendMode);
-      
-      m_context->setInputAssemblyState(m_iaState);
-      m_context->setInputLayout(0, nullptr, 0, nullptr);
-
-      m_context->bindResourceSampler(BindingIds::Image, fitSize ? m_samplerFitting : m_samplerScaling);
-      m_context->bindResourceSampler(BindingIds::Gamma, m_gammaSampler);
-
-      m_context->bindResourceView(BindingIds::Image, swapImageView, nullptr);
-      m_context->bindResourceView(BindingIds::Gamma, m_gammaTextureView, nullptr);
-
-      m_context->draw(3, 1, 0, 0);
+      m_blitter->presentImage(m_context.ptr(),
+        m_imageViews.at(imageIndex), dstRect,
+        swapImageView, srcRect);
 
       if (m_hud != nullptr)
         m_hud->render(m_context, info.format, info.imageExtent);
 
       if (i + 1 >= SyncInterval)
-        m_context->signal(m_frameLatencySignal, frameId);
+        m_context->signal(m_frameLatencySignal, m_frameId);
 
       SubmitPresent(sync, i);
     }
 
+    SyncFrameLatency();
+
     // Rotate swap chain buffers so that the back
     // buffer at index 0 becomes the front buffer.
     for (uint32_t i = 1; i < m_backBuffers.size(); i++)
@@ -901,8 +893,7 @@ namespace dxvk {
       if (cHud != nullptr && !cFrameId)
         cHud->update();
 
-      m_device->presentImage(m_presenter,
-        cSync.present, &m_presentStatus);
+      m_device->presentImage(m_presenter, &m_presentStatus);
     });
 
     m_parent->FlushCsChunk();
@@ -965,7 +956,9 @@ namespace dxvk {
       m_device->vkd(),
       presenterDevice,
       presenterDesc);
-    
+
+    m_presenter->setFrameRateLimit(m_parent->GetOptions()->maxFrameRate);
+
     CreateRenderTargetViews();
   }
 
@@ -989,6 +982,7 @@ namespace dxvk {
     imageInfo.access      = 0;
     imageInfo.tiling      = VK_IMAGE_TILING_OPTIMAL;
     imageInfo.layout      = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;
+    imageInfo.shared      = VK_TRUE;
 
     DxvkImageViewCreateInfo viewInfo;
     viewInfo.type         = VK_IMAGE_VIEW_TYPE_2D;
@@ -1004,7 +998,7 @@ namespace dxvk {
       VkImage imageHandle = m_presenter->getImage(i).image;
       
       Rc<DxvkImage> image = new DxvkImage(
-        m_device->vkd(), imageInfo, imageHandle);
+        m_device.ptr(), imageInfo, imageHandle);
 
       m_imageViews[i] = new DxvkImageView(
         m_device->vkd(), image, viewInfo);
@@ -1023,9 +1017,6 @@ namespace dxvk {
   void D3D9SwapChainEx::CreateBackBuffers(uint32_t NumBackBuffers) {
     // Explicitly destroy current swap image before
     // creating a new one to free up resources
-    m_resolveImage     = nullptr;
-    m_resolveImageView = nullptr;
-
     DestroyBackBuffers();
 
     int NumFrontBuffer = m_parent->GetOptions()->noExplicitFrontBuffer ? 0 : 1;
@@ -1044,52 +1035,14 @@ namespace dxvk {
     desc.Pool               = D3DPOOL_DEFAULT;
     desc.Usage              = D3DUSAGE_RENDERTARGET;
     desc.Discard            = FALSE;
+    desc.IsBackBuffer       = TRUE;
+    desc.IsAttachmentOnly   = FALSE;
 
     for (uint32_t i = 0; i < m_backBuffers.size(); i++)
-      m_backBuffers[i] = new D3D9Surface(m_parent, &desc, this);
+      m_backBuffers[i] = new D3D9Surface(m_parent, &desc, this, nullptr);
 
     auto swapImage = m_backBuffers[0]->GetCommonTexture()->GetImage();
 
-    // If the image is multisampled, we need to create
-    // another image which we'll use as a resolve target
-    if (swapImage->info().sampleCount != VK_SAMPLE_COUNT_1_BIT) {
-      DxvkImageCreateInfo resolveInfo;
-      resolveInfo.type          = VK_IMAGE_TYPE_2D;
-      resolveInfo.format        = swapImage->info().format;
-      resolveInfo.flags         = 0;
-      resolveInfo.sampleCount   = VK_SAMPLE_COUNT_1_BIT;
-      resolveInfo.extent        = swapImage->info().extent;
-      resolveInfo.numLayers     = 1;
-      resolveInfo.mipLevels     = 1;
-      resolveInfo.usage         = VK_IMAGE_USAGE_SAMPLED_BIT
-                                | VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT
-                                | VK_IMAGE_USAGE_TRANSFER_DST_BIT;
-      resolveInfo.stages        = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT
-                                | VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT
-                                | VK_PIPELINE_STAGE_TRANSFER_BIT;
-      resolveInfo.access        = VK_ACCESS_SHADER_READ_BIT
-                                | VK_ACCESS_TRANSFER_WRITE_BIT
-                                | VK_ACCESS_COLOR_ATTACHMENT_READ_BIT
-                                | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
-      resolveInfo.tiling        = VK_IMAGE_TILING_OPTIMAL;
-      resolveInfo.layout        = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
-      
-      m_resolveImage = m_device->createImage(
-        resolveInfo, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
-
-      DxvkImageViewCreateInfo viewInfo;
-      viewInfo.type       = VK_IMAGE_VIEW_TYPE_2D;
-      viewInfo.format     = m_resolveImage->info().format;
-      viewInfo.usage      = VK_IMAGE_USAGE_SAMPLED_BIT;
-      viewInfo.aspect     = VK_IMAGE_ASPECT_COLOR_BIT;
-      viewInfo.minLevel   = 0;
-      viewInfo.numLevels  = 1;
-      viewInfo.minLayer   = 0;
-      viewInfo.numLayers  = 1;
-
-      m_resolveImageView = m_device->createImageView(m_resolveImage, viewInfo);
-    }
-
     // Initialize the image so that we can use it. Clearing
     // to black prevents garbled output for the first frame.
     VkImageSubresourceRange subresources;
@@ -1099,19 +1052,13 @@ namespace dxvk {
     subresources.baseArrayLayer = 0;
     subresources.layerCount     = 1;
 
-    VkClearColorValue clearColor;
-    clearColor.float32[0] = 0.0f;
-    clearColor.float32[1] = 0.0f;
-    clearColor.float32[2] = 0.0f;
-    clearColor.float32[3] = 0.0f;
-
     m_context->beginRecording(
       m_device->createCommandList());
     
     for (uint32_t i = 0; i < m_backBuffers.size(); i++) {
-      m_context->clearColorImage(
+      m_context->initImage(
         m_backBuffers[i]->GetCommonTexture()->GetImage(),
-        clearColor, subresources);
+        subresources, VK_IMAGE_LAYOUT_UNDEFINED);
     }
 
     m_device->submitCommandList(
@@ -1121,63 +1068,8 @@ namespace dxvk {
   }
 
 
-  void D3D9SwapChainEx::CreateGammaTexture(
-            UINT                NumControlPoints,
-      const D3D9_VK_GAMMA_CP*   pControlPoints) {
-    if (m_gammaTexture == nullptr
-     || m_gammaTexture->info().extent.width != NumControlPoints) {
-      DxvkImageCreateInfo imgInfo;
-      imgInfo.type        = VK_IMAGE_TYPE_1D;
-      imgInfo.format      = VK_FORMAT_R16G16B16A16_UNORM;
-      imgInfo.flags       = 0;
-      imgInfo.sampleCount = VK_SAMPLE_COUNT_1_BIT;
-      imgInfo.extent      = { NumControlPoints, 1, 1 };
-      imgInfo.numLayers   = 1;
-      imgInfo.mipLevels   = 1;
-      imgInfo.usage       = VK_IMAGE_USAGE_TRANSFER_DST_BIT
-                          | VK_IMAGE_USAGE_SAMPLED_BIT;
-      imgInfo.stages      = VK_PIPELINE_STAGE_TRANSFER_BIT
-                          | VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
-      imgInfo.access      = VK_ACCESS_TRANSFER_WRITE_BIT
-                          | VK_ACCESS_SHADER_READ_BIT;
-      imgInfo.tiling      = VK_IMAGE_TILING_OPTIMAL;
-      imgInfo.layout      = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
-      
-      m_gammaTexture = m_device->createImage(
-        imgInfo, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
-
-      DxvkImageViewCreateInfo viewInfo;
-      viewInfo.type       = VK_IMAGE_VIEW_TYPE_1D;
-      viewInfo.format     = VK_FORMAT_R16G16B16A16_UNORM;
-      viewInfo.usage      = VK_IMAGE_USAGE_SAMPLED_BIT;
-      viewInfo.aspect     = VK_IMAGE_ASPECT_COLOR_BIT;
-      viewInfo.minLevel   = 0;
-      viewInfo.numLevels  = 1;
-      viewInfo.minLayer   = 0;
-      viewInfo.numLayers  = 1;
-      
-      m_gammaTextureView = m_device->createImageView(m_gammaTexture, viewInfo);
-    }
-
-    m_context->beginRecording(
-      m_device->createCommandList());
-    
-    m_context->updateImage(m_gammaTexture,
-      VkImageSubresourceLayers { VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1 },
-      VkOffset3D { 0, 0, 0 },
-      VkExtent3D { NumControlPoints, 1, 1 },
-      pControlPoints, 0, 0);
-    
-    m_device->submitCommandList(
-      m_context->endRecording(),
-      VK_NULL_HANDLE,
-      VK_NULL_HANDLE);
-  }
-
-
-  void D3D9SwapChainEx::DestroyGammaTexture() {
-    m_gammaTexture     = nullptr;
-    m_gammaTextureView = nullptr;
+  void D3D9SwapChainEx::CreateBlitter() {
+    m_blitter = new DxvkSwapchainBlitter(m_device);
   }
 
 
@@ -1191,108 +1083,6 @@ namespace dxvk {
   }
 
 
-  void D3D9SwapChainEx::InitRenderState() {
-    m_iaState.primitiveTopology = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP;
-    m_iaState.primitiveRestart  = VK_FALSE;
-    m_iaState.patchVertexCount  = 0;
-    
-    m_rsState.polygonMode        = VK_POLYGON_MODE_FILL;
-    m_rsState.cullMode           = VK_CULL_MODE_BACK_BIT;
-    m_rsState.frontFace          = VK_FRONT_FACE_COUNTER_CLOCKWISE;
-    m_rsState.depthClipEnable    = VK_FALSE;
-    m_rsState.depthBiasEnable    = VK_FALSE;
-    m_rsState.sampleCount        = VK_SAMPLE_COUNT_1_BIT;
-    
-    m_msState.sampleMask            = 0xffffffff;
-    m_msState.enableAlphaToCoverage = VK_FALSE;
-    
-    VkStencilOpState stencilOp;
-    stencilOp.failOp      = VK_STENCIL_OP_KEEP;
-    stencilOp.passOp      = VK_STENCIL_OP_KEEP;
-    stencilOp.depthFailOp = VK_STENCIL_OP_KEEP;
-    stencilOp.compareOp   = VK_COMPARE_OP_ALWAYS;
-    stencilOp.compareMask = 0xFFFFFFFF;
-    stencilOp.writeMask   = 0xFFFFFFFF;
-    stencilOp.reference   = 0;
-    
-    m_dsState.enableDepthTest   = VK_FALSE;
-    m_dsState.enableDepthWrite  = VK_FALSE;
-    m_dsState.enableStencilTest = VK_FALSE;
-    m_dsState.depthCompareOp    = VK_COMPARE_OP_ALWAYS;
-    m_dsState.stencilOpFront    = stencilOp;
-    m_dsState.stencilOpBack     = stencilOp;
-    
-    m_loState.enableLogicOp = VK_FALSE;
-    m_loState.logicOp       = VK_LOGIC_OP_NO_OP;
-
-    m_blendMode.enableBlending  = VK_FALSE;
-    m_blendMode.colorSrcFactor  = VK_BLEND_FACTOR_ONE;
-    m_blendMode.colorDstFactor  = VK_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
-    m_blendMode.colorBlendOp    = VK_BLEND_OP_ADD;
-    m_blendMode.alphaSrcFactor  = VK_BLEND_FACTOR_ONE;
-    m_blendMode.alphaDstFactor  = VK_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
-    m_blendMode.alphaBlendOp    = VK_BLEND_OP_ADD;
-    m_blendMode.writeMask       = VK_COLOR_COMPONENT_R_BIT
-                                | VK_COLOR_COMPONENT_G_BIT
-                                | VK_COLOR_COMPONENT_B_BIT
-                                | VK_COLOR_COMPONENT_A_BIT;
-  }
-
-
-  void D3D9SwapChainEx::InitSamplers() {
-    DxvkSamplerCreateInfo samplerInfo;
-    samplerInfo.magFilter       = VK_FILTER_NEAREST;
-    samplerInfo.minFilter       = VK_FILTER_NEAREST;
-    samplerInfo.mipmapMode      = VK_SAMPLER_MIPMAP_MODE_NEAREST;
-    samplerInfo.mipmapLodBias   = 0.0f;
-    samplerInfo.mipmapLodMin    = 0.0f;
-    samplerInfo.mipmapLodMax    = 0.0f;
-    samplerInfo.useAnisotropy   = VK_FALSE;
-    samplerInfo.maxAnisotropy   = 1.0f;
-    samplerInfo.addressModeU    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
-    samplerInfo.addressModeV    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
-    samplerInfo.addressModeW    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
-    samplerInfo.compareToDepth  = VK_FALSE;
-    samplerInfo.compareOp       = VK_COMPARE_OP_ALWAYS;
-    samplerInfo.borderColor     = VkClearColorValue();
-    samplerInfo.usePixelCoord   = VK_FALSE;
-    m_samplerFitting = m_device->createSampler(samplerInfo);
-
-    samplerInfo.magFilter       = VK_FILTER_LINEAR;
-    samplerInfo.minFilter       = VK_FILTER_LINEAR;
-    m_samplerScaling = m_device->createSampler(samplerInfo);
-
-    samplerInfo.addressModeU    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
-    samplerInfo.addressModeV    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
-    samplerInfo.addressModeW    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
-    m_gammaSampler = m_device->createSampler(samplerInfo);
-  }
-
-
-  void D3D9SwapChainEx::InitShaders() {
-    const SpirvCodeBuffer vsCode(d3d9_presenter_vert);
-    const SpirvCodeBuffer fsCode(d3d9_presenter_frag);
-    
-    const std::array<DxvkResourceSlot, 2> fsResourceSlots = {{
-      { BindingIds::Image, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_IMAGE_VIEW_TYPE_2D },
-      { BindingIds::Gamma, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_IMAGE_VIEW_TYPE_1D },
-    }};
-
-    m_vertShader = m_device->createShader(
-      VK_SHADER_STAGE_VERTEX_BIT,
-      0, nullptr,
-      { 0u, 1u,
-        0u, sizeof(D3D9PresentInfo) },
-      vsCode);
-    
-    m_fragShader = m_device->createShader(
-      VK_SHADER_STAGE_FRAGMENT_BIT,
-      fsResourceSlots.size(),
-      fsResourceSlots.data(),
-      { 1u, 1u }, fsCode);
-  }
-
-
   void D3D9SwapChainEx::InitRamp() {
     for (uint32_t i = 0; i < NumControlPoints; i++) {
       DWORD identity = DWORD(MapGammaControlPoint(float(i) / float(NumControlPoints - 1)));
@@ -1304,6 +1094,11 @@ namespace dxvk {
   }
 
 
+  void D3D9SwapChainEx::SyncFrameLatency() {
+    // Wait for the sync event so that we respect the maximum frame latency
+    m_frameLatencySignal->wait(m_frameId - GetActualFrameLatency());
+  }
+
 
   uint32_t D3D9SwapChainEx::GetActualFrameLatency() {
     uint32_t maxFrameLatency = m_parent->GetFrameLatency();
@@ -1323,8 +1118,9 @@ namespace dxvk {
 
     switch (Format) {
       default:
-        Logger::warn(str::format("D3D9SwapChainEx: Unexpected format: ", Format));
-        
+        Logger::warn(str::format("D3D9SwapChainEx: Unexpected format: ", Format));      
+     [[fallthrough]];
+
       case D3D9Format::A8R8G8B8:
       case D3D9Format::X8R8G8B8:
       case D3D9Format::A8B8G8R8:
@@ -1344,12 +1140,12 @@ namespace dxvk {
         pDstFormats[n++] = { VK_FORMAT_B5G5R5A1_UNORM_PACK16, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR };
         pDstFormats[n++] = { VK_FORMAT_R5G5B5A1_UNORM_PACK16, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR };
         pDstFormats[n++] = { VK_FORMAT_A1R5G5B5_UNORM_PACK16, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR };
-      }
+      } break;
 
       case D3D9Format::R5G6B5: {
         pDstFormats[n++] = { VK_FORMAT_B5G6R5_UNORM_PACK16, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR };
         pDstFormats[n++] = { VK_FORMAT_R5G6B5_UNORM_PACK16, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR };
-      }
+      } break;
     }
 
     return n;
@@ -1381,6 +1177,13 @@ namespace dxvk {
     return option > 0 ? uint32_t(option) : uint32_t(Preferred);
   }
 
+
+  void D3D9SwapChainEx::NotifyDisplayRefreshRate(
+          double                  RefreshRate) {
+    m_displayRefreshRate = RefreshRate;
+  }
+
+
   HRESULT D3D9SwapChainEx::EnterFullscreenMode(
           D3DPRESENT_PARAMETERS* pPresentParams,
     const D3DDISPLAYMODEEX*      pFullscreenDisplayMode) {    
@@ -1398,7 +1201,7 @@ namespace dxvk {
     // Some games restore window styles after we have changed it, so hooking is
     // also required. Doing it will allow us to create fullscreen windows
     // regardless of their style and it also appears to work on Windows.
-    HookWindowProc(m_window);
+    HookWindowProc(m_window, this);
 
     D3D9WindowMessageFilter filter(m_window);
     
@@ -1489,9 +1292,19 @@ namespace dxvk {
       devMode.dmDisplayFrequency = mode.RefreshRate;
     }
     
-    return SetMonitorDisplayMode(GetDefaultMonitor(), &devMode)
-      ? D3D_OK
-      : D3DERR_NOTAVAILABLE;
+    HMONITOR monitor = GetDefaultMonitor();
+
+    if (!SetMonitorDisplayMode(monitor, &devMode))
+      return D3DERR_NOTAVAILABLE;
+
+    devMode.dmFields = DM_DISPLAYFREQUENCY;
+    
+    if (GetMonitorDisplayMode(monitor, ENUM_CURRENT_SETTINGS, &devMode))
+      NotifyDisplayRefreshRate(double(devMode.dmDisplayFrequency));
+    else
+      NotifyDisplayRefreshRate(0.0);
+
+    return D3D_OK;
   }
   
   
@@ -1499,9 +1312,11 @@ namespace dxvk {
     if (hMonitor == nullptr)
       return D3DERR_INVALIDCALL;
     
-    return RestoreMonitorDisplayMode(hMonitor)
-      ? D3D_OK
-      : D3DERR_NOTAVAILABLE;
+    if (!RestoreMonitorDisplayMode())
+      return D3DERR_NOTAVAILABLE;
+
+    NotifyDisplayRefreshRate(0.0);
+    return D3D_OK;
   }
 
   bool    D3D9SwapChainEx::UpdatePresentRegion(const RECT* pSourceRect, const RECT* pDestRect) {
@@ -1557,4 +1372,4 @@ namespace dxvk {
     return this->GetParent()->IsExtended() ? "D3D9Ex" : "D3D9";
   }
 
-}
\ No newline at end of file
+}
