@@ -108,144 +108,158 @@ namespace dxvk {
     return queues;
   }
 
+  static bool log_info_and_return_false(const char *message) {
+    Logger::info(message);
+    return false;
+  }
 
   bool DxvkAdapter::checkFeatureSupport(const DxvkDeviceFeatures& required) const {
+#define REPORT(x) log_info_and_return_false("Missing feature: " #x)
     return (m_deviceFeatures.core.features.robustBufferAccess
-                || !required.core.features.robustBufferAccess)
+                || !required.core.features.robustBufferAccess || REPORT(robustBufferAccess))
         && (m_deviceFeatures.core.features.fullDrawIndexUint32
-                || !required.core.features.fullDrawIndexUint32)
+                || !required.core.features.fullDrawIndexUint32 || REPORT(fullDrawIndexUint32))
         && (m_deviceFeatures.core.features.imageCubeArray
-                || !required.core.features.imageCubeArray)
+                || !required.core.features.imageCubeArray || REPORT(imageCubeArray))
         && (m_deviceFeatures.core.features.independentBlend
-                || !required.core.features.independentBlend)
+                || !required.core.features.independentBlend || REPORT(independentBlend))
         && (m_deviceFeatures.core.features.geometryShader
-                || !required.core.features.geometryShader)
+                || !required.core.features.geometryShader || REPORT(geometryShader))
         && (m_deviceFeatures.core.features.tessellationShader
-                || !required.core.features.tessellationShader)
+                || !required.core.features.tessellationShader || REPORT(tessellationShader))
         && (m_deviceFeatures.core.features.sampleRateShading
-                || !required.core.features.sampleRateShading)
+                || !required.core.features.sampleRateShading || REPORT(sampleRateShading))
         && (m_deviceFeatures.core.features.dualSrcBlend
-                || !required.core.features.dualSrcBlend)
+                || !required.core.features.dualSrcBlend || REPORT(dualSrcBlend))
         && (m_deviceFeatures.core.features.logicOp
-                || !required.core.features.logicOp)
+                || !required.core.features.logicOp || REPORT(logicOp))
         && (m_deviceFeatures.core.features.multiDrawIndirect
-                || !required.core.features.multiDrawIndirect)
+                || !required.core.features.multiDrawIndirect || REPORT(multiDrawIndirect))
         && (m_deviceFeatures.core.features.drawIndirectFirstInstance
-                || !required.core.features.drawIndirectFirstInstance)
+                || !required.core.features.drawIndirectFirstInstance || REPORT(drawIndirectFirstInstance))
         && (m_deviceFeatures.core.features.depthClamp
-                || !required.core.features.depthClamp)
+                || !required.core.features.depthClamp || REPORT(depthClamp))
         && (m_deviceFeatures.core.features.depthBiasClamp
-                || !required.core.features.depthBiasClamp)
+                || !required.core.features.depthBiasClamp || REPORT(depthBiasClamp))
         && (m_deviceFeatures.core.features.fillModeNonSolid
-                || !required.core.features.fillModeNonSolid)
+                || !required.core.features.fillModeNonSolid || REPORT(fillModeNonSolid))
         && (m_deviceFeatures.core.features.depthBounds
-                || !required.core.features.depthBounds)
+                || !required.core.features.depthBounds || REPORT(depthBounds))
         && (m_deviceFeatures.core.features.wideLines
-                || !required.core.features.wideLines)
+                || !required.core.features.wideLines || REPORT(wideLines))
         && (m_deviceFeatures.core.features.largePoints
-                || !required.core.features.largePoints)
+                || !required.core.features.largePoints || REPORT(largePoints))
         && (m_deviceFeatures.core.features.alphaToOne
-                || !required.core.features.alphaToOne)
+                || !required.core.features.alphaToOne || REPORT(alphaToOne))
         && (m_deviceFeatures.core.features.multiViewport
-                || !required.core.features.multiViewport)
+                || !required.core.features.multiViewport || REPORT(multiViewport))
         && (m_deviceFeatures.core.features.samplerAnisotropy
-                || !required.core.features.samplerAnisotropy)
+                || !required.core.features.samplerAnisotropy || REPORT(samplerAnisotropy))
         && (m_deviceFeatures.core.features.textureCompressionETC2
-                || !required.core.features.textureCompressionETC2)
+                || !required.core.features.textureCompressionETC2 || REPORT(textureCompressionETC2))
         && (m_deviceFeatures.core.features.textureCompressionASTC_LDR
-                || !required.core.features.textureCompressionASTC_LDR)
+                || !required.core.features.textureCompressionASTC_LDR || REPORT(textureCompressionASTC_LDR))
         && (m_deviceFeatures.core.features.textureCompressionBC
-                || !required.core.features.textureCompressionBC)
+                || !required.core.features.textureCompressionBC || REPORT(textureCompressionBC))
         && (m_deviceFeatures.core.features.occlusionQueryPrecise
-                || !required.core.features.occlusionQueryPrecise)
+                || !required.core.features.occlusionQueryPrecise || REPORT(occlusionQueryPrecise))
         && (m_deviceFeatures.core.features.pipelineStatisticsQuery
-                || !required.core.features.pipelineStatisticsQuery)
+                || !required.core.features.pipelineStatisticsQuery || REPORT(pipelineStatisticsQuery))
         && (m_deviceFeatures.core.features.vertexPipelineStoresAndAtomics
-                || !required.core.features.vertexPipelineStoresAndAtomics)
+                || !required.core.features.vertexPipelineStoresAndAtomics || REPORT(vertexPipelineStoresAndAtomics))
         && (m_deviceFeatures.core.features.fragmentStoresAndAtomics
-                || !required.core.features.fragmentStoresAndAtomics)
+                || !required.core.features.fragmentStoresAndAtomics || REPORT(fragmentStoresAndAtomics))
         && (m_deviceFeatures.core.features.shaderTessellationAndGeometryPointSize
-                || !required.core.features.shaderTessellationAndGeometryPointSize)
+                || !required.core.features.shaderTessellationAndGeometryPointSize || REPORT(shaderTessellationAndGeometryPointSize))
         && (m_deviceFeatures.core.features.shaderImageGatherExtended
-                || !required.core.features.shaderImageGatherExtended)
+                || !required.core.features.shaderImageGatherExtended || REPORT(shaderImageGatherExtended))
         && (m_deviceFeatures.core.features.shaderStorageImageExtendedFormats
-                || !required.core.features.shaderStorageImageExtendedFormats)
+                || !required.core.features.shaderStorageImageExtendedFormats || REPORT(shaderStorageImageExtendedFormats))
         && (m_deviceFeatures.core.features.shaderStorageImageMultisample
-                || !required.core.features.shaderStorageImageMultisample)
+                || !required.core.features.shaderStorageImageMultisample || REPORT(shaderStorageImageMultisample))
         && (m_deviceFeatures.core.features.shaderStorageImageReadWithoutFormat
-                || !required.core.features.shaderStorageImageReadWithoutFormat)
+                || !required.core.features.shaderStorageImageReadWithoutFormat || REPORT(shaderStorageImageReadWithoutFormat))
         && (m_deviceFeatures.core.features.shaderStorageImageWriteWithoutFormat
-                || !required.core.features.shaderStorageImageWriteWithoutFormat)
+                || !required.core.features.shaderStorageImageWriteWithoutFormat || REPORT(shaderStorageImageWriteWithoutFormat))
         && (m_deviceFeatures.core.features.shaderUniformBufferArrayDynamicIndexing
-                || !required.core.features.shaderUniformBufferArrayDynamicIndexing)
+                || !required.core.features.shaderUniformBufferArrayDynamicIndexing || REPORT(shaderUniformBufferArrayDynamicIndexing))
         && (m_deviceFeatures.core.features.shaderSampledImageArrayDynamicIndexing
-                || !required.core.features.shaderSampledImageArrayDynamicIndexing)
+                || !required.core.features.shaderSampledImageArrayDynamicIndexing || REPORT(shaderSampledImageArrayDynamicIndexing))
         && (m_deviceFeatures.core.features.shaderStorageBufferArrayDynamicIndexing
-                || !required.core.features.shaderStorageBufferArrayDynamicIndexing)
+                || !required.core.features.shaderStorageBufferArrayDynamicIndexing || REPORT(shaderStorageBufferArrayDynamicIndexing))
         && (m_deviceFeatures.core.features.shaderStorageImageArrayDynamicIndexing
-                || !required.core.features.shaderStorageImageArrayDynamicIndexing)
+                || !required.core.features.shaderStorageImageArrayDynamicIndexing || REPORT(shaderStorageImageArrayDynamicIndexing))
         && (m_deviceFeatures.core.features.shaderClipDistance
-                || !required.core.features.shaderClipDistance)
+                || !required.core.features.shaderClipDistance || REPORT(shaderClipDistance))
         && (m_deviceFeatures.core.features.shaderCullDistance
-                || !required.core.features.shaderCullDistance)
+                || !required.core.features.shaderCullDistance || REPORT(shaderCullDistance))
         && (m_deviceFeatures.core.features.shaderFloat64
-                || !required.core.features.shaderFloat64)
+                || !required.core.features.shaderFloat64 || REPORT(shaderFloat64))
         && (m_deviceFeatures.core.features.shaderInt64
-                || !required.core.features.shaderInt64)
+                || !required.core.features.shaderInt64 || REPORT(shaderInt64))
         && (m_deviceFeatures.core.features.shaderInt16
-                || !required.core.features.shaderInt16)
+                || !required.core.features.shaderInt16 || REPORT(shaderInt16))
         && (m_deviceFeatures.core.features.shaderResourceResidency
-                || !required.core.features.shaderResourceResidency)
+                || !required.core.features.shaderResourceResidency || REPORT(shaderResourceResidency))
         && (m_deviceFeatures.core.features.shaderResourceMinLod
-                || !required.core.features.shaderResourceMinLod)
+                || !required.core.features.shaderResourceMinLod || REPORT(shaderResourceMinLod))
         && (m_deviceFeatures.core.features.sparseBinding
-                || !required.core.features.sparseBinding)
+                || !required.core.features.sparseBinding || REPORT(sparseBinding))
         && (m_deviceFeatures.core.features.sparseResidencyBuffer
-                || !required.core.features.sparseResidencyBuffer)
+                || !required.core.features.sparseResidencyBuffer || REPORT(sparseResidencyBuffer))
         && (m_deviceFeatures.core.features.sparseResidencyImage2D
-                || !required.core.features.sparseResidencyImage2D)
+                || !required.core.features.sparseResidencyImage2D || REPORT(sparseResidencyImage2D))
         && (m_deviceFeatures.core.features.sparseResidencyImage3D
-                || !required.core.features.sparseResidencyImage3D)
+                || !required.core.features.sparseResidencyImage3D || REPORT(sparseResidencyImage3D))
         && (m_deviceFeatures.core.features.sparseResidency2Samples
-                || !required.core.features.sparseResidency2Samples)
+                || !required.core.features.sparseResidency2Samples || REPORT(sparseResidency2Samples))
         && (m_deviceFeatures.core.features.sparseResidency4Samples
-                || !required.core.features.sparseResidency4Samples)
+                || !required.core.features.sparseResidency4Samples || REPORT(sparseResidency4Samples))
         && (m_deviceFeatures.core.features.sparseResidency8Samples
-                || !required.core.features.sparseResidency8Samples)
+                || !required.core.features.sparseResidency8Samples || REPORT(sparseResidency8Samples))
         && (m_deviceFeatures.core.features.sparseResidency16Samples
-                || !required.core.features.sparseResidency16Samples)
+                || !required.core.features.sparseResidency16Samples || REPORT(sparseResidency16Samples))
         && (m_deviceFeatures.core.features.sparseResidencyAliased
-                || !required.core.features.sparseResidencyAliased)
+                || !required.core.features.sparseResidencyAliased || REPORT(sparseResidencyAliased))
         && (m_deviceFeatures.core.features.variableMultisampleRate
-                || !required.core.features.variableMultisampleRate)
+                || !required.core.features.variableMultisampleRate || REPORT(variableMultisampleRate))
         && (m_deviceFeatures.core.features.inheritedQueries
-                || !required.core.features.inheritedQueries)
+                || !required.core.features.inheritedQueries || REPORT(inheritedQueries))
         && (m_deviceFeatures.shaderDrawParameters.shaderDrawParameters
-                || !required.shaderDrawParameters.shaderDrawParameters)
-        && (m_deviceFeatures.extConditionalRendering.conditionalRendering
-                || !required.extConditionalRendering.conditionalRendering)
+                || !required.shaderDrawParameters.shaderDrawParameters || REPORT(shaderDrawParameters))
+        && (m_deviceFeatures.ext4444Formats.formatA4R4G4B4
+                || !required.ext4444Formats.formatA4R4G4B4 || REPORT(formatA4R4G4B4))
+        && (m_deviceFeatures.ext4444Formats.formatA4B4G4R4
+                || !required.ext4444Formats.formatA4B4G4R4 || REPORT(formatA4B4G4R4))
         && (m_deviceFeatures.extCustomBorderColor.customBorderColors
-                || !required.extCustomBorderColor.customBorderColors)
+                || !required.extCustomBorderColor.customBorderColors || REPORT(customBorderColors))
         && (m_deviceFeatures.extCustomBorderColor.customBorderColorWithoutFormat
-                || !required.extCustomBorderColor.customBorderColorWithoutFormat)
+                || !required.extCustomBorderColor.customBorderColorWithoutFormat || REPORT(customBorderColorWithoutFormat))
         && (m_deviceFeatures.extDepthClipEnable.depthClipEnable
-                || !required.extDepthClipEnable.depthClipEnable)
+                || !required.extDepthClipEnable.depthClipEnable || REPORT(depthClipEnable))
+        && (m_deviceFeatures.extExtendedDynamicState.extendedDynamicState
+                || !required.extExtendedDynamicState.extendedDynamicState || REPORT(extendedDynamicState))
         && (m_deviceFeatures.extHostQueryReset.hostQueryReset
-                || !required.extHostQueryReset.hostQueryReset)
+                || !required.extHostQueryReset.hostQueryReset || REPORT(hostQueryReset))
         && (m_deviceFeatures.extMemoryPriority.memoryPriority
-                || !required.extMemoryPriority.memoryPriority)
+                || !required.extMemoryPriority.memoryPriority || REPORT(memoryPriority))
+        && (m_deviceFeatures.extNonSeamlessCubeMap.nonSeamlessCubeMap
+                || !required.extNonSeamlessCubeMap.nonSeamlessCubeMap || REPORT(nonSeamlessCubeMap))
         && (m_deviceFeatures.extRobustness2.robustBufferAccess2
-                || !required.extRobustness2.robustBufferAccess2)
+                || !required.extRobustness2.robustBufferAccess2 || REPORT(robustBufferAccess2))
         && (m_deviceFeatures.extRobustness2.robustImageAccess2
-                || !required.extRobustness2.robustImageAccess2)
+                || !required.extRobustness2.robustImageAccess2 || REPORT(robustImageAccess2))
         && (m_deviceFeatures.extRobustness2.nullDescriptor
-                || !required.extRobustness2.nullDescriptor)
+                || !required.extRobustness2.nullDescriptor || REPORT(nullDescriptor))
         && (m_deviceFeatures.extTransformFeedback.transformFeedback
-                || !required.extTransformFeedback.transformFeedback)
+                || !required.extTransformFeedback.transformFeedback || REPORT(transformFeedback))
         && (m_deviceFeatures.extVertexAttributeDivisor.vertexAttributeInstanceRateDivisor
-                || !required.extVertexAttributeDivisor.vertexAttributeInstanceRateDivisor)
+                || !required.extVertexAttributeDivisor.vertexAttributeInstanceRateDivisor || REPORT(vertexAttributeInstanceRateDivisor))
         && (m_deviceFeatures.extVertexAttributeDivisor.vertexAttributeInstanceRateZeroDivisor
-                || !required.extVertexAttributeDivisor.vertexAttributeInstanceRateZeroDivisor);
+                || !required.extVertexAttributeDivisor.vertexAttributeInstanceRateZeroDivisor || REPORT(vertexAttributeInstanceRateZeroDivisor))
+        && (m_deviceFeatures.khrTimelineSemaphore.timelineSemaphore
+                || !required.khrTimelineSemaphore.timelineSemaphore || REPORT(timelineSemaphore));
+#undef REPORT
   }
   
   
@@ -259,31 +273,57 @@ namespace dxvk {
           DxvkDeviceFeatures  enabledFeatures) {
     DxvkDeviceExtensions devExtensions;
 
-    std::array<DxvkExt*, 22> devExtensionList = {{
+    std::array<DxvkExt*, 32> devExtensionList = {{
       &devExtensions.amdMemoryOverallocationBehaviour,
       &devExtensions.amdShaderFragmentMask,
-      &devExtensions.extConditionalRendering,
+      &devExtensions.ext4444Formats,
+      &devExtensions.extConservativeRasterization,
       &devExtensions.extCustomBorderColor,
       &devExtensions.extDepthClipEnable,
+      &devExtensions.extExtendedDynamicState,
       &devExtensions.extFullScreenExclusive,
       &devExtensions.extHostQueryReset,
       &devExtensions.extMemoryBudget,
       &devExtensions.extMemoryPriority,
+      &devExtensions.extNonSeamlessCubeMap,
       &devExtensions.extRobustness2,
       &devExtensions.extShaderDemoteToHelperInvocation,
       &devExtensions.extShaderStencilExport,
       &devExtensions.extShaderViewportIndexLayer,
       &devExtensions.extTransformFeedback,
       &devExtensions.extVertexAttributeDivisor,
+      &devExtensions.khrBufferDeviceAddress,
       &devExtensions.khrCreateRenderPass2,
       &devExtensions.khrDepthStencilResolve,
       &devExtensions.khrDrawIndirectCount,
       &devExtensions.khrDriverProperties,
+      &devExtensions.khrExternalMemoryWin32,
+      &devExtensions.khrExternalSemaphoreWin32,
       &devExtensions.khrImageFormatList,
       &devExtensions.khrSamplerMirrorClampToEdge,
+      &devExtensions.khrShaderFloatControls,
       &devExtensions.khrSwapchain,
+      &devExtensions.khrTimelineSemaphore,
+      &devExtensions.nvxBinaryImport,
+      &devExtensions.nvxImageViewHandle,
     }};
 
+    // Only enable Cuda interop extensions in 64-bit builds in
+    // order to avoid potential driver or address space issues.
+    // VK_KHR_buffer_device_address is expensive on some drivers.
+    bool enableCudaInterop = !env::is32BitHostPlatform() &&
+      m_deviceExtensions.supports(devExtensions.nvxBinaryImport.name()) &&
+      m_deviceExtensions.supports(devExtensions.nvxImageViewHandle.name()) &&
+      m_deviceFeatures.khrBufferDeviceAddress.bufferDeviceAddress;
+
+    if (enableCudaInterop) {
+      devExtensions.nvxBinaryImport.setMode(DxvkExtMode::Optional);
+      devExtensions.nvxImageViewHandle.setMode(DxvkExtMode::Optional);
+      devExtensions.khrBufferDeviceAddress.setMode(DxvkExtMode::Optional);
+
+      enabledFeatures.khrBufferDeviceAddress.bufferDeviceAddress = VK_TRUE;
+    }
+
     DxvkNameSet extensionsEnabled;
 
     if (!m_deviceExtensions.enableExtensions(
@@ -295,6 +335,12 @@ namespace dxvk {
     // Enable additional extensions if necessary
     extensionsEnabled.merge(m_extraExtensions);
     DxvkNameList extensionNameList = extensionsEnabled.toNameList();
+
+    // Enable additional device features if supported
+    enabledFeatures.extExtendedDynamicState.extendedDynamicState = m_deviceFeatures.extExtendedDynamicState.extendedDynamicState;
+
+    enabledFeatures.ext4444Formats.formatA4B4G4R4 = m_deviceFeatures.ext4444Formats.formatA4B4G4R4;
+    enabledFeatures.ext4444Formats.formatA4R4G4B4 = m_deviceFeatures.ext4444Formats.formatA4R4G4B4;
     
     Logger::info(str::format("Device properties:"
       "\n  Device name:     : ", m_deviceInfo.core.properties.deviceName,
@@ -314,9 +360,9 @@ namespace dxvk {
     enabledFeatures.shaderDrawParameters.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DRAW_PARAMETERS_FEATURES;
     enabledFeatures.shaderDrawParameters.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.shaderDrawParameters);
 
-    if (devExtensions.extConditionalRendering) {
-      enabledFeatures.extConditionalRendering.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONDITIONAL_RENDERING_FEATURES_EXT;
-      enabledFeatures.extConditionalRendering.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extConditionalRendering);
+    if (devExtensions.ext4444Formats) {
+      enabledFeatures.ext4444Formats.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_4444_FORMATS_FEATURES_EXT;
+      enabledFeatures.ext4444Formats.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.ext4444Formats);
     }
 
     if (devExtensions.extCustomBorderColor) {
@@ -329,6 +375,11 @@ namespace dxvk {
       enabledFeatures.extDepthClipEnable.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extDepthClipEnable);
     }
 
+    if (devExtensions.extExtendedDynamicState) {
+      enabledFeatures.extExtendedDynamicState.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_FEATURES_EXT;
+      enabledFeatures.extExtendedDynamicState.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extExtendedDynamicState);
+    }
+
     if (devExtensions.extHostQueryReset) {
       enabledFeatures.extHostQueryReset.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES_EXT;
       enabledFeatures.extHostQueryReset.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extHostQueryReset);
@@ -339,6 +390,11 @@ namespace dxvk {
       enabledFeatures.extMemoryPriority.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extMemoryPriority);
     }
 
+    if (devExtensions.extNonSeamlessCubeMap) {
+      enabledFeatures.extNonSeamlessCubeMap.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NON_SEAMLESS_CUBE_MAP_FEATURES_EXT;
+      enabledFeatures.extNonSeamlessCubeMap.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extNonSeamlessCubeMap);
+    }
+
     if (devExtensions.extShaderDemoteToHelperInvocation) {
       enabledFeatures.extShaderDemoteToHelperInvocation.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DEMOTE_TO_HELPER_INVOCATION_FEATURES_EXT;
       enabledFeatures.extShaderDemoteToHelperInvocation.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extShaderDemoteToHelperInvocation);
@@ -359,6 +415,16 @@ namespace dxvk {
       enabledFeatures.extVertexAttributeDivisor.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.extVertexAttributeDivisor);
     }
 
+    if (devExtensions.khrBufferDeviceAddress) {
+      enabledFeatures.khrBufferDeviceAddress.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES_KHR;
+      enabledFeatures.khrBufferDeviceAddress.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.khrBufferDeviceAddress);
+    }
+
+    if (devExtensions.khrTimelineSemaphore) {
+      enabledFeatures.khrTimelineSemaphore.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES_KHR;
+      enabledFeatures.khrTimelineSemaphore.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.khrTimelineSemaphore);
+    }
+
     // Report the desired overallocation behaviour to the driver
     VkDeviceMemoryOverallocationCreateInfoAMD overallocInfo;
     overallocInfo.sType = VK_STRUCTURE_TYPE_DEVICE_MEMORY_OVERALLOCATION_CREATE_INFO_AMD;
@@ -403,8 +469,31 @@ namespace dxvk {
       overallocInfo.pNext = std::exchange(info.pNext, &overallocInfo);
     
     VkDevice device = VK_NULL_HANDLE;
-    
-    if (m_vki->vkCreateDevice(m_handle, &info, nullptr, &device) != VK_SUCCESS)
+    VkResult vr = m_vki->vkCreateDevice(m_handle, &info, nullptr, &device);
+
+    if (vr != VK_SUCCESS && enableCudaInterop) {
+      // Enabling certain Vulkan extensions can cause device creation to fail on
+      // Nvidia drivers if a certain kernel module isn't loaded, but we cannot know
+      // that in advance since the extensions are reported as supported anyway.
+      Logger::err("DxvkAdapter: Failed to create device, retrying without CUDA interop extensions");
+
+      extensionsEnabled.disableExtension(devExtensions.khrBufferDeviceAddress);
+      extensionsEnabled.disableExtension(devExtensions.nvxBinaryImport);
+      extensionsEnabled.disableExtension(devExtensions.nvxImageViewHandle);
+
+      enabledFeatures.khrBufferDeviceAddress.bufferDeviceAddress = VK_FALSE;
+
+      vk::removeStructFromPNextChain(&enabledFeatures.core.pNext,
+        VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES_KHR);
+
+      extensionNameList = extensionsEnabled.toNameList();
+      info.enabledExtensionCount      = extensionNameList.count();
+      info.ppEnabledExtensionNames    = extensionNameList.names();
+
+      vr = m_vki->vkCreateDevice(m_handle, &info, nullptr, &device);
+    }
+
+    if (vr != VK_SUCCESS)
       throw DxvkError("DxvkAdapter: Failed to create device");
     
     Rc<DxvkDevice> result = new DxvkDevice(instance, this,
@@ -479,6 +568,17 @@ namespace dxvk {
   }
   
   
+  bool DxvkAdapter::isUnifiedMemoryArchitecture() const {
+    auto memory = this->memoryProperties();
+    bool result = true;
+
+    for (uint32_t i = 0; i < memory.memoryHeapCount; i++)
+      result &= memory.memoryHeaps[i].flags & VK_MEMORY_HEAP_DEVICE_LOCAL_BIT;
+
+    return result;
+  }
+  
+  
   void DxvkAdapter::initHeapAllocInfo() {
     for (uint32_t i = 0; i < m_heapAlloc.size(); i++)
       m_heapAlloc[i] = 0;
@@ -504,6 +604,11 @@ namespace dxvk {
     m_deviceInfo.coreSubgroup.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES;
     m_deviceInfo.coreSubgroup.pNext = std::exchange(m_deviceInfo.core.pNext, &m_deviceInfo.coreSubgroup);
 
+    if (m_deviceExtensions.supports(VK_EXT_CONSERVATIVE_RASTERIZATION_EXTENSION_NAME)) {
+      m_deviceInfo.extConservativeRasterization.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONSERVATIVE_RASTERIZATION_PROPERTIES_EXT;
+      m_deviceInfo.extConservativeRasterization.pNext = std::exchange(m_deviceInfo.core.pNext, &m_deviceInfo.extConservativeRasterization);
+    }
+
     if (m_deviceExtensions.supports(VK_EXT_CUSTOM_BORDER_COLOR_EXTENSION_NAME)) {
       m_deviceInfo.extCustomBorderColor.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_PROPERTIES_EXT;
       m_deviceInfo.extCustomBorderColor.pNext = std::exchange(m_deviceInfo.core.pNext, &m_deviceInfo.extCustomBorderColor);
@@ -534,15 +639,35 @@ namespace dxvk {
       m_deviceInfo.khrDeviceDriverProperties.pNext = std::exchange(m_deviceInfo.core.pNext, &m_deviceInfo.khrDeviceDriverProperties);
     }
 
+    if (m_deviceExtensions.supports(VK_KHR_SHADER_FLOAT_CONTROLS_EXTENSION_NAME)) {
+      m_deviceInfo.khrShaderFloatControls.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FLOAT_CONTROLS_PROPERTIES_KHR;
+      m_deviceInfo.khrShaderFloatControls.pNext = std::exchange(m_deviceInfo.core.pNext, &m_deviceInfo.khrShaderFloatControls);
+    }
+
+    if (m_deviceExtensions.supports(VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME)) {
+      m_deviceInfo.khrTimelineSemaphore.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_PROPERTIES_KHR;
+      m_deviceInfo.khrTimelineSemaphore.pNext = std::exchange(m_deviceInfo.core.pNext, &m_deviceInfo.khrTimelineSemaphore);
+    }
+
     // Query full device properties for all enabled extensions
     m_vki->vkGetPhysicalDeviceProperties2(m_handle, &m_deviceInfo.core);
     
-    // Nvidia reports the driver version in a slightly different format
-    if (DxvkGpuVendor(m_deviceInfo.core.properties.vendorID) == DxvkGpuVendor::Nvidia) {
-      m_deviceInfo.core.properties.driverVersion = VK_MAKE_VERSION(
-        VK_VERSION_MAJOR(m_deviceInfo.core.properties.driverVersion),
-        VK_VERSION_MINOR(m_deviceInfo.core.properties.driverVersion >> 0) >> 2,
-        VK_VERSION_PATCH(m_deviceInfo.core.properties.driverVersion >> 2) >> 4);
+    // Some drivers reports the driver version in a slightly different format
+    switch (m_deviceInfo.khrDeviceDriverProperties.driverID) {
+      case VK_DRIVER_ID_NVIDIA_PROPRIETARY:
+        m_deviceInfo.core.properties.driverVersion = VK_MAKE_VERSION(
+          (m_deviceInfo.core.properties.driverVersion >> 22) & 0x3ff,
+          (m_deviceInfo.core.properties.driverVersion >> 14) & 0x0ff,
+          (m_deviceInfo.core.properties.driverVersion >>  6) & 0x0ff);
+        break;
+
+      case VK_DRIVER_ID_INTEL_PROPRIETARY_WINDOWS:
+        m_deviceInfo.core.properties.driverVersion = VK_MAKE_VERSION(
+          m_deviceInfo.core.properties.driverVersion >> 14,
+          m_deviceInfo.core.properties.driverVersion & 0x3fff, 0);
+        break;
+
+      default:;
     }
   }
 
@@ -555,9 +680,9 @@ namespace dxvk {
     m_deviceFeatures.shaderDrawParameters.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DRAW_PARAMETERS_FEATURES;
     m_deviceFeatures.shaderDrawParameters.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.shaderDrawParameters);
 
-    if (m_deviceExtensions.supports(VK_EXT_CONDITIONAL_RENDERING_EXTENSION_NAME)) {
-      m_deviceFeatures.extConditionalRendering.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONDITIONAL_RENDERING_FEATURES_EXT;
-      m_deviceFeatures.extConditionalRendering.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extConditionalRendering);
+    if (m_deviceExtensions.supports(VK_EXT_4444_FORMATS_EXTENSION_NAME)) {
+      m_deviceFeatures.ext4444Formats.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_4444_FORMATS_FEATURES_EXT;
+      m_deviceFeatures.ext4444Formats.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.ext4444Formats);
     }
 
     if (m_deviceExtensions.supports(VK_EXT_CUSTOM_BORDER_COLOR_EXTENSION_NAME)) {
@@ -570,6 +695,11 @@ namespace dxvk {
       m_deviceFeatures.extDepthClipEnable.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extDepthClipEnable);
     }
 
+    if (m_deviceExtensions.supports(VK_EXT_EXTENDED_DYNAMIC_STATE_EXTENSION_NAME)) {
+      m_deviceFeatures.extExtendedDynamicState.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_FEATURES_EXT;
+      m_deviceFeatures.extExtendedDynamicState.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extExtendedDynamicState);
+    }
+
     if (m_deviceExtensions.supports(VK_EXT_HOST_QUERY_RESET_EXTENSION_NAME)) {
       m_deviceFeatures.extHostQueryReset.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES_EXT;
       m_deviceFeatures.extHostQueryReset.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extHostQueryReset);
@@ -580,6 +710,11 @@ namespace dxvk {
       m_deviceFeatures.extMemoryPriority.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extMemoryPriority);
     }
 
+    if (m_deviceExtensions.supports(VK_EXT_NON_SEAMLESS_CUBE_MAP_EXTENSION_NAME)) {
+      m_deviceFeatures.extNonSeamlessCubeMap.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NON_SEAMLESS_CUBE_MAP_FEATURES_EXT;
+      m_deviceFeatures.extNonSeamlessCubeMap.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extNonSeamlessCubeMap);
+    }
+
     if (m_deviceExtensions.supports(VK_EXT_ROBUSTNESS_2_EXTENSION_NAME)) {
       m_deviceFeatures.extRobustness2.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_FEATURES_EXT;
       m_deviceFeatures.extRobustness2.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extRobustness2);
@@ -600,6 +735,16 @@ namespace dxvk {
       m_deviceFeatures.extVertexAttributeDivisor.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.extVertexAttributeDivisor);
     }
 
+    if (m_deviceExtensions.supports(VK_KHR_BUFFER_DEVICE_ADDRESS_EXTENSION_NAME)) {
+      m_deviceFeatures.khrBufferDeviceAddress.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES_KHR;
+      m_deviceFeatures.khrBufferDeviceAddress.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.khrBufferDeviceAddress);
+    }
+
+    if (m_deviceExtensions.supports(VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME)) {
+      m_deviceFeatures.khrTimelineSemaphore.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES_KHR;
+      m_deviceFeatures.khrTimelineSemaphore.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.khrTimelineSemaphore);
+    }
+
     m_vki->vkGetPhysicalDeviceFeatures2(m_handle, &m_deviceFeatures.core);
   }
 
@@ -666,17 +811,22 @@ namespace dxvk {
       "\n  shaderFloat64                          : ", features.core.features.shaderFloat64 ? "1" : "0",
       "\n  shaderInt64                            : ", features.core.features.shaderInt64 ? "1" : "0",
       "\n  variableMultisampleRate                : ", features.core.features.variableMultisampleRate ? "1" : "0",
-      "\n", VK_EXT_CONDITIONAL_RENDERING_EXTENSION_NAME,
-      "\n  conditionalRendering                   : ", features.extConditionalRendering.conditionalRendering ? "1" : "0",
+      "\n", VK_EXT_4444_FORMATS_EXTENSION_NAME,
+      "\n  formatA4R4G4B4                         : ", features.ext4444Formats.formatA4R4G4B4 ? "1" : "0",
+      "\n  formatA4B4G4R4                         : ", features.ext4444Formats.formatA4B4G4R4 ? "1" : "0",
       "\n", VK_EXT_CUSTOM_BORDER_COLOR_EXTENSION_NAME,
       "\n  customBorderColors                     : ", features.extCustomBorderColor.customBorderColors ? "1" : "0",
       "\n  customBorderColorWithoutFormat         : ", features.extCustomBorderColor.customBorderColorWithoutFormat ? "1" : "0",
       "\n", VK_EXT_DEPTH_CLIP_ENABLE_EXTENSION_NAME,
       "\n  depthClipEnable                        : ", features.extDepthClipEnable.depthClipEnable ? "1" : "0",
+      "\n", VK_EXT_EXTENDED_DYNAMIC_STATE_EXTENSION_NAME,
+      "\n  extendedDynamicState                   : ", features.extExtendedDynamicState.extendedDynamicState ? "1" : "0",
       "\n", VK_EXT_HOST_QUERY_RESET_EXTENSION_NAME,
       "\n  hostQueryReset                         : ", features.extHostQueryReset.hostQueryReset ? "1" : "0",
       "\n", VK_EXT_MEMORY_PRIORITY_EXTENSION_NAME,
       "\n  memoryPriority                         : ", features.extMemoryPriority.memoryPriority ? "1" : "0",
+      "\n", VK_EXT_NON_SEAMLESS_CUBE_MAP_EXTENSION_NAME,
+      "\n  nonSeamlessCubeMap                     : ", features.extNonSeamlessCubeMap.nonSeamlessCubeMap ? "1" : "0",
       "\n", VK_EXT_ROBUSTNESS_2_EXTENSION_NAME,
       "\n  robustBufferAccess2                    : ", features.extRobustness2.robustBufferAccess2 ? "1" : "0",
       "\n  robustImageAccess2                     : ", features.extRobustness2.robustImageAccess2 ? "1" : "0",
@@ -688,7 +838,11 @@ namespace dxvk {
       "\n  geometryStreams                        : ", features.extTransformFeedback.geometryStreams ? "1" : "0",
       "\n", VK_EXT_VERTEX_ATTRIBUTE_DIVISOR_EXTENSION_NAME,
       "\n  vertexAttributeInstanceRateDivisor     : ", features.extVertexAttributeDivisor.vertexAttributeInstanceRateDivisor ? "1" : "0",
-      "\n  vertexAttributeInstanceRateZeroDivisor : ", features.extVertexAttributeDivisor.vertexAttributeInstanceRateZeroDivisor ? "1" : "0"));
+      "\n  vertexAttributeInstanceRateZeroDivisor : ", features.extVertexAttributeDivisor.vertexAttributeInstanceRateZeroDivisor ? "1" : "0",
+      "\n", VK_KHR_BUFFER_DEVICE_ADDRESS_EXTENSION_NAME,
+      "\n  bufferDeviceAddress                    : ", features.khrBufferDeviceAddress.bufferDeviceAddress,
+      "\n", VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME,
+      "\n  timelineSemaphore                      : ", features.khrTimelineSemaphore.timelineSemaphore));
   }
 
 
