@@ -6,6 +6,7 @@
 #include "dxvk_context_state.h"
 #include "dxvk_data.h"
 #include "dxvk_objects.h"
+#include "dxvk_resource.h"
 #include "dxvk_util.h"
 
 namespace dxvk {
@@ -18,7 +19,7 @@ namespace dxvk {
    * recorded.
    */
   class DxvkContext : public RcObject {
-    
+    constexpr static VkDeviceSize StagingBufferSize = 4ull << 20;
   public:
     
     DxvkContext(const Rc<DxvkDevice>& device);
@@ -155,6 +156,7 @@ namespace dxvk {
     /**
      * \brief Binds vertex buffer
      * 
+     * When binding a null buffer, stride must be 0.
      * \param [in] binding Vertex buffer binding
      * \param [in] buffer New vertex buffer
      * \param [in] stride Stride between vertices
@@ -163,7 +165,7 @@ namespace dxvk {
             uint32_t              binding,
       const DxvkBufferSlice&      buffer,
             uint32_t              stride);
-    
+
     /**
      * \brief Binds transform feedback buffer
      * 
@@ -239,40 +241,6 @@ namespace dxvk {
             VkDeviceSize          length,
             VkClearColorValue     value);
     
-    /**
-     * \brief Clears subresources of a color image
-     * 
-     * \param [in] image The image to clear
-     * \param [in] value Clear value
-     * \param [in] subresources Subresources to clear
-     */
-    void clearColorImage(
-      const Rc<DxvkImage>&            image,
-      const VkClearColorValue&        value,
-      const VkImageSubresourceRange&  subresources);
-    
-    /**
-     * \brief Clears subresources of a depth-stencil image
-     * 
-     * \param [in] image The image to clear
-     * \param [in] value Clear value
-     * \param [in] subresources Subresources to clear
-     */
-    void clearDepthStencilImage(
-      const Rc<DxvkImage>&            image,
-      const VkClearDepthStencilValue& value,
-      const VkImageSubresourceRange&  subresources);
-    
-    /**
-     * \brief Clears a compressed image to black
-     *
-     * \param [in] image The image to clear
-     * \param [in] subresources Subresources to clear
-     */
-    void clearCompressedColorImage(
-      const Rc<DxvkImage>&            image,
-      const VkImageSubresourceRange&  subresources);
-    
     /**
      * \brief Clears an active render target
      * 
@@ -341,13 +309,15 @@ namespace dxvk {
     /**
      * \brief Copies data from a buffer to an image
      * 
+     * Source data must be packed, except for the row alignment.
      * \param [in] dstImage Destination image
      * \param [in] dstSubresource Destination subresource
      * \param [in] dstOffset Destination area offset
      * \param [in] dstExtent Destination area size
      * \param [in] srcBuffer Source buffer
      * \param [in] srcOffset Source offset, in bytes
-     * \param [in] srcExtent Source data extent
+     * \param [in] rowAlignment Row alignment, in bytes
+     * \param [in] sliceAlignment Slice alignment, in bytes
      */
     void copyBufferToImage(
       const Rc<DxvkImage>&        dstImage,
@@ -356,7 +326,8 @@ namespace dxvk {
             VkExtent3D            dstExtent,
       const Rc<DxvkBuffer>&       srcBuffer,
             VkDeviceSize          srcOffset,
-            VkExtent2D            srcExtent);
+            VkDeviceSize          rowAlignment,
+            VkDeviceSize          sliceAlignment);
     
     /**
      * \brief Copies data from one image to another
@@ -400,6 +371,8 @@ namespace dxvk {
      * \param [in] dstBuffer Destination buffer
      * \param [in] dstOffset Destination offset, in bytes
      * \param [in] dstExtent Destination data extent
+     * \param [in] rowAlignment Row alignment, in bytes
+     * \param [in] sliceAlignment Slice alignment, in bytes
      * \param [in] srcImage Source image
      * \param [in] srcSubresource Source subresource
      * \param [in] srcOffset Source area offset
@@ -408,7 +381,8 @@ namespace dxvk {
     void copyImageToBuffer(
       const Rc<DxvkBuffer>&       dstBuffer,
             VkDeviceSize          dstOffset,
-            VkExtent2D            dstExtent,
+            VkDeviceSize          rowAlignment,
+            VkDeviceSize          sliceAlignment,
       const Rc<DxvkImage>&        srcImage,
             VkImageSubresourceLayers srcSubresource,
             VkOffset3D            srcOffset,
@@ -422,7 +396,9 @@ namespace dxvk {
      * - \c VK_FORMAT_D24_UNORM_S8_UINT: 0xssdddddd
      * - \c VK_FORMAT_D32_SFLOAT_S8_UINT: 0xdddddddd 0x000000ss
      * \param [in] dstBuffer Destination buffer
-     * \param [in] dstOffset Destination offset, in bytes
+     * \param [in] dstBufferOffset Destination offset, in bytes
+     * \param [in] dstOffset Destination image offset
+     * \param [in] dstSize Destination image size
      * \param [in] srcImage Source image
      * \param [in] srcSubresource Source subresource
      * \param [in] srcOffset Source area offset
@@ -431,13 +407,43 @@ namespace dxvk {
      */
     void copyDepthStencilImageToPackedBuffer(
       const Rc<DxvkBuffer>&       dstBuffer,
-            VkDeviceSize          dstOffset,
+            VkDeviceSize          dstBufferOffset,
+            VkOffset2D            dstOffset,
+            VkExtent2D            dstExtent,
       const Rc<DxvkImage>&        srcImage,
             VkImageSubresourceLayers srcSubresource,
             VkOffset2D            srcOffset,
             VkExtent2D            srcExtent,
             VkFormat              format);
     
+    /**
+     * \brief Copies image data stored in a linear buffer to another
+     *
+     * The source and destination regions may overlap, in which case
+     * a temporary copy of the source buffer will be created.
+     * \param [in] dstBuffer Destination buffer
+     * \param [in] dstBufferOffset Destination subresource offset
+     * \param [in] dstOffset Destination image offset
+     * \param [in] dstSize Total size of the destination image
+     * \param [in] srcBuffer Source buffer
+     * \param [in] srcBufferOffset Source subresource offset
+     * \param [in] srcOffset Source image offset
+     * \param [in] srcSize Total size of the source image
+     * \param [in] extent Number of pixels to copy
+     * \param [in] elementSize Pixel size, in bytes
+     */
+    void copyPackedBufferImage(
+      const Rc<DxvkBuffer>&       dstBuffer,
+            VkDeviceSize          dstBufferOffset,
+            VkOffset3D            dstOffset,
+            VkExtent3D            dstSize,
+      const Rc<DxvkBuffer>&       srcBuffer,
+            VkDeviceSize          srcBufferOffset,
+            VkOffset3D            srcOffset,
+            VkExtent3D            srcSize,
+            VkExtent3D            extent,
+            VkDeviceSize          elementSize);
+
     /**
      * \brief Unpacks buffer data to a depth-stencil image
      * 
@@ -449,7 +455,9 @@ namespace dxvk {
      * \param [in] dstOffset Image area offset
      * \param [in] dstExtent Image area size
      * \param [in] srcBuffer Packed data buffer
-     * \param [in] srcOffset Packed data offset
+     * \param [in] srcBufferOffset Buffer offset of source image
+     * \param [in] srcOffset Offset into the source image
+     * \param [in] srcExtent Total size of the source image
      * \param [in] format Packed data format
      */
     void copyPackedBufferToDepthStencilImage(
@@ -458,7 +466,9 @@ namespace dxvk {
             VkOffset2D            dstOffset,
             VkExtent2D            dstExtent,
       const Rc<DxvkBuffer>&       srcBuffer,
-            VkDeviceSize          srcOffset,
+            VkDeviceSize          srcBufferOffset,
+            VkOffset2D            srcOffset,
+            VkExtent2D            srcExtent,
             VkFormat              format);
     
     /**
@@ -474,18 +484,18 @@ namespace dxvk {
       const Rc<DxvkBuffer>&       buffer);
     
     /**
-     * \brief Discards image subresources
+     * \brief Discards contents of an image view
      * 
      * Discards the current contents of the image
      * and performs a fast layout transition. This
      * may improve performance in some cases.
-     * \param [in] image The image to discard
-     * \param [in] subresources Image subresources
+     * \param [in] imageView View to discard
+     * \param [in] discardAspects Image aspects to discard
      */
-    void discardImage(
-      const Rc<DxvkImage>&          image,
-            VkImageSubresourceRange subresources);
-    
+    void discardImageView(
+      const Rc<DxvkImageView>&      imageView,
+            VkImageAspectFlags      discardAspects);
+
     /**
      * \brief Starts compute jobs
      * 
@@ -624,15 +634,28 @@ namespace dxvk {
      * Uses blitting to generate lower mip levels from
      * the top-most mip level passed to this method.
      * \param [in] imageView The image to generate mips for
+     * \param [in] filter The filter to use for generation
      */
     void generateMipmaps(
-      const Rc<DxvkImageView>&        imageView);
-    
+      const Rc<DxvkImageView>&        imageView,
+            VkFilter                  filter);
+
     /**
-     * \brief Initializes or invalidates an image
+     * \brief Initializes a buffer
+     *
+     * Clears the given buffer to zero. Only safe to call
+     * if the buffer is not currently in use by the GPU.
+     * \param [in] buffer Buffer to clear
+     */
+    void initBuffer(
+      const Rc<DxvkBuffer>&           buffer);
+
+    /**
+     * \brief Initializes an image
      * 
-     * Sets up the image layout for future operations
-     * while discarding any previous contents.
+     * Transitions the image into its default layout, and clears
+     * it to black unless the initial layout is preinitialized.
+     * Only safe to call if the image is not in use by the GPU.
      * \param [in] image The image to initialize
      * \param [in] subresources Image subresources
      * \param [in] initialLayout Initial image layout
@@ -737,27 +760,6 @@ namespace dxvk {
             VkDeviceSize              size,
       const void*                     data);
     
-    /**
-     * \brief Updates an image
-     * 
-     * Copies data from the host into an image.
-     * \param [in] image Destination image
-     * \param [in] subsresources Image subresources to update
-     * \param [in] imageOffset Offset of the image area to update
-     * \param [in] imageExtent Size of the image area to update
-     * \param [in] data Source data
-     * \param [in] pitchPerRow Row pitch of the source data
-     * \param [in] pitchPerLayer Layer pitch of the source data
-     */
-    void updateImage(
-      const Rc<DxvkImage>&            image,
-      const VkImageSubresourceLayers& subresources,
-            VkOffset3D                imageOffset,
-            VkExtent3D                imageExtent,
-      const void*                     data,
-            VkDeviceSize              pitchPerRow,
-            VkDeviceSize              pitchPerLayer);
-    
     /**
      * \brief Updates an depth-stencil image
      * 
@@ -934,21 +936,6 @@ namespace dxvk {
             uint32_t            index,
             uint32_t            value);
     
-    /**
-     * \brief Sets predicate
-     *
-     * Enables or disables conditional rendering,
-     * depending on whether the given buffer slice
-     * is defined or not. Draw calls and render
-     * target clear commands will get discarded if
-     * the predicate value is either zero or non-zero.
-     * \param [in] predicate The predicate buffer
-     * \param [in] flags Conditional rendering mode
-     */
-    void setPredicate(
-      const DxvkBufferSlice&    predicate,
-            VkConditionalRenderingFlagsEXT flags);
-    
     /**
      * \brief Sets barrier control flags
      *
@@ -959,6 +946,23 @@ namespace dxvk {
     void setBarrierControl(
             DxvkBarrierControlFlags control);
     
+    /**
+     * \brief Launches a Cuda kernel
+     *
+     * Since the kernel is launched with an opaque set of
+     * kernel-specific parameters which may reference
+     * resources bindlessly, such resources must be listed by
+     * the caller in the 'buffers' and 'images' parameters so
+     * that their access may be tracked appropriately.
+     * \param [in] nvxLaunchInfo Kernel launch parameter struct
+     * \param [in] buffers List of {buffer,read,write} used by kernel
+     * \param [in] images List of {image,read,write} used by kernel
+     */
+    void launchCuKernelNVX(
+      const VkCuLaunchInfoNVX& nvxLaunchInfo,
+      const std::vector<std::pair<Rc<DxvkBuffer>, DxvkAccessFlags>>& buffers,
+      const std::vector<std::pair<Rc<DxvkImage>, DxvkAccessFlags>>& images);
+    
     /**
      * \brief Signals a GPU event
      * \param [in] event The event
@@ -966,18 +970,6 @@ namespace dxvk {
     void signalGpuEvent(
       const Rc<DxvkGpuEvent>&   event);
     
-    /**
-     * \brief Copies query data to predicate buffer
-     * 
-     * The given buffer slice can then be passed
-     * to \c setPredicate to enable predication.
-     * \param [in] predicate Predicate buffer
-     * \param [in] query Source query
-     */
-    void writePredicate(
-      const DxvkBufferSlice&    predicate,
-      const Rc<DxvkGpuQuery>&   query);
-    
     /**
      * \brief Writes to a timestamp query
      * \param [in] query The timestamp query
@@ -997,16 +989,66 @@ namespace dxvk {
     void signal(
       const Rc<sync::Signal>&   signal,
             uint64_t            value);
-    
+
     /**
-     * \brief Trims staging buffers
-     * 
-     * Releases staging buffer resources. Calling
-     * this may be useful if data updates on a
-     * given context are rare.
+     * \brief Waits for fence
+     *
+     * Stalls current command list execution until
+     * the fence reaches the given value or higher.
+     * \param [in] fence Fence to wait on
+     * \param [in] value Value to wait on
      */
-    void trimStagingBuffers();
-    
+    void waitFence(const Rc<DxvkFence>& fence, uint64_t value);
+
+    /**
+     * \brief Signals fence
+     *
+     * Signals fence to the given value once the current
+     * command list execution completes on the GPU.
+     * \param [in] fence Fence to signal
+     * \param [in] value Value to signal
+     */
+    void signalFence(const Rc<DxvkFence>& fence, uint64_t value);
+
+    /**
+     * \brief Begins a debug label region
+     * \param [in] label The debug label
+     *
+     * Marks the start of a debug label region. Used by debugging/profiling
+     * tools to mark different workloads within a frame.
+     */
+    void beginDebugLabel(VkDebugUtilsLabelEXT *label);
+
+    /**
+     * \brief Ends a debug label region
+     *
+     * Marks the close of a debug label region. Used by debugging/profiling
+     * tools to mark different workloads within a frame.
+     */
+    void endDebugLabel();
+
+    /**
+     * \brief Inserts a debug label
+     * \param [in] label The debug label
+     *
+     * Inserts an instantaneous debug label. Used by debugging/profiling
+     * tools to mark different workloads within a frame.
+     */
+    void insertDebugLabel(VkDebugUtilsLabelEXT *label);
+
+    /**
+     * \brief Increments a given stat counter
+     *
+     * The stat counters will be merged into the global
+     * stat counters upon execution of the command list.
+     * \param [in] counter Stat counter to increment
+     * \param [in] value Increment value
+     */
+    void addStatCtr(DxvkStatCounter counter, uint64_t value) {
+      if (m_cmd != nullptr)
+        m_cmd->addStatCtr(counter, value);
+    }
+
   private:
     
     Rc<DxvkDevice>          m_device;
@@ -1029,8 +1071,10 @@ namespace dxvk {
     DxvkBarrierControlFlags m_barrierControl;
     
     DxvkGpuQueryManager     m_queryManager;
-    DxvkStagingDataAlloc    m_staging;
+    DxvkStagingBuffer       m_staging;
     
+    DxvkRenderTargetLayouts m_rtLayouts = { };
+
     VkPipeline m_gpActivePipeline = VK_NULL_HANDLE;
     VkPipeline m_cpActivePipeline = VK_NULL_HANDLE;
 
@@ -1045,12 +1089,10 @@ namespace dxvk {
     std::array<DxvkShaderResourceSlot, MaxNumResourceSlots>  m_rc;
     std::array<DxvkGraphicsPipeline*, 4096> m_gpLookupCache = { };
     std::array<DxvkComputePipeline*,   256> m_cpLookupCache = { };
+    std::array<Rc<DxvkFramebuffer>,    512> m_framebufferCache = { };
+
+    bool m_hackRenderAreaTrimmedToViewport = false;
 
-    std::unordered_map<
-      DxvkBufferSliceHandle,
-      DxvkGpuQueryHandle,
-      DxvkHash, DxvkEq>     m_predicateWrites;
-    
     void blitImageFb(
       const Rc<DxvkImage>&        dstImage,
       const Rc<DxvkImage>&        srcImage,
@@ -1064,6 +1106,28 @@ namespace dxvk {
       const VkImageBlit&          region,
             VkFilter              filter);
 
+    template<bool ToImage>
+    void copyImageBufferData(
+            DxvkCmdBuffer         cmd,
+      const Rc<DxvkImage>&        image,
+      const VkImageSubresourceLayers& imageSubresource,
+            VkOffset3D            imageOffset,
+            VkExtent3D            imageExtent,
+            VkImageLayout         imageLayout,
+      const DxvkBufferSliceHandle& bufferSlice,
+            VkDeviceSize          bufferRowAlignment,
+            VkDeviceSize          bufferSliceAlignment);
+
+    void copyImageHostData(
+            DxvkCmdBuffer         cmd,
+      const Rc<DxvkImage>&        image,
+      const VkImageSubresourceLayers& imageSubresource,
+            VkOffset3D            imageOffset,
+            VkExtent3D            imageExtent,
+      const void*                 hostData,
+            VkDeviceSize          rowPitch,
+            VkDeviceSize          slicePitch);
+
     void clearImageViewFb(
       const Rc<DxvkImageView>&    imageView,
             VkOffset3D            offset,
@@ -1095,6 +1159,14 @@ namespace dxvk {
             VkOffset3D            srcOffset,
             VkExtent3D            extent);
     
+    bool copyImageClear(
+      const Rc<DxvkImage>&        dstImage,
+            VkImageSubresourceLayers dstSubresource,
+            VkOffset3D            dstOffset,
+            VkExtent3D            dstExtent,
+      const Rc<DxvkImage>&        srcImage,
+            VkImageSubresourceLayers srcSubresource);
+
     void resolveImageHw(
       const Rc<DxvkImage>&            dstImage,
       const Rc<DxvkImage>&            srcImage,
@@ -1118,6 +1190,7 @@ namespace dxvk {
     void performClear(
       const Rc<DxvkImageView>&        imageView,
             int32_t                   attachmentIndex,
+            VkImageAspectFlags        discardAspects,
             VkImageAspectFlags        clearAspects,
             VkClearValue              clearValue);
 
@@ -1126,20 +1199,20 @@ namespace dxvk {
             VkImageAspectFlags        clearAspects,
             VkClearValue              clearValue);
 
+    void deferDiscard(
+      const Rc<DxvkImageView>&        imageView,
+            VkImageAspectFlags        discardAspects);
+
     void flushClears(
             bool                      useRenderPass);
 
-    void updatePredicate(
-      const DxvkBufferSliceHandle&    predicate,
-      const DxvkGpuQueryHandle&       query);
-    
-    void commitPredicateUpdates();
-    
+    void flushSharedImages();
+
     void startRenderPass();
-    void spillRenderPass(bool flushClears = true);
+    void spillRenderPass(bool suspend);
     
     void renderPassBindFramebuffer(
-      const Rc<DxvkFramebuffer>&  framebuffer,
+      const DxvkFramebufferInfo&  framebufferInfo,
       const DxvkRenderPassOps&    ops,
             uint32_t              clearValueCount,
       const VkClearValue*         clearValues);
@@ -1150,9 +1223,6 @@ namespace dxvk {
       const DxvkRenderTargets&    renderTargets,
             DxvkRenderPassOps&    renderPassOps);
 
-    void startConditionalRendering();
-    void pauseConditionalRendering();
-    
     void startTransformFeedback();
     void pauseTransformFeedback();
     
@@ -1176,16 +1246,45 @@ namespace dxvk {
             VkDescriptorSet         set,
       const DxvkPipelineLayout*     layout);
 
+    DxvkFramebufferInfo makeFramebufferInfo(
+      const DxvkRenderTargets&      renderTargets);
+
     void updateFramebuffer();
     
+    void applyRenderTargetLoadLayouts();
+
+    void applyRenderTargetStoreLayouts();
+
+    void transitionRenderTargetLayouts(
+            DxvkBarrierSet&         barriers,
+            bool                    sharedOnly);
+
+    void transitionColorAttachment(
+            DxvkBarrierSet&         barriers,
+      const DxvkAttachment&         attachment,
+            VkImageLayout           oldLayout);
+
+    void transitionDepthAttachment(
+            DxvkBarrierSet&         barriers,
+      const DxvkAttachment&         attachment,
+            VkImageLayout           oldLayout);
+
+    void updateRenderTargetLayouts(
+      const DxvkFramebufferInfo&    newFb,
+      const DxvkFramebufferInfo&    oldFb);
+
+    void prepareImage(
+            DxvkBarrierSet&         barriers,
+      const Rc<DxvkImage>&          image,
+      const VkImageSubresourceRange& subresources,
+            bool                    flushClears = true);
+
     bool updateIndexBufferBinding();
     void updateVertexBufferBindings();
 
     void updateTransformFeedbackBuffers();
     void updateTransformFeedbackState();
 
-    void updateConditionalRendering();
-    
     void updateDynamicState();
 
     template<VkPipelineBindPoint BindPoint>
@@ -1214,8 +1313,6 @@ namespace dxvk {
             VkPipelineStageFlags      stages,
             VkAccessFlags             access);
 
-    DxvkAccessFlags checkFramebufferBarrier();
-
     void emitMemoryBarrier(
             VkDependencyFlags         flags,
             VkPipelineStageFlags      srcStages,
@@ -1223,27 +1320,27 @@ namespace dxvk {
             VkPipelineStageFlags      dstStages,
             VkAccessFlags             dstAccess);
     
-    void initializeImage(
-      const Rc<DxvkImage>&            image,
-      const VkImageSubresourceRange&  subresources,
-            VkImageLayout             dstLayout,
-            VkPipelineStageFlags      dstStages,
-            VkAccessFlags             dstAccess);
-
     VkDescriptorSet allocateDescriptorSet(
             VkDescriptorSetLayout     layout);
 
     void trackDrawBuffer();
 
+    bool tryInvalidateDeviceLocalBuffer(
+      const Rc<DxvkBuffer>&           buffer,
+            VkDeviceSize              copySize);
+
     DxvkGraphicsPipeline* lookupGraphicsPipeline(
       const DxvkGraphicsPipelineShaders&  shaders);
 
     DxvkComputePipeline* lookupComputePipeline(
       const DxvkComputePipelineShaders&   shaders);
+    
+    Rc<DxvkFramebuffer> lookupFramebuffer(
+      const DxvkFramebufferInfo&      framebufferInfo);
 
     Rc<DxvkBuffer> createZeroBuffer(
             VkDeviceSize              size);
 
   };
   
-}
\ No newline at end of file
+}
