@@ -40,13 +40,36 @@ namespace dxvk {
     }
 
     template<typename T>
-    bool read(T& data) {
-      if (m_read + sizeof(T) > m_size)
-        return false;
-      
-      std::memcpy(&data, &m_data[m_read], sizeof(T));
-      m_read += sizeof(T);
-      return true;
+    bool read(T& data, uint32_t version) {
+      return read(data);
+    }
+
+    bool read(DxvkBindingMask& data, uint32_t version) {
+      if (version < 9) {
+        DxvkBindingMaskV8 v8;
+
+        if (!read(v8))
+          return false;
+
+        data = v8.convert();
+        return true;
+      }
+
+      return read(data);
+    }
+
+    bool read(DxvkIlBinding& data, uint32_t version) {
+      if (version < 10) {
+        DxvkIlBindingV9 v9;
+
+        if (!read(v9))
+          return false;
+
+        data = v9.convert();
+        return true;
+      }
+
+      return read(data);
     }
 
     template<typename T>
@@ -77,6 +100,16 @@ namespace dxvk {
     size_t m_read = 0;
     char   m_data[MaxSize];
 
+    template<typename T>
+    bool read(T& data) {
+      if (m_read + sizeof(T) > m_size)
+        return false;
+
+      std::memcpy(&data, &m_data[m_read], sizeof(T));
+      m_read += sizeof(T);
+      return true;
+    }
+
   };
 
 
@@ -117,23 +150,24 @@ namespace dxvk {
 
 
   DxvkStateCache::DxvkStateCache(
-    const DxvkDevice*           device,
+          DxvkDevice*           device,
           DxvkPipelineManager*  pipeManager,
           DxvkRenderPassPool*   passManager)
-  : m_pipeManager(pipeManager),
-    m_passManager(passManager) {
+  : m_device      (device),
+    m_pipeManager (pipeManager),
+    m_passManager (passManager) {
     bool newFile = !readCacheFile();
 
     if (newFile) {
       Logger::warn("DXVK: Creating new state cache file");
 
       // Start with an empty file
-      std::ofstream file(getCacheFileName(),
+      std::ofstream file(getCacheFileName().c_str(),
         std::ios_base::binary |
         std::ios_base::trunc);
 
       if (!file && env::createDirectory(getCacheDir())) {
-        file = std::ofstream(getCacheFileName(),
+        file = std::ofstream(getCacheFileName().c_str(),
           std::ios_base::binary |
           std::ios_base::trunc);
       }
@@ -151,45 +185,11 @@ namespace dxvk {
       for (auto& e : m_entries)
         writeCacheEntry(file, e);
     }
-
-    // Use half the available CPU cores for pipeline compilation
-    uint32_t numCpuCores = dxvk::thread::hardware_concurrency();
-    uint32_t numWorkers  = ((std::max(1u, numCpuCores) - 1) * 5) / 7;
-
-    if (numWorkers <  1) numWorkers =  1;
-    if (numWorkers > 32) numWorkers = 32;
-
-    if (device->config().numCompilerThreads > 0)
-      numWorkers = device->config().numCompilerThreads;
-    
-    Logger::info(str::format("DXVK: Using ", numWorkers, " compiler threads"));
-    
-    // Start the worker threads and the file writer
-    m_workerBusy.store(numWorkers);
-
-    for (uint32_t i = 0; i < numWorkers; i++) {
-      m_workerThreads.emplace_back([this] () { workerFunc(); });
-      m_workerThreads[i].set_priority(ThreadPriority::Lowest);
-    }
-    
-    m_writerThread = dxvk::thread([this] () { writerFunc(); });
   }
   
 
   DxvkStateCache::~DxvkStateCache() {
-    { std::lock_guard<std::mutex> workerLock(m_workerLock);
-      std::lock_guard<std::mutex> writerLock(m_writerLock);
-
-      m_stopThreads.store(true);
-
-      m_workerCond.notify_all();
-      m_writerCond.notify_all();
-    }
-
-    for (auto& worker : m_workerThreads)
-      worker.join();
-    
-    m_writerThread.join();
+    this->stopWorkerThreads();
   }
 
 
@@ -211,12 +211,14 @@ namespace dxvk {
     }
 
     // Queue a job to write this pipeline to the cache
-    std::unique_lock<std::mutex> lock(m_writerLock);
+    std::unique_lock<dxvk::mutex> lock(m_writerLock);
 
     m_writerQueue.push({ shaders, state,
       DxvkComputePipelineStateInfo(),
       format, g_nullHash });
     m_writerCond.notify_one();
+
+    createWriter();
   }
 
 
@@ -235,12 +237,14 @@ namespace dxvk {
     }
 
     // Queue a job to write this pipeline to the cache
-    std::unique_lock<std::mutex> lock(m_writerLock);
+    std::unique_lock<dxvk::mutex> lock(m_writerLock);
 
     m_writerQueue.push({ shaders,
       DxvkGraphicsPipelineStateInfo(), state,
       DxvkRenderPassFormat(), g_nullHash });
     m_writerCond.notify_one();
+
+    createWriter();
   }
 
 
@@ -251,11 +255,11 @@ namespace dxvk {
       return;
     
     // Add the shader so we can look it up by its key
-    std::unique_lock<std::mutex> entryLock(m_entryLock);
+    std::unique_lock<dxvk::mutex> entryLock(m_entryLock);
     m_shaderMap.insert({ key, shader });
 
     // Deferred lock, don't stall workers unless we have to
-    std::unique_lock<std::mutex> workerLock;
+    std::unique_lock<dxvk::mutex> workerLock;
 
     auto pipelines = m_pipelineMap.equal_range(key);
 
@@ -271,13 +275,34 @@ namespace dxvk {
         continue;
       
       if (!workerLock)
-        workerLock = std::unique_lock<std::mutex>(m_workerLock);
+        workerLock = std::unique_lock<dxvk::mutex>(m_workerLock);
       
       m_workerQueue.push(item);
     }
 
-    if (workerLock)
+    if (workerLock) {
       m_workerCond.notify_all();
+      createWorkers();
+    }
+  }
+
+
+  void DxvkStateCache::stopWorkerThreads() {
+    { std::lock_guard<dxvk::mutex> workerLock(m_workerLock);
+      std::lock_guard<dxvk::mutex> writerLock(m_writerLock);
+
+      if (m_stopThreads.exchange(true))
+        return;
+
+      m_workerCond.notify_all();
+      m_writerCond.notify_all();
+    }
+
+    for (auto& worker : m_workerThreads)
+      worker.join();
+    
+    if (m_writerThread.joinable())
+      m_writerThread.join();
   }
 
 
@@ -332,8 +357,10 @@ namespace dxvk {
       for (auto e = entries.first; e != entries.second; e++) {
         const auto& entry = m_entries[e->second];
 
-        auto rp = m_passManager->getRenderPass(entry.format);
-        pipeline->compilePipeline(entry.gpState, rp);
+        if (m_passManager->validateRenderPassFormat(entry.format)) {
+          auto rp = m_passManager->getRenderPass(entry.format);
+          pipeline->compilePipeline(entry.gpState, rp);
+        }
       }
     } else {
       auto pipeline = m_pipeManager->createComputePipeline(item.cp);
@@ -347,9 +374,9 @@ namespace dxvk {
   }
 
 
-  bool DxvkStateCache::realReadCacheFile(std::string name) {
+  bool DxvkStateCache::realReadCacheFile(std::wstring name) {
     // Open state file and just fail if it doesn't exist
-    std::ifstream ifile(name, std::ios_base::binary);
+    std::ifstream ifile(name.c_str(), std::ios_base::binary);
 
     if (!ifile) {
       Logger::warn("DXVK: No state cache file found");
@@ -524,13 +551,13 @@ namespace dxvk {
 
     for (uint32_t i = 0; i < 6; i++) {
       if (stageMask & VkShaderStageFlagBits(1 << i))
-        data.read(keys[i]);
+        data.read(keys[i], version);
       else
         keys[i] = g_nullShaderKey;
     }
 
     if (stageMask & VK_SHADER_STAGE_COMPUTE_BIT) {
-      if (!data.read(entry.cpState.bsBindingMask))
+      if (!data.read(entry.cpState.bsBindingMask, version))
         return false;
     } else {
       // Read packed render pass format
@@ -538,9 +565,9 @@ namespace dxvk {
       uint8_t imageFormat = 0;
       uint8_t imageLayout = 0;
 
-      if (!data.read(sampleCount)
-       || !data.read(imageFormat)
-       || !data.read(imageLayout))
+      if (!data.read(sampleCount, version)
+       || !data.read(imageFormat, version)
+       || !data.read(imageLayout, version))
         return false;
 
       entry.format.sampleCount = VkSampleCountFlagBits(sampleCount);
@@ -548,8 +575,8 @@ namespace dxvk {
       entry.format.depth.layout = unpackImageLayout(imageLayout);
 
       for (uint32_t i = 0; i < MaxNumRenderTargets; i++) {
-        if (!data.read(imageFormat)
-         || !data.read(imageLayout))
+        if (!data.read(imageFormat, version)
+         || !data.read(imageLayout, version))
           return false;
 
         entry.format.color[i].format = VkFormat(imageFormat);
@@ -560,15 +587,15 @@ namespace dxvk {
         return false;
 
       // Read common pipeline state
-      if (!data.read(entry.gpState.bsBindingMask)
-       || !data.read(entry.gpState.ia)
-       || !data.read(entry.gpState.il)
-       || !data.read(entry.gpState.rs)
-       || !data.read(entry.gpState.ms)
-       || !data.read(entry.gpState.ds)
-       || !data.read(entry.gpState.om)
-       || !data.read(entry.gpState.dsFront)
-       || !data.read(entry.gpState.dsBack))
+      if (!data.read(entry.gpState.bsBindingMask, version)
+       || !data.read(entry.gpState.ia, version)
+       || !data.read(entry.gpState.il, version)
+       || !data.read(entry.gpState.rs, version)
+       || !data.read(entry.gpState.ms, version)
+       || !data.read(entry.gpState.ds, version)
+       || !data.read(entry.gpState.om, version)
+       || !data.read(entry.gpState.dsFront, version)
+       || !data.read(entry.gpState.dsBack, version))
         return false;
 
       if (entry.gpState.il.attributeCount() > MaxNumVertexAttributes
@@ -577,25 +604,25 @@ namespace dxvk {
 
       // Read render target swizzles
       for (uint32_t i = 0; i < MaxNumRenderTargets; i++) {
-        if (!data.read(entry.gpState.omSwizzle[i]))
+        if (!data.read(entry.gpState.omSwizzle[i], version))
           return false;
       }
 
       // Read render target blend info
       for (uint32_t i = 0; i < MaxNumRenderTargets; i++) {
-        if (!data.read(entry.gpState.omBlend[i]))
+        if (!data.read(entry.gpState.omBlend[i], version))
           return false;
       }
 
       // Read defined vertex attributes
       for (uint32_t i = 0; i < entry.gpState.il.attributeCount(); i++) {
-        if (!data.read(entry.gpState.ilAttributes[i]))
+        if (!data.read(entry.gpState.ilAttributes[i], version))
           return false;
       }
 
       // Read defined vertex bindings
       for (uint32_t i = 0; i < entry.gpState.il.bindingCount(); i++) {
-        if (!data.read(entry.gpState.ilBindings[i]))
+        if (!data.read(entry.gpState.ilBindings[i], version))
           return false;
       }
     }
@@ -607,12 +634,12 @@ namespace dxvk {
 
     uint32_t specConstantMask = 0;
 
-    if (!data.read(specConstantMask))
+    if (!data.read(specConstantMask, version))
       return false;
 
     for (uint32_t i = 0; i < MaxNumSpecConstants; i++) {
       if (specConstantMask & (1 << i)) {
-        if (!data.read(sc.specConstants[i]))
+        if (!data.read(sc.specConstants[i], version))
           return false;
       }
     }
@@ -800,7 +827,7 @@ namespace dxvk {
 
     if (in.shaders.cs.eq(g_nullShaderKey)) {
       // Binding mask
-      out.gpState.bsBindingMask = in.gpState.bsBindingMask;
+      out.gpState.bsBindingMask = in.gpState.bsBindingMask.convert();
 
       // Graphics state
       out.gpState.ia = DxvkIaInfo(
@@ -835,7 +862,8 @@ namespace dxvk {
         in.gpState.rsCullMode,
         in.gpState.rsFrontFace,
         in.gpState.rsViewportCount,
-        in.gpState.rsSampleCount);
+        in.gpState.rsSampleCount,
+        VK_CONSERVATIVE_RASTERIZATION_MODE_DISABLED_EXT);
 
       out.gpState.ms = DxvkMsInfo(
         in.gpState.msSampleCount,
@@ -876,7 +904,7 @@ namespace dxvk {
         out.cpState.sc.specConstants[i] = in.cpState.scSpecConstants[i];
     } else {
       // Binding mask
-      out.cpState.bsBindingMask = in.cpState.bsBindingMask;
+      out.cpState.bsBindingMask = in.cpState.bsBindingMask.convert();
 
       for (uint32_t i = 0; i < 8 && i < MaxNumSpecConstants; i++)
         out.gpState.sc.specConstants[i] = in.gpState.scSpecConstants[i];
@@ -892,7 +920,7 @@ namespace dxvk {
     while (!m_stopThreads.load()) {
       WorkerItem item;
 
-      { std::unique_lock<std::mutex> lock(m_workerLock);
+      { std::unique_lock<dxvk::mutex> lock(m_workerLock);
 
         if (m_workerQueue.empty()) {
           m_workerBusy -= 1;
@@ -925,7 +953,7 @@ namespace dxvk {
     while (!m_stopThreads.load()) {
       DxvkStateCacheEntry entry;
 
-      { std::unique_lock<std::mutex> lock(m_writerLock);
+      { std::unique_lock<dxvk::mutex> lock(m_writerLock);
 
         m_writerCond.wait(lock, [this] () {
           return m_writerQueue.size()
@@ -939,8 +967,8 @@ namespace dxvk {
         m_writerQueue.pop();
       }
 
-      if (!file) {
-        file = std::ofstream(getCacheFileName(),
+      if (!file.is_open()) {
+        file.open(getCacheFileName().c_str(),
           std::ios_base::binary |
           std::ios_base::app);
       }
@@ -950,24 +978,50 @@ namespace dxvk {
   }
 
 
-  std::string DxvkStateCache::getCacheFileName() const {
+  void DxvkStateCache::createWorkers() {
+    if (m_workerThreads.empty()) {
+      // Use half the available CPU cores for pipeline compilation
+      uint32_t numCpuCores = dxvk::thread::hardware_concurrency();
+      uint32_t numWorkers  = ((std::max(1u, numCpuCores) - 1) * 5) / 7;
+
+      if (numWorkers <  1) numWorkers =  1;
+      if (numWorkers > 32) numWorkers = 32;
+
+      if (m_device->config().numCompilerThreads > 0)
+        numWorkers = m_device->config().numCompilerThreads;
+
+      Logger::info(str::format("DXVK: Using ", numWorkers, " compiler threads"));
+
+      // Start the worker threads and the file writer
+      m_workerBusy.store(numWorkers);
+
+      for (uint32_t i = 0; i < numWorkers; i++) {
+        m_workerThreads.emplace_back([this] () { workerFunc(); });
+        m_workerThreads[i].set_priority(ThreadPriority::Lowest);
+      }
+    }
+  }
+
+
+  void DxvkStateCache::createWriter() {
+    if (!m_writerThread.joinable())
+      m_writerThread = dxvk::thread([this] () { writerFunc(); });
+  }
+
+
+  std::wstring DxvkStateCache::getCacheFileName() const {
     std::string path = getCacheDir();
 
     if (!path.empty() && *path.rbegin() != '/')
       path += '/';
     
-    std::string exeName = env::getExeName();
-    auto extp = exeName.find_last_of('.');
-    
-    if (extp != std::string::npos && exeName.substr(extp + 1) == "exe")
-      exeName.erase(extp);
-    
+    std::string exeName = env::getExeBaseName();
     path += exeName + ".dxvk-cache";
-    return path;
+    return str::tows(path.c_str());
   }
 
 
-  std::string DxvkStateCache::getBaseCacheFileName() const {
+  std::wstring DxvkStateCache::getBaseCacheFileName() const {
     std::string path = getCacheDir();
 
     if (!path.empty() && *path.rbegin() != '/')
@@ -980,7 +1034,7 @@ namespace dxvk {
       exeName.erase(extp);
     
     path += exeName + ".dxvk-cache-base";
-    return path;
+    return str::tows(path.c_str());
   }
 
 
