@@ -837,6 +837,16 @@ static INCL_FILE *add_src_file( const char *name )
     return file;
 }
 
+static void reset_owner( INCL_FILE *pFile )
+{
+    int i;
+
+    if (!pFile->owner) return;
+    pFile->owner = NULL;
+
+    for (i = 0; i < MAX_INCLUDES; i++)
+        if (pFile->files[i]) reset_owner( pFile->files[i] );
+}
 
 /*******************************************************************
  *         output_include
@@ -1006,12 +1016,25 @@ static void output_dependencies(void)
     }
     LIST_FOR_EACH_ENTRY( pFile, &sources, INCL_FILE, entry )
     {
-        column = 0;
-        if (!output_src( file, pFile, &column )) continue;
-        for (i = 0; i < MAX_INCLUDES; i++)
-            if (pFile->files[i]) output_include( file, pFile->files[i],
-                                                 pFile, &column );
-        fprintf( file, "\n" );
+        static const char *dirs[] = {"", "win32", "win64", "wow64", NULL};
+        char *name = pFile->name;
+        const char **dir;
+        for (dir = dirs; *dir; dir++)
+        {
+            if (**dir) pFile->name = strmake( "%s/%s", *dir, name );
+
+            column = 0;
+            if (!output_src( file, pFile, &column )) continue;
+            for (i = 0; i < MAX_INCLUDES; i++)
+                if (pFile->files[i]) output_include( file, pFile->files[i],
+                                                     pFile, &column );
+            fprintf( file, "\n" );
+
+            for (i = 0; i < MAX_INCLUDES; i++)
+                if (pFile->files[i]) reset_owner( pFile->files[i] );
+            if (**dir) free( pFile->name );
+            pFile->name = name;
+        }
     }
     fclose( file );
 
@@ -1110,8 +1133,9 @@ int main( int argc, char *argv[] )
         if (path->name[0] != '/') continue;
         if (top_src_dir)
         {
-            if (!strncmp( path->name, top_src_dir, strlen(top_src_dir) ) &&
-                path->name[strlen(top_src_dir)] == '/') continue;
+            int len = (strrchr( top_src_dir, '/' ) - top_src_dir) / sizeof(*top_src_dir);
+            if (!strncmp( path->name, top_src_dir, len ) &&
+                path->name[len] == '/') continue;
         }
         list_remove( &path->entry );
         free( path );
