@@ -1,7 +1,7 @@
 /*
  * MVKCmdPipeline.mm
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -23,13 +23,62 @@
 #include "MVKBuffer.h"
 #include "MVKPipeline.h"
 #include "MVKFoundation.h"
-#include "MVKEnvironment.h"
 #include "mvk_datatypes.hpp"
 
 
+#pragma mark -
+#pragma mark MVKCmdExecuteCommands
+
+template <size_t N>
+VkResult MVKCmdExecuteCommands<N>::setContent(MVKCommandBuffer* cmdBuff,
+											  uint32_t commandBuffersCount,
+											  const VkCommandBuffer* pCommandBuffers) {
+	// Add clear values
+	_secondaryCommandBuffers.clear();	// Clear for reuse
+	_secondaryCommandBuffers.reserve(commandBuffersCount);
+	for (uint32_t cbIdx = 0; cbIdx < commandBuffersCount; cbIdx++) {
+		_secondaryCommandBuffers.push_back(MVKCommandBuffer::getMVKCommandBuffer(pCommandBuffers[cbIdx]));
+	}
+	cmdBuff->recordExecuteCommands(_secondaryCommandBuffers.contents());
+
+	return VK_SUCCESS;
+}
+
+template <size_t N>
+void MVKCmdExecuteCommands<N>::encode(MVKCommandEncoder* cmdEncoder) {
+    for (auto& cb : _secondaryCommandBuffers) { cmdEncoder->encodeSecondary(cb); }
+}
+
+template class MVKCmdExecuteCommands<1>;
+template class MVKCmdExecuteCommands<16>;
+
+
 #pragma mark -
 #pragma mark MVKCmdPipelineBarrier
 
+template <size_t N>
+VkResult MVKCmdPipelineBarrier<N>::setContent(MVKCommandBuffer* cmdBuff,
+											  const VkDependencyInfo* pDependencyInfo) {
+	_dependencyFlags = pDependencyInfo->dependencyFlags;
+
+	_barriers.clear();	// Clear for reuse
+	_barriers.reserve(pDependencyInfo->memoryBarrierCount + 
+					  pDependencyInfo->bufferMemoryBarrierCount +
+					  pDependencyInfo->imageMemoryBarrierCount);
+
+	for (uint32_t i = 0; i < pDependencyInfo->memoryBarrierCount; i++) {
+		_barriers.emplace_back(pDependencyInfo->pMemoryBarriers[i]);
+	}
+	for (uint32_t i = 0; i < pDependencyInfo->bufferMemoryBarrierCount; i++) {
+		_barriers.emplace_back(pDependencyInfo->pBufferMemoryBarriers[i]);
+	}
+	for (uint32_t i = 0; i < pDependencyInfo->imageMemoryBarrierCount; i++) {
+		_barriers.emplace_back(pDependencyInfo->pImageMemoryBarriers[i]);
+	}
+
+	return VK_SUCCESS;
+}
+
 template <size_t N>
 VkResult MVKCmdPipelineBarrier<N>::setContent(MVKCommandBuffer* cmdBuff,
 											  VkPipelineStageFlags srcStageMask,
@@ -41,21 +90,19 @@ VkResult MVKCmdPipelineBarrier<N>::setContent(MVKCommandBuffer* cmdBuff,
 											  const VkBufferMemoryBarrier* pBufferMemoryBarriers,
 											  uint32_t imageMemoryBarrierCount,
 											  const VkImageMemoryBarrier* pImageMemoryBarriers) {
-	_srcStageMask = srcStageMask;
-	_dstStageMask = dstStageMask;
 	_dependencyFlags = dependencyFlags;
 
 	_barriers.clear();	// Clear for reuse
 	_barriers.reserve(memoryBarrierCount + bufferMemoryBarrierCount + imageMemoryBarrierCount);
 
 	for (uint32_t i = 0; i < memoryBarrierCount; i++) {
-		_barriers.emplace_back(pMemoryBarriers[i]);
+		_barriers.emplace_back(pMemoryBarriers[i], srcStageMask, dstStageMask);
 	}
 	for (uint32_t i = 0; i < bufferMemoryBarrierCount; i++) {
-		_barriers.emplace_back(pBufferMemoryBarriers[i]);
+		_barriers.emplace_back(pBufferMemoryBarriers[i], srcStageMask, dstStageMask);
 	}
 	for (uint32_t i = 0; i < imageMemoryBarrierCount; i++) {
-		_barriers.emplace_back(pImageMemoryBarriers[i]);
+		_barriers.emplace_back(pImageMemoryBarriers[i], srcStageMask, dstStageMask);
 	}
 
 	return VK_SUCCESS;
@@ -63,18 +110,16 @@ VkResult MVKCmdPipelineBarrier<N>::setContent(MVKCommandBuffer* cmdBuff,
 
 template <size_t N>
 void MVKCmdPipelineBarrier<N>::encode(MVKCommandEncoder* cmdEncoder) {
+	
+	auto& mtlFeats = cmdEncoder->getMetalFeatures();
 
 #if MVK_MACOS
 	// Calls below invoke MTLBlitCommandEncoder so must apply this first.
 	// Check if pipeline barriers are available and we are in a renderpass.
-	if (cmdEncoder->getDevice()->_pMetalFeatures->memoryBarriers && cmdEncoder->_mtlRenderEncoder) {
-		MTLRenderStages srcStages = mvkMTLRenderStagesFromVkPipelineStageFlags(_srcStageMask, false);
-		MTLRenderStages dstStages = mvkMTLRenderStagesFromVkPipelineStageFlags(_dstStageMask, true);
-
-		id<MTLResource> resources[_barriers.size()];
-		uint32_t rezCnt = 0;
-
+	if (mtlFeats.memoryBarriers && cmdEncoder->_mtlRenderEncoder) {
 		for (auto& b : _barriers) {
+			MTLRenderStages srcStages = mvkMTLRenderStagesFromVkPipelineStageFlags(b.srcStageMask, false);
+			MTLRenderStages dstStages = mvkMTLRenderStagesFromVkPipelineStageFlags(b.dstStageMask, true);
 			switch (b.type) {
 				case MVKPipelineBarrier::Memory: {
 					MTLBarrierScope scope = (mvkMTLBarrierScopeFromVkAccessFlags(b.srcAccessMask) |
@@ -85,31 +130,30 @@ void MVKCmdPipelineBarrier<N>::encode(MVKCommandEncoder* cmdEncoder) {
 					break;
 				}
 
-				case MVKPipelineBarrier::Buffer:
-					resources[rezCnt++] = b.mvkBuffer->getMTLBuffer();
+				case MVKPipelineBarrier::Buffer: {
+					id<MTLResource> mtlRez = b.mvkBuffer->getMTLBuffer();
+					[cmdEncoder->_mtlRenderEncoder memoryBarrierWithResources: &mtlRez
+																		count: 1
+																  afterStages: srcStages
+																 beforeStages: dstStages];
 					break;
-
-				case MVKPipelineBarrier::Image:
-                    for (uint8_t planeIndex = 0; planeIndex < b.mvkImage->getPlaneCount(); planeIndex++) {
-                        resources[rezCnt++] = b.mvkImage->getMTLTexture(planeIndex);
-                    }
+				}
+				case MVKPipelineBarrier::Image: {
+					uint32_t plnCnt = b.mvkImage->getPlaneCount();
+					id<MTLResource> mtlRezs[plnCnt];
+					for (uint8_t plnIdx = 0; plnIdx < plnCnt; plnIdx++) {
+						mtlRezs[plnIdx] = b.mvkImage->getMTLTexture(plnIdx);
+					}
+					[cmdEncoder->_mtlRenderEncoder memoryBarrierWithResources: mtlRezs
+																		count: plnCnt
+																  afterStages: srcStages
+																 beforeStages: dstStages];
 					break;
-
+				}
 				default:
 					break;
 			}
 		}
-
-		if (rezCnt) {
-			[cmdEncoder->_mtlRenderEncoder memoryBarrierWithResources: resources
-																count: rezCnt
-														  afterStages: srcStages
-														 beforeStages: dstStages];
-		}
-	} else if (cmdEncoder->getDevice()->_pMetalFeatures->textureBarriers) {
-#if !MVK_MACCAT
-		if (coversTextures()) { [cmdEncoder->_mtlRenderEncoder textureBarrier]; }
-#endif
 	}
 #endif
 
@@ -119,7 +163,7 @@ void MVKCmdPipelineBarrier<N>::encode(MVKCommandEncoder* cmdEncoder) {
 	// into separate Metal renderpasses. Since this is a potentially expensive operation,
 	// verify that at least one attachment is being used both as an input and render attachment
 	// by checking for a VK_IMAGE_LAYOUT_GENERAL layout.
-	if (cmdEncoder->_mtlRenderEncoder && cmdEncoder->getDevice()->_pMetalFeatures->tileBasedDeferredRendering) {
+	if (cmdEncoder->_mtlRenderEncoder && mtlFeats.tileBasedDeferredRendering) {
 		bool needsRenderpassRestart = false;
 		for (auto& b : _barriers) {
 			if (b.type == MVKPipelineBarrier::Image && b.newLayout == VK_IMAGE_LAYOUT_GENERAL) {
@@ -139,15 +183,15 @@ void MVKCmdPipelineBarrier<N>::encode(MVKCommandEncoder* cmdEncoder) {
 	for (auto& b : _barriers) {
 		switch (b.type) {
 			case MVKPipelineBarrier::Memory:
-				mvkDvc->applyMemoryBarrier(_srcStageMask, _dstStageMask, b, cmdEncoder, cmdUse);
+				mvkDvc->applyMemoryBarrier(b, cmdEncoder, cmdUse);
 				break;
 
 			case MVKPipelineBarrier::Buffer:
-				b.mvkBuffer->applyBufferMemoryBarrier(_srcStageMask, _dstStageMask, b, cmdEncoder, cmdUse);
+				b.mvkBuffer->applyBufferMemoryBarrier(b, cmdEncoder, cmdUse);
 				break;
 
 			case MVKPipelineBarrier::Image:
-				b.mvkImage->applyImageMemoryBarrier(_srcStageMask, _dstStageMask, b, cmdEncoder, cmdUse);
+				b.mvkImage->applyImageMemoryBarrier(b, cmdEncoder, cmdUse);
 				break;
 
 			default:
@@ -313,6 +357,7 @@ void MVKCmdPushConstants<N>::encode(MVKCommandEncoder* cmdEncoder) {
         VK_SHADER_STAGE_VERTEX_BIT,
         VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT,
         VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT,
+        VK_SHADER_STAGE_GEOMETRY_BIT,
         VK_SHADER_STAGE_FRAGMENT_BIT,
         VK_SHADER_STAGE_COMPUTE_BIT
     };
@@ -346,7 +391,7 @@ VkResult MVKCmdPushDescriptorSet::setContent(MVKCommandBuffer* cmdBuff,
 	_pipelineLayout->retain();
 
 	// Add the descriptor writes
-	MVKDevice* mvkDvc = cmdBuff->getDevice();
+	auto& enabledExtns = cmdBuff->getEnabledExtensions();
 	clearDescriptorWrites();	// Clear for reuse
 	_descriptorWrites.reserve(descriptorWriteCount);
 	for (uint32_t dwIdx = 0; dwIdx < descriptorWriteCount; dwIdx++) {
@@ -368,7 +413,7 @@ VkResult MVKCmdPushDescriptorSet::setContent(MVKCommandBuffer* cmdBuff,
 			std::copy_n(descWrite.pTexelBufferView, descWrite.descriptorCount, pNewTexelBufferView);
 			descWrite.pTexelBufferView = pNewTexelBufferView;
 		}
-        if (mvkDvc->_enabledExtensions.vk_EXT_inline_uniform_block.enabled) {
+        if (enabledExtns.vk_EXT_inline_uniform_block.enabled) {
             const VkWriteDescriptorSetInlineUniformBlockEXT* pInlineUniformBlock = nullptr;
 			for (const auto* next = (VkBaseInStructure*)descWrite.pNext; next; next = next->pNext) {
                 switch (next->sType) {
@@ -394,7 +439,7 @@ VkResult MVKCmdPushDescriptorSet::setContent(MVKCommandBuffer* cmdBuff,
 }
 
 void MVKCmdPushDescriptorSet::encode(MVKCommandEncoder* cmdEncoder) {
-	_pipelineLayout->pushDescriptorSet(cmdEncoder, _descriptorWrites.contents(), _set);
+	_pipelineLayout->pushDescriptorSet(cmdEncoder, _pipelineBindPoint, _descriptorWrites.contents(), _set);
 }
 
 MVKCmdPushDescriptorSet::~MVKCmdPushDescriptorSet() {
@@ -429,7 +474,7 @@ void MVKCmdPushDescriptorSet::clearDescriptorWrites() {
 #pragma mark MVKCmdPushDescriptorSetWithTemplate
 
 VkResult MVKCmdPushDescriptorSetWithTemplate::setContent(MVKCommandBuffer* cmdBuff,
-														 VkDescriptorUpdateTemplateKHR descUpdateTemplate,
+														 VkDescriptorUpdateTemplate descUpdateTemplate,
 														 VkPipelineLayout layout,
 														 uint32_t set,
 														 const void* pData) {
@@ -443,7 +488,7 @@ VkResult MVKCmdPushDescriptorSetWithTemplate::setContent(MVKCommandBuffer* cmdBu
 
 	if (_pData) delete[] (char*)_pData;
 	// Work out how big the memory block in pData is.
-	const VkDescriptorUpdateTemplateEntryKHR* pEntry =
+	const VkDescriptorUpdateTemplateEntry* pEntry =
 		_descUpdateTemplate->getEntry(_descUpdateTemplate->getNumberOfEntries()-1);
 	size_t size = pEntry->offset;
 	// If we were given a stride, use that; otherwise, assume only one info
@@ -494,19 +539,23 @@ MVKCmdPushDescriptorSetWithTemplate::~MVKCmdPushDescriptorSetWithTemplate() {
 
 
 #pragma mark -
-#pragma mark MVKCmdSetResetEvent
+#pragma mark MVKCmdSetEvent
 
-VkResult MVKCmdSetResetEvent::setContent(MVKCommandBuffer* cmdBuff,
-										 VkEvent event,
-										 VkPipelineStageFlags stageMask) {
+VkResult MVKCmdSetEvent::setContent(MVKCommandBuffer* cmdBuff,
+									VkEvent event,
+									VkPipelineStageFlags stageMask) {
 	_mvkEvent = (MVKEvent*)event;
 
 	return VK_SUCCESS;
 }
 
+VkResult MVKCmdSetEvent::setContent(MVKCommandBuffer* cmdBuff,
+									VkEvent event,
+									const VkDependencyInfo* pDependencyInfo) {
+	_mvkEvent = (MVKEvent*)event;
 
-#pragma mark -
-#pragma mark MVKCmdSetEvent
+	return VK_SUCCESS;
+}
 
 void MVKCmdSetEvent::encode(MVKCommandEncoder* cmdEncoder) {
 	cmdEncoder->signalEvent(_mvkEvent, true);
@@ -516,6 +565,14 @@ void MVKCmdSetEvent::encode(MVKCommandEncoder* cmdEncoder) {
 #pragma mark -
 #pragma mark MVKCmdResetEvent
 
+VkResult MVKCmdResetEvent::setContent(MVKCommandBuffer* cmdBuff,
+									  VkEvent event,
+									  VkPipelineStageFlags2 stageMask) {
+	_mvkEvent = (MVKEvent*)event;
+
+	return VK_SUCCESS;
+}
+
 void MVKCmdResetEvent::encode(MVKCommandEncoder* cmdEncoder) {
 	cmdEncoder->signalEvent(_mvkEvent, false);
 }
@@ -524,6 +581,20 @@ void MVKCmdResetEvent::encode(MVKCommandEncoder* cmdEncoder) {
 #pragma mark -
 #pragma mark MVKCmdWaitEvents
 
+template <size_t N>
+VkResult MVKCmdWaitEvents<N>::setContent(MVKCommandBuffer* cmdBuff,
+										 uint32_t eventCount,
+										 const VkEvent* pEvents,
+										 const VkDependencyInfo* pDependencyInfos) {
+	_mvkEvents.clear();	// Clear for reuse
+	_mvkEvents.reserve(eventCount);
+	for (uint32_t i = 0; i < eventCount; i++) {
+		_mvkEvents.push_back((MVKEvent*)pEvents[i]);
+	}
+
+	return VK_SUCCESS;
+}
+
 template <size_t N>
 VkResult MVKCmdWaitEvents<N>::setContent(MVKCommandBuffer* cmdBuff,
 										 uint32_t eventCount,
