@@ -1,7 +1,7 @@
 /*
  * MVKCommandBuffer.h
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -65,6 +65,24 @@ private:
 } MVKCommandEncodingContext;
 
 
+#pragma mark -
+#pragma mark MVKCurrentSubpassInfo
+
+/** Tracks current render subpass information. */
+typedef struct MVKCurrentSubpassInfo {
+	MVKRenderPass* renderpass;
+	uint32_t subpassIndex;
+	uint32_t subpassViewMask;
+
+	void beginRenderpass(MVKRenderPass* rp);
+	void nextSubpass();
+	void beginRendering(uint32_t viewMask);
+
+private:
+	void updateViewMask();
+} MVKCurrentSubpassInfo;
+
+
 #pragma mark -
 #pragma mark MVKCommandBuffer
 
@@ -96,16 +114,25 @@ public:
 	void addCommand(MVKCommand* command);
 
 	/** Returns the number of commands currently in this command buffer. */
-	inline uint32_t getCommandCount() { return _commandCount; }
+	uint32_t getCommandCount() { return _commandCount; }
 
 	/** Returns the command pool backing this command buffer. */
-	inline MVKCommandPool* getCommandPool() { return _commandPool; }
+	MVKCommandPool* getCommandPool() { return _commandPool; }
 
 	/** Submit the commands in this buffer as part of the queue submission. */
 	void submit(MVKQueueCommandBufferSubmission* cmdBuffSubmit, MVKCommandEncodingContext* pEncodingContext);
 
     /** Returns whether this command buffer can be submitted to a queue more than once. */
-    inline bool getIsReusable() { return _isReusable; }
+    bool getIsReusable() { return _isReusable; }
+
+	/**
+	 * If this is a secondary command buffer, returns the number of views inherited
+	 * from the primary command buffer. If this is a primary command buffer, returns 1.
+	 */
+	uint32_t getViewCount() const;
+
+	/** Updated as renderpass commands are added. */
+	MVKCurrentSubpassInfo _currentSubpassInfo;
 
     /**
      * Metal requires that a visibility buffer is established when a render pass is created, 
@@ -116,7 +143,7 @@ public:
     bool _needsVisibilityResultMTLBuffer;
 
 	/** Called when a MVKCmdExecuteCommands is added to this command buffer. */
-	void recordExecuteCommands(const MVKArrayRef<MVKCommandBuffer*> secondaryCommandBuffers);
+	void recordExecuteCommands(MVKArrayRef<MVKCommandBuffer*const> secondaryCommandBuffers);
 
 	/** Called when a timestamp command is added. */
 	void recordTimestampCommand();
@@ -131,24 +158,6 @@ public:
 	MVKCmdBindPipeline* _lastTessellationPipeline;
 
 
-#pragma mark Multiview render pass command management
-
-	/** Update the last recorded multiview render pass */
-	void recordBeginRenderPass(MVKCmdBeginRenderPassBase* mvkBeginRenderPass);
-
-	/** Update the last recorded multiview subpass */
-	void recordNextSubpass();
-
-	/** Forget the last recorded multiview render pass */
-	void recordEndRenderPass();
-
-	/** The most recent recorded multiview render subpass */
-	MVKRenderSubpass* _lastMultiviewSubpass;
-
-	/** Returns the currently active multiview render subpass, even for secondary command buffers */
-	MVKRenderSubpass* getLastMultiviewSubpass();
-
-
 #pragma mark Construction
 
 	MVKCommandBuffer(MVKDevice* device) : MVKDeviceTrackingMixin(device) {}
@@ -159,13 +168,13 @@ public:
      * Returns a reference to this object suitable for use as a Vulkan API handle.
      * This is the compliment of the getMVKCommandBuffer() method.
      */
-    inline VkCommandBuffer getVkCommandBuffer() { return (VkCommandBuffer)getVkHandle(); }
+	VkCommandBuffer getVkCommandBuffer() { return (VkCommandBuffer)getVkHandle(); }
 
     /**
      * Retrieves the MVKCommandBuffer instance referenced by the VkCommandBuffer handle.
      * This is the compliment of the getVkCommandBuffer() method.
      */
-    static inline MVKCommandBuffer* getMVKCommandBuffer(VkCommandBuffer vkCommandBuffer) {
+    static MVKCommandBuffer* getMVKCommandBuffer(VkCommandBuffer vkCommandBuffer) {
         return (MVKCommandBuffer*)getDispatchableObject(vkCommandBuffer);
     }
 
@@ -173,23 +182,21 @@ protected:
 	friend class MVKCommandEncoder;
 	friend class MVKCommandPool;
 
-	MVKBaseObject* getBaseObject() override { return this; };
 	void propagateDebugName() override {}
 	void init(const VkCommandBufferAllocateInfo* pAllocateInfo);
 	bool canExecute();
-	bool canPrefill();
-	void prefill();
 	void clearPrefilledMTLCommandBuffer();
     void releaseCommands(MVKCommand* command);
 	void releaseRecordedCommands();
-    void flushImmediateCmdEncoder();
+	void flushImmediateCmdEncoder();
+	void checkDeferredEncoding();
 
 	MVKCommand* _head = nullptr;
 	MVKCommand* _tail = nullptr;
 	MVKSmallVector<VkFormat, kMVKDefaultAttachmentCount> _colorAttachmentFormats;
 	MVKCommandPool* _commandPool;
 	VkCommandBufferInheritanceInfo _secondaryInheritanceInfo;
-	VkCommandBufferInheritanceRenderingInfo _inerhitanceRenderingInfo;
+	VkCommandBufferInheritanceRenderingInfo _secondaryInheritanceRenderingInfo;
 	id<MTLCommandBuffer> _prefilledMTLCmdBuffer = nil;
     MVKCommandEncodingContext* _immediateCmdEncodingContext = nullptr;
     MVKCommandEncoder* _immediateCmdEncoder = nullptr;
@@ -243,23 +250,23 @@ public:
 						 const VkRect2D& renderArea,
 						 MVKArrayRef<VkClearValue> clearValues,
 						 MVKArrayRef<MVKImageView*> attachments,
-						 MVKArrayRef<MVKArrayRef<MTLSamplePosition>> subpassSamplePositions);
+						 MVKCommandUse cmdUse);
 
 	/** Begins the next render subpass. */
 	void beginNextSubpass(MVKCommand* subpassCmd, VkSubpassContents renderpassContents);
 
-	/** Begins the next multiview Metal render pass. */
-	void beginNextMultiviewPass();
-
-	/** Sets the dynamic custom sample positions to use when rendering. */
-	void setDynamicSamplePositions(MVKArrayRef<MTLSamplePosition> dynamicSamplePositions);
-
 	/** Begins dynamic rendering. */
 	void beginRendering(MVKCommand* rendCmd, const VkRenderingInfo* pRenderingInfo);
 
 	/** Begins a Metal render pass for the current render subpass. */
 	void beginMetalRenderPass(MVKCommandUse cmdUse);
 
+	/** 
+	 * If a Metal render pass has started, and it needs to be restarted,
+	 * then end the existing Metal render pass, and start a new one.
+	 */
+	void restartMetalRenderPassIfNeeded();
+
 	/** If a render encoder is active, encodes store actions for all attachments to it. */
 	void encodeStoreActions(bool storeOverride = false);
 
@@ -276,7 +283,7 @@ public:
 	uint32_t getFramebufferLayerCount();
 
 	/** Returns the index of the currently active multiview subpass, or zero if the current render pass is not multiview. */
-	uint32_t getMultiviewPassIndex();
+	uint32_t getMultiviewPassIndex() { return _multiviewPassIndex; }
 
 	/** Begins a Metal compute encoding. */
 	void beginMetalComputeEncoding(MVKCommandUse cmdUse);
@@ -295,14 +302,11 @@ public:
 	/** Encodes an operation to signal an event to a status. */
 	void signalEvent(MVKEvent* mvkEvent, bool status);
 
-    /**
-     * If a pipeline is currently bound, returns whether the current pipeline permits dynamic
-     * setting of the specified state. If no pipeline is currently bound, returns true.
-     */
-    bool supportsDynamicState(VkDynamicState state);
+	/** Clips the rect to ensure it fits inside the render area.  */
+	VkRect2D clipToRenderArea(VkRect2D rect);
 
 	/** Clips the scissor to ensure it fits inside the render area.  */
-	VkRect2D clipToRenderArea(VkRect2D scissor);
+	MTLScissorRect clipToRenderArea(MTLScissorRect scissor);
 
 	/** Called by each graphics draw command to establish any outstanding state just prior to performing the draw. */
 	void finalizeDrawState(MVKGraphicsStage stage);
@@ -331,10 +335,12 @@ public:
 	 * Returns the current Metal compute encoder for the specified use,
 	 * which determines the label assigned to the returned encoder.
 	 *
-	 * If the current encoder is not a compute encoder, this function ends current before 
-	 * beginning compute encoding.
+	 * If the current encoder is a compute encoder, the compute state being tracked can
+	 * optionally be marked dirty. Otherwise, if the current encoder is not a compute
+	 * encoder, this function ends the current encoder before beginning compute encoding.
 	 */
-	id<MTLComputeCommandEncoder> getMTLComputeEncoder(MVKCommandUse cmdUse);
+	id<MTLComputeCommandEncoder> getMTLComputeEncoder(MVKCommandUse cmdUse,
+													  bool markCurrentComputeStateDirty = false);
 
 	/**
 	 * Returns the current Metal BLIT encoder for the specified use,
@@ -351,17 +357,41 @@ public:
 	 */
 	id<MTLCommandEncoder> getMTLEncoder();
 
+	/** Returns the graphics pipeline. */
+	MVKGraphicsPipeline* getGraphicsPipeline() { return (MVKGraphicsPipeline*)_graphicsPipelineState.getPipeline(); }
+
+	/** Returns the compute pipeline. */
+	MVKComputePipeline* getComputePipeline() { return (MVKComputePipeline*)_computePipelineState.getPipeline(); }
+
 	/** Returns the push constants associated with the specified shader stage. */
 	MVKPushConstantsCommandEncoderState* getPushConstants(VkShaderStageFlagBits shaderStage);
 
-    /** Copy bytes into the Metal encoder at a Metal vertex buffer index. */
-    void setVertexBytes(id<MTLRenderCommandEncoder> mtlEncoder, const void* bytes, NSUInteger length, uint32_t mtlBuffIndex);
+	/** Encode the buffer binding as a vertex attribute buffer. */
+	void encodeVertexAttributeBuffer(MVKMTLBufferBinding& b, bool isDynamicStride);
 
-    /** Copy bytes into the Metal encoder at a Metal fragment buffer index. */
-    void setFragmentBytes(id<MTLRenderCommandEncoder> mtlEncoder, const void* bytes, NSUInteger length, uint32_t mtlBuffIndex);
+    /**
+	 * Copy bytes into the Metal encoder at a Metal vertex buffer index, and optionally indicate
+	 * that this binding might override a desriptor binding. If so, the descriptor binding will
+	 * be marked dirty so that it will rebind before the next usage.
+	 */
+    void setVertexBytes(id<MTLRenderCommandEncoder> mtlEncoder, const void* bytes,
+						NSUInteger length, uint32_t mtlBuffIndex, bool descOverride = false);
 
-    /** Copy bytes into the Metal encoder at a Metal compute buffer index. */
-    void setComputeBytes(id<MTLComputeCommandEncoder> mtlEncoder, const void* bytes, NSUInteger length, uint32_t mtlBuffIndex);
+	/**
+	 * Copy bytes into the Metal encoder at a Metal fragment buffer index, and optionally indicate
+	 * that this binding might override a desriptor binding. If so, the descriptor binding will
+	 * be marked dirty so that it will rebind before the next usage.
+	 */
+    void setFragmentBytes(id<MTLRenderCommandEncoder> mtlEncoder, const void* bytes,
+						  NSUInteger length, uint32_t mtlBuffIndex, bool descOverride = false);
+
+	/**
+	 * Copy bytes into the Metal encoder at a Metal compute buffer index, and optionally indicate
+	 * that this binding might override a desriptor binding. If so, the descriptor binding will
+	 * be marked dirty so that it will rebind before the next usage.
+	 */
+    void setComputeBytes(id<MTLComputeCommandEncoder> mtlEncoder, const void* bytes,
+						 NSUInteger length, uint32_t mtlBuffIndex, bool descOverride = false);
 
     /** Get a temporary MTLBuffer that will be returned to a pool after the command buffer is finished. */
     const MVKMTLBufferAllocation* getTempMTLBuffer(NSUInteger length, bool isPrivate = false, bool isDedicated = false);
@@ -391,18 +421,6 @@ public:
 	/** Context for tracking information across multiple encodings. */
 	MVKCommandEncodingContext* _pEncodingContext;
 
-    /** A reference to the Metal features supported by the device. */
-    const MVKPhysicalDeviceMetalFeatures* _pDeviceMetalFeatures;
-
-    /** A reference to the Vulkan features supported by the device. */
-    const VkPhysicalDeviceFeatures* _pDeviceFeatures;
-
-    /** Pointer to the properties of the device. */
-    const VkPhysicalDeviceProperties* _pDeviceProperties;
-
-    /** Pointer to the memory properties of the device. */
-    const VkPhysicalDeviceMemoryProperties* _pDeviceMemoryProperties;
-
 	/** The command buffer whose commands are being encoded. */
 	MVKCommandBuffer* _cmdBuffer;
 
@@ -413,37 +431,28 @@ public:
 	id<MTLRenderCommandEncoder> _mtlRenderEncoder;
 
     /** Tracks the current graphics pipeline bound to the encoder. */
-    MVKPipelineCommandEncoderState _graphicsPipelineState;
+	MVKPipelineCommandEncoderState _graphicsPipelineState;
+
+	/** Tracks the current graphics resources state of the encoder. */
+	MVKGraphicsResourcesCommandEncoderState _graphicsResourcesState;
 
     /** Tracks the current compute pipeline bound to the encoder. */
-    MVKPipelineCommandEncoderState _computePipelineState;
+	MVKPipelineCommandEncoderState _computePipelineState;
 
-    /** Tracks the current viewport state of the encoder. */
-    MVKViewportCommandEncoderState _viewportState;
+	/** Tracks the current compute resources state of the encoder. */
+	MVKComputeResourcesCommandEncoderState _computeResourcesState;
 
-    /** Tracks the current scissor state of the encoder. */
-    MVKScissorCommandEncoderState _scissorState;
-
-    /** Tracks the current depth bias state of the encoder. */
-    MVKDepthBiasCommandEncoderState _depthBiasState;
-
-    /** Tracks the current blend color state of the encoder. */
-    MVKBlendColorCommandEncoderState _blendColorState;
+	/** Tracks whether the GPU-addressable buffers need to be used. */
+	MVKGPUAddressableBuffersCommandEncoderState _gpuAddressableBuffersState;
 
     /** Tracks the current depth stencil state of the encoder. */
     MVKDepthStencilCommandEncoderState _depthStencilState;
 
-    /** Tracks the current stencil reference value state of the encoder. */
-    MVKStencilReferenceValueCommandEncoderState _stencilReferenceValueState;
+	/** Tracks the current rendering states of the encoder. */
+	MVKRenderingCommandEncoderState _renderingState;
 
-    /** Tracks the current graphics resources state of the encoder. */
-    MVKGraphicsResourcesCommandEncoderState _graphicsResourcesState;
-
-    /** Tracks the current compute resources state of the encoder. */
-    MVKComputeResourcesCommandEncoderState _computeResourcesState;
-
-	/** The type of primitive that will be rendered. */
-	MTLPrimitiveType _mtlPrimitiveType;
+	/** Tracks the occlusion query state of the encoder. */
+	MVKOcclusionQueryCommandEncoderState _occlusionQueryState;
 
     /** The size of the threadgroup for the compute shader. */
     MTLSize _mtlThreadgroupSize;
@@ -454,20 +463,31 @@ public:
 	/** Indicates whether the current draw is an indexed draw. */
 	bool _isIndexedDraw;
 
+	bool _transformFeedbackEnabled = false;
+	uint32_t _transformFeedbackBufferIndex = 19;
+	MVKMTLBufferBinding _transformFeedbackBinding;
+
 #pragma mark Construction
 
-	MVKCommandEncoder(MVKCommandBuffer* cmdBuffer);
+	MVKCommandEncoder(MVKCommandBuffer* cmdBuffer,
+					  MVKPrefillMetalCommandBuffersStyle prefillStyle = MVK_CONFIG_PREFILL_METAL_COMMAND_BUFFERS_STYLE_NO_PREFILL);
+
+	~MVKCommandEncoder() override;
 
 protected:
     void addActivatedQueries(MVKQueryPool* pQueryPool, uint32_t query, uint32_t queryCount);
     void finishQueries();
-	void setSubpass(MVKCommand* passCmd, VkSubpassContents subpassContents, uint32_t subpassIndex);
-	void clearRenderArea();
-    NSString* getMTLRenderCommandEncoderName(MVKCommandUse cmdUse);
+	void setSubpass(MVKCommand* passCmd, VkSubpassContents subpassContents, uint32_t subpassIndex, MVKCommandUse cmdUse);
+	void clearRenderArea(MVKCommandUse cmdUse);
+	bool hasMoreMultiviewPasses();
+	void beginNextMultiviewPass();
+	void encodeCommandsImpl(MVKCommand* command);
 	void encodeGPUCounterSample(MVKGPUCounterQueryPool* mvkQryPool, uint32_t sampleIndex, MVKCounterSamplingFlags samplingPoints);
 	void encodeTimestampStageCounterSamples();
 	id<MTLFence> getStageCountersMTLFence();
-	MVKArrayRef<MTLSamplePosition> getCustomSamplePositions();
+	NSString* getMTLRenderCommandEncoderName(MVKCommandUse cmdUse);
+	template<typename T> void retainIfImmediatelyEncoding(T& mtlEnc);
+	template<typename T> void endMetalEncoding(T& mtlEnc);
 
 	typedef struct GPUCounterQuery {
 		MVKGPUCounterQueryPool* queryPool = nullptr;
@@ -480,8 +500,6 @@ protected:
 	MVKSmallVector<GPUCounterQuery, 16> _timestampStageCounterQueries;
 	MVKSmallVector<VkClearValue, kMVKDefaultAttachmentCount> _clearValues;
 	MVKSmallVector<MVKImageView*, kMVKDefaultAttachmentCount> _attachments;
-	MVKSmallVector<MTLSamplePosition> _dynamicSamplePositions;
-	MVKSmallVector<MVKSmallVector<MTLSamplePosition>> _subpassSamplePositions;
 	id<MTLComputeCommandEncoder> _mtlComputeEncoder;
 	id<MTLBlitCommandEncoder> _mtlBlitEncoder;
 	id<MTLFence> _stageCountersMTLFence;
@@ -490,13 +508,14 @@ protected:
 	MVKPushConstantsCommandEncoderState _tessEvalPushConstants;
 	MVKPushConstantsCommandEncoderState _fragmentPushConstants;
 	MVKPushConstantsCommandEncoderState _computePushConstants;
-    MVKOcclusionQueryCommandEncoderState _occlusionQueryState;
+	MVKPushConstantsCommandEncoderState _geometryPushConstants;
+	MVKPrefillMetalCommandBuffersStyle _prefillStyle;
 	VkSubpassContents _subpassContents;
-	MVKCommandUse _mtlComputeEncoderUse;
-	MVKCommandUse _mtlBlitEncoderUse;
 	uint32_t _renderSubpassIndex;
 	uint32_t _multiviewPassIndex;
-    uint32_t _flushCount = 0;
+    uint32_t _flushCount;
+	MVKCommandUse _mtlComputeEncoderUse;
+	MVKCommandUse _mtlBlitEncoderUse;
 	bool _isRenderingEntireAttachment;
 };
 
@@ -504,9 +523,6 @@ protected:
 #pragma mark -
 #pragma mark Support functions
 
-/** Returns a name, suitable for use as a MTLCommandBuffer label, based on the MVKCommandUse. */
-NSString* mvkMTLCommandBufferLabel(MVKCommandUse cmdUse);
-
 /** Returns a name, suitable for use as a MTLRenderCommandEncoder label, based on the MVKCommandUse. */
 NSString* mvkMTLRenderCommandEncoderLabel(MVKCommandUse cmdUse);
 
