@@ -1,7 +1,7 @@
 /*
  * MVKCommandEncoderState.h
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,6 +21,7 @@
 #include "MVKMTLResourceBindings.h"
 #include "MVKCommandResourceFactory.h"
 #include "MVKDevice.h"
+#include "MVKPipeline.h"
 #include "MVKDescriptor.h"
 #include "MVKSmallVector.h"
 #include "MVKBitArray.h"
@@ -64,8 +65,7 @@ public:
     /**
      * Called automatically when a Metal render pass begins. If the contents have been
      * modified from the default values, this instance is marked as dirty, so the contents
-     * will be encoded to Metal, otherwise it is marked as clean, so the contents will NOT
-     * be encoded. Default state can be left unencoded on a new Metal encoder.
+     * will be encoded to Metal. Default state can be left unencoded on a new Metal encoder.
      */
 	virtual void beginMetalRenderPass() { if (_isModified) { markDirty(); } }
 
@@ -75,14 +75,18 @@ public:
 	/**
 	 * Called automatically when a Metal compute pass begins. If the contents have been
 	 * modified from the default values, this instance is marked as dirty, so the contents
-	 * will be encoded to Metal, otherwise it is marked as clean, so the contents will NOT
-	 * be encoded. Default state can be left unencoded on a new Metal encoder.
+	 * will be encoded to Metal. Default state can be left unencoded on a new Metal encoder.
 	 */
 	virtual void beginMetalComputeEncoding() { if (_isModified) { markDirty(); } }
 
+	/** Check whether this state is dirty */
+	bool isDirty() const { return _isDirty; }
+
     /**
      * If the content of this instance is dirty, marks this instance as no longer dirty
      * and calls the encodeImpl() function to encode the content onto the Metal encoder.
+	 * Marking clean is done in advance so that subclass encodeImpl() implementations
+	 * can override to leave this instance in a dirty state.
      * Subclasses must override the encodeImpl() function to do the actual work.
      */
     void encode(uint32_t stage = 0) {
@@ -96,8 +100,21 @@ public:
     MVKCommandEncoderState(MVKCommandEncoder* cmdEncoder) : _cmdEncoder(cmdEncoder) {}
 
 protected:
-    virtual void encodeImpl(uint32_t stage) = 0;
+	enum StateScope {
+		Static = 0,
+		Dynamic,
+		Count
+	};
+
+	virtual void encodeImpl(uint32_t stage) = 0;
 	MVKDevice* getDevice();
+	bool isDynamicState(MVKRenderStateType state);
+	template <typename T> T& getContent(T* iVarAry, bool isDynamic) {
+		return iVarAry[isDynamic ? StateScope::Dynamic : StateScope::Static];
+	}
+	template <typename T> T& getContent(T* iVarAry, MVKRenderStateType state) {
+		return getContent(iVarAry, isDynamicState(state));
+	}
 
     MVKCommandEncoder* _cmdEncoder;
 	bool _isDirty = false;
@@ -108,81 +125,23 @@ protected:
 #pragma mark -
 #pragma mark MVKPipelineCommandEncoderState
 
-/** Holds encoder state established by pipeline commands. */
+/** Abstract class to hold encoder state established by pipeline commands. */
 class MVKPipelineCommandEncoderState : public MVKCommandEncoderState {
 
 public:
+	void bindPipeline(MVKPipeline* pipeline);
 
-	/** Binds the pipeline. */
-    void bindPipeline(MVKPipeline* pipeline);
-
-    /** Returns the currently bound pipeline. */
     MVKPipeline* getPipeline();
 
     /** Constructs this instance for the specified command encoder. */
-    MVKPipelineCommandEncoderState(MVKCommandEncoder* cmdEncoder)
-        : MVKCommandEncoderState(cmdEncoder) {}
+    MVKPipelineCommandEncoderState(MVKCommandEncoder* cmdEncoder, VkPipelineBindPoint bindPoint)
+        : MVKCommandEncoderState(cmdEncoder), _bindPoint(bindPoint) {}
 
 protected:
     void encodeImpl(uint32_t stage) override;
 
     MVKPipeline* _pipeline = nullptr;
-};
-
-
-#pragma mark -
-#pragma mark MVKViewportCommandEncoderState
-
-/** Holds encoder state established by viewport commands. */
-class MVKViewportCommandEncoderState : public MVKCommandEncoderState {
-
-public:
-
-	/**
-	 * Sets one or more of the viewports, starting at the first index.
-	 * The isSettingDynamically indicates that the scissor is being changed dynamically,
-	 * which is only allowed if the pipeline was created as VK_DYNAMIC_STATE_SCISSOR.
-	 */
-	void setViewports(const MVKArrayRef<VkViewport> viewports,
-					  uint32_t firstViewport,
-					  bool isSettingDynamically);
-
-    /** Constructs this instance for the specified command encoder. */
-    MVKViewportCommandEncoderState(MVKCommandEncoder* cmdEncoder)
-        : MVKCommandEncoderState(cmdEncoder) {}
-
-protected:
-    void encodeImpl(uint32_t stage) override;
-
-    MVKSmallVector<VkViewport, kMVKCachedViewportScissorCount> _viewports, _dynamicViewports;
-};
-
-
-#pragma mark -
-#pragma mark MVKScissorCommandEncoderState
-
-/** Holds encoder state established by viewport commands. */
-class MVKScissorCommandEncoderState : public MVKCommandEncoderState {
-
-public:
-
-	/**
-	 * Sets one or more of the scissors, starting at the first index.
-	 * The isSettingDynamically indicates that the scissor is being changed dynamically,
-	 * which is only allowed if the pipeline was created as VK_DYNAMIC_STATE_SCISSOR.
-	 */
-	void setScissors(const MVKArrayRef<VkRect2D> scissors,
-					 uint32_t firstScissor,
-					 bool isSettingDynamically);
-
-    /** Constructs this instance for the specified command encoder. */
-    MVKScissorCommandEncoderState(MVKCommandEncoder* cmdEncoder)
-        : MVKCommandEncoderState(cmdEncoder) {}
-
-protected:
-    void encodeImpl(uint32_t stage) override;
-
-    MVKSmallVector<VkRect2D, kMVKCachedViewportScissorCount> _scissors, _dynamicScissors;
+    VkPipelineBindPoint _bindPoint;
 };
 
 
@@ -200,8 +159,7 @@ public:
     /** Sets the index of the Metal buffer used to hold the push constants. */
     void setMTLBufferIndex(uint32_t mtlBufferIndex, bool pipelineStageUsesPushConstants);
 
-    /** Constructs this instance for the specified command encoder. */
-    MVKPushConstantsCommandEncoderState(MVKCommandEncoder* cmdEncoder,
+	MVKPushConstantsCommandEncoderState(MVKCommandEncoder* cmdEncoder,
                                         VkShaderStageFlagBits shaderStage)
         : MVKCommandEncoderState(cmdEncoder), _shaderStage(shaderStage) {}
 
@@ -227,16 +185,29 @@ public:
     /** Sets the depth stencil state during pipeline binding. */
     void setDepthStencilState(const VkPipelineDepthStencilStateCreateInfo& vkDepthStencilInfo);
 
-    /** 
-     * Sets the stencil compare mask value of the indicated faces
-     * to the specified value, from explicit dynamic command.
-     */
+	/** Enables or disables depth testing, from explicit dynamic command. */
+	void setDepthTestEnable(VkBool32 depthTestEnable);
+
+	/** Enables or disables depth writing, from explicit dynamic command. */
+	void setDepthWriteEnable(VkBool32 depthWriteEnable);
+
+	/** Sets the depth compare operation, from explicit dynamic command. */
+	void setDepthCompareOp(VkCompareOp depthCompareOp);
+
+	/** Enables or disables stencil testing, from explicit dynamic command. */
+	void setStencilTestEnable(VkBool32 stencilTestEnable);
+
+	/** Sets the stencil operations of the indicated faces from explicit dynamic command. */
+	void setStencilOp(VkStencilFaceFlags faceMask,
+					  VkStencilOp failOp,
+					  VkStencilOp passOp,
+					  VkStencilOp depthFailOp,
+					  VkCompareOp compareOp);
+
+    /** Sets the stencil compare mask value of the indicated faces from explicit dynamic command. */
     void setStencilCompareMask(VkStencilFaceFlags faceMask, uint32_t stencilCompareMask);
 
-    /**
-     * Sets the stencil write mask value of the indicated faces
-     * to the specified value, from explicit dynamic command.
-     */
+    /** Sets the stencil write mask value of the indicated faces from explicit dynamic command. */
     void setStencilWriteMask(VkStencilFaceFlags faceMask, uint32_t stencilWriteMask);
 
 	void beginMetalRenderPass() override;
@@ -247,96 +218,139 @@ public:
 
 protected:
     void encodeImpl(uint32_t stage) override;
-    void setStencilState(MVKMTLStencilDescriptorData& stencilInfo,
-                         const VkStencilOpState& vkStencil,
-                         bool enabled);
+	MVKMTLDepthStencilDescriptorData& getData(MVKRenderStateType state) { return getContent(_depthStencilData, state); }
+	template <typename T> void setContent(T& content, T value) {
+		if (content != value) {
+			content = value;
+			markDirty();
+		}
+	}
+	void setStencilState(MVKMTLStencilDescriptorData& sData, const VkStencilOpState& vkStencil);
+	void setStencilOp(MVKMTLStencilDescriptorData& sData, VkStencilOp failOp,
+					  VkStencilOp passOp, VkStencilOp depthFailOp, VkCompareOp compareOp);
 
-    MVKMTLDepthStencilDescriptorData _depthStencilData = kMVKMTLDepthStencilDescriptorDataDefault;
+	MVKMTLDepthStencilDescriptorData _depthStencilData[StateScope::Count];
+	bool _depthTestEnabled[StateScope::Count] = {};
 	bool _hasDepthAttachment = false;
 	bool _hasStencilAttachment = false;
 };
 
 
 #pragma mark -
-#pragma mark MVKStencilReferenceValueCommandEncoderState
+#pragma mark MVKRenderingCommandEncoderState
 
-/** Holds encoder state established by stencil reference values commands. */
-class MVKStencilReferenceValueCommandEncoderState : public MVKCommandEncoderState {
-
-public:
-
-    /** Sets the stencil references during pipeline binding. */
-    void setReferenceValues(const VkPipelineDepthStencilStateCreateInfo& vkDepthStencilInfo);
-
-    /** Sets the stencil state from explicit dynamic command. */
-    void setReferenceValues(VkStencilFaceFlags faceMask, uint32_t stencilReference);
-
-    /** Constructs this instance for the specified command encoder. */
-    MVKStencilReferenceValueCommandEncoderState(MVKCommandEncoder* cmdEncoder)
-        : MVKCommandEncoderState(cmdEncoder) {}
-
-protected:
-    void encodeImpl(uint32_t stage) override;
-
-    uint32_t _frontFaceValue = 0;
-    uint32_t _backFaceValue = 0;
+struct MVKDepthBias {
+	float depthBiasConstantFactor;
+	float depthBiasClamp;
+	float depthBiasSlopeFactor;
 };
 
-
-#pragma mark -
-#pragma mark MVKDepthBiasCommandEncoderState
-
-/** Holds encoder state established by depth bias commands. */
-class MVKDepthBiasCommandEncoderState : public MVKCommandEncoderState {
-
-public:
-
-    /** Sets the depth bias during pipeline binding. */
-    void setDepthBias(const VkPipelineRasterizationStateCreateInfo& vkRasterInfo);
-
-    /** Sets the depth bias dynamically. */
-    void setDepthBias(float depthBiasConstantFactor,
-                      float depthBiasSlopeFactor,
-                      float depthBiasClamp);
-
-    /** Constructs this instance for the specified command encoder. */
-    MVKDepthBiasCommandEncoderState(MVKCommandEncoder* cmdEncoder)
-        : MVKCommandEncoderState(cmdEncoder) {}
-
-protected:
-    void encodeImpl(uint32_t stage) override;
-
-    float _depthBiasConstantFactor = 0;
-    float _depthBiasClamp = 0;
-    float _depthBiasSlopeFactor = 0;
-    bool _isEnabled = false;
+struct MVKDepthBounds {
+	float minDepthBound;
+	float maxDepthBound;
 };
 
+struct MVKStencilReference {
+	uint32_t frontFaceValue;
+	uint32_t backFaceValue;
+};
 
-#pragma mark -
-#pragma mark MVKBlendColorCommandEncoderState
+struct MVKMTLViewports {
+	MTLViewport viewports[kMVKMaxViewportScissorCount];
+	uint32_t viewportCount;
+};
 
-/** Holds encoder state established by blend color commands. */
-class MVKBlendColorCommandEncoderState : public MVKCommandEncoderState {
+struct MVKMTLScissors {
+	MTLScissorRect scissors[kMVKMaxViewportScissorCount];
+	uint32_t scissorCount;
+};
 
+/** Holds encoder state established by various rendering state commands. */
+class MVKRenderingCommandEncoderState : public MVKCommandEncoderState {
 public:
+	void setCullMode(VkCullModeFlags cullMode, bool isDynamic);
 
-    /** Sets the blend color, either as part of pipeline binding, or dynamically. */
-    void setBlendColor(float red, float green,
-                       float blue, float alpha,
-                       bool isDynamic);
+	void setFrontFace(VkFrontFace frontFace, bool isDynamic);
 
-    /** Constructs this instance for the specified command encoder. */
-    MVKBlendColorCommandEncoderState(MVKCommandEncoder* cmdEncoder)
-        : MVKCommandEncoderState(cmdEncoder) {}
+	void setPolygonMode(VkPolygonMode polygonMode, bool isDynamic);
+
+	void setLineWidth(float lineWidth, bool isDynamic);
+
+	void setBlendConstants(MVKColor32 blendConstants, bool isDynamic);
+
+	void setDepthBias(const VkPipelineRasterizationStateCreateInfo& vkRasterInfo);
+	void setDepthBias(MVKDepthBias depthBias, bool isDynamic);
+	void setDepthBiasEnable(VkBool32 depthBiasEnable, bool isDynamic);
+	void setDepthClipEnable(bool depthClip, bool isDynamic);
+	void setDepthBounds(MVKDepthBounds depthBounds, bool isDynamic);
+	void setDepthBoundsTestEnable(VkBool32 depthBoundsTestEnable, bool isDynamic);
+	void setStencilReferenceValues(const VkPipelineDepthStencilStateCreateInfo& vkDepthStencilInfo);
+	void setStencilReferenceValues(VkStencilFaceFlags faceMask, uint32_t stencilReference);
+
+	void setViewports(const MVKArrayRef<VkViewport> viewports, uint32_t firstViewport, bool isDynamic);
+	void setScissors(const MVKArrayRef<VkRect2D> scissors, uint32_t firstScissor, bool isDynamic);
+
+	void setPrimitiveRestartEnable(VkBool32 primitiveRestartEnable, bool isDynamic);
+
+	void setRasterizerDiscardEnable(VkBool32 rasterizerDiscardEnable, bool isDynamic);
+
+	void setPrimitiveTopology(VkPrimitiveTopology topology, bool isDynamic);
+	MTLPrimitiveType getPrimitiveType();
+
+	void setPatchControlPoints(uint32_t patchControlPoints, bool isDynamic);
+	uint32_t getPatchControlPoints();
+
+	void setSampleLocationsEnable(VkBool32 sampleLocationsEnable, bool isDynamic);
+	void setSampleLocations(const MVKArrayRef<VkSampleLocationEXT> sampleLocations, bool isDynamic);
+	MVKArrayRef<MTLSamplePosition> getSamplePositions();
+
+	void beginMetalRenderPass() override;
+	bool needsMetalRenderPassRestart();
+
+	bool isDirty(MVKRenderStateType state);
+	void markDirty() override;
+
+	MVKRenderingCommandEncoderState(MVKCommandEncoder* cmdEncoder) : MVKCommandEncoderState(cmdEncoder) {}
 
 protected:
-    void encodeImpl(uint32_t stage) override;
+	void encodeImpl(uint32_t stage) override;
+	bool isDrawingTriangles();
+	template <typename T> void setContent(MVKRenderStateType state, T* iVarAry, T* pVal, bool isDynamic) {
+		auto* pIVar = &iVarAry[isDynamic ? StateScope::Dynamic : StateScope::Static];
+		if( !mvkAreEqual(pVal, pIVar) ) {
+			*pIVar = *pVal;
+			_dirtyStates.enable(state);
+			_modifiedStates.enable(state);
+			MVKCommandEncoderState::markDirty();	// Avoid local markDirty() as it marks all states dirty.
+		}
+	}
+	template <typename T> void setContent(MVKRenderStateType state, T* iVarAry, T val, bool isDynamic) {
+		setContent(state, iVarAry, &val, isDynamic);
+	}
 
-    float _red = 0;
-    float _green = 0;
-    float _blue = 0;
-    float _alpha = 0;
+	MVKSmallVector<MTLSamplePosition, kMVKMaxSampleCount> _mtlSampleLocations[StateScope::Count] = {};
+	MVKMTLViewports _mtlViewports[StateScope::Count] = {};
+	MVKMTLScissors _mtlScissors[StateScope::Count] = {};
+	MVKColor32 _mtlBlendConstants[StateScope::Count] = {};
+	MVKDepthBias _mtlDepthBias[StateScope::Count] = {};
+	MVKDepthBounds _mtlDepthBounds[StateScope::Count] = {};
+	MVKStencilReference _mtlStencilReference[StateScope::Count] = {};
+	MTLCullMode _mtlCullMode[StateScope::Count] = { MTLCullModeNone, MTLCullModeNone };
+	MTLWinding _mtlFrontFace[StateScope::Count] = { MTLWindingClockwise, MTLWindingClockwise };
+	MTLPrimitiveType _mtlPrimitiveTopology[StateScope::Count] = { MTLPrimitiveTypePoint, MTLPrimitiveTypePoint };
+	MTLDepthClipMode _mtlDepthClipEnable[StateScope::Count] = { MTLDepthClipModeClip, MTLDepthClipModeClip };
+	MTLTriangleFillMode _mtlPolygonMode[StateScope::Count] = { MTLTriangleFillModeFill, MTLTriangleFillModeFill };
+	float _mtlLineWidth[StateScope::Count] = { 1, 1 };
+	uint32_t _mtlPatchControlPoints[StateScope::Count] = { 0, 0 };
+	MVKRenderStateFlags _dirtyStates;
+	MVKRenderStateFlags _modifiedStates;
+	bool _mtlSampleLocationsEnable[StateScope::Count] = {};
+	bool _mtlDepthBiasEnable[StateScope::Count] = {};
+	bool _mtlPrimitiveRestartEnable[StateScope::Count] = {};
+	bool _mtlRasterizerDiscardEnable[StateScope::Count] = {};
+	bool _mtlDepthBoundsTestEnable[StateScope::Count] = {};
+	bool _cullBothFaces[StateScope::Count] = {};
+	bool _isPolygonModePoint[StateScope::Count] = {};
 };
 
 
@@ -358,44 +372,71 @@ public:
 						   MVKArrayRef<uint32_t> dynamicOffsets,
 						   uint32_t& dynamicOffsetIndex);
 
-	/** Encodes the Metal resource to the Metal command encoder. */
-	virtual void encodeArgumentBufferResourceUsage(MVKShaderStage stage,
-												   id<MTLResource> mtlResource,
-												   MTLResourceUsage mtlUsage,
-												   MTLRenderStages mtlStages) = 0;
+	/** Encodes the indirect use of the Metal resource to the Metal command encoder. */
+	virtual void encodeResourceUsage(MVKShaderStage stage,
+									 id<MTLResource> mtlResource,
+									 MTLResourceUsage mtlUsage,
+									 MTLRenderStages mtlStages) = 0;
+	void updateBindings();
 
-    MVKResourcesCommandEncoderState(MVKCommandEncoder* cmdEncoder) :
-		MVKCommandEncoderState(cmdEncoder), _boundDescriptorSets{} {}
+	void markDirty() override;
+
+    MVKResourcesCommandEncoderState(MVKCommandEncoder* cmdEncoder, VkPipelineBindPoint bindPoint) :
+		MVKCommandEncoderState(cmdEncoder), _bindPoint(bindPoint), _boundDescriptorSets{} {}
 
 protected:
-	void markDirty() override;
 
     // Template function that marks both the vector and all binding elements in the vector as dirty.
     template<class T>
     void markDirty(T& bindings, bool& bindingsDirtyFlag) {
-        for (auto& b : bindings) { b.isDirty = true; }
+        for (auto& b : bindings) { b.markDirty(); }
         bindingsDirtyFlag = true;
     }
 
+	// Template function to find and mark as overridden the binding that uses the index.
+	template<class T>
+	void markBufferIndexOverridden(T& bufferBindings, uint32_t index) {
+		for (auto& b : bufferBindings) {
+			if (b.index == index) {
+				b.isOverridden = true;
+				return;
+			}
+		}
+	}
+
+	// Template function to mark any overridden bindings as dirty.
+	template<class T>
+	void markOverriddenBufferIndexesDirty(T& bufferBindings, bool& bindingsDirtyFlag) {
+		for (auto& b : bufferBindings) {
+			if (b.isOverridden) {
+				b.markDirty();
+				bindingsDirtyFlag = true;
+				MVKCommandEncoderState::markDirty();
+			}
+		}
+	}
+
     // Template function that updates an existing binding or adds a new binding to a vector
     // of bindings, and marks the binding, the vector, and this instance as dirty
     template<class T, class V>
     void bind(const T& b, V& bindings, bool& bindingsDirtyFlag) {
-
         if ( !b.mtlResource ) { return; }
 
-        T db = b;   // Copy that can be marked dirty
-        MVKCommandEncoderState::markDirty();
-        bindingsDirtyFlag = true;
-        db.isDirty = true;
-
-        for (auto iter = bindings.begin(), end = bindings.end(); iter != end; ++iter) {
-            if( iter->index == db.index ) {
-                *iter = db;
+        for (auto& rb : bindings) {
+			if (rb.index == b.index) {
+                rb.update(b);
+				if (rb.isDirty) {
+					bindingsDirtyFlag = true;
+					MVKCommandEncoderState::markDirty();
+				}
                 return;
             }
         }
-        bindings.push_back(db);
+
+        bindings.push_back(b);
+        bindings.back().markDirty();
+		bindingsDirtyFlag = true;
+		MVKCommandEncoderState::markDirty();
     }
 
 	// For texture bindings, we also keep track of whether any bindings need a texture swizzle
@@ -407,7 +448,8 @@ protected:
 
     // Template function that executes a lambda expression on each dirty element of
     // a vector of bindings, and marks the bindings and the vector as no longer dirty.
-	// Clear isDirty flag before operation to allow operation to possibly override.
+	// Clear binding isDirty flag before operation to allow operation to possibly override.
+	// If it does override, leave both the bindings and this instance as dirty.
 	template<class T, class V>
 	void encodeBinding(V& bindings,
 					   bool& bindingsDirtyFlag,
@@ -418,7 +460,7 @@ protected:
 				if (b.isDirty) {
 					b.isDirty = false;
 					mtlOperation(_cmdEncoder, b);
-					if (b.isDirty) { bindingsDirtyFlag = true; }
+					if (b.isDirty) { _isDirty = bindingsDirtyFlag = true; }
 				}
 			}
 		}
@@ -431,7 +473,7 @@ protected:
 		contents[index] = value;
 	}
 
-	void assertMissingSwizzles(bool needsSwizzle, const char* stageName, const MVKArrayRef<MVKMTLTextureBinding> texBindings);
+	void assertMissingSwizzles(bool needsSwizzle, const char* stageName, MVKArrayRef<const MVKMTLTextureBinding> texBindings);
 	void encodeMetalArgumentBuffer(MVKShaderStage stage);
 	virtual void bindMetalArgumentBuffer(MVKShaderStage stage, MVKMTLBufferBinding& buffBind) = 0;
 
@@ -442,11 +484,13 @@ protected:
 		MVKSmallVector<MVKMTLSamplerStateBinding, N> samplerStateBindings;
 		MVKSmallVector<uint32_t, N> swizzleConstants;
 		MVKSmallVector<uint32_t, N> bufferSizes;
+		MVKSmallVector<uint32_t, N> textureOffsets;
 
-		MVKMTLBufferBinding swizzleBufferBinding;
-		MVKMTLBufferBinding bufferSizeBufferBinding;
-		MVKMTLBufferBinding dynamicOffsetBufferBinding;
-		MVKMTLBufferBinding viewRangeBufferBinding;
+		MVKMTLBufferBinding swizzleBufferBinding = { .isInline = true };
+		MVKMTLBufferBinding bufferSizeBufferBinding = { .isInline = true };
+		MVKMTLBufferBinding dynamicOffsetBufferBinding = { .isInline = true };
+		MVKMTLBufferBinding textureOffsetBufferBinding = { .isInline = true };
+		MVKMTLBufferBinding viewRangeBufferBinding = { .isInline = true };
 
 		bool areBufferBindingsDirty = false;
 		bool areTextureBindingsDirty = false;
@@ -455,6 +499,7 @@ protected:
 		bool needsSwizzle = false;
 	};
 
+	VkPipelineBindPoint _bindPoint;
 	MVKDescriptorSet* _boundDescriptorSets[kMVKMaxDescriptorSetCount];
 	MVKBitArray _metalUsageDirtyDescriptors[kMVKMaxDescriptorSetCount];
 
@@ -512,6 +557,13 @@ public:
 								 bool needTessEvalDynanicOffsetBuffer,
 								 bool needFragmentDynanicOffsetBuffer);
 
+	/** Sets the current dynamic offset buffer state. */
+	void bindTextureOffsetBuffer(const MVKShaderImplicitRezBinding& binding,
+	                             bool needVertexTextureOffsetBuffer,
+	                             bool needTessCtlTextureOffsetBuffer,
+	                             bool needTessEvalTextureOffsetBuffer,
+	                             bool needFragmentTextureOffsetBuffer);
+
     /** Sets the current view range buffer state. */
     void bindViewRangeBuffer(const MVKShaderImplicitRezBinding& binding,
                              bool needVertexViewBuffer,
@@ -521,29 +573,44 @@ public:
                         const char* pStageName,
                         bool fullImageViewSwizzle,
                         std::function<void(MVKCommandEncoder*, MVKMTLBufferBinding&)> bindBuffer,
-                        std::function<void(MVKCommandEncoder*, MVKMTLBufferBinding&, const MVKArrayRef<uint32_t>)> bindImplicitBuffer,
+                        std::function<void(MVKCommandEncoder*, MVKMTLBufferBinding&, MVKArrayRef<const uint32_t>)> bindImplicitBuffer,
                         std::function<void(MVKCommandEncoder*, MVKMTLTextureBinding&)> bindTexture,
                         std::function<void(MVKCommandEncoder*, MVKMTLSamplerStateBinding&)> bindSampler);
 
-	void encodeArgumentBufferResourceUsage(MVKShaderStage stage,
-										   id<MTLResource> mtlResource,
-										   MTLResourceUsage mtlUsage,
-										   MTLRenderStages mtlStages) override;
+	void encodeResourceUsage(MVKShaderStage stage,
+							 id<MTLResource> mtlResource,
+							 MTLResourceUsage mtlUsage,
+							 MTLRenderStages mtlStages) override;
 
 	/** Offset all buffers for vertex attribute bindings with zero divisors by the given number of strides. */
 	void offsetZeroDivisorVertexBuffers(MVKGraphicsStage stage, MVKGraphicsPipeline* pipeline, uint32_t firstInstance);
 
+	/**
+	 * Marks the buffer binding using the index as having been overridden,
+	 * such as by push constants or internal rendering in some transfers.
+	 * */
+	void markBufferIndexOverridden(MVKShaderStage stage, uint32_t mtlBufferIndex);
+
+	/** Marks any overridden buffer indexes as dirty. */
+	void markOverriddenBufferIndexesDirty();
+
+	void endMetalRenderPass() override;
+
+	void markDirty() override;
+	void markDirty(MVKShaderStage stage);
+
 #pragma mark Construction
-    
-    /** Constructs this instance for the specified command encoder. */
-    MVKGraphicsResourcesCommandEncoderState(MVKCommandEncoder* cmdEncoder) : MVKResourcesCommandEncoderState(cmdEncoder) {}
+
+	/** Constructs this instance for the specified command encoder. */
+	MVKGraphicsResourcesCommandEncoderState(MVKCommandEncoder* cmdEncoder)
+		: MVKResourcesCommandEncoderState(cmdEncoder, VK_PIPELINE_BIND_POINT_GRAPHICS) {}
 
 protected:
     void encodeImpl(uint32_t stage) override;
-    void markDirty() override;
 	void bindMetalArgumentBuffer(MVKShaderStage stage, MVKMTLBufferBinding& buffBind) override;
 
-    ResourceBindings<8> _shaderStageResourceBindings[4];
+    ResourceBindings<8> _shaderStageResourceBindings[kMVKShaderStageFragment + 1];
+	std::unordered_map<id<MTLResource>, MTLRenderStages> _renderUsageStages;
 };
 
 
@@ -576,17 +643,30 @@ public:
 	/** Sets the current dynamic offset buffer state. */
 	void bindDynamicOffsetBuffer(const MVKShaderImplicitRezBinding& binding, bool needDynamicOffsetBuffer);
 
-	void encodeArgumentBufferResourceUsage(MVKShaderStage stage,
-										   id<MTLResource> mtlResource,
-										   MTLResourceUsage mtlUsage,
-										   MTLRenderStages mtlStages) override;
+	/** Sets the current texture offset buffer state. */
+	void bindTextureOffsetBuffer(const MVKShaderImplicitRezBinding& binding, bool needTextureOffsetBuffer);
+
+	void encodeResourceUsage(MVKShaderStage stage,
+							 id<MTLResource> mtlResource,
+							 MTLResourceUsage mtlUsage,
+							 MTLRenderStages mtlStages) override;
+
+	/**
+	 * Marks the buffer binding using the index as having been overridden,
+	 * such as by push constants or internal rendering in some transfers.
+	 * */
+	void markBufferIndexOverridden(uint32_t mtlBufferIndex);
+
+	/** Marks any overridden buffer indexes as dirty. */
+	void markOverriddenBufferIndexesDirty();
 
     void markDirty() override;
 
 #pragma mark Construction
 
-    /** Constructs this instance for the specified command encoder. */
-    MVKComputeResourcesCommandEncoderState(MVKCommandEncoder* cmdEncoder) : MVKResourcesCommandEncoderState(cmdEncoder) {}
+	/** Constructs this instance for the specified command encoder. */
+	MVKComputeResourcesCommandEncoderState(MVKCommandEncoder* cmdEncoder)
+		: MVKResourcesCommandEncoderState(cmdEncoder, VK_PIPELINE_BIND_POINT_COMPUTE) {}
 
 protected:
     void encodeImpl(uint32_t) override;
@@ -596,6 +676,26 @@ protected:
 };
 
 
+#pragma mark -
+#pragma mark MVKGPUAddressableBuffersCommandEncoderState
+
+/** Tracks whether the GPU-addressable buffers need to be used. */
+class MVKGPUAddressableBuffersCommandEncoderState : public MVKCommandEncoderState {
+
+public:
+
+	/** Marks that GPU addressable buffers may be needed in the specified shader stage. */
+	void useGPUAddressableBuffersInStage(MVKShaderStage shaderStage);
+
+	MVKGPUAddressableBuffersCommandEncoderState(MVKCommandEncoder* cmdEncoder) : MVKCommandEncoderState(cmdEncoder) {}
+
+protected:
+	void encodeImpl(uint32_t stage) override;
+
+	bool _usageStages[kMVKShaderStageCount] = {};
+};
+
+
 #pragma mark -
 #pragma mark MVKOcclusionQueryCommandEncoderState
 
@@ -629,6 +729,7 @@ protected:
 
 	MVKSmallVector<OcclusionQueryLocation> _mtlRenderPassQueries;
     MTLVisibilityResultMode _mtlVisibilityResultMode = MTLVisibilityResultModeDisabled;
+	bool _hasRasterized = false;
 };
 
 
