@@ -1,7 +1,7 @@
 /*
  * MVKCommandResourceFactory.h
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -37,20 +37,31 @@ class MVKQueryPool;
  * This structure can be used as a key in a std::map and std::unordered_map.
  */
 typedef struct MVKRPSKeyBlitImg {
-	uint16_t srcMTLPixelFormat = 0;			/**< as MTLPixelFormat */
-	uint16_t dstMTLPixelFormat = 0;			/**< as MTLPixelFormat */
-	uint8_t srcMTLTextureType = 0;			/**< as MTLTextureType */
+	uint16_t srcMTLPixelFormat : 12;		/**< as MTLPixelFormat */
+	uint16_t dstMTLPixelFormat : 12;		/**< as MTLPixelFormat */
+	uint8_t srcMTLTextureType : 4;			/**< as MTLTextureType */
+	uint8_t srcFilter : 4;					/**< as MTLSamplerMinMagFilter */
 	uint8_t srcAspect = 0;					/**< as VkImageAspectFlags */
-	uint8_t srcFilter = 0;					/**< as MTLSamplerMinMagFilter */
 	uint8_t dstSampleCount = 0;
+	uint8_t srcSwizzleR : 4;				/**< as VkComponentSwizzle */
+	uint8_t srcSwizzleG : 4;				/**< as VkComponentSwizzle */
+	uint8_t srcSwizzleB : 4;				/**< as VkComponentSwizzle */
+	uint8_t srcSwizzleA : 4;				/**< as VkComponentSwizzle */
+
+	MVKRPSKeyBlitImg() : srcMTLPixelFormat(0), dstMTLPixelFormat(0), srcMTLTextureType(0), srcFilter(0),
+		srcSwizzleR(0), srcSwizzleG(0), srcSwizzleB(0), srcSwizzleA(0) {}
 
 	bool operator==(const MVKRPSKeyBlitImg& rhs) const {
 		if (srcMTLPixelFormat != rhs.srcMTLPixelFormat) { return false; }
 		if (dstMTLPixelFormat != rhs.dstMTLPixelFormat) { return false; }
 		if (srcMTLTextureType != rhs.srcMTLTextureType) { return false; }
-		if (srcAspect != rhs.srcAspect) { return false; }
 		if (srcFilter != rhs.srcFilter) { return false; }
+		if (srcAspect != rhs.srcAspect) { return false; }
 		if (dstSampleCount != rhs.dstSampleCount) { return false; }
+		if (srcSwizzleR != rhs.srcSwizzleR) { return false; }
+		if (srcSwizzleG != rhs.srcSwizzleG) { return false; }
+		if (srcSwizzleB != rhs.srcSwizzleB) { return false; }
+		if (srcSwizzleA != rhs.srcSwizzleA) { return false; }
 		return true;
 	}
 
@@ -70,23 +81,40 @@ typedef struct MVKRPSKeyBlitImg {
 				srcMTLTextureType == MTLTextureType1DArray);
 	}
 
+	VkComponentMapping getSrcSwizzle() {
+		return { (VkComponentSwizzle)srcSwizzleR, (VkComponentSwizzle)srcSwizzleG,
+			 (VkComponentSwizzle)srcSwizzleB, (VkComponentSwizzle)srcSwizzleA };
+	}
+
 	std::size_t hash() const {
 		std::size_t hash = srcMTLPixelFormat;
 
-		hash <<= 16;
+		hash <<= 12;
 		hash |= dstMTLPixelFormat;
 
-		hash <<= 8;
+		hash <<= 4;
 		hash |= srcMTLTextureType;
 
+		hash <<= 4;
+		hash |= srcFilter;
+
 		hash <<= 8;
 		hash |= srcAspect;
 
-		hash <<= 8;
-		hash |= srcFilter;
-
 		hash <<= 8;
 		hash |= dstSampleCount;
+
+		hash <<= 4;
+		hash |= srcSwizzleR;
+
+		hash <<= 4;
+		hash |= srcSwizzleG;
+
+		hash <<= 4;
+		hash |= srcSwizzleB;
+
+		hash <<= 4;
+		hash |= srcSwizzleA;
 		return hash;
 	}
 
@@ -107,9 +135,11 @@ namespace std {
 #pragma mark -
 #pragma mark MVKRPSKeyClearAtt
 
-#define kMVKClearAttachmentCount						(kMVKCachedColorAttachmentCount + 1)
-#define kMVKClearAttachmentDepthStencilIndex			(kMVKClearAttachmentCount - 1)
-#define kMVKClearAttachmentLayeredRenderingBitIndex		kMVKClearAttachmentCount
+const static uint32_t kMVKClearColorAttachmentCount = kMVKMaxColorAttachmentCount;
+const static uint32_t kMVKClearAttachmentDepthIndex = kMVKClearColorAttachmentCount;
+const static uint32_t kMVKClearAttachmentStencilIndex = kMVKClearAttachmentDepthIndex + 1;
+const static uint32_t kMVKClearAttachmentCount = kMVKClearAttachmentStencilIndex + 1;
+const static uint32_t kMVKClearAttachmentLayeredRenderingBitIndex = kMVKClearAttachmentStencilIndex + 1;
 
 /**
  * Key to use for looking up cached MTLRenderPipelineState instances.
@@ -180,27 +210,24 @@ namespace std {
  * change as early as possible.
  */
 typedef struct MVKMTLStencilDescriptorData {
-    bool enabled;                       /**< Indicates whether stencil testing for this face is enabled. */
+	uint32_t readMask;					/**< The bit-mask to apply when comparing the stencil buffer value to the reference value. */
+	uint32_t writeMask;					/**< The bit-mask to apply when writing values to the stencil buffer. */
     uint8_t stencilCompareFunction;		/**< The stencil compare function (interpreted as MTLCompareFunction). */
     uint8_t stencilFailureOperation;	/**< The operation to take when the stencil test fails (interpreted as MTLStencilOperation). */
     uint8_t depthFailureOperation;		/**< The operation to take when the stencil test passes, but the depth test fails (interpreted as MTLStencilOperation). */
     uint8_t depthStencilPassOperation;	/**< The operation to take when both the stencil and depth tests pass (interpreted as MTLStencilOperation). */
-    uint32_t readMask;					/**< The bit-mask to apply when comparing the stencil buffer value to the reference value. */
-    uint32_t writeMask;					/**< The bit-mask to apply when writing values to the stencil buffer. */
+
+	bool operator==(const MVKMTLStencilDescriptorData& rhs) const { return mvkAreEqual(this, &rhs); }
+	bool operator!=(const MVKMTLStencilDescriptorData& rhs) const { return !(*this == rhs); }
 
     MVKMTLStencilDescriptorData() {
-
-        // Start with all zeros to ensure memory comparisons will work,
-        // even if the structure contains alignment gaps.
-        mvkClear(this);
-
-        enabled = false;
+        mvkClear(this);  // Clear all memory to ensure memory comparisons will work.
+		mvkEnableAllFlags(readMask);
+		mvkEnableAllFlags(writeMask);
         stencilCompareFunction = MTLCompareFunctionAlways;
         stencilFailureOperation = MTLStencilOperationKeep;
         depthFailureOperation = MTLStencilOperationKeep;
         depthStencilPassOperation = MTLStencilOperationKeep;
-        readMask = static_cast<uint32_t>(~0);
-        writeMask = static_cast<uint32_t>(~0);
     }
 
 } MVKMTLStencilDescriptorData;
@@ -217,34 +244,32 @@ const MVKMTLStencilDescriptorData kMVKMTLStencilDescriptorDataDefault;
  * change as early as possible.
  */
 typedef struct MVKMTLDepthStencilDescriptorData {
-    uint8_t depthCompareFunction;		/**< The depth compare function (interpreted as MTLCompareFunction). */
-    bool depthWriteEnabled;				/**< Indicates whether depth writing is enabled. */
     MVKMTLStencilDescriptorData frontFaceStencilData;
     MVKMTLStencilDescriptorData backFaceStencilData;
+	uint8_t depthCompareFunction;		/**< The depth compare function (interpreted as MTLCompareFunction). */
+	bool depthWriteEnabled;				/**< Indicates whether depth writing is enabled. */
+	bool stencilTestEnabled;			/**< Indicates whether stencil testing is enabled. */
 
 	bool operator==(const MVKMTLDepthStencilDescriptorData& rhs) const { return mvkAreEqual(this, &rhs); }
+	bool operator!=(const MVKMTLDepthStencilDescriptorData& rhs) const { return !(*this == rhs); }
 
 	std::size_t hash() const {
 		return mvkHash((uint64_t*)this, sizeof(*this) / sizeof(uint64_t));
 	}
-
-	/** Disable depth and/or stencil testing. */
-	void disable(bool disableDepth, bool disableStencil) {
-		if (disableDepth) {
-			depthCompareFunction = MTLCompareFunctionAlways;
-			depthWriteEnabled = false;
-		}
-		if (disableStencil) {
-			frontFaceStencilData = kMVKMTLStencilDescriptorDataDefault;
-			backFaceStencilData = kMVKMTLStencilDescriptorDataDefault;
-		}
+	void disableDepth() {
+		depthCompareFunction = MTLCompareFunctionAlways;
+		depthWriteEnabled = false;
+	}
+	void disableStencil() {
+		stencilTestEnabled = false;
+		frontFaceStencilData = kMVKMTLStencilDescriptorDataDefault;
+		backFaceStencilData = kMVKMTLStencilDescriptorDataDefault;
 	}
 
 	MVKMTLDepthStencilDescriptorData() {
-		// Start with all zeros to ensure memory comparisons will work,
-		// even if the structure contains alignment gaps.
-		mvkClear(this);
-		disable(true, true);
+		mvkClear(this);  // Clear all memory to ensure memory comparisons will work.
+		disableDepth();
+		disableStencil();
 	}
 
 } __attribute__((aligned(sizeof(uint64_t)))) MVKMTLDepthStencilDescriptorData;
@@ -429,19 +454,24 @@ public:
 
 	/** Returns a new MTLComputePipelineState for clearing an image. */
 	id<MTLComputePipelineState> newCmdClearColorImageMTLComputePipelineState(MVKFormatType type,
-																			 MVKVulkanAPIDeviceObject* owner);
+																			 MVKVulkanAPIDeviceObject* owner,
+																			 bool isTextureArray);
 
 	/** Returns a new MTLComputePipelineState for resolving an image. */
 	id<MTLComputePipelineState> newCmdResolveColorImageMTLComputePipelineState(MVKFormatType type,
-																			   MVKVulkanAPIDeviceObject* owner);
+																			   MVKVulkanAPIDeviceObject* owner,
+																			   bool isTextureArray);
 
 	/** Returns a new MTLComputePipelineState for copying between a buffer holding compressed data and a 3D image. */
 	id<MTLComputePipelineState> newCmdCopyBufferToImage3DDecompressMTLComputePipelineState(bool needTempBuf,
 																						   MVKVulkanAPIDeviceObject* owner);
 
-	/** Returns a new MTLComputePipelineState for converting an indirect buffer for use in a multiview draw. */
-	id<MTLComputePipelineState> newCmdDrawIndirectMultiviewConvertBuffersMTLComputePipelineState(bool indexed,
-																								 MVKVulkanAPIDeviceObject* owner);
+	/** Returns a new MTLComputePipelineState for populating an indirect index buffer from a non-indexed indirect buffer. */
+	id<MTLComputePipelineState> newCmdDrawIndirectPopulateIndexesMTLComputePipelineState(MVKVulkanAPIDeviceObject* owner);
+
+	/** Returns a new MTLComputePipelineState for converting the contents of an indirect buffer. */
+	id<MTLComputePipelineState> newCmdDrawIndirectConvertBuffersMTLComputePipelineState(bool indexed,
+																						MVKVulkanAPIDeviceObject* owner);
 
 	/** Returns a new MTLComputePipelineState for converting an indirect buffer for use in a tessellated draw. */
 	id<MTLComputePipelineState> newCmdDrawIndirectTessConvertBuffersMTLComputePipelineState(bool indexed,
