@@ -1,7 +1,7 @@
 /*
  * MVKDescriptor.h
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -28,6 +28,10 @@ class MVKCommandEncoder;
 class MVKResourcesCommandEncoderState;
 
 
+/** Magic number to indicate the variable descriptor count is currently unknown. */
+static uint32_t kMVKVariableDescriptorCountUnknown = std::numeric_limits<uint32_t>::max();
+
+
 #pragma mark MVKShaderStageResourceBinding
 
 /** Indicates the Metal resource indexes used by a single shader stage in a descriptor. */
@@ -35,7 +39,6 @@ typedef struct MVKShaderStageResourceBinding {
 	uint32_t bufferIndex = 0;
 	uint32_t textureIndex = 0;
 	uint32_t samplerIndex = 0;
-	uint32_t resourceIndex = 0;
 	uint32_t dynamicOffsetBufferIndex = 0;
 
 	MVKShaderStageResourceBinding operator+ (const MVKShaderStageResourceBinding& rhs);
@@ -51,10 +54,9 @@ typedef struct MVKShaderStageResourceBinding {
 typedef struct MVKShaderResourceBinding {
 	MVKShaderStageResourceBinding stages[kMVKShaderStageCount];
 
-	uint16_t getMaxResourceIndex();
-	uint16_t getMaxBufferIndex();
-	uint16_t getMaxTextureIndex();
-	uint16_t getMaxSamplerIndex();
+	uint32_t getMaxBufferIndex();
+	uint32_t getMaxTextureIndex();
+	uint32_t getMaxSamplerIndex();
 
 	MVKShaderResourceBinding operator+ (const MVKShaderResourceBinding& rhs);
 	MVKShaderResourceBinding& operator+= (const MVKShaderResourceBinding& rhs);
@@ -75,7 +77,8 @@ void mvkPopulateShaderConversionConfig(mvk::SPIRVToMSLConversionConfiguration& s
 									   uint32_t bindingIndex,
 									   uint32_t count,
 									   VkDescriptorType descType,
-									   MVKSampler* immutableSampler);
+									   MVKSampler* immutableSampler,
+									   bool usingNativeTextureAtomics);
 
 
 #pragma mark -
@@ -90,34 +93,36 @@ public:
 	MVKVulkanAPIObject* getVulkanAPIObject() override;
 
 	/** Returns the binding number of this layout. */
-	inline uint32_t getBinding() { return _info.binding; }
+	uint32_t getBinding() { return _info.binding; }
 
 	/** Returns whether this binding has a variable descriptor count. */
-	inline bool hasVariableDescriptorCount() const {
-		return mvkIsAnyFlagEnabled(_flags, VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT_EXT);
+	bool hasVariableDescriptorCount() const {
+		return mvkIsAnyFlagEnabled(_flags, VK_DESCRIPTOR_BINDING_VARIABLE_DESCRIPTOR_COUNT_BIT);
 	}
 
 	/**
 	 * Returns the number of descriptors in this layout.
 	 *
 	 * If this is an inline block data descriptor, always returns 1. If this descriptor
-	 * has a variable descriptor count, and descSet is not null, the variable descriptor
-	 * count provided to that descriptor set is returned. Otherwise returns the value
-	 * defined in VkDescriptorSetLayoutBinding::descriptorCount.
+	 * has a variable descriptor count, and it is provided here, it is returned.
+	 * Otherwise returns the value defined in VkDescriptorSetLayoutBinding::descriptorCount.
 	 */
-	uint32_t getDescriptorCount(MVKDescriptorSet* descSet = nullptr) const;
+	uint32_t getDescriptorCount(uint32_t variableDescriptorCount = kMVKVariableDescriptorCountUnknown) const;
 
 	/** Returns the descriptor type of this layout. */
-	inline VkDescriptorType getDescriptorType() { return _info.descriptorType; }
+	VkDescriptorType getDescriptorType() { return _info.descriptorType; }
 
 	/** Returns whether this binding uses immutable samplers. */
 	bool usesImmutableSamplers() { return !_immutableSamplers.empty(); }
 
 	/** Returns the immutable sampler at the index, or nullptr if immutable samplers are not used. */
-	MVKSampler* getImmutableSampler(uint32_t index);
+	MVKSampler* getImmutableSampler(uint32_t index) {
+		return (index < _immutableSamplers.size()) ? _immutableSamplers[index] : nullptr;
+	}
 
 	/** Encodes the descriptors in the descriptor set that are specified by this layout, */
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSet* descSet,
 			  MVKShaderResourceBinding& dslMTLRezIdxOffsets,
 			  MVKArrayRef<uint32_t> dynamicOffsets,
@@ -125,6 +130,7 @@ public:
 
     /** Encodes this binding layout and the specified descriptor on the specified command encoder immediately. */
     void push(MVKCommandEncoder* cmdEncoder,
+              VkPipelineBindPoint pipelineBindPoint,
               uint32_t& dstArrayElement,
               uint32_t& descriptorCount,
               uint32_t& descriptorsPushed,
@@ -155,7 +161,8 @@ public:
 								  MVKDescriptorSetLayout* layout,
 								  const VkDescriptorSetLayoutBinding* pBinding,
 								  VkDescriptorBindingFlagsEXT bindingFlags,
-								  uint32_t descriptorIndex);
+								  uint32_t& dslDescCnt,
+								  uint32_t& dslMTLRezCnt);
 
 	MVKDescriptorSetLayoutBinding(const MVKDescriptorSetLayoutBinding& binding);
 
@@ -163,19 +170,27 @@ public:
 
 protected:
 	friend class MVKDescriptorSetLayout;
+	friend class MVKDescriptorSet;
     friend class MVKInlineUniformBlockDescriptor;
 	
-	void initMetalResourceIndexOffsets(const VkDescriptorSetLayoutBinding* pBinding, uint32_t stage);
-	void addMTLArgumentDescriptors(NSMutableArray<MTLArgumentDescriptor*>* args);
+	void initMetalResourceIndexOffsets(const VkDescriptorSetLayoutBinding* pBinding,
+									   uint32_t stage,
+									   uint32_t dslMTLRezCnt);
+	void addMTLArgumentDescriptors(NSMutableArray<MTLArgumentDescriptor*>* args,
+								   uint32_t variableDescriptorCount);
 	void addMTLArgumentDescriptor(NSMutableArray<MTLArgumentDescriptor*>* args,
+								  uint32_t variableDescriptorCount,
 								  uint32_t argIndex,
 								  MTLDataType dataType,
 								  MTLArgumentAccess access);
-	bool isUsingMetalArgumentBuffer();
 	void populateShaderConversionConfig(mvk::SPIRVToMSLConversionConfiguration& shaderConfig,
 										MVKShaderResourceBinding& dslMTLRezIdxOffsets,
 										uint32_t dslIndex);
 	bool validate(MVKSampler* mvkSampler);
+	void encodeImmutableSamplersToMetalArgumentBuffer(MVKDescriptorSet* mvkDescSet);
+	uint8_t getMaxPlaneCount();
+	uint32_t getMTLResourceCount(uint32_t variableDescriptorCount = kMVKVariableDescriptorCountUnknown);
+	std::string getLogDescription();
 
 	MVKDescriptorSetLayout* _layout;
 	VkDescriptorSetLayoutBinding _info;
@@ -205,6 +220,7 @@ public:
 
 	/** Encodes this descriptor (based on its layout binding index) on the the command encoder. */
 	virtual void bind(MVKCommandEncoder* cmdEncoder,
+					  VkPipelineBindPoint pipelineBindPoint,
 					  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 					  uint32_t elementIndex,
 					  bool stages[],
@@ -212,16 +228,6 @@ public:
 					  MVKArrayRef<uint32_t> dynamicOffsets,
 					  uint32_t& dynamicOffsetIndex) = 0;
 
-	/** Encodes this descriptor to the Metal argument buffer. */
-	virtual void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-											 id<MTLArgumentEncoder> mtlArgEncoder,
-											 uint32_t descSetIndex,
-											 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-											 uint32_t elementIndex,
-											 MVKShaderStage stage,
-											 bool encodeToArgBuffer,
-											 bool encodeUsage) = 0;
-
 	/**
 	 * Updates the internal binding from the specified content. The format of the content depends
 	 * on the descriptor type, and is extracted from pData at the location given by index * stride.
@@ -229,8 +235,9 @@ public:
 	 */
 	virtual void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 					   MVKDescriptorSet* mvkDescSet,
-					   uint32_t index,
-					   size_t stride,
+					   uint32_t dstIdx,
+					   uint32_t srcIdx,
+					   size_t srcStride,
 					   const void* pData) = 0;
 
 	/**
@@ -252,6 +259,11 @@ public:
 					  VkBufferView* pTexelBufferView,
 					  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock) = 0;
 
+	/** Encodes the usage of this resource to the Metal command encoder. */
+	virtual void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+									 MVKShaderStage stage) = 0;
+
 	/** Resets any internal content. */
 	virtual void reset() {}
 
@@ -271,6 +283,7 @@ class MVKBufferDescriptor : public MVKDescriptor {
 
 public:
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			  uint32_t elementIndex,
 			  bool stages[],
@@ -278,19 +291,11 @@ public:
 			  MVKArrayRef<uint32_t> dynamicOffsets,
 			  uint32_t& dynamicOffsetIndex) override;
 
-	void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-									 id<MTLArgumentEncoder> mtlArgEncoder,
-									 uint32_t descSetIndex,
-									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-									 uint32_t elementIndex,
-									 MVKShaderStage stage,
-									 bool encodeToArgBuffer,
-									 bool encodeUsage) override;
-
 	void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			   MVKDescriptorSet* mvkDescSet,
-			   uint32_t srcIndex,
-			   size_t stride,
+			   uint32_t dstIdx,
+			   uint32_t srcIdx,
+			   size_t srcStride,
 			   const void* pData) override;
 
 	void read(MVKDescriptorSetLayoutBinding* mvkDSLBind,
@@ -301,11 +306,17 @@ public:
 			  VkBufferView* pTexelBufferView,
 			  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock) override;
 
+	void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+							 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+							 MVKShaderStage stage) override;
+
 	void reset() override;
 
 	~MVKBufferDescriptor() { reset(); }
 
 protected:
+	uint32_t getBufferSize(VkDeviceSize dynamicOffset = 0);
+
 	MVKBuffer* _mvkBuffer = nullptr;
 	VkDeviceSize _buffOffset = 0;
 	VkDeviceSize _buffRange = 0;
@@ -360,6 +371,7 @@ public:
 	VkDescriptorType getDescriptorType() override { return VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK_EXT; }
 
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			  uint32_t elementIndex,
 			  bool stages[],
@@ -367,19 +379,11 @@ public:
 			  MVKArrayRef<uint32_t> dynamicOffsets,
 			  uint32_t& dynamicOffsetIndex) override;
 
-	void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-									 id<MTLArgumentEncoder> mtlArgEncoder,
-									 uint32_t descSetIndex,
-									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-									 uint32_t elementIndex,
-									 MVKShaderStage stage,
-									 bool encodeToArgBuffer,
-									 bool encodeUsage) override;
-
 	void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			   MVKDescriptorSet* mvkDescSet,
-			   uint32_t dstOffset, // For inline buffers we are using this parameter as dst offset not as src descIdx
-			   size_t stride,
+			   uint32_t dstOffset, 	// Inline buffers use this parameter as an offset, not an index
+			   uint32_t srcIdx,
+			   size_t srcStride,
 			   const void* pData) override;
 
 	void read(MVKDescriptorSetLayoutBinding* mvkDSLBind,
@@ -389,13 +393,17 @@ public:
 			  VkDescriptorBufferInfo* pBufferInfo,
 			  VkBufferView* pTexelBufferView,
 			  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock) override;
-    
+
+	void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+							 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+							 MVKShaderStage stage) override;
+
 	void reset() override;
 
 	~MVKInlineUniformBlockDescriptor() { reset(); }
 
 protected:
-	inline uint8_t* getData() { return _mvkMTLBufferAllocation ? (uint8_t*)_mvkMTLBufferAllocation->getContents() : nullptr; }
+	uint8_t* getData() { return _mvkMTLBufferAllocation ? (uint8_t*)_mvkMTLBufferAllocation->getContents() : nullptr; }
 
 	MVKMTLBufferAllocation* _mvkMTLBufferAllocation = nullptr;
 };
@@ -409,6 +417,7 @@ class MVKImageDescriptor : public MVKDescriptor {
 
 public:
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			  uint32_t elementIndex,
 			  bool stages[],
@@ -416,19 +425,11 @@ public:
 			  MVKArrayRef<uint32_t> dynamicOffsets,
 			  uint32_t& dynamicOffsetIndex) override;
 
-	void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-									 id<MTLArgumentEncoder> mtlArgEncoder,
-									 uint32_t descSetIndex,
-									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-									 uint32_t elementIndex,
-									 MVKShaderStage stage,
-									 bool encodeToArgBuffer,
-									 bool encodeUsage) override;
-
 	void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			   MVKDescriptorSet* mvkDescSet,
-			   uint32_t srcIndex,
-			   size_t stride,
+			   uint32_t dstIdx,
+			   uint32_t srcIdx,
+			   size_t srcStride,
 			   const void* pData) override;
 
 	void read(MVKDescriptorSetLayoutBinding* mvkDSLBind,
@@ -439,6 +440,10 @@ public:
 			  VkBufferView* pTexelBufferView,
 			  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock) override;
 
+	void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+							 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+							 MVKShaderStage stage) override;
+
 	void reset() override;
 
 	~MVKImageDescriptor() { reset(); }
@@ -489,6 +494,7 @@ class MVKSamplerDescriptorMixin {
 
 protected:
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			  uint32_t elementIndex,
 			  bool stages[],
@@ -496,18 +502,11 @@ protected:
 			  MVKArrayRef<uint32_t> dynamicOffsets,
 			  uint32_t& dynamicOffsetIndex);
 
-	void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-									 id<MTLArgumentEncoder> mtlArgEncoder,
-									 uint32_t descSetIndex,
-									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-									 uint32_t elementIndex,
-									 MVKShaderStage stage,
-									 bool encodeToArgBuffer);
-
 	void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			   MVKDescriptorSet* mvkDescSet,
-			   uint32_t srcIndex,
-			   size_t stride,
+			   uint32_t dstIdx,
+			   uint32_t srcIdx,
+			   size_t srcStride,
 			   const void* pData);
 
 	void read(MVKDescriptorSetLayoutBinding* mvkDSLBind,
@@ -518,6 +517,10 @@ protected:
 			  VkBufferView* pTexelBufferView,
 			  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock);
 
+	void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+							 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+							 MVKShaderStage stage) {}
+
 	void reset();
 
 	~MVKSamplerDescriptorMixin() { reset(); }
@@ -536,6 +539,7 @@ public:
 	VkDescriptorType getDescriptorType() override { return VK_DESCRIPTOR_TYPE_SAMPLER; }
 
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			  uint32_t elementIndex,
 			  bool stages[],
@@ -543,19 +547,11 @@ public:
 			  MVKArrayRef<uint32_t> dynamicOffsets,
 			  uint32_t& dynamicOffsetIndex) override;
 
-	void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-									 id<MTLArgumentEncoder> mtlArgEncoder,
-									 uint32_t descSetIndex,
-									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-									 uint32_t elementIndex,
-									 MVKShaderStage stage,
-									 bool encodeToArgBuffer,
-									 bool encodeUsage) override;
-
 	void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			   MVKDescriptorSet* mvkDescSet,
-			   uint32_t srcIndex,
-			   size_t stride,
+			   uint32_t dstIdx,
+			   uint32_t srcIdx,
+			   size_t srcStride,
 			   const void* pData) override;
 
 	void read(MVKDescriptorSetLayoutBinding* mvkDSLBind,
@@ -566,6 +562,10 @@ public:
 			  VkBufferView* pTexelBufferView,
 			  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock) override;
 
+	void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+							 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+							 MVKShaderStage stage) override {}
+
 	void reset() override;
 
 	~MVKSamplerDescriptor() { reset(); }
@@ -583,6 +583,7 @@ public:
 	VkDescriptorType getDescriptorType() override { return VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER; }
 
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			  uint32_t elementIndex,
 			  bool stages[],
@@ -590,19 +591,11 @@ public:
 			  MVKArrayRef<uint32_t> dynamicOffsets,
 			  uint32_t& dynamicOffsetIndex) override;
 
-	void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-									 id<MTLArgumentEncoder> mtlArgEncoder,
-									 uint32_t descSetIndex,
-									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-									 uint32_t elementIndex,
-									 MVKShaderStage stage,
-									 bool encodeToArgBuffer,
-									 bool encodeUsage) override;
-
 	void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			   MVKDescriptorSet* mvkDescSet,
-			   uint32_t srcIndex,
-			   size_t stride,
+			   uint32_t dstIdx,
+			   uint32_t srcIdx,
+			   size_t srcStride,
 			   const void* pData) override;
 
 	void read(MVKDescriptorSetLayoutBinding* mvkDSLBind,
@@ -613,6 +606,10 @@ public:
 			  VkBufferView* pTexelBufferView,
 			  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock) override;
 
+	void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+							 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+							 MVKShaderStage stage) override;
+
 	void reset() override;
 
 	~MVKCombinedImageSamplerDescriptor() { reset(); }
@@ -628,6 +625,7 @@ class MVKTexelBufferDescriptor : public MVKDescriptor {
 
 public:
 	void bind(MVKCommandEncoder* cmdEncoder,
+			  VkPipelineBindPoint pipelineBindPoint,
 			  MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			  uint32_t elementIndex,
 			  bool stages[],
@@ -635,19 +633,11 @@ public:
 			  MVKArrayRef<uint32_t> dynamicOffsets,
 			  uint32_t& dynamicOffsetIndex) override;
 
-	void encodeToMetalArgumentBuffer(MVKResourcesCommandEncoderState* rezEncState,
-									 id<MTLArgumentEncoder> mtlArgEncoder,
-									 uint32_t descSetIndex,
-									 MVKDescriptorSetLayoutBinding* mvkDSLBind,
-									 uint32_t elementIndex,
-									 MVKShaderStage stage,
-									 bool encodeToArgBuffer,
-									 bool encodeUsage) override;
-
 	void write(MVKDescriptorSetLayoutBinding* mvkDSLBind,
 			   MVKDescriptorSet* mvkDescSet,
-			   uint32_t srcIndex,
-			   size_t stride,
+			   uint32_t dstIdx,
+			   uint32_t srcIdx,
+			   size_t srcStride,
 			   const void* pData) override;
 
 	void read(MVKDescriptorSetLayoutBinding* mvkDSLBind,
@@ -658,6 +648,10 @@ public:
 			  VkBufferView* pTexelBufferView,
 			  VkWriteDescriptorSetInlineUniformBlockEXT* inlineUniformBlock) override;
 
+	void encodeResourceUsage(MVKResourcesCommandEncoderState* rezEncState,
+							 MVKDescriptorSetLayoutBinding* mvkDSLBind,
+							 MVKShaderStage stage) override;
+
 	void reset() override;
 
 	~MVKTexelBufferDescriptor() { reset(); }
@@ -683,3 +677,17 @@ class MVKStorageTexelBufferDescriptor : public MVKTexelBufferDescriptor {
 public:
 	VkDescriptorType getDescriptorType() override { return VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER; }
 };
+
+
+#pragma mark -
+#pragma mark Support functions
+
+/** 
+ * If the binding defines a buffer type, returns whether there are buffers, and
+ * therefore an auxilliary buffer is required to hold the lengths of those buffers.
+ * Returns false if the binding does not define a buffer type.
+ */
+bool mvkNeedsBuffSizeAuxBuffer(const VkDescriptorSetLayoutBinding* pBinding);
+
+/** Returns the name of the descriptor type. */
+const char* mvkVkDescriptorTypeName(VkDescriptorType vkDescType);
