@@ -1,7 +1,7 @@
 /*
  * MVKImage.h
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -31,6 +31,7 @@
 class MVKImage;
 class MVKImageView;
 class MVKSwapchain;
+class MVKQueue;
 class MVKCommandEncoder;
 
 
@@ -73,9 +74,7 @@ protected:
 	bool overlaps(VkSubresourceLayout& imgLayout, VkDeviceSize offset, VkDeviceSize size);
     void propagateDebugName();
     MVKImageMemoryBinding* getMemoryBinding() const;
-	void applyImageMemoryBarrier(VkPipelineStageFlags srcStageMask,
-								 VkPipelineStageFlags dstStageMask,
-								 MVKPipelineBarrier& barrier,
+	void applyImageMemoryBarrier(MVKPipelineBarrier& barrier,
 								 MVKCommandEncoder* cmdEncoder,
 								 MVKCommandUse cmdUse);
 	void pullFromDeviceOnCompletion(MVKCommandEncoder* cmdEncoder,
@@ -118,9 +117,7 @@ public:
     VkResult bindDeviceMemory(MVKDeviceMemory* mvkMem, VkDeviceSize memOffset) override;
 
     /** Applies the specified global memory barrier. */
-    void applyMemoryBarrier(VkPipelineStageFlags srcStageMask,
-                            VkPipelineStageFlags dstStageMask,
-                            MVKPipelineBarrier& barrier,
+    void applyMemoryBarrier(MVKPipelineBarrier& barrier,
                             MVKCommandEncoder* cmdEncoder,
                             MVKCommandUse cmdUse) override;
 
@@ -132,9 +129,7 @@ protected:
     friend MVKImage;
 
     void propagateDebugName() override;
-    bool needsHostReadSync(VkPipelineStageFlags srcStageMask,
-                           VkPipelineStageFlags dstStageMask,
-                           MVKPipelineBarrier& barrier);
+    bool needsHostReadSync(MVKPipelineBarrier& barrier);
     bool shouldFlushHostMemory();
     VkResult flushToDevice(VkDeviceSize offset, VkDeviceSize size);
     VkResult pullFromDevice(VkDeviceSize offset, VkDeviceSize size);
@@ -224,10 +219,17 @@ public:
     /** Returns the number of planes of this image view. */
     uint8_t getPlaneCount() { return _planes.size(); }
 
+	/** Returns whether or not the image format requires swizzling. */
+	bool needsSwizzle() { return getPixelFormats()->needsSwizzle(_vkFormat); }
+
 	/** Populates the specified layout for the specified sub-resource. */
 	VkResult getSubresourceLayout(const VkImageSubresource* pSubresource,
 								  VkSubresourceLayout* pLayout);
 
+	/** Populates the specified layout for the specified sub-resource. */
+	VkResult getSubresourceLayout(const VkImageSubresource2KHR* pSubresource,
+								  VkSubresourceLayout2KHR* pLayout);
+
     /** Populates the specified transfer image descriptor data structure. */
     void getTransferDescriptorData(MVKImageDescriptorData& imgData);
 
@@ -247,15 +249,23 @@ public:
 	virtual VkResult bindDeviceMemory2(const VkBindImageMemoryInfo* pBindInfo);
 
 	/** Applies the specified image memory barrier. */
-	void applyImageMemoryBarrier(VkPipelineStageFlags srcStageMask,
-								 VkPipelineStageFlags dstStageMask,
-								 MVKPipelineBarrier& barrier,
+	void applyImageMemoryBarrier(MVKPipelineBarrier& barrier,
 								 MVKCommandEncoder* cmdEncoder,
 								 MVKCommandUse cmdUse);
 
     /** Flush underlying buffer memory into the image if necessary */
     void flushToDevice(VkDeviceSize offset, VkDeviceSize size);
 
+	/** Host-copy the content of an image to another using the CPU. */
+	static VkResult copyImageToImage(const VkCopyImageToImageInfoEXT* pCopyImageToImageInfo);
+
+	/** Host-copy the content of an image to memory using the CPU. */
+	VkResult copyImageToMemory(const VkCopyImageToMemoryInfoEXT* pCopyImageToMemoryInfo);
+
+	/** Host-copy the content of an image from memory using the CPU. */
+	VkResult copyMemoryToImage(const VkCopyMemoryToImageInfoEXT* pCopyMemoryToImageInfo);
+
+
 #pragma mark Metal
 
 	/** Returns the Metal texture underlying this image. */
@@ -314,7 +324,10 @@ public:
      * attempting to copy a depth image with a substituted format to and from a buffer.
      */
     inline bool hasExpectedTexelSize() { return _hasExpectedTexelSize; }
-    
+
+	/** Returns whether the texture has the PixelFormatView usage flag, allowing it to be reinterpreted. */
+	inline bool hasPixelFormatView(uint32_t planeIndex) { return mvkIsAnyFlagEnabled(getMTLTexture(planeIndex).usage, MTLTextureUsagePixelFormatView); }
+
 	/** Returns the Metal resource options for this image. */
     MTLStorageMode getMTLStorageMode();
 
@@ -346,6 +359,16 @@ protected:
 	bool getIsValidViewFormat(VkFormat viewFormat);
 	VkImageUsageFlags getCombinedUsage() { return _usage | _stencilUsage; }
 	MTLTextureUsage getMTLTextureUsage(MTLPixelFormat mtlPixFmt);
+	uint8_t getMemoryBindingCount() const { return (uint8_t)_memoryBindings.size(); }
+	uint8_t getMemoryBindingIndex(uint8_t planeIndex) const;
+	MVKImageMemoryBinding* getMemoryBinding(uint8_t planeIndex);
+	template<typename CopyInfo> VkResult copyContent(const CopyInfo* pCopyInfo);
+	VkResult copyContent(id<MTLTexture> mtlTex,
+						 VkMemoryToImageCopyEXT imgRgn, uint32_t mipLevel, uint32_t slice,
+						 void* pImgBytes, size_t rowPitch, size_t depthPitch);
+	VkResult copyContent(id<MTLTexture> mtlTex,
+						 VkImageToMemoryCopyEXT imgRgn, uint32_t mipLevel, uint32_t slice,
+						 void* pImgBytes, size_t rowPitch, size_t depthPitch);
 
     MVKSmallVector<MVKImageMemoryBinding*, 3> _memoryBindings;
     MVKSmallVector<MVKImagePlane*, 3> _planes;
@@ -370,6 +393,7 @@ protected:
 	bool _isAliasable;
 	bool _hasExtendedUsage;
 	bool _hasMutableFormat;
+	bool _shouldSupportAtomics;
 	bool _isLinearForAtomics;
 };
 
@@ -382,30 +406,24 @@ class MVKSwapchainImage : public MVKImage {
 
 public:
 
-	/** Binds this resource to the specified offset within the specified memory allocation. */
 	VkResult bindDeviceMemory(MVKDeviceMemory* mvkMem, VkDeviceSize memOffset, uint8_t planeIndex) override;
 
-#pragma mark Metal
-
-	/** Returns the Metal texture used by the CAMetalDrawable underlying this image. */
-	id<MTLTexture> getMTLTexture(uint8_t planeIndex) override;
-
 
 #pragma mark Construction
 
-	/** Constructs an instance for the specified device and swapchain. */
 	MVKSwapchainImage(MVKDevice* device,
 					  const VkImageCreateInfo* pCreateInfo,
 					  MVKSwapchain* swapchain,
 					  uint32_t swapchainIndex);
 
-	~MVKSwapchainImage() override;
+	void destroy() override;
 
 protected:
 	friend class MVKPeerSwapchainImage;
 
-	virtual id<CAMetalDrawable> getCAMetalDrawable() = 0;
+	void detachSwapchain();
 
+	std::mutex _detachmentLock;
 	MVKSwapchain* _swapchain;
 	uint32_t _swapchainIndex;
 };
@@ -422,13 +440,16 @@ typedef struct MVKSwapchainImageAvailability {
 	bool operator< (const MVKSwapchainImageAvailability& rhs) const;
 } MVKSwapchainImageAvailability;
 
-/** VK_GOOGLE_display_timing extension info */
+/** Presentation info. */
 typedef struct  {
 	MVKPresentableSwapchainImage* presentableImage;
-	bool hasPresentTime;      		// Keep track of whether presentation included VK_GOOGLE_display_timing
-	uint32_t presentID;           	// VK_GOOGLE_display_timing presentID
+	MVKQueue* queue;				// The queue on which the vkQueuePresentKHR() command was executed.
+	MVKFence* fence;				// VK_EXT_swapchain_maintenance1 fence signaled when resources can be destroyed
 	uint64_t desiredPresentTime;  	// VK_GOOGLE_display_timing desired presentation time in nanoseconds
-} MVKPresentTimingInfo;
+	uint32_t presentID;           	// VK_GOOGLE_display_timing presentID
+	VkPresentModeKHR presentMode;	// VK_EXT_swapchain_maintenance1 present mode specialization
+	bool hasPresentTime;      		// Keep track of whether presentation included VK_GOOGLE_display_timing
+} MVKImagePresentInfo;
 
 /** Tracks a semaphore and fence for later signaling. */
 struct MVKSwapchainSignaler {
@@ -445,42 +466,46 @@ public:
 
 #pragma mark Metal
 
-	/** Presents the contained drawable to the OS. */
-	void presentCAMetalDrawable(id<MTLCommandBuffer> mtlCmdBuff,
-								MVKPresentTimingInfo presentTimingInfo);
+	id<MTLTexture> getMTLTexture(uint8_t planeIndex) override;
 
+	/** Presents the contained drawable to the OS. */
+	VkResult presentCAMetalDrawable(id<MTLCommandBuffer> mtlCmdBuff, MVKImagePresentInfo presentInfo);
+
+	/** Called when the presentation begins. */
+	void beginPresentation(const MVKImagePresentInfo& presentInfo);
+
+	/** Called via callback when the presentation completes. */
+	void endPresentation(const MVKImagePresentInfo& presentInfo,
+						 const MVKSwapchainSignaler& signaler,
+						 uint64_t actualPresentTime = 0);
 
 #pragma mark Construction
 
-	/** Constructs an instance for the specified device and swapchain. */
-	MVKPresentableSwapchainImage(MVKDevice* device,
-								 const VkImageCreateInfo* pCreateInfo,
-								 MVKSwapchain* swapchain,
-								 uint32_t swapchainIndex);
+	MVKPresentableSwapchainImage(MVKDevice* device, const VkImageCreateInfo* pCreateInfo,
+								 MVKSwapchain* swapchain, uint32_t swapchainIndex);
+
+	void destroy() override;
 
 	~MVKPresentableSwapchainImage() override;
 
 protected:
 	friend MVKSwapchain;
 
-	id<CAMetalDrawable> getCAMetalDrawable() override;
-	void presentCAMetalDrawable(id<CAMetalDrawable> mtlDrawable, MVKPresentTimingInfo presentTimingInfo);
+	id<CAMetalDrawable> getCAMetalDrawable();
+	void addPresentedHandler(id<CAMetalDrawable> mtlDrawable, MVKImagePresentInfo presentInfo, MVKSwapchainSignaler signaler);
 	void releaseMetalDrawable();
 	MVKSwapchainImageAvailability getAvailability();
-	void makeAvailable(const MVKSwapchainSignaler& signaler);
-	void acquireAndSignalWhenAvailable(MVKSemaphore* semaphore, MVKFence* fence);
-	void signal(const MVKSwapchainSignaler& signaler, id<MTLCommandBuffer> mtlCmdBuff);
-	void signalPresentationSemaphore(const MVKSwapchainSignaler& signaler, id<MTLCommandBuffer> mtlCmdBuff);
-	static void markAsTracked(const MVKSwapchainSignaler& signaler);
-	static void unmarkAsTracked(const MVKSwapchainSignaler& signaler);
-	void renderWatermark(id<MTLCommandBuffer> mtlCmdBuff);
+	void makeAvailable();
+	VkResult acquireAndSignalWhenAvailable(MVKSemaphore* semaphore, MVKFence* fence);
+	MVKSwapchainSignaler getPresentationSignaler();
 
-	id<CAMetalDrawable> _mtlDrawable;
-	id<MTLCommandBuffer> _presentingMTLCmdBuff;
+	id<CAMetalDrawable> _mtlDrawable = nil;
+	id<MTLTexture> _mtlTextureHeadless = nil;
 	MVKSwapchainImageAvailability _availability;
 	MVKSmallVector<MVKSwapchainSignaler, 1> _availabilitySignalers;
-	MVKSwapchainSignaler _preSignaler;
+	MVKSwapchainSignaler _preSignaler = {};
 	std::mutex _availabilityLock;
+	uint64_t _presentationStartTime = 0;
 };
 
 
@@ -492,21 +517,17 @@ class MVKPeerSwapchainImage : public MVKSwapchainImage {
 
 public:
 
-	/** Binds this resource according to the specified bind information. */
+	id<MTLTexture> getMTLTexture(uint8_t planeIndex) override;
+
 	VkResult bindDeviceMemory2(const VkBindImageMemoryInfo* pBindInfo) override;
 
 
 #pragma mark Construction
 
-	/** Constructs an instance for the specified device and swapchain. */
 	MVKPeerSwapchainImage(MVKDevice* device,
 						  const VkImageCreateInfo* pCreateInfo,
 						  MVKSwapchain* swapchain,
 						  uint32_t swapchainIndex);
-
-protected:
-	id<CAMetalDrawable> getCAMetalDrawable() override;
-
 };
 
 
@@ -565,6 +586,9 @@ public:
 	/**  Returns the 3D extent of this image at the specified mipmap level. */
 	VkExtent3D getExtent3D(uint8_t planeIndex = 0, uint32_t mipLevel = 0) { return _image->getExtent3D(planeIndex, mipLevel); }
 
+	/** Return the underlying image. */
+	MVKImage* getImage() { return _image; }
+
 #pragma mark Metal
 
 	/** Returns the Metal texture underlying this image view.  */
@@ -638,7 +662,7 @@ public:
     VkDebugReportObjectTypeEXT getVkDebugReportObjectType() override { return VK_DEBUG_REPORT_OBJECT_TYPE_SAMPLER_YCBCR_CONVERSION_EXT; }
     
     /** Returns the number of planes of this ycbcr conversion. */
-    inline uint8_t getPlaneCount() { return _planes; }
+    uint8_t getPlaneCount() { return _planes; }
 
     /** Writes this conversion settings to a MSL constant sampler */
     void updateConstExprSampler(SPIRV_CROSS_NAMESPACE::MSLConstexprSampler& constExprSampler) const;
@@ -676,10 +700,10 @@ public:
 	VkDebugReportObjectTypeEXT getVkDebugReportObjectType() override { return VK_DEBUG_REPORT_OBJECT_TYPE_SAMPLER_EXT; }
 
 	/** Returns the Metal sampler state. */
-	inline id<MTLSamplerState> getMTLSamplerState() { return _mtlSamplerState; }
+	id<MTLSamplerState> getMTLSamplerState() { return _mtlSamplerState; }
     
     /** Returns the number of planes if this is a ycbcr conversion or 0 otherwise. */
-    inline uint8_t getPlaneCount() { return (_ycbcrConversion) ? _ycbcrConversion->getPlaneCount() : 0; }
+    uint8_t getPlaneCount() { return (_ycbcrConversion) ? _ycbcrConversion->getPlaneCount() : 0; }
 
 	/**
 	 * If this sampler requires hardcoding in MSL, populates the hardcoded sampler in the resource binding.
