@@ -1,7 +1,7 @@
 /*
  * MVKPixelFormats.h
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -18,10 +18,10 @@
 
 #pragma once
 
-#include "mvk_datatypes.h"
-#include "MVKEnvironment.h"
-#include "MVKOSExtensions.h"
 #include "MVKBaseObject.h"
+#include "MVKOSExtensions.h"
+#include "MVKInflectionMap.h"
+#include "mvk_datatypes.hpp"
 #include <spirv_msl.hpp>
 #include <unordered_map>
 
@@ -30,14 +30,6 @@
 class MVKPhysicalDevice;
 
 
-// Validate these values periodically as new formats are added over time.
-static const uint32_t _vkFormatCount = 256;
-static const uint32_t _vkFormatCoreCount = VK_FORMAT_ASTC_12x12_SRGB_BLOCK + 1;
-static const uint32_t _mtlPixelFormatCount = 256;
-static const uint32_t _mtlPixelFormatCoreCount = MTLPixelFormatX32_Stencil8 + 2;     // The actual last enum value is not available on iOS
-static const uint32_t _mtlVertexFormatCount = MTLVertexFormatHalf + 1;
-
-
 #pragma mark -
 #pragma mark Metal format capabilities
 
@@ -148,17 +140,25 @@ typedef struct MVKVkFormatDesc {
 	VkExtent2D blockTexelSize;
 	uint32_t bytesPerBlock;
 	MVKFormatType formatType;
-	VkFormatProperties properties;
+    VkFormatProperties3 properties;
+	VkComponentMapping componentMapping;
 	const char* name;
 	bool hasReportedSubstitution;
     
     inline double bytesPerTexel() const { return (double)bytesPerBlock / (double)(blockTexelSize.width * blockTexelSize.height); };
 
-	inline bool isSupported() const { return (mtlPixelFormat != MTLPixelFormatInvalid || chromaSubsamplingPlaneCount > 0); };
+	inline bool isSupported() const { return (mtlPixelFormat != MTLPixelFormatInvalid || chromaSubsamplingPlaneCount > 1); };
 	inline bool isSupportedOrSubstitutable() const { return isSupported() || (mtlPixelFormatSubstitute != MTLPixelFormatInvalid); };
 
 	inline bool vertexIsSupported() const { return (mtlVertexFormat != MTLVertexFormatInvalid); };
 	inline bool vertexIsSupportedOrSubstitutable() const { return vertexIsSupported() || (mtlVertexFormatSubstitute != MTLVertexFormatInvalid); };
+
+	bool needsSwizzle() const {
+		return (componentMapping.r != VK_COMPONENT_SWIZZLE_IDENTITY ||
+				componentMapping.g != VK_COMPONENT_SWIZZLE_IDENTITY ||
+				componentMapping.b != VK_COMPONENT_SWIZZLE_IDENTITY ||
+				componentMapping.a != VK_COMPONENT_SWIZZLE_IDENTITY);
+	}
 } MVKVkFormatDesc;
 
 /** Describes the properties of a MTLPixelFormat or MTLVertexFormat. */
@@ -321,8 +321,24 @@ public:
 	 */
 	size_t getBytesPerLayer(MTLPixelFormat mtlFormat, size_t bytesPerRow, uint32_t texelRowsPerLayer);
 
+	/** Returns whether or not the specified Vulkan format requires swizzling to use with Metal. */
+	bool needsSwizzle(VkFormat vkFormat);
+
+	/** Returns any VkComponentMapping needed to use the specified Vulkan format. */
+	VkComponentMapping getVkComponentMapping(VkFormat vkFormat);
+
+	/**
+	 * Returns the inverse of the VkComponentMapping needed to use the specified Vulkan format.
+	 * If the original mapping is not a one-to-one function, the behaviour is undefined.
+	 */
+	VkComponentMapping getInverseComponentMapping(VkFormat vkFormat);
+
+	/** Returns any MTLTextureSwizzleChannels needed to use the specified Vulkan format. */
+	MTLTextureSwizzleChannels getMTLTextureSwizzleChannels(VkFormat vkFormat);
+
 	/** Returns the default properties for the specified Vulkan format. */
-	VkFormatProperties& getVkFormatProperties(VkFormat vkFormat);
+    VkFormatProperties getVkFormatProperties(VkFormat format);
+	VkFormatProperties3& getVkFormatProperties3(VkFormat vkFormat);
 
 	/** Returns the Metal format capabilities supported by the specified Vulkan format, without substitution. */
 	MVKMTLFmtCaps getCapabilities(VkFormat vkFormat, bool isExtended = false);
@@ -372,18 +388,19 @@ public:
 									   VkSampleCountFlagBits samples = VK_SAMPLE_COUNT_1_BIT,
                                        bool isLinear = false,
                                        bool needsReinterpretation = true,
-                                       bool isExtended = false);
+                                       bool isExtended = false,
+									   bool supportAtomics = false);
 
 	/** Enumerates all formats that support the given features, calling a specified function for each one. */
-	void enumerateSupportedFormats(VkFormatProperties properties, bool any, std::function<bool(VkFormat)> func);
+	void enumerateSupportedFormats(const VkFormatProperties3& properties, bool any, std::function<bool(VkFormat)> func);
 
 	/**
 	 * Returns the Metal MTLVertexFormat corresponding to the specified
 	 * Vulkan VkFormat as used as a vertex attribute format.
 	 */
 	MTLVertexFormat getMTLVertexFormat(VkFormat vkFormat);
-
-
+    
+    static VkFormatFeatureFlags convertFormatPropertiesFlagBits(VkFormatFeatureFlags2 flags);
 #pragma mark Construction
 
 	MVKPixelFormats(MVKPhysicalDevice* physicalDevice = nullptr);
@@ -393,13 +410,12 @@ protected:
 	MVKVkFormatDesc& getVkFormatDesc(MTLPixelFormat mtlFormat);
 	MVKMTLFormatDesc& getMTLPixelFormatDesc(MTLPixelFormat mtlFormat);
 	MVKMTLFormatDesc& getMTLVertexFormatDesc(MTLVertexFormat mtlFormat);
+	id<MTLDevice> getMTLDevice();
 	void initVkFormatCapabilities();
 	void initMTLPixelFormatCapabilities();
 	void initMTLVertexFormatCapabilities();
-	void buildMTLFormatMaps();
-	void buildVkFormatMaps();
-	void setFormatProperties(MVKVkFormatDesc& vkDesc);
-	void modifyMTLFormatCapabilities();
+	void buildVkFormatMaps(id<MTLDevice> mtlDevice);
+	void setFormatProperties(id<MTLDevice> mtlDevice, MVKVkFormatDesc& vkDesc);
 	void modifyMTLFormatCapabilities(id<MTLDevice> mtlDevice);
 	void addMTLPixelFormatCapabilities(id<MTLDevice> mtlDevice,
 									   MTLFeatureSet mtlFeatSet,
@@ -424,19 +440,7 @@ protected:
 										MVKMTLFmtCaps mtlFmtCaps);
 
 	MVKPhysicalDevice* _physicalDevice;
-	MVKVkFormatDesc _vkFormatDescriptions[_vkFormatCount];
-	MVKMTLFormatDesc _mtlPixelFormatDescriptions[_mtlPixelFormatCount];
-	MVKMTLFormatDesc _mtlVertexFormatDescriptions[_mtlVertexFormatCount];
-
-	// Vulkan core formats have small values and are mapped by simple lookup array.
-	// Vulkan extension formats have larger values and are mapped by a map.
-	uint16_t _vkFormatDescIndicesByVkFormatsCore[_vkFormatCoreCount];
-	std::unordered_map<uint32_t, uint32_t> _vkFormatDescIndicesByVkFormatsExt;
-
-	// Most Metal formats have small values and are mapped by simple lookup array.
-	// Outliers are mapped by a map.
-	uint16_t _mtlFormatDescIndicesByMTLPixelFormatsCore[_mtlPixelFormatCoreCount];
-	std::unordered_map<NSUInteger, uint32_t> _mtlFormatDescIndicesByMTLPixelFormatsExt;
-
-	uint16_t _mtlFormatDescIndicesByMTLVertexFormats[_mtlVertexFormatCount];
+	MVKInflectionMap<VkFormat, MVKVkFormatDesc, VK_FORMAT_ASTC_12x12_SRGB_BLOCK + 1> _vkFormatDescriptions;
+	MVKInflectionMap<uint16_t, MVKMTLFormatDesc, MTLPixelFormatX32_Stencil8 + 2> _mtlPixelFormatDescriptions;  // The actual last enum value is not available on iOS
+	MVKSmallVector<MVKMTLFormatDesc> _mtlVertexFormatDescriptions;
 };
