@@ -1,7 +1,7 @@
 /*
  * MVKSync.h
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -63,6 +63,9 @@ public:
 	/** Returns whether this instance is in a reserved state. */
 	bool isReserved();
 
+	/** Returns the number of outstanding reservations. */
+	uint32_t getReservationCount();
+
 	/**
 	 * Blocks processing on the current thread until any or all (depending on configuration) outstanding
      * reservations have been released, or until the specified timeout interval in nanoseconds expires.
@@ -89,20 +92,19 @@ public:
 	 * require a separate call to the release() function to cause the semaphore to stop blocking.
 	 */
     MVKSemaphoreImpl(bool waitAll = true, uint32_t reservationCount = 0)
-        : _shouldWaitAll(waitAll), _reservationCount(reservationCount) {}
+        : _reservationCount(reservationCount), _shouldWaitAll(waitAll) {}
 
-    /** Destructor. */
     ~MVKSemaphoreImpl();
 
 
 private:
 	bool operator()();
-    inline bool isClear() { return _reservationCount == 0; }    // Not thread-safe
+    bool isClear() { return _reservationCount == 0; }    // Not thread-safe
 
 	std::mutex _lock;
 	std::condition_variable _blocker;
-	bool _shouldWaitAll;
 	uint32_t _reservationCount;
+	bool _shouldWaitAll;
 };
 
 
@@ -183,13 +185,18 @@ public:
 	/** Returns whether this semaphore uses command encoding. */
 	virtual bool isUsingCommandEncoding() = 0;
 
+	/**
+	 * Returns the MTLSharedEvent underlying this Vulkan semaphore,
+	 * or nil if this semaphore is not underpinned by a MTLSharedEvent.
+	 */
+	virtual id<MTLSharedEvent> getMTLSharedEvent() { return nil; };
+
 	/** Encodes an operation to block command buffer operation until this semaphore is signaled. */
 	void encodeWait(id<MTLCommandBuffer> cmdBuff);
 
 	/** Encodes an operation to signal the semaphore. */
 	void encodeSignal(id<MTLCommandBuffer> cmdBuff);
 
-
 #pragma mark Construction
 
     MVKSemaphore(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo) : MVKVulkanAPIDeviceObject(device) {}
@@ -201,24 +208,29 @@ protected:
 
 
 #pragma mark -
-#pragma mark MVKSemaphoreMTLFence
+#pragma mark MVKSemaphoreSingleQueue
 
-/** An MVKSemaphore that uses MTLFence to provide synchronization. */
-class MVKSemaphoreMTLFence : public MVKSemaphore {
+/**
+ * An MVKSemaphore that uses Metal's built-in guarantees on single-queue submission to provide semaphore-like guarantees.
+ *
+ * Relies on Metal's enabled-by-default hazard tracking, and will need to start doing things with MTLFences
+ * if we start using things with MTLHazardTrackingModeUntracked
+ */
+class MVKSemaphoreSingleQueue : public MVKSemaphore {
 
 public:
 	void encodeWait(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) override;
 	void encodeSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) override;
 	uint64_t deferSignal() override;
 	void encodeDeferredSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) override;
-	bool isUsingCommandEncoding() override { return true; }
+	bool isUsingCommandEncoding() override { return false; }
 
-	MVKSemaphoreMTLFence(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo);
+	MVKSemaphoreSingleQueue(MVKDevice* device,
+	                        const VkSemaphoreCreateInfo* pCreateInfo,
+	                        const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+	                        const VkImportMetalSharedEventInfoEXT* pImportInfo);
 
-	~MVKSemaphoreMTLFence() override;
-
-protected:
-	id<MTLFence> _mtlFence;
+	~MVKSemaphoreSingleQueue() override;
 };
 
 
@@ -235,7 +247,10 @@ public:
 	void encodeDeferredSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t deferToken) override;
 	bool isUsingCommandEncoding() override { return true; }
 
-	MVKSemaphoreMTLEvent(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo);
+	MVKSemaphoreMTLEvent(MVKDevice* device,
+						 const VkSemaphoreCreateInfo* pCreateInfo,
+						 const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+						 const VkImportMetalSharedEventInfoEXT* pImportInfo);
 
 	~MVKSemaphoreMTLEvent() override;
 
@@ -258,7 +273,10 @@ public:
 	void encodeDeferredSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) override;
 	bool isUsingCommandEncoding() override { return false; }
 
-	MVKSemaphoreEmulated(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo);
+	MVKSemaphoreEmulated(MVKDevice* device,
+						 const VkSemaphoreCreateInfo* pCreateInfo,
+						 const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+						 const VkImportMetalSharedEventInfoEXT* pImportInfo);
 
 protected:
 	MVKSemaphoreImpl _blocker;
@@ -298,8 +316,7 @@ public:
 
 #pragma mark Construction
 
-    MVKTimelineSemaphore(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo, const VkSemaphoreTypeCreateInfo* pTypeCreateInfo)
-        : MVKSemaphore(device, pCreateInfo) {}
+	MVKTimelineSemaphore(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo) : MVKSemaphore(device, pCreateInfo) {}
 
 };
 
@@ -314,18 +331,23 @@ public:
 	void encodeWait(id<MTLCommandBuffer> mtlCmdBuff, uint64_t value) override;
 	void encodeSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t value) override;
 	bool isUsingCommandEncoding() override { return true; }
+	id<MTLSharedEvent> getMTLSharedEvent() override { return _mtlEvent; };
 
 	uint64_t getCounterValue() override { return _mtlEvent.signaledValue; }
 	void signal(const VkSemaphoreSignalInfo* pSignalInfo) override;
 	bool registerWait(MVKFenceSitter* sitter, const VkSemaphoreWaitInfo* pWaitInfo, uint32_t index) override;
 	void unregisterWait(MVKFenceSitter* sitter) override;
 
-	MVKTimelineSemaphoreMTLEvent(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo, const VkSemaphoreTypeCreateInfo* pTypeCreateInfo);
+	MVKTimelineSemaphoreMTLEvent(MVKDevice* device,
+								 const VkSemaphoreCreateInfo* pCreateInfo,
+								 const VkSemaphoreTypeCreateInfo* pTypeCreateInfo,
+								 const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+								 const VkImportMetalSharedEventInfoEXT* pImportInfo);
 
 	~MVKTimelineSemaphoreMTLEvent() override;
 
 protected:
-	id<MTLSharedEvent> _mtlEvent;
+	id<MTLSharedEvent> _mtlEvent = nil;
 	std::mutex _lock;
 	std::unordered_set<MVKFenceSitter*> _sitters;
 };
@@ -347,7 +369,11 @@ public:
 	bool registerWait(MVKFenceSitter* sitter, const VkSemaphoreWaitInfo* pWaitInfo, uint32_t index) override;
 	void unregisterWait(MVKFenceSitter* sitter) override;
 
-	MVKTimelineSemaphoreEmulated(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo, const VkSemaphoreTypeCreateInfo* pTypeCreateInfo);
+	MVKTimelineSemaphoreEmulated(MVKDevice* device,
+								 const VkSemaphoreCreateInfo* pCreateInfo,
+								 const VkSemaphoreTypeCreateInfo* pTypeCreateInfo,
+								 const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+								 const VkImportMetalSharedEventInfoEXT* pImportInfo);
 
 protected:
 	void signalImpl(uint64_t value);
@@ -442,6 +468,8 @@ public:
 
 	MVKFenceSitter(bool waitAll) : _blocker(waitAll, 0) {}
 
+	~MVKFenceSitter() override { [_listener release]; }
+
 private:
 	friend class MVKFence;
 	friend class MVKTimelineSemaphoreMTLEvent;
@@ -483,10 +511,16 @@ public:
 	/** Encodes an operation to block command buffer operation until this event is signaled. */
 	virtual void encodeWait(id<MTLCommandBuffer> mtlCmdBuff) = 0;
 
+	/** Returns the MTLSharedEvent underlying this Vulkan event. */
+	virtual id<MTLSharedEvent> getMTLSharedEvent() = 0;
+
 
 #pragma mark Construction
 
-	MVKEvent(MVKDevice* device, const VkEventCreateInfo* pCreateInfo) : MVKVulkanAPIDeviceObject(device) {}
+	MVKEvent(MVKDevice* device,
+			 const VkEventCreateInfo* pCreateInfo,
+			 const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+			 const VkImportMetalSharedEventInfoEXT* pImportInfo) : MVKVulkanAPIDeviceObject(device) {}
 
 protected:
 	void propagateDebugName() override {}
@@ -505,8 +539,12 @@ public:
 	void signal(bool status) override;
 	void encodeSignal(id<MTLCommandBuffer> mtlCmdBuff, bool status) override;
 	void encodeWait(id<MTLCommandBuffer> mtlCmdBuff) override;
+	id<MTLSharedEvent> getMTLSharedEvent() override { return _mtlEvent; };
 
-	MVKEventNative(MVKDevice* device, const VkEventCreateInfo* pCreateInfo);
+	MVKEventNative(MVKDevice* device,
+				   const VkEventCreateInfo* pCreateInfo,
+				   const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+				   const VkImportMetalSharedEventInfoEXT* pImportInfo);
 
 	~MVKEventNative() override;
 
@@ -526,8 +564,12 @@ public:
 	void signal(bool status) override;
 	void encodeSignal(id<MTLCommandBuffer> mtlCmdBuff, bool status) override;
 	void encodeWait(id<MTLCommandBuffer> mtlCmdBuff) override;
+	id<MTLSharedEvent> getMTLSharedEvent() override { return nil; };
 
-	MVKEventEmulated(MVKDevice* device, const VkEventCreateInfo* pCreateInfo);
+	MVKEventEmulated(MVKDevice* device,
+					 const VkEventCreateInfo* pCreateInfo,
+					 const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+					 const VkImportMetalSharedEventInfoEXT* pImportInfo);
 
 protected:
 	MVKSemaphoreImpl _blocker;
@@ -570,7 +612,7 @@ VkResult mvkWaitSemaphores(MVKDevice* device,
  *
  * Instances of this class are one-shot, and can only be used for a single compilation.
  */
-class MVKMetalCompiler : public MVKBaseObject {
+class MVKMetalCompiler : public MVKBaseDeviceObject {
 
 public:
 
@@ -583,7 +625,7 @@ public:
 
 #pragma mark Construction
 
-	MVKMetalCompiler(MVKVulkanAPIDeviceObject* owner) : _owner(owner) {}
+	MVKMetalCompiler(MVKVulkanAPIDeviceObject* owner) : MVKBaseDeviceObject(owner->getDevice()), _owner(owner) {}
 
 	~MVKMetalCompiler() override;
 
@@ -603,3 +645,89 @@ protected:
 	std::string _compilerType = "Unknown";
 	MVKPerformanceTracker* _pPerformanceTracker = nullptr;
 };
+
+#pragma mark -
+#pragma mark MVKDeferredOperation
+
+/** The maximum number of function parameters for all functions added to MVKDeferredOperationFunctionPointer. */
+static const int kMVKMaxDeferredFunctionParameters = 3;
+
+/** Defines the function pointer for each dependent function. */
+union MVKDeferredOperationFunctionPointer {
+    // Empty until deferred functions from other extensions have been defined.
+	// Planning to use std::functions
+	// Each function should take a pointer to a MVKDeferredOperation instance, to allow
+	// the function to call setMaxConcurrency() and setOperationResult(), and should
+	// return a VkResult representing the value returned by vkDeferredOperationJoinKHR().
+};
+
+/** Indicates what kind of function is being deferred. */
+enum MVKDeferredOperationFunctionType {
+    // Empty until deferred functions from other extensions have been defined
+};
+
+/**
+ * Holds and invokes a deferred operation.
+ *
+ * Deferred operations are added by as required by future extensions that require them,
+ * and are implemented by each of those extensions as a function pointer added to
+ * MVKDeferredOperationFunctionPointer, plus a corresponding type identifier added
+ * to MVKDeferredOperationFunctionType.
+ */
+class MVKDeferredOperation : public MVKVulkanAPIDeviceObject {
+public:
+    VkObjectType getVkObjectType() override { return VK_OBJECT_TYPE_DEFERRED_OPERATION_KHR; }
+    
+    VkDebugReportObjectTypeEXT getVkDebugReportObjectType() override { return VK_DEBUG_REPORT_OBJECT_TYPE_UNKNOWN_EXT; }
+    
+    /** Begins executing the deferred operation on the current thread. */
+    VkResult join();
+
+	/** Gets the result of the execution of the deferred operation */
+	VkResult getOperationResult();
+
+	/**
+	 * Sets the result of the execution of the deferred operation.
+	 * Called from the function that is implementing the deferred operation.
+	 */
+	void setOperationResult(VkResult opResult);
+
+    /** Gets the max number of threads that can execute the deferred operation concurrently. */
+	uint32_t getMaxConcurrency();
+
+	/**
+	 * Sets the max number of threads that can execute the deferred operation concurrently.
+	 * Called from the function that is implementing the deferred operation.
+	 */
+	void setMaxConcurrency(uint32_t maxConCurr);
+
+	/**
+	 * Sets the result of the execution of the deferred operation, and updates the max concurrently.
+	 * Equivalent to calling setOperationResult() and setMaxConcurrency(), but with a single concurrency mutex lock.
+	 * Called from the function that is implementing the deferred operation.
+	 */
+	void updateResults(VkResult opResult, uint32_t maxConCurr);
+
+    /**
+	 * Sets all the variables needed for a deferred operation.
+	 * Called from the code that implements a deferred operation.
+	 */
+    void deferOperation(const MVKDeferredOperationFunctionPointer& pointer,
+						MVKDeferredOperationFunctionType type,
+						void** parameters,
+						uint32_t paramCount);
+
+#pragma mark Construction
+    MVKDeferredOperation(MVKDevice* device) : MVKVulkanAPIDeviceObject(device) {}
+
+protected:
+	void propagateDebugName() override {}
+
+	std::mutex _lock;
+	MVKSmallVector<void*, kMVKMaxDeferredFunctionParameters>  _functionParameters;
+	MVKDeferredOperationFunctionPointer _functionPointer = {};
+	MVKDeferredOperationFunctionType _functionType;
+	VkResult _operationResult = VK_SUCCESS;
+    uint32_t _maxConcurrency = 0;
+
+};
