@@ -1,7 +1,7 @@
 /*
  * MVKSync.mm
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -18,6 +18,7 @@
 
 #include "MVKSync.h"
 #include "MVKFoundation.h"
+#include "MVKInstance.h"
 
 using namespace std;
 
@@ -50,6 +51,11 @@ bool MVKSemaphoreImpl::isReserved() {
 	return !isClear();
 }
 
+uint32_t MVKSemaphoreImpl::getReservationCount() {
+	lock_guard<mutex> lock(_lock);
+	return _reservationCount;
+}
+
 bool MVKSemaphoreImpl::wait(uint64_t timeout, bool reserveAgain) {
     unique_lock<mutex> lock(_lock);
 
@@ -77,40 +83,35 @@ MVKSemaphoreImpl::~MVKSemaphoreImpl() {
 
 
 #pragma mark -
-#pragma mark MVKSemaphoreMTLFence
+#pragma mark MVKSemaphoreSingleQueue
 
-// Could use any encoder. Assume BLIT is fastest and lightest.
-// Nil mtlCmdBuff will do nothing.
-void MVKSemaphoreMTLFence::encodeWait(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) {
-	id<MTLBlitCommandEncoder> mtlCmdEnc = mtlCmdBuff.blitCommandEncoder;
-	[mtlCmdEnc waitForFence: _mtlFence];
-	[mtlCmdEnc endEncoding];
+void MVKSemaphoreSingleQueue::encodeWait(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) {
+	// Metal will handle all synchronization for us automatically
 }
 
-// Could use any encoder. Assume BLIT is fastest and lightest.
-// Nil mtlCmdBuff will do nothing.
-void MVKSemaphoreMTLFence::encodeSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) {
-	id<MTLBlitCommandEncoder> mtlCmdEnc = mtlCmdBuff.blitCommandEncoder;
-	[mtlCmdEnc updateFence: _mtlFence];
-	[mtlCmdEnc endEncoding];
+void MVKSemaphoreSingleQueue::encodeSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) {
+	// Metal will handle all synchronization for us automatically
 }
 
-uint64_t MVKSemaphoreMTLFence::deferSignal() {
+uint64_t MVKSemaphoreSingleQueue::deferSignal() {
 	return 0;
 }
 
-void MVKSemaphoreMTLFence::encodeDeferredSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) {
+void MVKSemaphoreSingleQueue::encodeDeferredSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t) {
 	encodeSignal(mtlCmdBuff, 0);
 }
 
-MVKSemaphoreMTLFence::MVKSemaphoreMTLFence(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo) :
-	MVKSemaphore(device, pCreateInfo),
-	_mtlFence([device->getMTLDevice() newFence]) {}		//retained
-
-MVKSemaphoreMTLFence::~MVKSemaphoreMTLFence() {
-	[_mtlFence release];
+MVKSemaphoreSingleQueue::MVKSemaphoreSingleQueue(MVKDevice* device,
+                                                 const VkSemaphoreCreateInfo* pCreateInfo,
+                                                 const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+                                                 const VkImportMetalSharedEventInfoEXT* pImportInfo) : MVKSemaphore(device, pCreateInfo) {
+	if ((pImportInfo && pImportInfo->mtlSharedEvent) || (pExportInfo && pExportInfo->exportObjectType == VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT)) {
+		setConfigurationResult(reportError(VK_ERROR_INITIALIZATION_FAILED, "vkCreateEvent(): MTLSharedEvent is not available with VkSemaphores that use implicit synchronization."));
+	}
 }
 
+MVKSemaphoreSingleQueue::~MVKSemaphoreSingleQueue() = default;
+
 
 #pragma mark -
 #pragma mark MVKSemaphoreMTLEvent
@@ -128,13 +129,26 @@ uint64_t MVKSemaphoreMTLEvent::deferSignal() {
 }
 
 void MVKSemaphoreMTLEvent::encodeDeferredSignal(id<MTLCommandBuffer> mtlCmdBuff, uint64_t deferToken) {
-	if (mtlCmdBuff) { [mtlCmdBuff encodeSignalEvent: _mtlEvent value: deferToken]; }
+	[mtlCmdBuff encodeSignalEvent: _mtlEvent value: deferToken];
 }
 
-MVKSemaphoreMTLEvent::MVKSemaphoreMTLEvent(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo) :
-	MVKSemaphore(device, pCreateInfo),
-	_mtlEvent([device->getMTLDevice() newEvent]),	//retained
-	_mtlEventValue(1) {}
+MVKSemaphoreMTLEvent::MVKSemaphoreMTLEvent(MVKDevice* device,
+										   const VkSemaphoreCreateInfo* pCreateInfo,
+										   const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+										   const VkImportMetalSharedEventInfoEXT* pImportInfo) : MVKSemaphore(device, pCreateInfo) {
+	// In order of preference, import a MTLSharedEvent,
+	// create a MTLSharedEvent, or create a MTLEvent.
+	if (pImportInfo && pImportInfo->mtlSharedEvent) {
+		_mtlEvent = [pImportInfo->mtlSharedEvent retain];		// retained
+		_mtlEventValue = pImportInfo->mtlSharedEvent.signaledValue + 1;
+	} else if (pExportInfo && pExportInfo->exportObjectType == VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT) {
+		_mtlEvent = [getMTLDevice() newSharedEvent];	//retained
+		_mtlEventValue = ((id<MTLSharedEvent>)_mtlEvent).signaledValue + 1;
+	} else {
+		_mtlEvent = [getMTLDevice() newEvent];			//retained
+		_mtlEventValue = 1;
+	}
+}
 
 MVKSemaphoreMTLEvent::~MVKSemaphoreMTLEvent() {
     [_mtlEvent release];
@@ -164,9 +178,17 @@ void MVKSemaphoreEmulated::encodeDeferredSignal(id<MTLCommandBuffer> mtlCmdBuff,
 	encodeSignal(mtlCmdBuff, 0);
 }
 
-MVKSemaphoreEmulated::MVKSemaphoreEmulated(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo) :
+MVKSemaphoreEmulated::MVKSemaphoreEmulated(MVKDevice* device,
+										   const VkSemaphoreCreateInfo* pCreateInfo,
+										   const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+										   const VkImportMetalSharedEventInfoEXT* pImportInfo) :
 	MVKSemaphore(device, pCreateInfo),
-	_blocker(false, 1) {}
+	_blocker(false, 1) {
+
+	if ((pImportInfo && pImportInfo->mtlSharedEvent) || (pExportInfo && pExportInfo->exportObjectType == VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT)) {
+		setConfigurationResult(reportError(VK_ERROR_INITIALIZATION_FAILED, "vkCreateEvent(): MTLSharedEvent is not available with VkSemaphores that use CPU emulation."));
+	}
+}
 
 
 #pragma mark -
@@ -211,11 +233,20 @@ void MVKTimelineSemaphoreMTLEvent::unregisterWait(MVKFenceSitter* sitter) {
 	_sitters.erase(sitter);
 }
 
-MVKTimelineSemaphoreMTLEvent::MVKTimelineSemaphoreMTLEvent(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo, const VkSemaphoreTypeCreateInfo* pTypeCreateInfo) :
-	MVKTimelineSemaphore(device, pCreateInfo, pTypeCreateInfo),
-	_mtlEvent([device->getMTLDevice() newSharedEvent]) { 	//retained
+MVKTimelineSemaphoreMTLEvent::MVKTimelineSemaphoreMTLEvent(MVKDevice* device,
+														   const VkSemaphoreCreateInfo* pCreateInfo,
+														   const VkSemaphoreTypeCreateInfo* pTypeCreateInfo,
+														   const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+														   const VkImportMetalSharedEventInfoEXT* pImportInfo) : MVKTimelineSemaphore(device, pCreateInfo) {
 
-	_mtlEvent.signaledValue = pTypeCreateInfo->initialValue;
+	// Import or create a Metal event
+	_mtlEvent = (pImportInfo && pImportInfo->mtlSharedEvent
+				 ? [pImportInfo->mtlSharedEvent retain]
+				 : [getMTLDevice() newSharedEvent]);	//retained
+
+	if (pTypeCreateInfo) {
+		_mtlEvent.signaledValue = pTypeCreateInfo->initialValue;
+	}
 }
 
 MVKTimelineSemaphoreMTLEvent::~MVKTimelineSemaphoreMTLEvent() {
@@ -284,9 +315,18 @@ void MVKTimelineSemaphoreEmulated::unregisterWait(MVKFenceSitter* sitter) {
 	for (auto value : emptySets) { _sitters.erase(value); }
 }
 
-MVKTimelineSemaphoreEmulated::MVKTimelineSemaphoreEmulated(MVKDevice* device, const VkSemaphoreCreateInfo* pCreateInfo, const VkSemaphoreTypeCreateInfo* pTypeCreateInfo) :
-	MVKTimelineSemaphore(device, pCreateInfo, pTypeCreateInfo),
-	_value(pTypeCreateInfo->initialValue) {}
+MVKTimelineSemaphoreEmulated::MVKTimelineSemaphoreEmulated(MVKDevice* device,
+														   const VkSemaphoreCreateInfo* pCreateInfo,
+														   const VkSemaphoreTypeCreateInfo* pTypeCreateInfo,
+														   const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+														   const VkImportMetalSharedEventInfoEXT* pImportInfo) :
+	MVKTimelineSemaphore(device, pCreateInfo),
+	_value(pTypeCreateInfo ? pTypeCreateInfo->initialValue : 0) {
+
+	if ((pImportInfo && pImportInfo->mtlSharedEvent) || (pExportInfo && pExportInfo->exportObjectType == VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT)) {
+		setConfigurationResult(reportError(VK_ERROR_INITIALIZATION_FAILED, "vkCreateEvent(): MTLSharedEvent is not available on this platform."));
+	}
+}
 
 
 #pragma mark -
@@ -375,8 +415,16 @@ void MVKEventNative::encodeWait(id<MTLCommandBuffer> mtlCmdBuff) {
 	}
 }
 
-MVKEventNative::MVKEventNative(MVKDevice* device, const VkEventCreateInfo* pCreateInfo) : MVKEvent(device, pCreateInfo) {
-	_mtlEvent = [_device->getMTLDevice() newSharedEvent];	// retained
+MVKEventNative::MVKEventNative(MVKDevice* device,
+							   const VkEventCreateInfo* pCreateInfo,
+							   const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+							   const VkImportMetalSharedEventInfoEXT* pImportInfo) :
+	MVKEvent(device, pCreateInfo, pExportInfo, pImportInfo) {
+
+	// Import or create a Metal event
+	_mtlEvent = (pImportInfo
+				 ? [pImportInfo->mtlSharedEvent retain]
+				 : [getMTLDevice() newSharedEvent]);	//retained
 }
 
 MVKEventNative::~MVKEventNative() {
@@ -421,8 +469,16 @@ void MVKEventEmulated::encodeWait(id<MTLCommandBuffer> mtlCmdBuff) {
 	}
 }
 
-MVKEventEmulated::MVKEventEmulated(MVKDevice* device, const VkEventCreateInfo* pCreateInfo) :
-	MVKEvent(device, pCreateInfo), _blocker(false, 1), _inlineSignalStatus(false) {}
+MVKEventEmulated::MVKEventEmulated(MVKDevice* device,
+								   const VkEventCreateInfo* pCreateInfo,
+								   const VkExportMetalObjectCreateInfoEXT* pExportInfo,
+								   const VkImportMetalSharedEventInfoEXT* pImportInfo) :
+	MVKEvent(device, pCreateInfo, pExportInfo, pImportInfo), _blocker(false, 1), _inlineSignalStatus(false) {
+
+	if (pExportInfo && pExportInfo->exportObjectType == VK_EXPORT_METAL_OBJECT_TYPE_METAL_SHARED_EVENT_BIT_EXT) {
+		setConfigurationResult(reportError(VK_ERROR_INITIALIZATION_FAILED, "vkCreateEvent(): MTLSharedEvent is not available on this platform."));
+	}
+}
 
 
 #pragma mark -
@@ -515,28 +571,33 @@ VkResult mvkWaitSemaphores(MVKDevice* device,
 void MVKMetalCompiler::compile(unique_lock<mutex>& lock, dispatch_block_t block) {
 	MVKAssert( _startTime == 0, "%s compile occurred already in this instance. Instances of %s should only be used for a single compile activity.", _compilerType.c_str(), getClassName().c_str());
 
-	MVKDevice* mvkDev = _owner->getDevice();
-	_startTime = mvkDev->getPerformanceTimestamp();
+	_startTime = getPerformanceTimestamp();
 
 	dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{ @autoreleasepool { block(); } });
 
 	// Limit timeout to avoid overflow since wait_for() uses wait_until()
-	chrono::nanoseconds nanoTimeout(min(mvkConfig().metalCompileTimeout, kMVKUndefinedLargeUInt64));
+	chrono::nanoseconds nanoTimeout(min(getMVKConfig().metalCompileTimeout, kMVKUndefinedLargeUInt64));
 	_blocker.wait_for(lock, nanoTimeout, [this]{ return _isCompileDone; });
 
 	if ( !_isCompileDone ) {
 		@autoreleasepool {
 			NSString* errDesc = [NSString stringWithFormat: @"Timeout after %.3f milliseconds. Likely internal Metal compiler error", (double)nanoTimeout.count() / 1e6];
-			_compileError = [[NSError alloc] initWithDomain: @"MoltenVK" code: 1 userInfo: @{NSLocalizedDescriptionKey : errDesc}];	// retained
+			_compileError = [[NSError alloc] initWithDomain: @(kMVKMoltenVKDriverLayerName) code: 1 userInfo: @{NSLocalizedDescriptionKey : errDesc}];	// retained
 		}
 	}
 
 	if (_compileError) { handleError(); }
 
-	mvkDev->addActivityPerformance(*_pPerformanceTracker, _startTime);
+	addPerformanceInterval(*_pPerformanceTracker, _startTime);
 }
 
 void MVKMetalCompiler::handleError() {
+	MVKInstance* instance = _owner->getInstance();
+	if (instance && instance->getEngine() == MVKInstance::EngineID::Angle) {
+		MVKLogWarn("%s compile failed but Angle is special so we'll pretend everything is happy: %s",
+		           _compilerType.c_str(), _compileError.localizedDescription.UTF8String);
+		return;
+	}
 	_owner->setConfigurationResult(reportError(VK_ERROR_INITIALIZATION_FAILED,
 											   "%s compile failed (Error code %li):\n%s.",
 											   _compilerType.c_str(), (long)_compileError.code,
@@ -570,5 +631,57 @@ MVKMetalCompiler::~MVKMetalCompiler() {
 	[_compileError release];
 }
 
+#pragma mark -
+#pragma mark MVKDeferredOperation
 
+// Call appropriate function from MVKDeferredOperationFunctionPointer with parameters.
+// While executing, the function can call setOperationResult() and setMaxConcurrency()
+// to update status of the operation, and should return a VkResult that is returned here.
+VkResult MVKDeferredOperation::join() {
+	switch(_functionType) {
+		// .....
+        default: return VK_SUCCESS;
+    };
+}
+
+void MVKDeferredOperation::deferOperation(const MVKDeferredOperationFunctionPointer& pointer,
+										  MVKDeferredOperationFunctionType type,
+										  void** parameters,
+										  uint32_t paramCount) {
+    _functionPointer = pointer;
+    _functionType = type;
+
+	_functionParameters.reserve(paramCount);
+	for(int i = 0; i < paramCount; i++) {
+        _functionParameters.push_back(parameters[i]);
+    }
+
+	updateResults(VK_SUCCESS, mvkGetAvaliableCPUCores());
+}
+
+VkResult MVKDeferredOperation::getOperationResult() {
+	lock_guard<mutex> lock(_lock);
+	return _operationResult;
+}
+
+void MVKDeferredOperation::setOperationResult(VkResult opResult) {
+	lock_guard<mutex> lock(_lock);
+	_operationResult = opResult;
+}
+
+uint32_t MVKDeferredOperation::getMaxConcurrency() {
+	lock_guard<mutex> lock(_lock);
+	return _maxConcurrency;
+}
+
+void MVKDeferredOperation::setMaxConcurrency(uint32_t maxConCurr) {
+	lock_guard<mutex> lock(_lock);
+	_maxConcurrency = maxConCurr;
+}
+
+void MVKDeferredOperation::updateResults(VkResult opResult, uint32_t maxConCurr) {
+	lock_guard<mutex> lock(_lock);
+	_operationResult = opResult;
+	_maxConcurrency = maxConCurr;
+}
 
