@@ -1,7 +1,7 @@
 /*
  * MVKSmallVector.h
  *
- * Copyright (c) 2012-2022 Dr. Torsten Hans (hans@ipacs.de)
+ * Copyright (c) 2012-2024 Dr. Torsten Hans (hans@ipacs.de)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -62,6 +62,7 @@ class MVKSmallVectorImpl
   Allocator  alc;
   
 public:
+  using value_type = Type;
   class iterator
   {
     const MVKSmallVectorImpl *vector;
@@ -85,9 +86,9 @@ public:
       return *this;
     }
 
-    Type *operator->() { return &vector->alc.ptr[index]; }
-    Type &operator*()  { return  vector->alc.ptr[index]; }
-    operator Type*()   { return &vector->alc.ptr[index]; }
+    Type *operator->() const { return &vector->alc.ptr[index]; }
+    Type &operator*()  const { return  vector->alc.ptr[index]; }
+    operator Type*()   const { return &vector->alc.ptr[index]; }
 
     bool operator==( const iterator &it ) const { return vector == it.vector && index == it.index; }
     bool operator!=( const iterator &it ) const { return vector != it.vector || index != it.index; }
@@ -97,24 +98,24 @@ public:
     iterator& operator--()      {                 --index; return *this; }
     iterator  operator--( int ) { auto t = *this; --index; return t; }
 
-    iterator operator+ (const diff_type n)   { return iterator( index + n, *vector ); }
-    iterator& operator+= (const diff_type n) { index += n; return *this; }
-    iterator operator- (const diff_type n)   { return iterator( index - n, *vector ); }
-    iterator& operator-= (const diff_type n) { index -= n; return *this; }
+    iterator operator+ (const diff_type n)   const { return iterator( index + n, *vector ); }
+    iterator& operator+= (const diff_type n)       { index += n; return *this; }
+    iterator operator- (const diff_type n)   const { return iterator( index - n, *vector ); }
+    iterator& operator-= (const diff_type n)       { index -= n; return *this; }
 
-    diff_type operator- (const iterator& it) { return index - it.index; }
+    diff_type operator- (const iterator& it) const { return index - it.index; }
 
-    bool operator< (const iterator& it)  { return index < it.index; }
-    bool operator<= (const iterator& it) { return index <= it.index; }
-    bool operator> (const iterator& it)  { return index > it.index; }
-    bool operator>= (const iterator& it) { return index >= it.index; }
+    bool operator< (const iterator& it)  const { return index < it.index; }
+    bool operator<= (const iterator& it) const { return index <= it.index; }
+    bool operator> (const iterator& it)  const { return index > it.index; }
+    bool operator>= (const iterator& it) const { return index >= it.index; }
 
-    const Type &operator[]( const diff_type i ) const { return vector->alc.ptr[index + i]; }
-    Type &operator[]( const diff_type i )             { return vector->alc.ptr[index + i]; }
+    Type &operator[]( const diff_type i ) const { return vector->alc.ptr[index + i]; }
 
     bool   is_valid()     const { return index < vector->alc.size(); }
     size_t get_position() const { return index; }
   };
+  using reverse_iterator = std::reverse_iterator<iterator>;
 
 private:
   // this is the growth strategy -> adjust to your needs
@@ -293,12 +294,15 @@ public:
   iterator begin() const { return iterator( 0, *this ); }
   iterator end()   const { return iterator( alc.num_elements_used, *this ); }
 
-  const MVKArrayRef<Type> contents() const { return MVKArrayRef<Type>(data(), size()); }
-        MVKArrayRef<Type> contents()       { return MVKArrayRef<Type>(data(), size()); }
+  reverse_iterator rbegin() const { return reverse_iterator( end() ); }
+  reverse_iterator rend()   const { return reverse_iterator( begin() ); }
 
-  const Type &operator[]( const size_t i ) const { return alc[i]; }
+  MVKArrayRef<const Type> contents() const { return MVKArrayRef<const Type>(data(), size()); }
+  MVKArrayRef<      Type> contents()       { return MVKArrayRef<      Type>(data(), size()); }
+
+  const Type &operator[]( const size_t i ) const  { return alc[i]; }
         Type &operator[]( const size_t i )        { return alc[i]; }
-  const Type &at( const size_t i )         const { return alc[i]; }
+  const Type &at( const size_t i )         const  { return alc[i]; }
         Type &at( const size_t i )                { return alc[i]; }
   const Type &front()                      const  { return alc[0]; }
         Type &front()                             { return alc[0]; }
@@ -521,6 +525,7 @@ class MVKSmallVectorImpl<Type*, Allocator>
   Allocator  alc;
 
 public:
+  using value_type = Type*;
   class iterator
   {
     MVKSmallVectorImpl *vector;
@@ -572,6 +577,7 @@ public:
     bool   is_valid()     const { return index < vector->alc.size(); }
     size_t get_position() const { return index; }
   };
+  using reverse_iterator = std::reverse_iterator<iterator>;
 
 private:
   // this is the growth strategy -> adjust to your needs
@@ -728,6 +734,9 @@ public:
   iterator begin()        { return iterator( 0, *this ); }
   iterator end()          { return iterator( alc.num_elements_used, *this ); }
 
+  reverse_iterator rbegin()       { return reverse_iterator( end() ); }
+  reverse_iterator rend()         { return reverse_iterator( rbegin() ); }
+
   const MVKArrayRef<Type*> contents() const { return MVKArrayRef<Type*>(data(), size()); }
         MVKArrayRef<Type*> contents()       { return MVKArrayRef<Type*>(data(), size()); }
 
