@@ -1,7 +1,7 @@
 /*
  * SPIRVToMSLConverter.cpp
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -94,17 +94,17 @@ MVK_PUBLIC_SYMBOL SPIRVToMSLConversionOptions::SPIRVToMSLConversionOptions() {
 	mslOptions.pad_fragment_output_components = true;
 }
 
-MVK_PUBLIC_SYMBOL bool mvk::MSLShaderInput::matches(const mvk::MSLShaderInput& other) const {
-	if (memcmp(&shaderInput, &other.shaderInput, sizeof(shaderInput)) != 0) { return false; }
+MVK_PUBLIC_SYMBOL bool mvk::MSLShaderInterfaceVariable::matches(const mvk::MSLShaderInterfaceVariable& other) const {
+	if (memcmp(&shaderVar, &other.shaderVar, sizeof(shaderVar)) != 0) { return false; }
 	if (binding != other.binding) { return false; }
 	return true;
 }
 
-MVK_PUBLIC_SYMBOL mvk::MSLShaderInput::MSLShaderInput() {
-	// Explicitly set shaderInput to defaults over cleared memory to ensure all instances
+MVK_PUBLIC_SYMBOL mvk::MSLShaderInterfaceVariable::MSLShaderInterfaceVariable() {
+	// Explicitly set shaderVar to defaults over cleared memory to ensure all instances
 	// have exactly the same memory layout when using memory comparison in matches().
-	memset(&shaderInput, 0, sizeof(shaderInput));
-	shaderInput = SPIRV_CROSS_NAMESPACE::MSLShaderInput();
+	memset(&shaderVar, 0, sizeof(shaderVar));
+	shaderVar = SPIRV_CROSS_NAMESPACE::MSLShaderInterfaceVariable();
 }
 
 // If requiresConstExprSampler is false, constExprSampler can be ignored
@@ -143,7 +143,21 @@ MVK_PUBLIC_SYMBOL bool SPIRVToMSLConversionConfiguration::stageSupportsVertexAtt
 // Check them all in case inactive VA's duplicate locations used by active VA's.
 MVK_PUBLIC_SYMBOL bool SPIRVToMSLConversionConfiguration::isShaderInputLocationUsed(uint32_t location) const {
     for (auto& si : shaderInputs) {
-        if ((si.shaderInput.location == location) && si.outIsUsedByShader) { return true; }
+        if ((si.shaderVar.location == location) && si.outIsUsedByShader) { return true; }
+    }
+    return false;
+}
+
+MVK_PUBLIC_SYMBOL bool SPIRVToMSLConversionConfiguration::isShaderInputBuiltInUsed(spv::BuiltIn builtin) const {
+    for (auto& si : shaderInputs) {
+        if ((si.shaderVar.builtin == builtin) && si.outIsUsedByShader) { return true; }
+    }
+    return false;
+}
+
+MVK_PUBLIC_SYMBOL bool SPIRVToMSLConversionConfiguration::isShaderOutputLocationUsed(uint32_t location) const {
+    for (auto& so : shaderOutputs) {
+        if ((so.shaderVar.location == location) && so.outIsUsedByShader) { return true; }
     }
     return false;
 }
@@ -166,8 +180,9 @@ MVK_PUBLIC_SYMBOL bool SPIRVToMSLConversionConfiguration::isResourceUsed(Executi
 	return false;
 }
 
-MVK_PUBLIC_SYMBOL void SPIRVToMSLConversionConfiguration::markAllInputsAndResourcesUsed() {
+MVK_PUBLIC_SYMBOL void SPIRVToMSLConversionConfiguration::markAllInterfaceVarsAndResourcesUsed() {
 	for (auto& si : shaderInputs) { si.outIsUsedByShader = true; }
+	for (auto& so : shaderOutputs) { so.outIsUsedByShader = true; }
 	for (auto& rb : resourceBindings) { rb.outIsUsedByShader = true; }
 }
 
@@ -175,7 +190,7 @@ MVK_PUBLIC_SYMBOL void SPIRVToMSLConversionConfiguration::markAllInputsAndResour
 // and the resources can be spread across these shader stages. To improve cache hits when using
 // this function to find a cached shader for a particular shader stage, only consider the resources
 // that are used in that shader stage. By contrast, discreteDescriptorSet apply across all stages,
-// and shaderInputs are populated before each stage, so neither needs to be filtered by stage here.
+// and shaderInputs and shaderOutputs are populated before each stage, so neither needs to be filtered by stage here.
 MVK_PUBLIC_SYMBOL bool SPIRVToMSLConversionConfiguration::matches(const SPIRVToMSLConversionConfiguration& other) const {
 
     if ( !options.matches(other.options) ) { return false; }
@@ -184,6 +199,10 @@ MVK_PUBLIC_SYMBOL bool SPIRVToMSLConversionConfiguration::matches(const SPIRVToM
 		if (si.outIsUsedByShader && !containsMatching(other.shaderInputs, si)) { return false; }
 	}
 
+	for (const auto& so : shaderOutputs) {
+		if (so.outIsUsedByShader && !containsMatching(other.shaderOutputs, so)) { return false; }
+	}
+
     for (const auto& rb : resourceBindings) {
         if (rb.resourceBinding.stage == options.entryPointStage &&
 			rb.outIsUsedByShader &&
@@ -212,6 +231,13 @@ MVK_PUBLIC_SYMBOL void SPIRVToMSLConversionConfiguration::alignWith(const SPIRVT
 		}
 	}
 
+	for (auto& so : shaderOutputs) {
+		so.outIsUsedByShader = false;
+		for (auto& srcSO : srcContext.shaderOutputs) {
+			if (so.matches(srcSO)) { so.outIsUsedByShader = srcSO.outIsUsedByShader; }
+		}
+	}
+
     for (auto& rb : resourceBindings) {
         rb.outIsUsedByShader = false;
         for (auto& srcRB : srcContext.resourceBindings) {
@@ -235,6 +261,7 @@ MVK_PUBLIC_SYMBOL void SPIRVToMSLConverter::setSPIRV(const uint32_t* spirvCode,
 }
 
 MVK_PUBLIC_SYMBOL bool SPIRVToMSLConverter::convert(SPIRVToMSLConversionConfiguration& shaderConfig,
+													SPIRVToMSLConversionResult& conversionResult,
 													bool shouldLogSPIRV,
 													bool shouldLogMSL,
                                                     bool shouldLogGLSL) {
@@ -244,14 +271,10 @@ MVK_PUBLIC_SYMBOL bool SPIRVToMSLConverter::convert(SPIRVToMSLConversionConfigur
 //	spvFile.write((char*)_spirv.data(), _spirv.size() << 2);
 //	spvFile.close();
 
-	_wasConverted = true;
-	_resultLog.clear();
-	_msl.clear();
-	_shaderConversionResults.reset();
-
-	if (shouldLogSPIRV) { logSPIRV("Converting"); }
+	if (shouldLogSPIRV) { logSPIRV(conversionResult.resultLog, "Converting"); }
 
 	CompilerMSL* pMSLCompiler = nullptr;
+	bool wasConverted = true;
 
 #ifndef SPIRV_CROSS_EXCEPTIONS_TO_ASSERTIONS
 	try {
@@ -281,18 +304,23 @@ MVK_PUBLIC_SYMBOL bool SPIRVToMSLConverter::convert(SPIRVToMSLConversionConfigur
 		scOpts.vertex.flip_vert_y = shaderConfig.options.shouldFlipVertexY;
 		pMSLCompiler->set_common_options(scOpts);
 
-		// Add shader inputs
+		// Add shader inputs and outputs
 		for (auto& si : shaderConfig.shaderInputs) {
-			pMSLCompiler->add_msl_shader_input(si.shaderInput);
+			pMSLCompiler->add_msl_shader_input(si.shaderVar);
+		}
+
+		for (auto& so : shaderConfig.shaderOutputs) {
+			pMSLCompiler->add_msl_shader_output(so.shaderVar);
 		}
 
 		// Add resource bindings and hardcoded constexpr samplers
 		for (auto& rb : shaderConfig.resourceBindings) {
 			auto& rbb = rb.resourceBinding;
-			pMSLCompiler->add_msl_resource_binding(rbb);
-
-			if (rb.requiresConstExprSampler) {
-				pMSLCompiler->remap_constexpr_sampler_by_binding(rbb.desc_set, rbb.binding, rb.constExprSampler);
+			if (rbb.stage == shaderConfig.options.entryPointStage) {
+				pMSLCompiler->add_msl_resource_binding(rbb);
+				if (rb.requiresConstExprSampler) {
+					pMSLCompiler->remap_constexpr_sampler_by_binding(rbb.desc_set, rbb.binding, rb.constExprSampler);
+				}
 			}
 		}
 
@@ -311,58 +339,75 @@ MVK_PUBLIC_SYMBOL bool SPIRVToMSLConverter::convert(SPIRVToMSLConversionConfigur
 				}
 			}
 		}
-		_msl = pMSLCompiler->compile();
+		conversionResult.msl = pMSLCompiler->compile();
 
-        if (shouldLogMSL) { logSource(_msl, "MSL", "Converted"); }
+        if (shouldLogMSL) { logSource(conversionResult.resultLog, conversionResult.msl, "MSL", "Converted"); }
 
 #ifndef SPIRV_CROSS_EXCEPTIONS_TO_ASSERTIONS
 	} catch (CompilerError& ex) {
-		string errMsg("MSL conversion error: ");
+		string errMsg("SPIR-V to MSL conversion error: ");
 		errMsg += ex.what();
-		logError(errMsg.data());
+		logError(conversionResult.resultLog, errMsg.data());
         if (shouldLogMSL && pMSLCompiler) {
-            _msl = pMSLCompiler->get_partial_source();
-            logSource(_msl, "MSL", "Partially converted");
+			auto partialMSL = pMSLCompiler->get_partial_source();
+            logSource(conversionResult.resultLog, partialMSL, "MSL", "Partially converted");
         }
 	}
 #endif
 
 	// Populate the shader conversion results with info from the compilation run,
 	// and mark which vertex attributes and resource bindings are used by the shader
-	populateEntryPoint(pMSLCompiler, shaderConfig.options);
-	_shaderConversionResults.isRasterizationDisabled = pMSLCompiler && pMSLCompiler->get_is_rasterization_disabled();
-	_shaderConversionResults.isPositionInvariant = pMSLCompiler && pMSLCompiler->is_position_invariant();
-	_shaderConversionResults.needsSwizzleBuffer = pMSLCompiler && pMSLCompiler->needs_swizzle_buffer();
-	_shaderConversionResults.needsOutputBuffer = pMSLCompiler && pMSLCompiler->needs_output_buffer();
-	_shaderConversionResults.needsPatchOutputBuffer = pMSLCompiler && pMSLCompiler->needs_patch_output_buffer();
-	_shaderConversionResults.needsBufferSizeBuffer = pMSLCompiler && pMSLCompiler->needs_buffer_size_buffer();
-	_shaderConversionResults.needsInputThreadgroupMem = pMSLCompiler && pMSLCompiler->needs_input_threadgroup_mem();
-	_shaderConversionResults.needsDispatchBaseBuffer = pMSLCompiler && pMSLCompiler->needs_dispatch_base_buffer();
-	_shaderConversionResults.needsViewRangeBuffer = pMSLCompiler && pMSLCompiler->needs_view_mask_buffer();
+	populateEntryPoint(pMSLCompiler, shaderConfig.options, conversionResult.resultInfo.entryPoint);
+	conversionResult.resultInfo.entryPoint.supportsFastMath &= !pMSLCompiler || pMSLCompiler->can_compile_with_fast_math();
+	conversionResult.resultInfo.isRasterizationDisabled = pMSLCompiler && pMSLCompiler->get_is_rasterization_disabled();
+	conversionResult.resultInfo.isPositionInvariant = pMSLCompiler && pMSLCompiler->is_position_invariant();
+	conversionResult.resultInfo.needsSwizzleBuffer = pMSLCompiler && pMSLCompiler->needs_swizzle_buffer();
+	conversionResult.resultInfo.needsOutputBuffer = pMSLCompiler && pMSLCompiler->needs_output_buffer();
+	conversionResult.resultInfo.needsPatchOutputBuffer = pMSLCompiler && pMSLCompiler->needs_patch_output_buffer();
+	conversionResult.resultInfo.needsBufferSizeBuffer = pMSLCompiler && pMSLCompiler->needs_buffer_size_buffer();
+	conversionResult.resultInfo.needsTextureOffsetBuffer = pMSLCompiler && pMSLCompiler->needs_texture_offset_buffer();
+	conversionResult.resultInfo.needsInputThreadgroupMem = pMSLCompiler && pMSLCompiler->needs_input_threadgroup_mem();
+	conversionResult.resultInfo.needsDispatchBaseBuffer = pMSLCompiler && pMSLCompiler->needs_dispatch_base_buffer();
+	conversionResult.resultInfo.needsViewRangeBuffer = pMSLCompiler && pMSLCompiler->needs_view_mask_buffer();
+	conversionResult.resultInfo.needsXfbBuffer = pMSLCompiler && pMSLCompiler->needs_xfb_buffer();
+	conversionResult.resultInfo.usesPhysicalStorageBufferAddressesCapability = usesPhysicalStorageBufferAddressesCapability(pMSLCompiler);
 
 	// When using Metal argument buffers, if the shader is provided with dynamic buffer offsets,
 	// then it needs a buffer to hold these dynamic offsets.
-	_shaderConversionResults.needsDynamicOffsetBuffer = false;
+	conversionResult.resultInfo.needsDynamicOffsetBuffer = false;
 	if (shaderConfig.options.mslOptions.argument_buffers) {
 		for (auto& db : shaderConfig.dynamicBufferDescriptors) {
 			if (db.stage == shaderConfig.options.entryPointStage) {
-				_shaderConversionResults.needsDynamicOffsetBuffer = true;
+				conversionResult.resultInfo.needsDynamicOffsetBuffer = true;
 			}
 		}
 	}
 
-	for (auto& ctxSI : shaderConfig.shaderInputs) {
-		ctxSI.outIsUsedByShader = pMSLCompiler->is_msl_shader_input_used(ctxSI.shaderInput.location);
-	}
-	for (auto& ctxRB : shaderConfig.resourceBindings) {
-		if (ctxRB.resourceBinding.stage == shaderConfig.options.entryPointStage) {
-			ctxRB.outIsUsedByShader = pMSLCompiler->is_msl_resource_binding_used(ctxRB.resourceBinding.stage,
-																				 ctxRB.resourceBinding.desc_set,
-																				 ctxRB.resourceBinding.binding);
-		}
-	}
-
-	delete pMSLCompiler;
+    if (pMSLCompiler) {
+        for (auto& ctxSI : shaderConfig.shaderInputs) {
+            if (ctxSI.shaderVar.builtin != spv::BuiltInMax) {
+                ctxSI.outIsUsedByShader = pMSLCompiler->has_active_builtin(ctxSI.shaderVar.builtin, spv::StorageClassInput);
+            } else {
+                ctxSI.outIsUsedByShader = pMSLCompiler->is_msl_shader_input_used(ctxSI.shaderVar.location);
+            }
+        }
+        for (auto& ctxSO : shaderConfig.shaderOutputs) {
+            if (ctxSO.shaderVar.builtin != spv::BuiltInMax) {
+                ctxSO.outIsUsedByShader = pMSLCompiler->has_active_builtin(ctxSO.shaderVar.builtin, spv::StorageClassOutput);
+            } else {
+                ctxSO.outIsUsedByShader = pMSLCompiler->is_msl_shader_output_used(ctxSO.shaderVar.location);
+            }
+        }
+        for (auto& ctxRB : shaderConfig.resourceBindings) {
+            if (ctxRB.resourceBinding.stage == shaderConfig.options.entryPointStage) {
+                ctxRB.outIsUsedByShader = pMSLCompiler->is_msl_resource_binding_used(ctxRB.resourceBinding.stage,
+                                                                                     ctxRB.resourceBinding.desc_set,
+                                                                                     ctxRB.resourceBinding.binding);
+            }
+        }
+        
+        delete pMSLCompiler;
+    }
 
     // To check GLSL conversion
     if (shouldLogGLSL) {
@@ -377,68 +422,65 @@ MVK_PUBLIC_SYMBOL bool SPIRVToMSLConverter::convert(SPIRVToMSLConversionConfigur
 			options.separate_shader_objects = true;
 			pGLSLCompiler->set_common_options(options);
 			string glsl = pGLSLCompiler->compile();
-            logSource(glsl, "GLSL", "Estimated original");
+            logSource(conversionResult.resultLog, glsl, "GLSL", "Estimated original");
 #ifndef SPIRV_CROSS_EXCEPTIONS_TO_ASSERTIONS
         } catch (CompilerError& ex) {
             string errMsg("Original GLSL extraction error: ");
             errMsg += ex.what();
-            logMsg(errMsg.data());
+            logMsg(conversionResult.resultLog, errMsg.data());
 			if (pGLSLCompiler) {
 				string glsl = pGLSLCompiler->get_partial_source();
-				logSource(glsl, "GLSL", "Partially converted");
+				logSource(conversionResult.resultLog, glsl, "GLSL", "Partially converted");
 			}
         }
 #endif
 		delete pGLSLCompiler;
 	}
 
-	return _wasConverted;
+	return wasConverted;
 }
 
 // Appends the message text to the result log.
-void SPIRVToMSLConverter::logMsg(const char* logMsg) {
+void SPIRVToMSLConverter::logMsg(string& log, const char* logMsg) {
 	string trimMsg = trim(logMsg);
 	if ( !trimMsg.empty() ) {
-		_resultLog += trimMsg;
-		_resultLog += "\n\n";
+		log += trimMsg;
+		log += "\n\n";
 	}
 }
 
-// Appends the error text to the result log, sets the wasConverted property to false, and returns it.
-bool SPIRVToMSLConverter::logError(const char* errMsg) {
-	logMsg(errMsg);
-	_wasConverted = false;
-	return _wasConverted;
+// Appends the error text to the result log, and returns false to indicate an error.
+bool SPIRVToMSLConverter::logError(string& log, const char* errMsg) {
+	logMsg(log, errMsg);
+	fprintf(stderr, "[mvk-error] %s\n", errMsg);
+	return false;
 }
 
 // Appends the SPIR-V to the result log, indicating whether it is being converted or was converted.
-void SPIRVToMSLConverter::logSPIRV(const char* opDesc) {
+void SPIRVToMSLConverter::logSPIRV(string& log, const char* opDesc) {
 
-	string spvLog;
-	mvk::logSPIRV(_spirv, spvLog);
-
-	_resultLog += opDesc;
-	_resultLog += " SPIR-V:\n";
-	_resultLog += spvLog;
-	_resultLog += "\nEnd SPIR-V\n\n";
+	log += opDesc;
+	log += " SPIR-V:\n";
+	mvk::logSPIRV(_spirv, log);
+	log += "\nEnd SPIR-V\n\n";
 
 	// Uncomment one or both of the following lines to get additional debugging and tracability capabilities.
 	// The SPIR-V can be written in binary form to a file, and/or logged in human readable form to the console.
 	// These can be helpful if errors occur during conversion of SPIR-V to MSL.
-//	writeSPIRVToFile("spvout.spv");
-//	printf("\n%s\n", getResultLog().c_str());
+//	writeSPIRVToFile("spvout.spv", log);
+//	printf("\n%s\n", log.c_str());
 }
 
 // Writes the SPIR-V code to a file. This can be useful for debugging
 // when the SPRIR-V did not originally come from a known file
-void SPIRVToMSLConverter::writeSPIRVToFile(string spvFilepath) {
+void SPIRVToMSLConverter::writeSPIRVToFile(string spvFilepath, string& log) {
 	vector<char> fileContents;
 	spirvToBytes(_spirv, fileContents);
 	string errMsg;
 	if (writeFile(spvFilepath, fileContents, errMsg)) {
-		_resultLog += "Saved SPIR-V to file: " + absolutePath(spvFilepath) + "\n\n";
+		log += "Saved SPIR-V to file: " + absolutePath(spvFilepath) + "\n\n";
 	} else {
-		_resultLog += "Could not write SPIR-V file. " + errMsg + "\n\n";
+		log += "Could not write SPIR-V file. " + errMsg + "\n\n";
 	}
 }
 
@@ -451,15 +493,15 @@ bool SPIRVToMSLConverter::validateSPIRV() {
 }
 
 // Appends the source to the result log, prepending with the operation.
-void SPIRVToMSLConverter::logSource(string& src, const char* srcLang, const char* opDesc) {
-    _resultLog += opDesc;
-    _resultLog += " ";
-    _resultLog += srcLang;
-    _resultLog += ":\n";
-    _resultLog += src;
-    _resultLog += "\nEnd ";
-    _resultLog += srcLang;
-    _resultLog += "\n\n";
+void SPIRVToMSLConverter::logSource(string& log, string& src, const char* srcLang, const char* opDesc) {
+    log += opDesc;
+    log += " ";
+    log += srcLang;
+    log += ":\n";
+    log += src;
+    log += "\nEnd ";
+    log += srcLang;
+    log += "\n\n";
 }
 
 void SPIRVToMSLConverter::populateWorkgroupDimension(SPIRVWorkgroupSizeDimension& wgDim,
@@ -472,7 +514,8 @@ void SPIRVToMSLConverter::populateWorkgroupDimension(SPIRVWorkgroupSizeDimension
 
 // Populates the entry point with info extracted from the SPRI-V compiler.
 void SPIRVToMSLConverter::populateEntryPoint(Compiler* pCompiler,
-											 SPIRVToMSLConversionOptions& options) {
+											 SPIRVToMSLConversionOptions& options,
+											 SPIRVEntryPoint& entryPoint) {
 
 	if ( !pCompiler ) { return; }
 
@@ -487,15 +530,26 @@ void SPIRVToMSLConverter::populateEntryPoint(Compiler* pCompiler,
 		}
 	}
 
-	auto& ep = _shaderConversionResults.entryPoint;
-	ep.mtlFunctionName = spvEP.name;
-	ep.supportsFastMath = !spvEP.flags.get(ExecutionModeSignedZeroInfNanPreserve);
+	entryPoint.mtlFunctionName = spvEP.name;
+	entryPoint.supportsFastMath = !spvEP.flags.get(ExecutionModeSignedZeroInfNanPreserve);
 
 	SpecializationConstant widthSC, heightSC, depthSC;
 	pCompiler->get_work_group_size_specialization_constants(widthSC, heightSC, depthSC);
 
-	auto& wgSize = ep.workgroupSize;
+	auto& wgSize = entryPoint.workgroupSize;
 	populateWorkgroupDimension(wgSize.width, spvEP.workgroup_size.x, widthSC);
 	populateWorkgroupDimension(wgSize.height, spvEP.workgroup_size.y, heightSC);
 	populateWorkgroupDimension(wgSize.depth, spvEP.workgroup_size.z, depthSC);
 }
+
+bool SPIRVToMSLConverter::usesPhysicalStorageBufferAddressesCapability(Compiler* pCompiler) {
+	if (pCompiler) {
+		auto& declaredCapabilities = pCompiler->get_declared_capabilities();
+		for(auto dc: declaredCapabilities) {
+			if (dc == CapabilityPhysicalStorageBufferAddresses) {
+				return true;
+			}
+		}
+	}
+	return false;
+}
