@@ -1,7 +1,7 @@
 /*
  * MoltenVKShaderConverterTool.cpp
  *
- * Copyright (c) 2015-2022 The Brenwill Workshop Ltd. (http://www.brenwill.com)
+ * Copyright (c) 2015-2024 The Brenwill Workshop Ltd. (http://www.brenwill.com)
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -137,24 +137,23 @@ bool MoltenVKShaderConverterTool::convertGLSL(string& glslInFile,
 	}
 
 	// Convert GLSL to SPIR-V
+	GLSLToSPIRVConversionResult conversionResult;
 	GLSLToSPIRVConverter glslConverter;
 	glslConverter.setGLSL(glslCode);
 
 	uint64_t startTime = _glslConversionPerformance.getTimestamp();
-	bool wasConverted = glslConverter.convert(shaderStage, _shouldLogConversions, _shouldLogConversions);
+	bool wasConverted = glslConverter.convert(shaderStage, conversionResult, _shouldLogConversions, _shouldLogConversions);
 	_glslConversionPerformance.accumulate(startTime);
 
 	if (wasConverted) {
-		if (_shouldLogConversions) { log(glslConverter.getResultLog().data()); }
+		if (_shouldLogConversions) { log(conversionResult.resultLog.data()); }
 	} else {
 		string logMsg = "Could not convert GLSL in file: " + absolutePath(path);
 		log(logMsg.data());
-		log(glslConverter.getResultLog().data());
+		log(conversionResult.resultLog.data());
 		return false;
 	}
 
-	const vector<uint32_t>& spv = glslConverter.getSPIRV();
-
 	// Write the SPIR-V code to a file.
 	// If no file has been supplied, create one from the GLSL file name.
 	if (_shouldWriteSPIRV) {
@@ -162,9 +161,9 @@ bool MoltenVKShaderConverterTool::convertGLSL(string& glslInFile,
 		if (path.empty()) { path = pathWithExtension(glslInFile, _shouldOutputAsHeaders ? "h" : "spv",
 													 _shouldIncludeOrigPathExtn, _origPathExtnSep); }
 		if (_shouldOutputAsHeaders) {
-			spirvToHeaderBytes(spv, fileContents, fileName(path, false));
+			spirvToHeaderBytes(conversionResult.spirv, fileContents, fileName(path, false));
 		} else {
-			spirvToBytes(spv, fileContents);
+			spirvToBytes(conversionResult.spirv, fileContents);
 		}
 
 		if (writeFile(path, fileContents, errMsg)) {
@@ -177,7 +176,7 @@ bool MoltenVKShaderConverterTool::convertGLSL(string& glslInFile,
 		}
 	}
 
-	return convertSPIRV(spv, glslInFile, mslOutFile, false);
+	return convertSPIRV(conversionResult.spirv, glslInFile, mslOutFile, false);
 }
 
 // Read SPIR-V code from a SPIR-V file, convert to MSL, and write the MSL code to files.
@@ -219,30 +218,35 @@ bool MoltenVKShaderConverterTool::convertSPIRV(const vector<uint32_t>& spv,
 	mslContext.options.mslOptions.platform = _mslPlatform;
 	mslContext.options.mslOptions.set_msl_version(_mslVersionMajor, _mslVersionMinor, _mslVersionPatch);
 	mslContext.options.shouldFlipVertexY = _shouldFlipVertexY;
+	mslContext.options.mslOptions.argument_buffers = _useMetalArgumentBuffers;
+	mslContext.options.mslOptions.force_active_argument_buffer_resources = false;
+	mslContext.options.mslOptions.pad_argument_buffer_resources = false;
+	mslContext.options.mslOptions.argument_buffers_tier = SPIRV_CROSS_NAMESPACE::CompilerMSL::Options::ArgumentBuffersTier::Tier2;
+	mslContext.options.mslOptions.replace_recursive_inputs = mvkOSVersionIsAtLeast(14.0, 17.0, 1.0);
 
 	SPIRVToMSLConverter spvConverter;
 	spvConverter.setSPIRV(spv);
 
 	uint64_t startTime = _spvConversionPerformance.getTimestamp();
-	bool wasConverted = spvConverter.convert(mslContext, shouldLogSPV, _shouldLogConversions, (_shouldLogConversions && shouldLogSPV));
+	SPIRVToMSLConversionResult conversionResult;
+	bool wasConverted = spvConverter.convert(mslContext, conversionResult, shouldLogSPV, _shouldLogConversions, (_shouldLogConversions && shouldLogSPV));
 	_spvConversionPerformance.accumulate(startTime);
 
 	if (wasConverted) {
-		if (_shouldLogConversions) { log(spvConverter.getResultLog().data()); }
+		if (_shouldLogConversions) { log(conversionResult.resultLog.data()); }
 	} else {
 		string errMsg = "Could not convert SPIR-V in file: " + absolutePath(inFile);
 		log(errMsg.data());
-		log(spvConverter.getResultLog().data());
+		log(conversionResult.resultLog.data());
 		return false;
 	}
 
 	// Write the MSL to file
 	string path = mslOutFile;
 	if (mslOutFile.empty()) { path = pathWithExtension(inFile, "metal", _shouldIncludeOrigPathExtn, _origPathExtnSep); }
-	const string& msl = spvConverter.getMSL();
 
 	string compileErrMsg;
-	bool wasCompiled = compile(msl, compileErrMsg, _mslVersionMajor, _mslVersionMinor, _mslVersionPatch);
+	bool wasCompiled = compile(conversionResult.msl, compileErrMsg, _mslVersionMajor, _mslVersionMinor, _mslVersionPatch);
 	if (compileErrMsg.size() > 0) {
 		string preamble = wasCompiled ? "is valid but the validation compilation produced warnings: " : "failed a validation compilation: ";
 		compileErrMsg = "Generated MSL " + preamble + compileErrMsg;
@@ -252,7 +256,7 @@ bool MoltenVKShaderConverterTool::convertSPIRV(const vector<uint32_t>& spv,
 	}
 
 	vector<char> fileContents;
-	fileContents.insert(fileContents.end(), msl.begin(), msl.end());
+	fileContents.insert(fileContents.end(), conversionResult.msl.begin(), conversionResult.msl.end());
 	string writeErrMsg;
 	if (writeFile(path, fileContents, writeErrMsg)) {
 		string logMsg = "Saved MSL to file: " + fileName(path);
@@ -364,6 +368,7 @@ void MoltenVKShaderConverterTool::showUsage() {
     log("                       May be omitted for defaults (\"cp cmp comp compute kn kl krn kern kernel\").");
 	log("  -sx \"fileExtns\"    - List of SPIR-V shader file extensions.");
 	log("                       May be omitted for defaults (\"spv spirv\").");
+	log("  -mab               - Use Metal Argument Buffers to hold resources in the shaders.");
 	log("  -l                 - Log the conversion results to the console (to aid debugging).");
 	log("  -p                 - Log the performance of the shader conversions.");
 	log("  -q                 - Quiet mode. Stops logging of informational messages.");
@@ -419,18 +424,28 @@ MoltenVKShaderConverterTool::MoltenVKShaderConverterTool(int argc, const char* a
 	_shouldReportPerformance = false;
 	_shouldOutputAsHeaders = false;
 	_quietMode = false;
+	_useMetalArgumentBuffers = false;
 
-	_mslVersionMajor = 2;
-
-	if (mvkOSVersionIsAtLeast(12.0)) {
+	if (mvkOSVersionIsAtLeast(14.0)) {
+		_mslVersionMajor = 3;
+		_mslVersionMinor = 1;
+	} else 	if (mvkOSVersionIsAtLeast(13.0)) {
+		_mslVersionMajor = 3;
+		_mslVersionMinor = 0;
+	} else if (mvkOSVersionIsAtLeast(12.0)) {
+		_mslVersionMajor = 2;
 		_mslVersionMinor = 4;
 	} else if (mvkOSVersionIsAtLeast(11.0)) {
+		_mslVersionMajor = 2;
 		_mslVersionMinor = 3;
 	} else if (mvkOSVersionIsAtLeast(10.15)) {
+		_mslVersionMajor = 2;
 		_mslVersionMinor = 2;
 	} else if (mvkOSVersionIsAtLeast(10.14)) {
+		_mslVersionMajor = 2;
 		_mslVersionMinor = 1;
 	} else if (mvkOSVersionIsAtLeast(10.13)) {
+		_mslVersionMajor = 2;
 		_mslVersionMinor = 0;
 	} else if (mvkOSVersionIsAtLeast(10.12)) {
 		_mslVersionMajor = 1;
@@ -633,6 +648,11 @@ bool MoltenVKShaderConverterTool::parseArgs(int argc, const char* argv[]) {
 			continue;
 		}
 
+		if(equal(arg, "-mab", true)) {
+			_useMetalArgumentBuffers = true;
+			continue;
+		}
+
 		if(equal(arg, "-l", true)) {
 			_shouldLogConversions = true;
 			continue;
