@@ -1,5 +1,7 @@
 #!/bin/bash
 
+if [ "${SKIP_PACKAGING}" = "Y" ]; then exit 0; fi
+
 . "${PROJECT_DIR}/Scripts/create_xcframework_func.sh"
 
 export MVK_EXT_DIR="${PROJECT_DIR}/External"
@@ -15,6 +17,6 @@ ln -sfn "${abs_ext_dir}/SPIRV-Cross" "${hdr_dir}/SPIRVCross"
 ln -sfn "${abs_ext_dir}/glslang/External/spirv-tools/include/spirv-tools" "${hdr_dir}/SPIRVTools"
 ln -sfn "${abs_ext_dir}/glslang" "${hdr_dir}/glslang"
 
-create_xcframework "SPIRVCross"
-create_xcframework "SPIRVTools"
-create_xcframework "glslang"
+create_xcframework "SPIRVCross" "library"
+create_xcframework "SPIRVTools" "library"
+create_xcframework "glslang" "library"
