@@ -1,8 +1,9 @@
 #!/bin/bash
 
-. "${SRCROOT}/Scripts/package_moltenvk.sh"
 . "${SRCROOT}/Scripts/package_moltenvk_xcframework.sh"
 . "${SRCROOT}/Scripts/package_shader_converter_xcframework.sh"
+. "${SRCROOT}/Scripts/package_dylibs.sh"
+. "${SRCROOT}/Scripts/package_headers.sh"
 . "${SRCROOT}/Scripts/package_docs.sh"
 . "${SRCROOT}/Scripts/package_update_latest.sh"
 
