@@ -29,7 +29,7 @@
 #include <unistd.h>
 #include <dlfcn.h>
 #include <sys/stat.h>
-#ifdef HAVE_SECURITY_SECURITY_H
+#ifdef __APPLE__
 #include <Security/Security.h>
 #endif
 #ifdef SONAME_LIBGNUTLS
@@ -95,19 +95,18 @@ static NTSTATUS process_attach( void *args )
         setenv("GNUTLS_SYSTEM_PRIORITY_FILE", "/dev/null", 0);
     }
 
-if (1) { /* CROSSOVER HACK - bug 10151 */
-    const char *libgnutls_name_candidates[] = {SONAME_LIBGNUTLS,
-                                               "libgnutls.so.30",
-                                               "libgnutls.so.28",
-                                               "libgnutls-deb0.so.28",
-                                               "libgnutls.so.26",
-                                               NULL};
-    int i;
-    for (i=0; libgnutls_name_candidates[i] && !libgnutls_handle; i++)
-        libgnutls_handle = dlopen(libgnutls_name_candidates[i], RTLD_NOW);
-}
-else
-    libgnutls_handle = dlopen( SONAME_LIBGNUTLS, RTLD_NOW );
+    /* CROSSOVER HACK - bug 10151 */
+    {
+        const char *libgnutls_name_candidates[] = {SONAME_LIBGNUTLS,
+                                                   "libgnutls.so.30",
+                                                   "libgnutls.so.28",
+                                                   "libgnutls-deb0.so.28",
+                                                   "libgnutls.so.26",
+                                                   NULL};
+        int i;
+        for (i = 0; libgnutls_name_candidates[i] && !libgnutls_handle; i++)
+            libgnutls_handle = dlopen(libgnutls_name_candidates[i], RTLD_NOW);
+    }
 
     if (!libgnutls_handle)
     {
@@ -144,7 +143,9 @@ else
 
     if (TRACE_ON( crypt ))
     {
-        pgnutls_global_set_log_level( 4 );
+        char *env = getenv("GNUTLS_DEBUG_LEVEL");
+        int level = env ? atoi(env) : 4;
+        pgnutls_global_set_log_level(level);
         pgnutls_global_set_log_function( gnutls_log );
     }
 
@@ -158,9 +159,14 @@ fail:
 
 static NTSTATUS process_detach( void *args )
 {
-    pgnutls_global_deinit();
-    dlclose( libgnutls_handle );
-    libgnutls_handle = NULL;
+    if (libgnutls_handle)
+    {
+        if (TRACE_ON( crypt ))
+            pgnutls_global_set_log_function( NULL );
+        pgnutls_global_deinit();
+        dlclose( libgnutls_handle );
+        libgnutls_handle = NULL;
+    }
     return STATUS_SUCCESS;
 }
 #define RSA_MAGIC_KEY  ('R' | ('S' << 8) | ('A' << 16) | ('2' << 24))
@@ -639,7 +645,7 @@ static void load_root_certs(void)
 {
     unsigned int i;
 
-#ifdef HAVE_SECURITY_SECURITY_H
+#ifdef __APPLE__
     const SecTrustSettingsDomain domains[] = {
         kSecTrustSettingsDomainSystem,
         kSecTrustSettingsDomainAdmin,
@@ -709,6 +715,8 @@ const unixlib_entry_t __wine_unix_call_funcs[] =
     enum_root_certs,
 };
 
+C_ASSERT( ARRAYSIZE(__wine_unix_call_funcs) == unix_funcs_count );
+
 #ifdef _WIN64
 
 typedef ULONG PTR32;
@@ -810,4 +818,6 @@ const unixlib_entry_t __wine_unix_call_wow64_funcs[] =
     wow64_enum_root_certs,
 };
 
+C_ASSERT( ARRAYSIZE(__wine_unix_call_wow64_funcs) == unix_funcs_count );
+
 #endif  /* _WIN64 */
