@@ -1,7 +1,7 @@
 /*
- * Direct3D X 9 main file
+ * D3DX10 main file
  *
- * Copyright (C) 2007 David Adam
+ * Copyright (c) 2010 Owen Rudge for CodeWeavers
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -19,59 +19,26 @@
  *
  */
 
-#include "config.h"
-#include "wine/port.h"
-#include "initguid.h"
 #include <stdarg.h>
 
+#define COBJMACROS
+
 #include "windef.h"
 #include "winbase.h"
-#include "wingdi.h"
 #include "winuser.h"
+#include "objbase.h"
 
-#include "d3dx9.h"
-#include "wine/debug.h"
-
-WINE_DEFAULT_DEBUG_CHANNEL(d3dx);
+#include "d3d10.h"
 
 /***********************************************************************
- * DllMain.
- */
-BOOL WINAPI DllMain(HINSTANCE inst, DWORD reason, LPVOID reserved)
-{
-    switch(reason)
-    {
-    case DLL_WINE_PREATTACH:
-        return FALSE; /* prefer native version */
-    case DLL_PROCESS_ATTACH:
-        DisableThreadLibraryCalls(inst);
-        break;
-    }
-    return TRUE;
-}
-
-/***********************************************************************
- * D3DXCheckVersion
+ * D3DX10CheckVersion
+ *
  * Checks whether we are compiling against the correct d3d and d3dx library.
  */
-BOOL WINAPI D3DXCheckVersion(UINT d3dsdkvers, UINT d3dxsdkvers)
+BOOL WINAPI D3DX10CheckVersion(UINT d3dsdkvers, UINT d3dxsdkvers)
 {
-    if(d3dsdkvers==D3D_SDK_VERSION && d3dxsdkvers==30)
+    if ((d3dsdkvers == D3D10_SDK_VERSION) && (d3dxsdkvers == 37))
         return TRUE;
-    else
-        return FALSE;
-}
 
-typedef enum _D3DX_CPU_OPTIMIZATION
-{
-    D3DX_NOT_OPTIMIZED      = 0,
-    D3DX_3DNOW_OPTIMIZED    = 1,
-    D3DX_SSE2_OPTIMIZED     = 2,
-    D3DX_SSE_OPTIMIZED      = 3
-} D3DX_CPU_OPTIMIZATION;
-
-D3DX_CPU_OPTIMIZATION WINAPI D3DXCpuOptimizations(BOOL enable)
-{
-    FIXME("%i - stub\n", enable);
-    return D3DX_NOT_OPTIMIZED;
+    return FALSE;
 }
