@@ -29,10 +29,10 @@
 #include "winerror.h"
 #include "winternl.h"
 #include "winioctl.h"
+#include "ddk/ntddk.h"
 
 #include "kernelbase.h"
 #include "wine/debug.h"
-#include "wine/heap.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(security);
 
@@ -139,26 +139,6 @@ static const WELLKNOWNRID WellKnownRids[] =
     { WinAccountRasAndIasServersSid, DOMAIN_ALIAS_RID_RAS_SERVERS },
 };
 
-static const SID world_sid = { SID_REVISION, 1, { SECURITY_WORLD_SID_AUTHORITY} , { SECURITY_WORLD_RID } };
-static const DWORD world_access_acl_size = sizeof(ACL) + sizeof(ACCESS_ALLOWED_ACE) + sizeof(world_sid) - sizeof(DWORD);
-
-static void get_world_access_acl( PACL acl )
-{
-    PACCESS_ALLOWED_ACE ace = (PACCESS_ALLOWED_ACE)(acl + 1);
-
-    acl->AclRevision = ACL_REVISION;
-    acl->Sbz1 = 0;
-    acl->AclSize = world_access_acl_size;
-    acl->AceCount = 1;
-    acl->Sbz2 = 0;
-    ace->Header.AceType = ACCESS_ALLOWED_ACE_TYPE;
-    ace->Header.AceFlags = CONTAINER_INHERIT_ACE;
-    ace->Header.AceSize = sizeof(ACCESS_ALLOWED_ACE) + sizeof(world_sid) - sizeof(DWORD);
-    ace->Mask = 0xf3ffffff; /* Everything except reserved bits */
-    memcpy( &ace->SidStart, &world_sid, sizeof(world_sid) );
-}
-
-
 static NTSTATUS open_file( LPCWSTR name, DWORD access, HANDLE *file )
 {
     UNICODE_STRING file_nameW;
@@ -167,11 +147,7 @@ static NTSTATUS open_file( LPCWSTR name, DWORD access, HANDLE *file )
     NTSTATUS status;
 
     if ((status = RtlDosPathNameToNtPathName_U_WithStatus( name, &file_nameW, NULL, NULL ))) return status;
-    attr.Length = sizeof(attr);
-    attr.RootDirectory = 0;
-    attr.Attributes = OBJ_CASE_INSENSITIVE;
-    attr.ObjectName = &file_nameW;
-    attr.SecurityDescriptor = NULL;
+    InitializeObjectAttributes( &attr, &file_nameW, OBJ_CASE_INSENSITIVE, 0, NULL );
     status = NtCreateFile( file, access|SYNCHRONIZE, &attr, &io, NULL, FILE_FLAG_BACKUP_SEMANTICS,
                            FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE, FILE_OPEN,
                            FILE_OPEN_FOR_BACKUP_INTENT, NULL, 0 );
@@ -431,7 +407,7 @@ BOOL WINAPI GetWindowsAccountDomainSid( PSID sid, PSID domain_sid, DWORD *size )
  */
 BOOL WINAPI InitializeSid ( PSID sid, PSID_IDENTIFIER_AUTHORITY auth, BYTE count )
 {
-    return RtlInitializeSid( sid, auth, count );
+    return set_ntstatus(RtlInitializeSid( sid, auth, count ));
 }
 
 /******************************************************************************
@@ -611,7 +587,7 @@ BOOL WINAPI CheckTokenMembership( HANDLE token, PSID sid_to_check, PBOOL is_memb
     if (!ret && GetLastError() != ERROR_INSUFFICIENT_BUFFER)
         goto exit;
 
-    token_groups = heap_alloc(size);
+    token_groups = HeapAlloc(GetProcessHeap(), 0, size);
     if (!token_groups)
     {
         ret = FALSE;
@@ -637,7 +613,7 @@ BOOL WINAPI CheckTokenMembership( HANDLE token, PSID sid_to_check, PBOOL is_memb
     }
 
 exit:
-    heap_free(token_groups);
+    HeapFree(GetProcessHeap(), 0, token_groups);
     if (thread_token != NULL) CloseHandle(thread_token);
     return ret;
 }
@@ -660,21 +636,27 @@ BOOL WINAPI CreateRestrictedToken( HANDLE token, DWORD flags,
 
     if (disable_sid_count)
     {
-        if (!(nt_disable_sids = heap_alloc( offsetof( TOKEN_GROUPS, Groups[disable_sid_count] ) ))) goto out;
+        if (!(nt_disable_sids = HeapAlloc( GetProcessHeap(), 0,
+                                           offsetof( TOKEN_GROUPS, Groups[disable_sid_count] ) )))
+            goto out;
         nt_disable_sids->GroupCount = disable_sid_count;
         memcpy( nt_disable_sids->Groups, disable_sids, disable_sid_count * sizeof(*disable_sids) );
     }
 
     if (delete_priv_count)
     {
-        if (!(nt_privs = heap_alloc( offsetof( TOKEN_PRIVILEGES, Privileges[delete_priv_count] ) ))) goto out;
+        if (!(nt_privs = HeapAlloc( GetProcessHeap(), 0,
+                                    offsetof( TOKEN_PRIVILEGES, Privileges[delete_priv_count] ) )))
+            goto out;
         nt_privs->PrivilegeCount = delete_priv_count;
         memcpy( nt_privs->Privileges, delete_privs, delete_priv_count * sizeof(*delete_privs) );
     }
 
     if (restrict_sid_count)
     {
-        if (!(nt_restrict_sids = heap_alloc( offsetof( TOKEN_GROUPS, Groups[restrict_sid_count] ) ))) goto out;
+        if (!(nt_restrict_sids = HeapAlloc( GetProcessHeap(), 0,
+                                            offsetof( TOKEN_GROUPS, Groups[restrict_sid_count] ) )))
+            goto out;
         nt_restrict_sids->GroupCount = restrict_sid_count;
         memcpy( nt_restrict_sids->Groups, restrict_sids, restrict_sid_count * sizeof(*restrict_sids) );
     }
@@ -682,9 +664,9 @@ BOOL WINAPI CreateRestrictedToken( HANDLE token, DWORD flags,
     status = NtFilterToken(token, flags, nt_disable_sids, nt_privs, nt_restrict_sids, ret);
 
 out:
-    heap_free(nt_disable_sids);
-    heap_free(nt_privs);
-    heap_free(nt_restrict_sids);
+    HeapFree(GetProcessHeap(), 0, nt_disable_sids);
+    HeapFree(GetProcessHeap(), 0, nt_privs);
+    HeapFree(GetProcessHeap(), 0, nt_restrict_sids);
     return set_ntstatus( status );
 }
 
@@ -702,13 +684,19 @@ BOOL WINAPI DuplicateToken( HANDLE token, SECURITY_IMPERSONATION_LEVEL level, PH
 BOOL WINAPI DuplicateTokenEx( HANDLE token, DWORD access, LPSECURITY_ATTRIBUTES sa,
                               SECURITY_IMPERSONATION_LEVEL level, TOKEN_TYPE type, PHANDLE ret )
 {
+    SECURITY_QUALITY_OF_SERVICE qos;
     OBJECT_ATTRIBUTES attr;
 
     TRACE("%p 0x%08lx 0x%08x 0x%08x %p\n", token, access, level, type, ret );
 
+    qos.Length = sizeof(qos);
+    qos.ImpersonationLevel = level;
+    qos.ContextTrackingMode = SECURITY_STATIC_TRACKING;
+    qos.EffectiveOnly = FALSE;
     InitializeObjectAttributes( &attr, NULL, (sa && sa->bInheritHandle) ? OBJ_INHERIT : 0,
                                 NULL, sa ? sa->lpSecurityDescriptor : NULL );
-    return set_ntstatus( NtDuplicateToken( token, access, &attr, level, type, ret ));
+    attr.SecurityQualityOfService = &qos;
+    return set_ntstatus( NtDuplicateToken( token, access, &attr, FALSE, type, ret ));
 }
 
 /******************************************************************************
@@ -717,8 +705,8 @@ BOOL WINAPI DuplicateTokenEx( HANDLE token, DWORD access, LPSECURITY_ATTRIBUTES
 BOOL WINAPI GetTokenInformation( HANDLE token, TOKEN_INFORMATION_CLASS class,
                                  LPVOID info, DWORD len, LPDWORD retlen )
 {
-    TRACE("(%p, %s, %p, %ld, %p):\n",
-          token,
+    TRACE("(%p, %d [%s], %p, %ld, %p):\n",
+          token, class,
           (class == TokenUser) ? "TokenUser" :
           (class == TokenGroups) ? "TokenGroups" :
           (class == TokenPrivileges) ? "TokenPrivileges" :
@@ -734,6 +722,9 @@ BOOL WINAPI GetTokenInformation( HANDLE token, TOKEN_INFORMATION_CLASS class,
           (class == TokenGroupsAndPrivileges) ? "TokenGroupsAndPrivileges" :
           (class == TokenSessionReference) ? "TokenSessionReference" :
           (class == TokenSandBoxInert) ? "TokenSandBoxInert" :
+          (class == TokenElevation) ? "TokenElevation" :
+          (class == TokenElevationType) ? "TokenElevationType" :
+          (class == TokenLinkedToken) ? "TokenLinkedToken" :
           "Unknown",
           info, len, retlen);
 
@@ -812,7 +803,7 @@ BOOL WINAPI IsTokenRestricted( HANDLE token )
     status = NtQueryInformationToken(token, TokenRestrictedSids, NULL, 0, &size);
     if (status != STATUS_BUFFER_TOO_SMALL) return set_ntstatus(status);
 
-    groups = heap_alloc(size);
+    groups = HeapAlloc(GetProcessHeap(), 0, size);
     if (!groups)
     {
         SetLastError(ERROR_OUTOFMEMORY);
@@ -822,12 +813,12 @@ BOOL WINAPI IsTokenRestricted( HANDLE token )
     status = NtQueryInformationToken(token, TokenRestrictedSids, groups, size, &size);
     if (status != STATUS_SUCCESS)
     {
-        heap_free(groups);
+        HeapFree(GetProcessHeap(), 0, groups);
         return set_ntstatus(status);
     }
 
     restricted = groups->GroupCount > 0;
-    heap_free(groups);
+    HeapFree(GetProcessHeap(), 0, groups);
 
     return restricted;
 }
@@ -938,7 +929,7 @@ BOOL WINAPI CreatePrivateObjectSecurity( PSECURITY_DESCRIPTOR parent, PSECURITY_
                                          PSECURITY_DESCRIPTOR *descr, BOOL is_container, HANDLE token,
                                          PGENERIC_MAPPING mapping )
 {
-    return CreatePrivateObjectSecurityEx( parent, creator, descr, NULL, is_container, 0, token, mapping );
+    return set_ntstatus( RtlNewSecurityObject( parent, creator, descr, is_container, token, mapping ));
 }
 
 /******************************************************************************
@@ -948,46 +939,7 @@ BOOL WINAPI CreatePrivateObjectSecurityEx( PSECURITY_DESCRIPTOR parent, PSECURIT
                                            PSECURITY_DESCRIPTOR *descr, GUID *type, BOOL is_container,
                                            ULONG flags, HANDLE token, PGENERIC_MAPPING mapping )
 {
-    SECURITY_DESCRIPTOR_RELATIVE *relative;
-    DWORD needed, offset;
-    BYTE *buffer;
-
-    FIXME( "%p %p %p %p %d %lu %p %p - returns fake SECURITY_DESCRIPTOR\n",
-           parent, creator, descr, type, is_container, flags, token, mapping );
-
-    needed = sizeof(SECURITY_DESCRIPTOR_RELATIVE);
-    needed += sizeof(world_sid);
-    needed += sizeof(world_sid);
-    needed += world_access_acl_size;
-    needed += world_access_acl_size;
-
-    if (!(buffer = heap_alloc( needed ))) return FALSE;
-    relative = (SECURITY_DESCRIPTOR_RELATIVE *)buffer;
-    if (!InitializeSecurityDescriptor( relative, SECURITY_DESCRIPTOR_REVISION ))
-    {
-        heap_free( buffer );
-        return FALSE;
-    }
-    relative->Control |= SE_SELF_RELATIVE;
-    offset = sizeof(SECURITY_DESCRIPTOR_RELATIVE);
-
-    memcpy( buffer + offset, &world_sid, sizeof(world_sid) );
-    relative->Owner = offset;
-    offset += sizeof(world_sid);
-
-    memcpy( buffer + offset, &world_sid, sizeof(world_sid) );
-    relative->Group = offset;
-    offset += sizeof(world_sid);
-
-    get_world_access_acl( (ACL *)(buffer + offset) );
-    relative->Dacl = offset;
-    offset += world_access_acl_size;
-
-    get_world_access_acl( (ACL *)(buffer + offset) );
-    relative->Sacl = offset;
-
-    *descr = relative;
-    return TRUE;
+    return set_ntstatus( RtlNewSecurityObjectEx( parent, creator, descr, type, is_container, flags, token, mapping ));
 }
 
 /******************************************************************************
@@ -1000,9 +952,8 @@ BOOL WINAPI CreatePrivateObjectSecurityWithMultipleInheritance( PSECURITY_DESCRI
                                                                 BOOL is_container, ULONG flags,
                                                                 HANDLE token, PGENERIC_MAPPING mapping )
 {
-    FIXME(": semi-stub\n");
-    return CreatePrivateObjectSecurityEx( parent, creator, descr, NULL, is_container,
-                                          flags, token, mapping );
+    return set_ntstatus( RtlNewSecurityObjectWithMultipleInheritance( parent, creator, descr, types, count,
+            is_container, flags, token, mapping ));
 }
 
 /******************************************************************************
@@ -1010,9 +961,7 @@ BOOL WINAPI CreatePrivateObjectSecurityWithMultipleInheritance( PSECURITY_DESCRI
  */
 BOOL WINAPI DestroyPrivateObjectSecurity( PSECURITY_DESCRIPTOR *descr )
 {
-    FIXME("%p - stub\n", descr);
-    heap_free( *descr );
-    return TRUE;
+    return set_ntstatus( RtlDeleteSecurityObject( descr ));
 }
 
 /******************************************************************************
@@ -1167,7 +1116,10 @@ BOOL WINAPI InitializeSecurityDescriptor( PSECURITY_DESCRIPTOR descr, DWORD revi
  */
 BOOL WINAPI IsValidSecurityDescriptor( PSECURITY_DESCRIPTOR descr )
 {
-    return set_ntstatus( RtlValidSecurityDescriptor( descr ));
+    if (!RtlValidSecurityDescriptor( descr ))
+        return set_ntstatus(STATUS_INVALID_SECURITY_DESCR);
+
+    return TRUE;
 }
 
 /******************************************************************************
@@ -1610,3 +1562,49 @@ BOOL WINAPI SetCachedSigningLevel( PHANDLE source, ULONG count, ULONG flags, HAN
     FIXME( "%p %lu %lu %p - stub\n", source, count, flags, file );
     return TRUE;
 }
+
+/******************************************************************************
+ * DeriveCapabilitySidsFromName    (kernelbase.@)
+ */
+BOOL WINAPI DeriveCapabilitySidsFromName( const WCHAR *cap_name, PSID **cap_group_sids, DWORD *cap_group_sid_count,
+                                          PSID **cap_sids, DWORD *cap_sid_count )
+{
+    NTSTATUS status = STATUS_NO_MEMORY;
+    UNICODE_STRING name_us;
+    DWORD size;
+
+    TRACE( "cap_name %s, cap_group_sids %p, cap_group_sid_count %p, cap_sids %p, cap_sid_count %p.\n",
+           debugstr_w(cap_name), cap_group_sids, cap_group_sid_count, cap_sids, cap_sid_count );
+
+    *cap_group_sid_count = 1;
+    *cap_sid_count = 1;
+    *cap_group_sids = NULL;
+    *cap_sids = NULL;
+
+    if (!(*cap_group_sids = RtlAllocateHeap( GetProcessHeap(), 0, *cap_group_sid_count * sizeof(**cap_group_sids) )))
+        goto done;
+    if (!(*cap_sids = RtlAllocateHeap( GetProcessHeap(), 0, *cap_sid_count * sizeof(**cap_sids) )))
+        goto done;
+
+    /* There is no obvious way to query returned subauthority count for RtlDeriveCapabilitySidsFromName,
+     * so let's reserve a bit more. */
+    size = RtlLengthRequiredSid( 16 );
+    if (!(**cap_group_sids = RtlAllocateHeap( GetProcessHeap(), 0, size ))) goto done;
+    if (!(**cap_sids = RtlAllocateHeap( GetProcessHeap(), 0, size ))) goto done;
+    RtlInitUnicodeString( &name_us, cap_name );
+    status = RtlDeriveCapabilitySidsFromName( &name_us, **cap_group_sids, **cap_sids );
+
+done:
+    if (!status) return TRUE;
+    if (*cap_group_sids)
+    {
+        RtlFreeHeap( GetProcessHeap(), 0, **cap_group_sids );
+        RtlFreeHeap( GetProcessHeap(), 0, *cap_group_sids );
+    }
+    if (*cap_sids)
+    {
+        RtlFreeHeap( GetProcessHeap(), 0, **cap_sids );
+        RtlFreeHeap( GetProcessHeap(), 0, *cap_sids );
+    }
+    return set_ntstatus( status );
+}
