@@ -24,7 +24,6 @@
 
 #include "ntstatus.h"
 #define WIN32_NO_STATUS
-#define NONAMELESSUNION
 #include "windef.h"
 #include "winbase.h"
 #include "winnls.h"
@@ -34,7 +33,6 @@
 #include "wine/exception.h"
 #include "wine/asm.h"
 #include "wine/debug.h"
-#include "wine/heap.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(thread);
 
@@ -65,6 +63,22 @@ HANDLE WINAPI DECLSPEC_HOTPATCH CreateRemoteThread( HANDLE process, SECURITY_ATT
     return CreateRemoteThreadEx( process, sa, stack, start, param, flags, NULL, id );
 }
 
+struct proc_thread_attr
+{
+    DWORD_PTR attr;
+    SIZE_T size;
+    void *value;
+};
+
+struct _PROC_THREAD_ATTRIBUTE_LIST
+{
+    DWORD mask;  /* bitmask of items in list */
+    DWORD size;  /* max number of items in list */
+    DWORD count; /* number of items in list */
+    DWORD pad;
+    DWORD_PTR unk;
+    struct proc_thread_attr attrs[];
+};
 
 /***************************************************************************
  *           CreateRemoteThreadEx   (kernelbase.@)
@@ -74,33 +88,78 @@ HANDLE WINAPI DECLSPEC_HOTPATCH CreateRemoteThreadEx( HANDLE process, SECURITY_A
                                                       LPVOID param, DWORD flags,
                                                       LPPROC_THREAD_ATTRIBUTE_LIST attributes, DWORD *id )
 {
+    ULONG_PTR buffer[offsetof( PS_ATTRIBUTE_LIST, Attributes[3] ) / sizeof(ULONG_PTR)];
+    PS_ATTRIBUTE_LIST *attr_list = (PS_ATTRIBUTE_LIST *)buffer;
+    struct _ACTIVATION_CONTEXT *actctx;
     HANDLE handle;
     CLIENT_ID client_id;
+    TEB *teb;
+    ULONG count = 0, ret;
+    OBJECT_ATTRIBUTES attr;
     SIZE_T stack_reserve = 0, stack_commit = 0;
+    GROUP_AFFINITY *group_affinity = NULL;
 
-    if (attributes) FIXME("thread attributes ignored\n");
+    if (attributes)
+    {
+        DWORD i;
+        for (i = 0; i < attributes->count; i++)
+            switch (attributes->attrs[i].attr)
+            {
+            case PROC_THREAD_ATTRIBUTE_GROUP_AFFINITY:
+                group_affinity = attributes->attrs[i].value;
+                break;
+            default:
+                FIXME("thread attributes %Ix ignored\n", attributes->attrs[i].attr);
+                break;
+            }
+    }
 
     if (flags & STACK_SIZE_PARAM_IS_A_RESERVATION) stack_reserve = stack;
     else stack_commit = stack;
 
-    if (!set_ntstatus( RtlCreateUserThread( process, sa ? sa->lpSecurityDescriptor : NULL, TRUE,
-                                            0, stack_reserve, stack_commit,
-                                            (PRTL_THREAD_START_ROUTINE)start, param, &handle, &client_id )))
+    attr_list->Attributes[count].Attribute    = PS_ATTRIBUTE_CLIENT_ID;
+    attr_list->Attributes[count].Size         = sizeof(client_id);
+    attr_list->Attributes[count].ValuePtr     = &client_id;
+    attr_list->Attributes[count].ReturnLength = NULL;
+    count++;
+    attr_list->Attributes[count].Attribute    = PS_ATTRIBUTE_TEB_ADDRESS;
+    attr_list->Attributes[count].Size         = sizeof(teb);
+    attr_list->Attributes[count].ValuePtr     = &teb;
+    attr_list->Attributes[count].ReturnLength = NULL;
+    count++;
+    if (group_affinity)
+    {
+        attr_list->Attributes[count].Attribute    = PS_ATTRIBUTE_GROUP_AFFINITY;
+        attr_list->Attributes[count].Size         = sizeof(*group_affinity);
+        attr_list->Attributes[count].ValuePtr     = group_affinity;
+        attr_list->Attributes[count].ReturnLength = NULL;
+        count++;
+    }
+    attr_list->TotalLength = offsetof( PS_ATTRIBUTE_LIST, Attributes[count] );
+
+    InitializeObjectAttributes( &attr, NULL, 0, NULL, sa ? sa->lpSecurityDescriptor : NULL );
+    if (sa && sa->bInheritHandle) attr.Attributes |= OBJ_INHERIT;
+
+    RtlGetActiveActivationContext( &actctx );
+
+    if (!set_ntstatus( NtCreateThreadEx( &handle, THREAD_ALL_ACCESS, &attr, process,
+                                         (PRTL_THREAD_START_ROUTINE)start, param,
+                                         THREAD_CREATE_FLAGS_CREATE_SUSPENDED, 0,
+                                         stack_commit, stack_reserve, attr_list )))
+    {
+        if (actctx) RtlReleaseActivationContext( actctx );
         return 0;
+    }
+
+    if (actctx)
+    {
+        ULONG_PTR cookie;
+        RtlActivateActivationContextEx( 0, teb, actctx, &cookie );
+        RtlReleaseActivationContext( actctx );
+    }
 
     if (id) *id = HandleToULong( client_id.UniqueThread );
-    if (sa && sa->nLength >= sizeof(*sa) && sa->bInheritHandle)
-        SetHandleInformation( handle, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT );
-    if (!(flags & CREATE_SUSPENDED))
-    {
-        ULONG ret;
-        if (NtResumeThread( handle, &ret ))
-        {
-            NtClose( handle );
-            SetLastError( ERROR_NOT_ENOUGH_MEMORY );
-            handle = 0;
-        }
-    }
+    if (!(flags & CREATE_SUSPENDED)) NtResumeThread( handle, &ret );
     return handle;
 }
 
@@ -248,11 +307,9 @@ DWORD WINAPI DECLSPEC_HOTPATCH GetThreadId( HANDLE thread )
 /***********************************************************************
  *           GetThreadIdealProcessorEx   (kernelbase.@)
  */
-BOOL WINAPI /* DECLSPEC_HOTPATCH */ GetThreadIdealProcessorEx( HANDLE thread, PROCESSOR_NUMBER *ideal )
+BOOL WINAPI DECLSPEC_HOTPATCH GetThreadIdealProcessorEx( HANDLE thread, PROCESSOR_NUMBER *ideal )
 {
-    FIXME( "(%p %p): stub\n", thread, ideal );
-    SetLastError( ERROR_CALL_NOT_IMPLEMENTED );
-    return FALSE;
+    return set_ntstatus( NtQueryInformationThread( thread, ThreadIdealProcessorEx, ideal, sizeof(*ideal), NULL));
 }
 
 
@@ -277,7 +334,7 @@ INT WINAPI DECLSPEC_HOTPATCH GetThreadPriority( HANDLE thread )
     if (!set_ntstatus( NtQueryInformationThread( thread, ThreadBasicInformation,
                                                  &info, sizeof(info), NULL )))
         return THREAD_PRIORITY_ERROR_RETURN;
-    return info.Priority;
+    return info.BasePriority;
 }
 
 
@@ -286,8 +343,7 @@ INT WINAPI DECLSPEC_HOTPATCH GetThreadPriority( HANDLE thread )
  */
 BOOL WINAPI DECLSPEC_HOTPATCH GetThreadPriorityBoost( HANDLE thread, BOOL *state )
 {
-    if (state) *state = FALSE;
-    return TRUE;
+    return set_ntstatus( NtQueryInformationThread( thread, ThreadPriorityBoost, state, sizeof(*state), NULL ));
 }
 
 
@@ -348,13 +404,7 @@ HANDLE WINAPI DECLSPEC_HOTPATCH OpenThread( DWORD access, BOOL inherit, DWORD id
     OBJECT_ATTRIBUTES attr;
     CLIENT_ID cid;
 
-    attr.Length = sizeof(attr);
-    attr.RootDirectory = 0;
-    attr.Attributes = inherit ? OBJ_INHERIT : 0;
-    attr.ObjectName = NULL;
-    attr.SecurityDescriptor = NULL;
-    attr.SecurityQualityOfService = NULL;
-
+    InitializeObjectAttributes( &attr, NULL, inherit ? OBJ_INHERIT : 0, 0, NULL );
     cid.UniqueProcess = 0;
     cid.UniqueThread = ULongToHandle( id );
 
@@ -379,6 +429,15 @@ DWORD WINAPI DECLSPEC_HOTPATCH QueueUserAPC( PAPCFUNC func, HANDLE thread, ULONG
 }
 
 
+/***********************************************************************
+ *	     QueueUserAPC2   (kernelbase.@)
+ */
+DWORD WINAPI DECLSPEC_HOTPATCH QueueUserAPC2( PAPCFUNC func, HANDLE thread, ULONG_PTR data, QUEUE_USER_APC_FLAGS flags )
+{
+    return set_ntstatus( NtQueueApcThreadEx2( thread, NULL, flags, call_user_apc, (ULONG_PTR)func, data, 0 ));
+}
+
+
 /***********************************************************************
  *           QueryThreadCycleTime   (kernelbase.@)
  */
@@ -451,7 +510,7 @@ HRESULT WINAPI DECLSPEC_HOTPATCH GetThreadDescription( HANDLE thread, WCHAR **de
     if (status != STATUS_BUFFER_TOO_SMALL)
         return HRESULT_FROM_NT(status);
 
-    if (!(info = heap_alloc( length )))
+    if (!(info = HeapAlloc( GetProcessHeap(), 0, length )))
         return HRESULT_FROM_NT(STATUS_NO_MEMORY);
 
     status = NtQueryInformationThread( thread, ThreadNameInformation, info, length, &length );
@@ -467,7 +526,7 @@ HRESULT WINAPI DECLSPEC_HOTPATCH GetThreadDescription( HANDLE thread, WCHAR **de
         }
     }
 
-    heap_free(info);
+    HeapFree(GetProcessHeap(), 0, info);
 
     return HRESULT_FROM_NT(status);
 }
@@ -512,13 +571,13 @@ BOOL WINAPI DECLSPEC_HOTPATCH SetThreadGroupAffinity( HANDLE thread, const GROUP
  */
 DWORD WINAPI DECLSPEC_HOTPATCH SetThreadIdealProcessor( HANDLE thread, DWORD proc )
 {
-    FIXME( "(%p %lu): stub\n", thread, proc );
-    if (proc > MAXIMUM_PROCESSORS)
-    {
-        SetLastError( ERROR_INVALID_PARAMETER );
-        return ~0u;
-    }
-    return 0;
+    NTSTATUS status;
+
+    status = NtSetInformationThread( thread, ThreadIdealProcessor, &proc, sizeof(proc) );
+    if (NT_SUCCESS(status)) return status;
+
+    SetLastError( RtlNtStatusToDosError( status ));
+    return ~0u;
 }
 
 
@@ -568,7 +627,7 @@ BOOL WINAPI DECLSPEC_HOTPATCH SetThreadPriority( HANDLE thread, INT priority )
  */
 BOOL WINAPI DECLSPEC_HOTPATCH SetThreadPriorityBoost( HANDLE thread, BOOL disable )
 {
-    return TRUE;
+    return set_ntstatus( NtSetInformationThread( thread, ThreadPriorityBoost, &disable, sizeof(disable) ));
 }
 
 
@@ -606,6 +665,25 @@ LANGID WINAPI DECLSPEC_HOTPATCH SetThreadUILanguage( LANGID langid )
 }
 
 
+/**********************************************************************
+ *            SetThreadInformation   (kernelbase.@)
+ */
+BOOL WINAPI DECLSPEC_HOTPATCH SetThreadInformation( HANDLE thread, THREAD_INFORMATION_CLASS info_class,
+        VOID *info, DWORD size )
+{
+    switch (info_class)
+    {
+        case ThreadMemoryPriority:
+            return set_ntstatus( NtSetInformationThread( thread, ThreadPagePriority, info, size ));
+        case ThreadPowerThrottling:
+            return set_ntstatus( NtSetInformationThread( thread, ThreadPowerThrottlingState, info, size ));
+        default:
+            FIXME("Unsupported class %u.\n", info_class);
+            return FALSE;
+    }
+}
+
+
 /**********************************************************************
  *           SuspendThread   (kernelbase.@)
  */
@@ -613,6 +691,10 @@ DWORD WINAPI DECLSPEC_HOTPATCH SuspendThread( HANDLE thread )
 {
     DWORD ret;
 
+    /* NT in Win9x mode returns 0 for current thread. */
+    if ((GetVersion() & 0x80000000) && GetThreadId( thread ) == GetCurrentThreadId())
+        return 0;
+
     if (!set_ntstatus( NtSuspendThread( thread, &ret ))) ret = ~0U;
     return ret;
 }
@@ -747,11 +829,47 @@ BOOL WINAPI DECLSPEC_HOTPATCH TlsSetValue( DWORD index, LPVOID value )
 }
 
 
+/***********************************************************************
+ *           Wow64GetThreadContext   (kernelbase.@)
+ */
+BOOL WINAPI Wow64GetThreadContext( HANDLE handle, WOW64_CONTEXT *context)
+{
+#ifdef __i386__
+    return set_ntstatus( NtGetContextThread( handle, (CONTEXT *)context ));
+#elif defined(__x86_64__)
+    return set_ntstatus( RtlWow64GetThreadContext( handle, context ));
+#else
+    return set_ntstatus( STATUS_NOT_IMPLEMENTED );
+#endif
+}
+
+
+/***********************************************************************
+ *           Wow64SetThreadContext   (kernelbase.@)
+ */
+BOOL WINAPI Wow64SetThreadContext( HANDLE handle, const WOW64_CONTEXT *context)
+{
+#ifdef __i386__
+    return set_ntstatus( NtSetContextThread( handle, (const CONTEXT *)context ));
+#elif defined(__x86_64__)
+    return set_ntstatus( RtlWow64SetThreadContext( handle, context ));
+#else
+    return set_ntstatus( STATUS_NOT_IMPLEMENTED );
+#endif
+}
+
+
 /***********************************************************************
  * Fibers
  ***********************************************************************/
 
 
+struct fiber_actctx
+{
+    ACTIVATION_CONTEXT_STACK stack_space;    /* activation context stack space */
+    ACTIVATION_CONTEXT_STACK *stack_ptr;     /* last value of ActivationContextStackPointer */
+};
+
 struct fiber_data
 {
     LPVOID                param;             /* 00/00 fiber param */
@@ -763,10 +881,11 @@ struct fiber_data
     DWORD                 flags;             /*       fiber flags */
     LPFIBER_START_ROUTINE start;             /*       start routine */
     void                 *fls_slots;         /*       fiber storage slots */
+    struct fiber_actctx   actctx;            /*       activation context state */
 };
 
-extern void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new );
 #ifdef __i386__
+extern void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new );
 __ASM_STDCALL_FUNC( switch_fiber, 8,
                     "movl 4(%esp),%ecx\n\t"     /* old */
                     "movl %edi,0x9c(%ecx)\n\t"  /* old->Edi */
@@ -784,8 +903,37 @@ __ASM_STDCALL_FUNC( switch_fiber, 8,
                     "movl 0xb4(%ecx),%ebp\n\t"  /* new->Ebp */
                     "movl 0xc4(%ecx),%esp\n\t"  /* new->Esp */
                     "jmp *0xb8(%ecx)" )         /* new->Eip */
+#elif defined(__arm64ec__)
+static void __attribute__((naked)) WINAPI switch_fiber( CONTEXT *old, CONTEXT *new )
+{
+    asm( "mov x2, sp\n\t"
+         "stp x27, x2,  [x0, #0x90]\n\t"  /* old->Rbx,Rsp */
+         "str x29,      [x0, #0xa0]\n\t"  /* old->Rbp */
+         "stp x25, x26, [x0, #0xa8]\n\t"  /* old->Rsi,Rdi */
+         "stp x19, x20, [x0, #0xd8]\n\t"  /* old->R12,R13 */
+         "stp x21, x22, [x0, #0xe8]\n\t"  /* old->R14,R15 */
+         "str x30,      [x0, #0xf8]\n\t"  /* old->Rip */
+         "stp q8,  q9,  [x0, #0x220]\n\t" /* old->Xmm8,Xmm9 */
+         "stp q10, q11, [x0, #0x240]\n\t" /* old->Xmm10,Xmm11 */
+         "stp q12, q13, [x0, #0x260]\n\t" /* old->Xmm12,Xmm13 */
+         "stp q14, q15, [x0, #0x280]\n\t" /* old->Xmm14,Xmm15 */
+         /* FIXME: MxCsr */
+         "ldp x27, x2,  [x1, #0x90]\n\t"  /* old->Rbx,Rsp */
+         "ldr x29,      [x1, #0xa0]\n\t"  /* old->Rbp */
+         "ldp x25, x26, [x1, #0xa8]\n\t"  /* old->Rsi,Rdi */
+         "ldp x19, x20, [x1, #0xd8]\n\t"  /* old->R12,R13 */
+         "ldp x21, x22, [x1, #0xe8]\n\t"  /* old->R14,R15 */
+         "ldr x30,      [x1, #0xf8]\n\t"  /* old->Rip */
+         "ldp q8,  q9,  [x1, #0x220]\n\t" /* old->Xmm8,Xmm9 */
+         "ldp q10, q11, [x1, #0x240]\n\t" /* old->Xmm10,Xmm11 */
+         "ldp q12, q13, [x1, #0x260]\n\t" /* old->Xmm12,Xmm13 */
+         "ldp q14, q15, [x1, #0x280]\n\t" /* old->Xmm14,Xmm15 */
+         "mov sp, x2\n\t"
+         "ret" );
+}
 #elif defined(__x86_64__)
-__ASM_STDCALL_FUNC( switch_fiber, 8,
+extern void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new );
+__ASM_GLOBAL_FUNC( switch_fiber,
                     "movq %rbx,0x90(%rcx)\n\t"       /* old->Rbx */
                     "leaq 0x8(%rsp),%rax\n\t"
                     "movq %rax,0x98(%rcx)\n\t"       /* old->Rsp */
@@ -829,7 +977,8 @@ __ASM_STDCALL_FUNC( switch_fiber, 8,
                     "movq 0x98(%rdx),%rsp\n\t"       /* new->Rsp */
                     "jmp *0xf8(%rdx)" )              /* new->Rip */
 #elif defined(__arm__)
-__ASM_STDCALL_FUNC( switch_fiber, 8,
+extern void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new );
+__ASM_GLOBAL_FUNC( switch_fiber,
                    "str r4, [r0, #0x14]\n\t"   /* old->R4 */
                    "str r5, [r0, #0x18]\n\t"   /* old->R5 */
                    "str r6, [r0, #0x1c]\n\t"   /* old->R6 */
@@ -852,7 +1001,8 @@ __ASM_STDCALL_FUNC( switch_fiber, 8,
                    "ldr r2, [r1, #0x40]\n\t"   /* new->Pc */
                    "bx r2" )
 #elif defined(__aarch64__)
-__ASM_STDCALL_FUNC( switch_fiber, 8,
+extern void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new );
+__ASM_GLOBAL_FUNC( switch_fiber,
                    "stp x19, x20, [x0, #0xa0]\n\t"  /* old->X19,X20 */
                    "stp x21, x22, [x0, #0xb0]\n\t"  /* old->X21,X22 */
                    "stp x23, x24, [x0, #0xc0]\n\t"  /* old->X23,X24 */
@@ -873,7 +1023,7 @@ __ASM_STDCALL_FUNC( switch_fiber, 8,
                    "mov sp, x2\n\t"
                    "ret" )
 #else
-void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new )
+static void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new )
 {
     FIXME( "not implemented\n" );
     DbgBreakPoint();
@@ -883,7 +1033,7 @@ void WINAPI switch_fiber( CONTEXT *old, CONTEXT *new )
 /* call the fiber initial function once we have switched stack */
 static void CDECL start_fiber(void)
 {
-    struct fiber_data *fiber = NtCurrentTeb()->Tib.u.FiberData;
+    struct fiber_data *fiber = NtCurrentTeb()->Tib.FiberData;
     LPFIBER_START_ROUTINE start = fiber->start;
 
     __TRY
@@ -903,6 +1053,9 @@ static void init_fiber_context( struct fiber_data *fiber )
 #ifdef __i386__
     fiber->context.Esp = (ULONG_PTR)fiber->stack_base - 4;
     fiber->context.Eip = (ULONG_PTR)start_fiber;
+#elif defined(__arm64ec__)
+    fiber->context.Rsp = (ULONG_PTR)fiber->stack_base;
+    fiber->context.Rip = (ULONG_PTR)start_fiber;
 #elif defined(__x86_64__)
     fiber->context.Rsp = (ULONG_PTR)fiber->stack_base - 0x28;
     fiber->context.Rip = (ULONG_PTR)start_fiber;
@@ -915,6 +1068,38 @@ static void init_fiber_context( struct fiber_data *fiber )
 #endif
 }
 
+static void move_list( LIST_ENTRY *dest, LIST_ENTRY *src )
+{
+    LIST_ENTRY *head = src->Flink;
+    LIST_ENTRY *tail = src->Blink;
+
+    if (src != head)
+    {
+        dest->Flink = head;
+        dest->Blink = tail;
+        head->Blink = dest;
+        tail->Flink = dest;
+    }
+    else InitializeListHead( dest );
+}
+
+static void relocate_thread_actctx_stack( ACTIVATION_CONTEXT_STACK *dest )
+{
+    ACTIVATION_CONTEXT_STACK *src = NtCurrentTeb()->ActivationContextStackPointer;
+
+    C_ASSERT(sizeof(*dest) == sizeof(dest->ActiveFrame) + sizeof(dest->FrameListCache) +
+                              sizeof(dest->Flags) + sizeof(dest->NextCookieSequenceNumber) +
+                              sizeof(dest->StackId));
+
+    dest->ActiveFrame = src->ActiveFrame;
+    move_list( &dest->FrameListCache, &src->FrameListCache );
+    dest->Flags = src->Flags;
+    dest->NextCookieSequenceNumber = src->NextCookieSequenceNumber;
+    dest->StackId = src->StackId;
+
+    NtCurrentTeb()->ActivationContextStackPointer = dest;
+}
+
 
 /***********************************************************************
  *           CreateFiber   (kernelbase.@)
@@ -953,6 +1138,8 @@ LPVOID WINAPI DECLSPEC_HOTPATCH CreateFiberEx( SIZE_T stack_commit, SIZE_T stack
     fiber->except      = (void *)-1;
     fiber->start       = start;
     fiber->flags       = flags;
+    InitializeListHead( &fiber->actctx.stack_space.FrameListCache );
+    fiber->actctx.stack_ptr = &fiber->actctx.stack_space;
     init_fiber_context( fiber );
     return fiber;
 }
@@ -963,11 +1150,12 @@ LPVOID WINAPI DECLSPEC_HOTPATCH CreateFiberEx( SIZE_T stack_commit, SIZE_T stack
  */
 BOOL WINAPI DECLSPEC_HOTPATCH ConvertFiberToThread(void)
 {
-    struct fiber_data *fiber = NtCurrentTeb()->Tib.u.FiberData;
+    struct fiber_data *fiber = NtCurrentTeb()->Tib.FiberData;
 
     if (fiber)
     {
-        NtCurrentTeb()->Tib.u.FiberData = NULL;
+        relocate_thread_actctx_stack( &NtCurrentTeb()->ActivationContextStack );
+        NtCurrentTeb()->Tib.FiberData = NULL;
         HeapFree( GetProcessHeap(), 0, fiber );
     }
     return TRUE;
@@ -977,7 +1165,7 @@ BOOL WINAPI DECLSPEC_HOTPATCH ConvertFiberToThread(void)
 /***********************************************************************
  *           ConvertThreadToFiber   (kernelbase.@)
  */
-LPVOID WINAPI DECLSPEC_HOTPATCH ConvertThreadToFiber( LPVOID param )
+LPVOID WINAPI /* DECLSPEC_HOTPATCH */ ConvertThreadToFiber( LPVOID param )
 {
     return ConvertThreadToFiberEx( param, 0 );
 }
@@ -990,7 +1178,7 @@ LPVOID WINAPI DECLSPEC_HOTPATCH ConvertThreadToFiberEx( LPVOID param, DWORD flag
 {
     struct fiber_data *fiber;
 
-    if (NtCurrentTeb()->Tib.u.FiberData)
+    if (NtCurrentTeb()->Tib.FiberData)
     {
         /* CrossOver hack for bug 20987 */
         static int is_halo_mcc = -1;
@@ -1008,7 +1196,7 @@ LPVOID WINAPI DECLSPEC_HOTPATCH ConvertThreadToFiberEx( LPVOID param, DWORD flag
         if (is_halo_mcc)
         {
             TRACE("CrossOver hack: faking success\n");
-            return NtCurrentTeb()->Tib.u.FiberData;
+            return NtCurrentTeb()->Tib.FiberData;
         }
 
         SetLastError( ERROR_ALREADY_FIBER );
@@ -1028,7 +1216,8 @@ LPVOID WINAPI DECLSPEC_HOTPATCH ConvertThreadToFiberEx( LPVOID param, DWORD flag
     fiber->start            = NULL;
     fiber->flags            = flags;
     fiber->fls_slots        = NtCurrentTeb()->FlsSlots;
-    NtCurrentTeb()->Tib.u.FiberData = fiber;
+    relocate_thread_actctx_stack( &fiber->actctx.stack_space );
+    NtCurrentTeb()->Tib.FiberData = fiber;
     return fiber;
 }
 
@@ -1041,13 +1230,15 @@ void WINAPI DECLSPEC_HOTPATCH DeleteFiber( LPVOID fiber_ptr )
     struct fiber_data *fiber = fiber_ptr;
 
     if (!fiber) return;
-    if (fiber == NtCurrentTeb()->Tib.u.FiberData)
+    if (fiber == NtCurrentTeb()->Tib.FiberData)
     {
+        relocate_thread_actctx_stack( &NtCurrentTeb()->ActivationContextStack );
         HeapFree( GetProcessHeap(), 0, fiber );
         RtlExitUserThread( 1 );
     }
     RtlFreeUserStack( fiber->stack_allocation );
     RtlProcessFlsData( fiber->fls_slots, 3 );
+    RtlFreeActivationContextStack( &fiber->actctx.stack_space );
     HeapFree( GetProcessHeap(), 0, fiber );
 }
 
@@ -1057,7 +1248,7 @@ void WINAPI DECLSPEC_HOTPATCH DeleteFiber( LPVOID fiber_ptr )
  */
 BOOL WINAPI DECLSPEC_HOTPATCH IsThreadAFiber(void)
 {
-    return NtCurrentTeb()->Tib.u.FiberData != NULL;
+    return NtCurrentTeb()->Tib.FiberData != NULL;
 }
 
 
@@ -1067,20 +1258,22 @@ BOOL WINAPI DECLSPEC_HOTPATCH IsThreadAFiber(void)
 void WINAPI DECLSPEC_HOTPATCH SwitchToFiber( LPVOID fiber )
 {
     struct fiber_data *new_fiber = fiber;
-    struct fiber_data *current_fiber = NtCurrentTeb()->Tib.u.FiberData;
+    struct fiber_data *current_fiber = NtCurrentTeb()->Tib.FiberData;
 
     current_fiber->except      = NtCurrentTeb()->Tib.ExceptionList;
     current_fiber->stack_limit = NtCurrentTeb()->Tib.StackLimit;
     current_fiber->fls_slots   = NtCurrentTeb()->FlsSlots;
+    current_fiber->actctx.stack_ptr = NtCurrentTeb()->ActivationContextStackPointer;
     /* stack_allocation and stack_base never change */
 
     /* FIXME: should save floating point context if requested in fiber->flags */
-    NtCurrentTeb()->Tib.u.FiberData   = new_fiber;
+    NtCurrentTeb()->Tib.FiberData     = new_fiber;
     NtCurrentTeb()->Tib.ExceptionList = new_fiber->except;
     NtCurrentTeb()->Tib.StackBase     = new_fiber->stack_base;
     NtCurrentTeb()->Tib.StackLimit    = new_fiber->stack_limit;
     NtCurrentTeb()->DeallocationStack = new_fiber->stack_allocation;
     NtCurrentTeb()->FlsSlots          = new_fiber->fls_slots;
+    NtCurrentTeb()->ActivationContextStackPointer = new_fiber->actctx.stack_ptr;
     switch_fiber( &current_fiber->context, &new_fiber->context );
 }
 
@@ -1169,7 +1362,7 @@ PTP_CLEANUP_GROUP WINAPI DECLSPEC_HOTPATCH CreateThreadpoolCleanupGroup(void)
 static void WINAPI tp_io_callback( TP_CALLBACK_INSTANCE *instance, void *userdata, void *cvalue, IO_STATUS_BLOCK *iosb, TP_IO *io )
 {
     PTP_WIN32_IO_CALLBACK callback = *(void **)io;
-    callback( instance, userdata, cvalue, RtlNtStatusToDosError( iosb->u.Status ), iosb->Information, io );
+    callback( instance, userdata, cvalue, RtlNtStatusToDosError( iosb->Status ), iosb->Information, io );
 }
 
 
