@@ -21,8 +21,6 @@
 #include <stdarg.h>
 #include <stdlib.h>
 
-#define NONAMELESSUNION
-
 #include "mountmgr.h"
 #include "winreg.h"
 #include "unixlib.h"
@@ -46,13 +44,11 @@ struct mount_point
 static struct list mount_points_list = LIST_INIT(mount_points_list);
 static HKEY mount_key;
 
-unixlib_handle_t mountmgr_handle = 0;
-
 void set_mount_point_id( struct mount_point *mount, const void *id, unsigned int id_len )
 {
-    RtlFreeHeap( GetProcessHeap(), 0, mount->id );
+    free( mount->id );
     mount->id_len = max( MIN_ID_LEN, id_len );
-    if ((mount->id = RtlAllocateHeap( GetProcessHeap(), HEAP_ZERO_MEMORY, mount->id_len )))
+    if ((mount->id = calloc( mount->id_len, 1 )))
     {
         memcpy( mount->id, id, id_len );
         RegSetValueExW( mount_key, mount->link.Buffer, 0, REG_BINARY, mount->id, mount->id_len );
@@ -67,7 +63,7 @@ static struct mount_point *add_mount_point( DEVICE_OBJECT *device, UNICODE_STRIN
     WCHAR *str;
     UINT len = (lstrlenW(link) + 1) * sizeof(WCHAR) + device_name->Length + sizeof(WCHAR);
 
-    if (!(mount = RtlAllocateHeap( GetProcessHeap(), 0, sizeof(*mount) + len ))) return NULL;
+    if (!(mount = malloc( sizeof(*mount) + len ))) return NULL;
 
     str = (WCHAR *)(mount + 1);
     lstrcpyW( str, link );
@@ -118,8 +114,8 @@ void delete_mount_point( struct mount_point *mount )
     list_remove( &mount->entry );
     RegDeleteValueW( mount_key, mount->link.Buffer );
     IoDeleteSymbolicLink( &mount->link );
-    RtlFreeHeap( GetProcessHeap(), 0, mount->id );
-    RtlFreeHeap( GetProcessHeap(), 0, mount );
+    free( mount->id );
+    free( mount );
 }
 
 /* check if a given mount point matches the requested specs */
@@ -188,7 +184,7 @@ static NTSTATUS query_mount_points( void *buff, SIZE_T insize,
         return STATUS_BUFFER_OVERFLOW;
     }
 
-    input = HeapAlloc( GetProcessHeap(), 0, insize );
+    input = malloc( insize );
     if (!input)
         return STATUS_NO_MEMORY;
     memcpy( input, buff, insize );
@@ -219,7 +215,7 @@ static NTSTATUS query_mount_points( void *buff, SIZE_T insize,
     }
     info->Size = pos;
     iosb->Information = pos;
-    HeapFree( GetProcessHeap(), 0, input );
+    free( input );
     return STATUS_SUCCESS;
 }
 
@@ -261,7 +257,7 @@ static NTSTATUS define_unix_drive( const void *in_buff, SIZE_T insize )
         case DRIVE_RAMDISK:   type = DEVICE_RAMDISK; break;
         case DRIVE_FIXED:     type = DEVICE_HARDDISK_VOL; break;
         }
-        return add_dos_device( letter - 'a', NULL, device, mount_point, type, NULL, NULL );
+        return add_dos_device( letter - 'a', NULL, device, mount_point, type, NULL, NULL, NULL );
     }
     else
     {
@@ -275,11 +271,8 @@ static NTSTATUS define_shell_folder( const void *in_buff, SIZE_T insize )
 {
     const struct mountmgr_shell_folder *input = in_buff;
     const char *link = NULL;
-    OBJECT_ATTRIBUTES attr;
-    UNICODE_STRING name;
     NTSTATUS status;
-    ULONG size = 256;
-    char *buffer = NULL, *backup = NULL;
+    WCHAR *folder;
     struct set_shell_folder_params params;
 
     if (input->folder_offset >= insize || input->folder_size > insize - input->folder_offset ||
@@ -294,43 +287,15 @@ static NTSTATUS define_shell_folder( const void *in_buff, SIZE_T insize )
         if (!link[0]) link = NULL;
     }
 
-    name.Buffer = (WCHAR *)((char *)in_buff + input->folder_offset);
-    name.Length = input->folder_size;
-    InitializeObjectAttributes( &attr, &name, 0, 0, NULL );
+    if (!(folder = malloc( input->folder_size + sizeof(WCHAR) ))) return STATUS_NO_MEMORY;
+    memcpy( folder, (char *)in_buff + input->folder_offset, input->folder_size );
+    folder[input->folder_size / sizeof(WCHAR)] = 0;
 
-    for (;;)
-    {
-        if (!(buffer = HeapAlloc( GetProcessHeap(), 0, size )))
-        {
-            status = STATUS_NO_MEMORY;
-            goto done;
-        }
-        status = wine_nt_to_unix_file_name( &attr, buffer, &size, FILE_OPEN_IF );
-        if (status == STATUS_NO_SUCH_FILE) status = STATUS_SUCCESS;
-        if (status == STATUS_SUCCESS) break;
-        if (status != STATUS_BUFFER_TOO_SMALL) goto done;
-        HeapFree( GetProcessHeap(), 0, buffer );
-    }
-
-    if (input->create_backup)
-    {
-        if (!(backup = HeapAlloc( GetProcessHeap(), 0, strlen(buffer) + sizeof(".backup" ) )))
-        {
-            status = STATUS_NO_MEMORY;
-            goto done;
-        }
-        strcpy( backup, buffer );
-        strcat( backup, ".backup" );
-    }
-
-    params.folder = buffer;
-    params.backup = backup;
+    params.folder = folder;
     params.link = link;
+    params.create_backup = input->create_backup;
     status = MOUNTMGR_CALL( set_shell_folder, &params );
-
-done:
-    HeapFree( GetProcessHeap(), 0, buffer );
-    HeapFree( GetProcessHeap(), 0, backup );
+    free( folder );
     return status;
 }
 
@@ -338,32 +303,18 @@ done:
 static NTSTATUS query_shell_folder( void *buff, SIZE_T insize, SIZE_T outsize, IO_STATUS_BLOCK *iosb )
 {
     char *output = buff;
-    OBJECT_ATTRIBUTES attr;
-    UNICODE_STRING name;
     NTSTATUS status;
-    ULONG size = 256;
-    char *buffer;
+    WCHAR *folder;
+    struct get_shell_folder_params params = { .buffer = output, .size = outsize };
 
-    name.Buffer = buff;
-    name.Length = insize;
-    InitializeObjectAttributes( &attr, &name, 0, 0, NULL );
+    if (!(folder = malloc( insize + sizeof(WCHAR) ))) return STATUS_NO_MEMORY;
+    memcpy( folder, (char *)buff, insize );
+    folder[insize / sizeof(WCHAR)] = 0;
 
-    for (;;)
-    {
-        if (!(buffer = HeapAlloc( GetProcessHeap(), 0, size ))) return STATUS_NO_MEMORY;
-        status = wine_nt_to_unix_file_name( &attr, buffer, &size, FILE_OPEN );
-        if (!status)
-        {
-            struct get_shell_folder_params params = { buffer, output, outsize };
-            status = MOUNTMGR_CALL( get_shell_folder, &params );
-            if (!status) iosb->Information = strlen(output) + 1;
-            break;
-        }
-        if (status != STATUS_BUFFER_TOO_SMALL) break;
-        HeapFree( GetProcessHeap(), 0, buffer );
-    }
-
-    HeapFree( GetProcessHeap(), 0, buffer );
+    params.folder = folder;
+    status = MOUNTMGR_CALL( get_shell_folder, &params );
+    if (!status) iosb->Information = strlen(output) + 1;
+    free( folder );
     return status;
 }
 
@@ -380,20 +331,20 @@ static void WINAPI query_dhcp_request_params( TP_CALLBACK_INSTANCE *instance, vo
     /* sanity checks */
     if (FIELD_OFFSET(struct mountmgr_dhcp_request_params, params[query->count]) > insize)
     {
-        irp->IoStatus.u.Status = STATUS_INVALID_PARAMETER;
+        irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
         goto err;
     }
 
     for (i = 0; i < query->count; i++)
         if (query->params[i].offset + query->params[i].size > insize)
         {
-            irp->IoStatus.u.Status = STATUS_INVALID_PARAMETER;
+            irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
             goto err;
         }
 
     if (!memchr( query->unix_name, 0, sizeof(query->unix_name) ))
     {
-        irp->IoStatus.u.Status = STATUS_INVALID_PARAMETER;
+        irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
         goto err;
     }
 
@@ -409,11 +360,11 @@ static void WINAPI query_dhcp_request_params( TP_CALLBACK_INSTANCE *instance, vo
         {
             if (offset >= sizeof(query->size)) query->size = offset;
             offset = sizeof(query->size);
-            irp->IoStatus.u.Status = STATUS_BUFFER_OVERFLOW;
+            irp->IoStatus.Status = STATUS_BUFFER_OVERFLOW;
             goto err;
         }
     }
-    irp->IoStatus.u.Status = STATUS_SUCCESS;
+    irp->IoStatus.Status = STATUS_SUCCESS;
 
 err:
     irp->IoStatus.Information = offset;
@@ -432,7 +383,7 @@ static void WINAPI query_symbol_file_callback( TP_CALLBACK_INSTANCE *instance, v
     NTSTATUS status = MOUNTMGR_CALL( query_symbol_file, &params );
 
     irp->IoStatus.Information = info;
-    irp->IoStatus.u.Status = status;
+    irp->IoStatus.Status = status;
     IoCompleteRequest( irp, IO_NO_INCREMENT );
 }
 
@@ -447,11 +398,11 @@ static void device_op( void )
     {
     case ADD_DOS_DEVICE:
         add_dos_device( -1, info.udi, info.device, info.mount_point,
-                        info.type, info.guid, info.scsi_info );
+                        info.type, info.guid, info.label, info.scsi_info );
         break;
     case ADD_VOLUME:
         add_volume( info.udi, info.device, info.mount_point, DEVICE_HARDDISK_VOL,
-                    info.guid, info.serial, info.scsi_info );
+                    info.guid, info.serial, info.label, info.scsi_info );
         break;
     case REMOVE_DEVICE:
         if (!remove_dos_device( -1, info.udi )) remove_volume( info.udi );
@@ -529,7 +480,7 @@ static NTSTATUS WINAPI mountmgr_ioctl( DEVICE_OBJECT *device, IRP *irp )
         }
 
         if (TrySubmitThreadpoolCallback( query_dhcp_request_params, irp, NULL ))
-            return (irp->IoStatus.u.Status = STATUS_PENDING);
+            return (irp->IoStatus.Status = STATUS_PENDING);
         status = STATUS_NO_MEMORY;
         break;
     case IOCTL_MOUNTMGR_QUERY_SYMBOL_FILE:
@@ -539,7 +490,7 @@ static NTSTATUS WINAPI mountmgr_ioctl( DEVICE_OBJECT *device, IRP *irp )
             break;
         }
         if (TrySubmitThreadpoolCallback( query_symbol_file_callback, irp, NULL ))
-            return (irp->IoStatus.u.Status = STATUS_PENDING);
+            return (irp->IoStatus.Status = STATUS_PENDING);
         status = STATUS_NO_MEMORY;
         break;
     case IOCTL_MOUNTMGR_READ_CREDENTIAL:
@@ -595,7 +546,7 @@ static NTSTATUS WINAPI mountmgr_ioctl( DEVICE_OBJECT *device, IRP *irp )
         status = STATUS_NOT_SUPPORTED;
         break;
     }
-    irp->IoStatus.u.Status = status;
+    irp->IoStatus.Status = status;
     IoCompleteRequest( irp, IO_NO_INCREMENT );
     return status;
 }
@@ -618,8 +569,11 @@ NTSTATUS WINAPI DriverEntry( DRIVER_OBJECT *driver, UNICODE_STRING *path )
 #ifdef _WIN64
     HKEY wow64_ports_key = NULL;
 #endif
-    void *instance;
-    UNICODE_STRING nameW, linkW;
+    UNICODE_STRING device_mount_point_manager = RTL_CONSTANT_STRING( L"\\Device\\MountPointManager" );
+    UNICODE_STRING object_mount_point_manager = RTL_CONSTANT_STRING( L"\\??\\MountPointManager" );
+    UNICODE_STRING driver_harddisk = RTL_CONSTANT_STRING( L"\\Driver\\Harddisk" );
+    UNICODE_STRING driver_serial = RTL_CONSTANT_STRING( L"\\Driver\\Serial" );
+    UNICODE_STRING driver_parallel = RTL_CONSTANT_STRING( L"\\Driver\\Parallel" );
     DEVICE_OBJECT *device;
     HKEY devicemap_key;
     NTSTATUS status;
@@ -627,17 +581,13 @@ NTSTATUS WINAPI DriverEntry( DRIVER_OBJECT *driver, UNICODE_STRING *path )
 
     TRACE( "%s\n", debugstr_w(path->Buffer) );
 
-    RtlPcToFileHeader( DriverEntry, &instance );
-    status = NtQueryVirtualMemory( GetCurrentProcess(), instance, MemoryWineUnixFuncs,
-                                   &mountmgr_handle, sizeof(mountmgr_handle), NULL );
+    status = __wine_init_unix_call();
     if (status) return status;
 
     driver->MajorFunction[IRP_MJ_DEVICE_CONTROL] = mountmgr_ioctl;
 
-    RtlInitUnicodeString( &nameW, L"\\Device\\MountPointManager" );
-    RtlInitUnicodeString( &linkW, L"\\??\\MountPointManager" );
-    if (!(status = IoCreateDevice( driver, 0, &nameW, 0, 0, FALSE, &device )))
-        status = IoCreateSymbolicLink( &linkW, &nameW );
+    if (!(status = IoCreateDevice( driver, 0, &device_mount_point_manager, 0, 0, FALSE, &device )))
+        status = IoCreateSymbolicLink( &object_mount_point_manager, &device_mount_point_manager );
     if (status)
     {
         FIXME( "failed to create device error %lx\n", status );
@@ -651,8 +601,7 @@ NTSTATUS WINAPI DriverEntry( DRIVER_OBJECT *driver, UNICODE_STRING *path )
                           KEY_ALL_ACCESS, NULL, &devicemap_key, NULL ))
         RegCloseKey( devicemap_key );
 
-    RtlInitUnicodeString( &nameW, L"\\Driver\\Harddisk" );
-    status = IoCreateDriver( &nameW, harddisk_driver_entry );
+    status = IoCreateDriver( &driver_harddisk, harddisk_driver_entry );
 
     thread = CreateThread( NULL, 0, device_op_thread, NULL, 0, NULL );
     CloseHandle( CreateThread( NULL, 0, run_loop_thread, thread, 0, NULL ));
@@ -667,11 +616,8 @@ NTSTATUS WINAPI DriverEntry( DRIVER_OBJECT *driver, UNICODE_STRING *path )
     RegCloseKey( wow64_ports_key );
 #endif
 
-    RtlInitUnicodeString( &nameW, L"\\Driver\\Serial" );
-    IoCreateDriver( &nameW, serial_driver_entry );
-
-    RtlInitUnicodeString( &nameW, L"\\Driver\\Parallel" );
-    IoCreateDriver( &nameW, parallel_driver_entry );
+    IoCreateDriver( &driver_serial, serial_driver_entry );
+    IoCreateDriver( &driver_parallel, parallel_driver_entry );
 
     return status;
 }
