@@ -30,6 +30,12 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/stat.h>
+#ifdef HAVE_SYS_STATFS_H
+#include <sys/statfs.h>
+#endif
+#ifdef HAVE_SYS_STATVFS_H
+# include <sys/statvfs.h>
+#endif
 #include <unistd.h>
 
 #include <pthread.h>
@@ -37,25 +43,62 @@
 #include <wine/list.h>
 
 #include "unixlib.h"
+#include "wine/debug.h"
+
+WINE_DEFAULT_DEBUG_CHANNEL(mountmgr);
+
+static NTSTATUS errno_to_status( int err )
+{
+    TRACE( "errno = %d\n", err );
+    switch (err)
+    {
+    case EAGAIN:    return STATUS_SHARING_VIOLATION;
+    case EBADF:     return STATUS_INVALID_HANDLE;
+    case EBUSY:     return STATUS_DEVICE_BUSY;
+    case ENOSPC:    return STATUS_DISK_FULL;
+    case EPERM:
+    case EROFS:
+    case EACCES:    return STATUS_ACCESS_DENIED;
+    case ENOTDIR:   return STATUS_OBJECT_PATH_NOT_FOUND;
+    case ENOENT:    return STATUS_OBJECT_NAME_NOT_FOUND;
+    case EISDIR:    return STATUS_INVALID_DEVICE_REQUEST;
+    case EMFILE:
+    case ENFILE:    return STATUS_TOO_MANY_OPENED_FILES;
+    case EINVAL:    return STATUS_INVALID_PARAMETER;
+    case ENOTEMPTY: return STATUS_DIRECTORY_NOT_EMPTY;
+    case EPIPE:     return STATUS_PIPE_DISCONNECTED;
+    case EIO:       return STATUS_DEVICE_NOT_READY;
+#ifdef ENOMEDIUM
+    case ENOMEDIUM: return STATUS_NO_MEDIA_IN_DEVICE;
+#endif
+    case ENXIO:     return STATUS_NO_SUCH_DEVICE;
+    case ENOTTY:
+    case EOPNOTSUPP:return STATUS_NOT_SUPPORTED;
+    case ECONNRESET:return STATUS_PIPE_DISCONNECTED;
+    case EFAULT:    return STATUS_ACCESS_VIOLATION;
+    case ESPIPE:    return STATUS_ILLEGAL_FUNCTION;
+    case ELOOP:     return STATUS_REPARSE_POINT_NOT_RESOLVED;
+#ifdef ETIME /* Missing on FreeBSD */
+    case ETIME:     return STATUS_IO_TIMEOUT;
+#endif
+    case ENOEXEC:   /* ?? */
+    case EEXIST:    /* ?? */
+    default:
+        FIXME( "Converting errno %d to STATUS_UNSUCCESSFUL\n", err );
+        return STATUS_UNSUCCESSFUL;
+    }
+}
 
 static char *get_dosdevices_path( const char *dev )
 {
-    const char *home = getenv( "HOME" );
     const char *prefix = getenv( "WINEPREFIX" );
-    size_t len = (prefix ? strlen(prefix) : strlen(home) + strlen("/.wine")) + sizeof("/dosdevices/") + strlen(dev);
-    char *path = malloc( len );
+    char *path = NULL;
+
+    if (prefix)
+        asprintf( &path, "%s/dosdevices/%s", prefix, dev );
+    else
+        asprintf( &path, "%s/.wine/dosdevices/%s", getenv( "HOME" ), dev );
 
-    if (path)
-    {
-        if (prefix) strcpy( path, prefix );
-        else
-        {
-            strcpy( path, home );
-            strcat( path, "/.wine" );
-        }
-        strcat( path, "/dosdevices/" );
-        strcat( path, dev );
-    }
     return path;
 }
 
@@ -78,7 +121,7 @@ static void detect_devices( const char **paths, char *names, ULONG size )
 
         for (;;)
         {
-            int len = sprintf( unix_path, *paths, i++ );
+            int len = snprintf( unix_path, sizeof(unix_path), *paths, i++ );
             if (len + 2 > size) break;
             if (access( unix_path, F_OK ) != 0) break;
             strcpy( names, unix_path );
@@ -101,7 +144,7 @@ struct device_info_list
 
 void queue_device_op( enum device_op op, const char *udi, const char *device,
                       const char *mount_point, enum device_type type, const GUID *guid,
-                      const char *serial, const struct scsi_info *scsi_info )
+                      const char *serial, const char *label, const struct scsi_info *scsi_info )
 {
     struct device_info_list *info;
     char *str, *end;
@@ -120,6 +163,7 @@ void queue_device_op( enum device_op op, const char *udi, const char *device,
     ADD_STR(device);
     ADD_STR(mount_point);
     ADD_STR(serial);
+    ADD_STR(label);
 #undef ADD_STR
     if (guid)
     {
@@ -164,6 +208,7 @@ static NTSTATUS dequeue_device_op( void *args )
     if (dst->device) dst->device = (char *)dst + (src->device_info.device - (char *)&src->device_info);
     if (dst->mount_point) dst->mount_point = (char *)dst + (src->device_info.mount_point - (char *)&src->device_info);
     if (dst->serial) dst->serial = (char *)dst + (src->device_info.serial - (char *)&src->device_info);
+    if (dst->label) dst->label = (char *)dst + (src->device_info.label - (char *)&src->device_info);
     if (dst->guid) dst->guid = &dst->guid_buffer;
     if (dst->scsi_info) dst->scsi_info = &dst->scsi_buffer;
 
@@ -285,6 +330,85 @@ static NTSTATUS set_dosdev_symlink( void *args )
     return status;
 }
 
+static NTSTATUS get_volume_size_info( void *args )
+{
+    const struct get_volume_size_info_params *params = args;
+    const char *unix_mount = params->unix_mount;
+    struct size_info *info = params->info;
+
+    struct stat st;
+    ULONGLONG bsize;
+    NTSTATUS status;
+    int fd = -1;
+
+#if !defined(linux) || !defined(HAVE_FSTATFS)
+    struct statvfs stfs;
+#else
+    struct statfs stfs;
+#endif
+
+    if (!unix_mount) return STATUS_NO_SUCH_DEVICE;
+
+    if (unix_mount[0] != '/')
+    {
+        char *path = get_dosdevices_path( unix_mount );
+        if (path) fd = open( path, O_RDONLY );
+        free( path );
+    }
+    else fd = open( unix_mount, O_RDONLY );
+
+    if (fstat( fd, &st ) < 0)
+    {
+        status = errno_to_status( errno );
+        goto done;
+    }
+    if (!S_ISREG(st.st_mode) && !S_ISDIR(st.st_mode))
+    {
+        status = STATUS_INVALID_DEVICE_REQUEST;
+        goto done;
+    }
+
+    /* Linux's fstatvfs is buggy */
+#if !defined(linux) || !defined(HAVE_FSTATFS)
+    if (fstatvfs( fd, &stfs ) < 0)
+    {
+        status = errno_to_status( errno );
+        goto done;
+    }
+    bsize = stfs.f_frsize;
+#else
+    if (fstatfs( fd, &stfs ) < 0)
+    {
+        status = errno_to_status( errno );
+        goto done;
+    }
+    bsize = stfs.f_bsize;
+#endif
+    if (bsize == 2048)  /* assume CD-ROM */
+    {
+        info->bytes_per_sector = 2048;
+        info->sectors_per_allocation_unit = 1;
+    }
+    else
+    {
+        info->bytes_per_sector = 512;
+        info->sectors_per_allocation_unit = 8;
+    }
+
+    info->total_allocation_units =
+        bsize * stfs.f_blocks / (info->bytes_per_sector * info->sectors_per_allocation_unit);
+    info->caller_available_allocation_units =
+        bsize * stfs.f_bavail / (info->bytes_per_sector * info->sectors_per_allocation_unit);
+    info->actual_available_allocation_units =
+        bsize * stfs.f_bfree / (info->bytes_per_sector * info->sectors_per_allocation_unit);
+
+    status = STATUS_SUCCESS;
+
+done:
+    close( fd );
+    return status;
+}
+
 static NTSTATUS get_volume_dos_devices( void *args )
 {
     const struct get_volume_dos_devices_params *params = args;
@@ -309,9 +433,9 @@ static NTSTATUS read_volume_file( void *args )
 {
     const struct read_volume_file_params *params = args;
     int ret, fd = -1;
-    char *name = malloc( strlen(params->volume) + strlen(params->file) + 2 );
+    char *name = NULL;
 
-    sprintf( name, "%s/%s", params->volume, params->file );
+    asprintf( &name, "%s/%s", params->volume, params->file );
 
     if (name[0] != '/')
     {
@@ -386,23 +510,16 @@ static NTSTATUS detect_parallel_ports( void *args )
 static NTSTATUS set_shell_folder( void *args )
 {
     const struct set_shell_folder_params *params = args;
-    const char *folder = params->folder;
-    const char *backup = params->backup;
     const char *link = params->link;
     struct stat st;
     const char *home;
-    char *homelink = NULL;
+    char *folder = NULL, *backup = NULL, *homelink = NULL;
     NTSTATUS status = STATUS_SUCCESS;
 
     if (link && (!strcmp( link, "$HOME" ) || !strncmp( link, "$HOME/", 6 )) && (home = getenv( "HOME" )))
     {
-#ifdef __ANDROID__
-        home = "/sdcard/Download";
-#endif
         link += 5;
-        homelink = malloc( strlen(home) + strlen(link) + 1 );
-        strcpy( homelink, home );
-        strcat( homelink, link );
+        asprintf( &homelink, "%s%s", home, link );
         link = homelink;
     }
 
@@ -413,6 +530,12 @@ static NTSTATUS set_shell_folder( void *args )
         goto done;
     }
 
+    status = ntdll_get_unix_file_name( params->folder, &folder, FILE_OPEN_IF );
+    if (status == STATUS_NO_SUCH_FILE) status = STATUS_SUCCESS;
+    if (status) goto done;
+
+    if (params->create_backup) asprintf( &backup, "%s.backup", folder );
+
     if (!lstat( folder, &st )) /* move old folder/link out of the way */
     {
         if (S_ISLNK( st.st_mode ))
@@ -441,6 +564,8 @@ static NTSTATUS set_shell_folder( void *args )
     }
 
 done:
+    free( backup );
+    free( folder );
     free( homelink );
     return status;
 }
@@ -448,11 +573,17 @@ done:
 static NTSTATUS get_shell_folder( void *args )
 {
     const struct get_shell_folder_params *params = args;
-    int ret = readlink( params->folder, params->buffer, params->size - 1 );
+    char *folder = NULL;
+    NTSTATUS status = ntdll_get_unix_file_name( params->folder, &folder, FILE_OPEN );
 
-    if (ret < 0) return STATUS_OBJECT_NAME_NOT_FOUND;
-    params->buffer[ret] = 0;
-    return STATUS_SUCCESS;
+    if (!status)
+    {
+        int ret = readlink( folder, params->buffer, params->size - 1 );
+        free( folder );
+        if (ret < 0) return STATUS_OBJECT_NAME_NOT_FOUND;
+        params->buffer[ret] = 0;
+    }
+    return status;
 }
 
 const unixlib_entry_t __wine_unix_call_funcs[] =
@@ -462,6 +593,7 @@ const unixlib_entry_t __wine_unix_call_funcs[] =
     add_drive,
     get_dosdev_symlink,
     set_dosdev_symlink,
+    get_volume_size_info,
     get_volume_dos_devices,
     read_volume_file,
     match_unixdev,
@@ -478,6 +610,8 @@ const unixlib_entry_t __wine_unix_call_funcs[] =
     enumerate_credentials,
 };
 
+C_ASSERT( ARRAYSIZE(__wine_unix_call_funcs) == unix_funcs_count );
+
 #ifdef _WIN64
 
 typedef ULONG PTR32;
@@ -493,6 +627,7 @@ static NTSTATUS wow64_dequeue_device_op(void *args)
         PTR32 device;
         PTR32 mount_point;
         PTR32 serial;
+        PTR32 label;
         PTR32 guid;
         PTR32 scsi_info;
 
@@ -535,6 +670,10 @@ static NTSTATUS wow64_dequeue_device_op(void *args)
         dst->serial = PtrToUlong(dst->str_buffer + (src.serial - src.str_buffer));
     else
         dst->serial = 0;
+    if (src.label)
+        dst->label = PtrToUlong(dst->str_buffer + (src.label - src.str_buffer));
+    else
+        dst->label = 0;
     if (src.guid)
         dst->guid = PtrToUlong(&dst->guid_buffer);
     else
@@ -599,6 +738,21 @@ static NTSTATUS wow64_set_dosdev_symlink(void *args)
     return set_dosdev_symlink(&params);
 }
 
+static NTSTATUS wow64_get_volume_size_info(void *args)
+{
+    struct
+    {
+        PTR32 unix_mount;
+        PTR32 info;
+    } *params32 = args;
+    struct get_volume_size_info_params params =
+    {
+        ULongToPtr(params32->unix_mount),
+        ULongToPtr(params32->info),
+    };
+    return get_volume_size_info(&params);
+}
+
 static NTSTATUS wow64_get_volume_dos_devices(void *args)
 {
     struct
@@ -692,14 +846,14 @@ static NTSTATUS wow64_set_shell_folder(void *args)
     struct
     {
         PTR32 folder;
-        PTR32 backup;
         PTR32 link;
+        ULONG create_backup;
     } *params32 = args;
     struct set_shell_folder_params params =
     {
         ULongToPtr(params32->folder),
-        ULongToPtr(params32->backup),
         ULongToPtr(params32->link),
+        params32->create_backup,
     };
     return set_shell_folder(&params);
 }
@@ -802,6 +956,7 @@ const unixlib_entry_t __wine_unix_call_wow64_funcs[] =
     wow64_add_drive,
     wow64_get_dosdev_symlink,
     wow64_set_dosdev_symlink,
+    wow64_get_volume_size_info,
     wow64_get_volume_dos_devices,
     wow64_read_volume_file,
     wow64_match_unixdev,
@@ -818,4 +973,6 @@ const unixlib_entry_t __wine_unix_call_wow64_funcs[] =
     wow64_enumerate_credentials,
 };
 
+C_ASSERT( ARRAYSIZE(__wine_unix_call_funcs) == unix_funcs_count );
+
 #endif  /* _WIN64 */
