@@ -75,6 +75,28 @@ struct ip_hdr
     uint32_t daddr;
 };
 
+struct ipv6_hdr
+{
+    uint8_t v_prio; /* version << 4 | priority */
+    uint8_t flow_lbl[3];
+    uint16_t next_len;
+    uint8_t next_hdr;
+    uint8_t hop_limit;
+    struct in6_addr saddr;
+    struct in6_addr daddr;
+};
+
+C_ASSERT( sizeof(struct ipv6_hdr) == 40 );
+
+struct ipv6_pseudo_header
+{
+    struct in6_addr src;
+    struct in6_addr dst;
+    UINT32 next_len; /* incapsulated packet length in network byte order */
+    BYTE zero[3];
+    BYTE next_header;
+};
+
 struct icmp_hdr
 {
     uint8_t type;
@@ -113,6 +135,12 @@ struct icmp_data
     unsigned short id;
     unsigned short seq;
     const struct family_ops *ops;
+    struct sockaddr_storage src_storage;
+    struct sockaddr_storage dst_storage;
+    int src_len;
+    int dst_len;
+    int hop_limit;
+    BOOL ping_socket;
 };
 
 #define MAX_HANDLES 256 /* Max number of simultaneous pings - could become dynamic if need be */
@@ -185,11 +213,13 @@ static void ipv4_init_icmp_hdr( struct icmp_data *data, struct icmp_hdr *icmp_hd
 }
 
 /* rfc 1071 checksum */
-static unsigned short chksum( BYTE *data, unsigned int count )
+static unsigned short chksum( struct icmp_data *icmp_data, BYTE *data, unsigned int count )
 {
     unsigned int sum = 0, carry = 0;
     unsigned short check, s;
 
+    if (icmp_data->ping_socket) return 0;
+
     while (count > 1)
     {
         s = *(unsigned short *)data;
@@ -212,13 +242,6 @@ static unsigned short chksum( BYTE *data, unsigned int count )
     return check;
 }
 
-#ifdef __linux__
-static unsigned short null_chksum( BYTE *data, unsigned int count )
-{
-    return 0;
-}
-#endif
-
 static int ipv4_set_reply_ip_status( IP_STATUS ip_status, unsigned int bits, void *out )
 {
     if (bits == 32)
@@ -322,8 +345,8 @@ static BOOL ipv4_linux_ping_parse_ip_hdr( struct msghdr *msg, int recvd, int *ip
 }
 #endif
 
-static int ipv4_parse_icmp_hdr_( struct icmp_data *data, struct icmp_hdr *icmp, int icmp_size,
-                                 struct icmp_reply_ctx *ctx, int ping_socket )
+static int ipv4_parse_icmp_hdr( struct icmp_data *data, struct icmp_hdr *icmp, int icmp_size,
+                                struct icmp_reply_ctx *ctx )
 {
     static const IP_STATUS unreach_codes[] =
     {
@@ -352,7 +375,7 @@ static int ipv4_parse_icmp_hdr_( struct icmp_data *data, struct icmp_hdr *icmp,
     switch (icmp->type)
     {
     case ICMP4_ECHO_REPLY:
-        if ((!ping_socket && icmp->un.echo.id != data->id) ||
+        if ((!data->ping_socket && icmp->un.echo.id != data->id) ||
             icmp->un.echo.sequence != data->seq) return -1;
 
         ctx->status = IP_SUCCESS;
@@ -384,6 +407,8 @@ static int ipv4_parse_icmp_hdr_( struct icmp_data *data, struct icmp_hdr *icmp,
         return -1;
     }
 
+    if (data->ping_socket) return 0;
+
     /* Check that the appended packet is really ours -
      * all handled icmp replies have an 8-byte header
      * followed by the original ip hdr. */
@@ -395,27 +420,13 @@ static int ipv4_parse_icmp_hdr_( struct icmp_data *data, struct icmp_hdr *icmp,
     orig_icmp_hdr = (const struct icmp_hdr *)((const BYTE *)orig_ip_hdr + orig_ip_hdr_len);
     if (orig_icmp_hdr->type != ICMP4_ECHO_REQUEST ||
         orig_icmp_hdr->code != 0 ||
-        (!ping_socket && orig_icmp_hdr->un.echo.id != data->id) ||
+        (!data->ping_socket && orig_icmp_hdr->un.echo.id != data->id) ||
         orig_icmp_hdr->un.echo.sequence != data->seq) return -1;
 
     ctx->status = status;
     return 0;
 }
 
-static int ipv4_parse_icmp_hdr( struct icmp_data *data, struct icmp_hdr *icmp,
-                                int icmp_size, struct icmp_reply_ctx *ctx)
-{
-    return ipv4_parse_icmp_hdr_( data, icmp, icmp_size, ctx, 0 );
-}
-
-#ifdef __linux__
-static int ipv4_linux_ping_parse_icmp_hdr( struct icmp_data *data, struct icmp_hdr *icmp,
-                                           int icmp_size, struct icmp_reply_ctx *ctx )
-{
-    return ipv4_parse_icmp_hdr_( data, icmp, icmp_size, ctx, 1 );
-}
-#endif
-
 static void ipv4_fill_reply( struct icmp_listen_params *params, struct icmp_reply_ctx *ctx)
 {
     void *options_data;
@@ -468,7 +479,7 @@ struct family_ops
     int family;
     int icmp_protocol;
     void (*init_icmp_hdr)( struct icmp_data *data, struct icmp_hdr *icmp_hdr );
-    unsigned short (*chksum)( BYTE *data, unsigned int count );
+    unsigned short (*chksum)( struct icmp_data *icmp_data, BYTE *data, unsigned int count );
     int (*set_reply_ip_status)( IP_STATUS ip_status, unsigned int bits, void *out );
     void (*set_socket_opts)( struct icmp_data *data, struct icmp_send_echo_params *params );
     int (*reply_buffer_len)( struct icmp_listen_params *params );
@@ -498,16 +509,213 @@ static const struct family_ops ipv4_linux_ping =
     AF_INET,
     IPPROTO_ICMP,
     ipv4_init_icmp_hdr,
-    null_chksum,
+    chksum,
     ipv4_set_reply_ip_status,
     ipv4_linux_ping_set_socket_opts,
     ipv4_linux_ping_reply_buffer_len,
     ipv4_linux_ping_parse_ip_hdr,
-    ipv4_linux_ping_parse_icmp_hdr,
+    ipv4_parse_icmp_hdr,
     ipv4_fill_reply,
 };
 #endif
 
+static void ipv6_init_icmp_hdr( struct icmp_data *data, struct icmp_hdr *icmp_hdr )
+{
+    icmp_hdr->type = ICMP6_ECHO_REQUEST;
+    icmp_hdr->code = 0;
+    icmp_hdr->checksum = 0;
+    icmp_hdr->un.echo.id = data->id = getpid() & 0xffff; /* will be overwritten for linux ping socks */
+    icmp_hdr->un.echo.sequence = data->seq = InterlockedIncrement( &icmp_sequence ) & 0xffff;
+}
+
+static unsigned short ipv6_chksum( struct icmp_data *icmp_data, BYTE *data, unsigned int count )
+{
+    struct ipv6_pseudo_header *ip_h;
+    struct sockaddr_in6 addr;
+    unsigned short sum;
+    socklen_t slen;
+    int s, ret;
+
+    if (icmp_data->ping_socket) return 0;
+
+    /* Determine source address. Do it on a separate socket or raw socket won't receive ICMP replies
+     * originating not from the destination address. */
+    s = socket( AF_INET6, SOCK_RAW, IPPROTO_ICMPV6 );
+    if (s < 0)
+    {
+        TRACE( "failed to open raw sock, trying a dgram sock\n" );
+        s = socket( AF_INET6, SOCK_DGRAM, IPPROTO_ICMPV6 );
+        if (s < 0)
+        {
+            WARN( "Unable to create socket\n" );
+            return 0;
+        }
+    }
+    if (bind( s, (void *)&icmp_data->src_storage, icmp_data->src_len ))
+    {
+        close( s );
+        return 0;
+    }
+    if (connect( s, (void *)&icmp_data->dst_storage, icmp_data->dst_len ))
+    {
+        close( s );
+        return 0;
+    }
+    slen = sizeof(addr);
+    ret = getsockname( s, (void *)&addr, &slen );
+    close( s );
+    if (ret) return 0;
+
+    ip_h = malloc( sizeof(*ip_h) + count );
+    if (!ip_h) return 0;
+    memset( ip_h, 0, sizeof(*ip_h) );
+    ip_h->src = addr.sin6_addr;
+    ip_h->dst = ((struct sockaddr_in6 *)&icmp_data->dst_storage)->sin6_addr;
+    ip_h->next_len = htonl( count );
+    ip_h->next_header = IPPROTO_ICMPV6;
+    memcpy( ip_h + 1, data, count );
+    sum = chksum( icmp_data, (BYTE *)ip_h, sizeof(*ip_h) + count );
+    free( ip_h );
+    return sum;
+}
+
+static int ipv6_set_reply_ip_status( IP_STATUS ip_status, unsigned int bits, void *out )
+{
+    ICMPV6_ECHO_REPLY *reply = out;
+
+    memset( reply, 0, sizeof(*reply) );
+    reply->Status = ip_status;
+    return sizeof(*reply);
+}
+
+static void ipv6_set_socket_opts( struct icmp_data *data, struct icmp_send_echo_params *params )
+{
+#ifdef IPV6_UNICAST_HOPS
+{
+    int val = params->hop_limit;
+
+    setsockopt( data->socket, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &val, sizeof(val) );
+}
+#endif
+}
+
+static int ipv6_reply_buffer_len( struct icmp_listen_params *params )
+{
+    return sizeof(struct icmp_hdr) + params->reply_len - sizeof(ICMPV6_ECHO_REPLY);
+}
+
+static BOOL ipv6_parse_ip_hdr( struct msghdr *msg, int recvd, int *ip_hdr_len,
+                                          struct icmp_reply_ctx *ctx )
+{
+    *ip_hdr_len = 0;
+    ctx->options_data = NULL;
+    ctx->ttl = 0;
+    ctx->tos = 0;
+    ctx->flags = 0;
+    ctx->options_size = 0;
+    return TRUE;
+}
+
+static int ipv6_parse_icmp_hdr( struct icmp_data *data, struct icmp_hdr *icmp,
+                                int icmp_size, struct icmp_reply_ctx *ctx )
+{
+    static const IP_STATUS unreach_codes[] =
+    {
+        IP_DEST_NO_ROUTE,
+        IP_DEST_PROHIBITED,
+        IP_DEST_SCOPE_MISMATCH,
+        IP_DEST_ADDR_UNREACHABLE,
+        IP_DEST_PORT_UNREACHABLE,
+        IP_ICMP_ERROR,
+        IP_DEST_UNREACHABLE,
+        IP_BAD_HEADER,
+    };
+    const struct ipv6_hdr *orig_ip_hdr;
+    const struct icmp_hdr *orig_icmp_hdr;
+    IP_STATUS status;
+
+    switch (icmp->type)
+    {
+    case ICMP6_ECHO_REPLY:
+        if ((!data->ping_socket && icmp->un.echo.id != data->id) ||
+            icmp->un.echo.sequence != data->seq) return -1;
+
+        ctx->status = IP_SUCCESS;
+        return icmp_size - sizeof(*icmp);
+
+    case ICMP6_DST_UNREACH:
+        if (icmp->code < ARRAY_SIZE(unreach_codes))
+            status = unreach_codes[icmp->code];
+        else
+            status = IP_DEST_HOST_UNREACHABLE;
+        break;
+
+    case ICMP6_PACKET_TOO_BIG:
+        status = IP_PACKET_TOO_BIG;
+        break;
+
+    case ICMP6_TIME_EXCEEDED:
+        switch(icmp->code)
+        {
+        case 0: status = IP_HOP_LIMIT_EXCEEDED; break;
+        case 1: status = IP_REASSEMBLY_TIME_EXCEEDED; break;
+        default: status = IP_TIME_EXCEEDED; break;
+        }
+        break;
+
+    case ICMP6_PARAM_PROB:
+        status = IP_PARAMETER_PROBLEM;
+        break;
+
+    default:
+        return -1;
+    }
+
+    if (data->ping_socket) return 0;
+    /* Check that the appended packet is really ours. */
+    if (icmp_size < sizeof(*icmp) + sizeof(*orig_ip_hdr)) return -1;
+    orig_ip_hdr = (struct ipv6_hdr *)(icmp + 1);
+    if ((orig_ip_hdr->v_prio >> 4) != 6 || orig_ip_hdr->next_hdr != IPPROTO_ICMPV6) return -1;
+    if (icmp_size < sizeof(*icmp) + sizeof(*orig_ip_hdr) + sizeof(*orig_icmp_hdr)) return -1;
+    orig_icmp_hdr = (const struct icmp_hdr *)((const BYTE *)orig_ip_hdr + sizeof(*orig_ip_hdr));
+    if (orig_icmp_hdr->type != ICMP6_ECHO_REQUEST ||
+        orig_icmp_hdr->code != 0 ||
+        (!data->ping_socket && orig_icmp_hdr->un.echo.id != data->id) ||
+        orig_icmp_hdr->un.echo.sequence != data->seq) return -1;
+
+    ctx->status = status;
+
+    return 0;
+}
+
+static void ipv6_fill_reply( struct icmp_listen_params *params, struct icmp_reply_ctx *ctx)
+{
+    ICMPV6_ECHO_REPLY *reply = params->reply;
+
+    reply->Status = ctx->status;
+    memcpy( reply->Address.sin6_addr, &ctx->addr.Ipv6.sin6_addr, sizeof(ctx->addr.Ipv6.sin6_addr) );
+    reply->Address.sin6_flowinfo = ctx->addr.Ipv6.sin6_flowinfo;
+    reply->Address.sin6_port = ctx->addr.Ipv6.sin6_port;
+    reply->Address.sin6_scope_id = ctx->addr.Ipv6.sin6_scope_id;
+    reply->RoundTripTime = ctx->round_trip_time;
+    memcpy( reply + 1, ctx->data, ctx->data_size );
+    params->reply_len = sizeof(*reply) + ctx->data_size;
+}
+
+static const struct family_ops ipv6 =
+{
+    AF_INET6,
+    IPPROTO_ICMPV6,
+    ipv6_init_icmp_hdr,
+    ipv6_chksum,
+    ipv6_set_reply_ip_status,
+    ipv6_set_socket_opts,
+    ipv6_reply_buffer_len,
+    ipv6_parse_ip_hdr,
+    ipv6_parse_icmp_hdr,
+    ipv6_fill_reply,
+};
+
 static IP_STATUS errno_to_ip_status( int err )
 {
     switch( err )
@@ -580,12 +788,14 @@ static NTSTATUS icmp_data_create( ADDRESS_FAMILY win_family, struct icmp_data **
     struct icmp_data *data;
     const struct family_ops *ops;
 
-    if (win_family == WS_AF_INET) ops = &ipv4;
+    if (win_family == WS_AF_INET6)     ops = &ipv6;
+    else if (win_family == WS_AF_INET) ops = &ipv4;
     else return STATUS_INVALID_PARAMETER;
 
     data = malloc( sizeof(*data) );
     if (!data) return STATUS_NO_MEMORY;
 
+    data->ping_socket = FALSE;
     data->socket = socket( ops->family, SOCK_RAW, ops->icmp_protocol );
     if (data->socket < 0) /* Try a ping-socket */
     {
@@ -599,6 +809,7 @@ static NTSTATUS icmp_data_create( ADDRESS_FAMILY win_family, struct icmp_data **
         }
 #ifdef __linux__
         if (ops->family == AF_INET) ops = &ipv4_linux_ping;
+        data->ping_socket = TRUE;
 #endif
     }
     if (pipe( data->cancel_pipe ))
@@ -625,14 +836,27 @@ NTSTATUS icmp_send_echo( void *args )
 {
     struct icmp_send_echo_params *params = args;
     struct icmp_hdr *icmp_hdr; /* this is the same for both ipv4 and ipv6 */
-    struct sockaddr_storage dst_storage;
-    struct sockaddr *dst = (struct sockaddr *)&dst_storage;
     struct icmp_data *data;
-    int dst_len, ret;
+    int ret;
+    struct sockaddr *src, *dst;
+
     NTSTATUS status;
 
     status = icmp_data_create( params->dst->si_family, &data );
     if (status) return status;
+
+    data->hop_limit = params->hop_limit;
+    src = (struct sockaddr *)&data->src_storage;
+    dst = (struct sockaddr *)&data->dst_storage;
+    data->src_len = SOCKADDR_INET_to_sockaddr( params->src, src, sizeof(data->src_storage) );
+    data->dst_len = SOCKADDR_INET_to_sockaddr( params->dst, dst, sizeof(data->dst_storage) );
+
+    if (bind( data->socket, src, data->src_len ))
+    {
+        icmp_data_free( data );
+        return STATUS_INVALID_ADDRESS_COMPONENT;
+    }
+
     data->ops->set_socket_opts( data, params );
 
     icmp_hdr = malloc( sizeof(*icmp_hdr) + params->request_size );
@@ -643,12 +867,10 @@ NTSTATUS icmp_send_echo( void *args )
     }
     data->ops->init_icmp_hdr( data, icmp_hdr );
     memcpy( icmp_hdr + 1, params->request, params->request_size );
-    icmp_hdr->checksum = data->ops->chksum( (BYTE *)icmp_hdr, sizeof(*icmp_hdr) + params->request_size );
-
-    dst_len = SOCKADDR_INET_to_sockaddr( params->dst, dst, sizeof(dst_storage) );
+    icmp_hdr->checksum = data->ops->chksum( data, (BYTE *)icmp_hdr, sizeof(*icmp_hdr) + params->request_size );
 
     NtQueryPerformanceCounter( &data->send_time, NULL );
-    ret = sendto( data->socket, icmp_hdr, sizeof(*icmp_hdr) + params->request_size, 0, dst, dst_len );
+    ret = sendto( data->socket, icmp_hdr, sizeof(*icmp_hdr) + params->request_size, 0, dst, data->dst_len );
     free( icmp_hdr );
 
     if (ret < 0)
@@ -664,7 +886,7 @@ NTSTATUS icmp_send_echo( void *args )
     return *params->handle ? STATUS_PENDING : STATUS_NO_MEMORY;
 }
 
-static int get_timeout( LARGE_INTEGER start, DWORD timeout )
+static int get_timeout( LARGE_INTEGER start, UINT timeout )
 {
     LARGE_INTEGER now, end;
 
@@ -745,6 +967,13 @@ NTSTATUS icmp_listen( void *args )
     data = handle_data( params->handle );
     if (!data) return STATUS_INVALID_PARAMETER;
 
+    if (data->dst_storage.ss_family == AF_INET6 && !data->hop_limit)
+    {
+        TRACE( "Invalid hop_limit.\n" );
+        params->reply_len = data->ops->set_reply_ip_status( IP_GENERAL_FAILURE, params->bits, params->reply );
+        return STATUS_SUCCESS;
+    }
+
     fds[0].fd = data->socket;
     fds[0].events = POLLIN;
     fds[1].fd = data->cancel_pipe[0];
