@@ -31,8 +31,6 @@
 #include <string.h>
 
 #define COBJMACROS
-#define NONAMELESSUNION
-
 #include "windef.h"
 #include "winbase.h"
 #include "winerror.h"
@@ -735,6 +733,7 @@ HRESULT WINAPI DoDragDrop (
   TrackerWindowInfo trackerInfo;
   HWND            hwndTrackWindow;
   MSG             msg;
+  HCURSOR         cursor;
 
   TRACE("%p, %p, %#lx, %p.\n", pDataObject, pDropSource, dwOKEffect, pdwEffect);
 
@@ -769,6 +768,9 @@ HRESULT WINAPI DoDragDrop (
 
     msg.message = 0;
 
+    /* save cursor */
+    cursor = GetCursor();
+
     /*
      * Pump messages. All mouse input should go to the capture window.
      */
@@ -815,6 +817,9 @@ HRESULT WINAPI DoDragDrop (
      */
     DestroyWindow(hwndTrackWindow);
 
+    /* restore cursor */
+    SetCursor(cursor);
+
     return trackerInfo.returnValue;
   }
 
@@ -2010,64 +2015,64 @@ void WINAPI ReleaseStgMedium(
     case TYMED_HGLOBAL:
     {
       if ( (pmedium->pUnkForRelease==0) &&
-	   (pmedium->u.hGlobal!=0) )
-	GlobalFree(pmedium->u.hGlobal);
+	   (pmedium->hGlobal!=0) )
+	GlobalFree(pmedium->hGlobal);
       break;
     }
     case TYMED_FILE:
     {
-      if (pmedium->u.lpszFileName!=0)
+      if (pmedium->lpszFileName!=0)
       {
 	if (pmedium->pUnkForRelease==0)
 	{
-	  DeleteFileW(pmedium->u.lpszFileName);
+	  DeleteFileW(pmedium->lpszFileName);
 	}
 
-	CoTaskMemFree(pmedium->u.lpszFileName);
+	CoTaskMemFree(pmedium->lpszFileName);
       }
       break;
     }
     case TYMED_ISTREAM:
     {
-      if (pmedium->u.pstm!=0)
+      if (pmedium->pstm!=0)
       {
-	IStream_Release(pmedium->u.pstm);
+	IStream_Release(pmedium->pstm);
       }
       break;
     }
     case TYMED_ISTORAGE:
     {
-      if (pmedium->u.pstg!=0)
+      if (pmedium->pstg!=0)
       {
-	IStorage_Release(pmedium->u.pstg);
+	IStorage_Release(pmedium->pstg);
       }
       break;
     }
     case TYMED_GDI:
     {
       if ( (pmedium->pUnkForRelease==0) &&
-	   (pmedium->u.hBitmap!=0) )
-	DeleteObject(pmedium->u.hBitmap);
+	   (pmedium->hBitmap!=0) )
+	DeleteObject(pmedium->hBitmap);
       break;
     }
     case TYMED_MFPICT:
     {
       if ( (pmedium->pUnkForRelease==0) &&
-	   (pmedium->u.hMetaFilePict!=0) )
+	   (pmedium->hMetaFilePict!=0) )
       {
-	LPMETAFILEPICT pMP = GlobalLock(pmedium->u.hMetaFilePict);
+	LPMETAFILEPICT pMP = GlobalLock(pmedium->hMetaFilePict);
 	DeleteMetaFile(pMP->hMF);
-	GlobalUnlock(pmedium->u.hMetaFilePict);
-	GlobalFree(pmedium->u.hMetaFilePict);
+	GlobalUnlock(pmedium->hMetaFilePict);
+	GlobalFree(pmedium->hMetaFilePict);
       }
       break;
     }
     case TYMED_ENHMF:
     {
       if ( (pmedium->pUnkForRelease==0) &&
-	   (pmedium->u.hEnhMetaFile!=0) )
+	   (pmedium->hEnhMetaFile!=0) )
       {
-	DeleteEnhMetaFile(pmedium->u.hEnhMetaFile);
+	DeleteEnhMetaFile(pmedium->hEnhMetaFile);
       }
       break;
     }
@@ -2794,7 +2799,7 @@ static HRESULT get_storage(IDataObject *data, IStorage *stg, UINT *src_cf, BOOL
     /* CF_EMBEDEDOBJECT */
     init_fmtetc(&fmt, embedded_object_clipboard_format, TYMED_ISTORAGE);
     med.tymed = TYMED_ISTORAGE;
-    med.u.pstg = stg;
+    med.pstg = stg;
     med.pUnkForRelease = NULL;
     hr = IDataObject_GetDataHere(data, &fmt, &med);
     if(SUCCEEDED(hr))
@@ -2806,7 +2811,7 @@ static HRESULT get_storage(IDataObject *data, IStorage *stg, UINT *src_cf, BOOL
     /* CF_EMBEDSOURCE */
     init_fmtetc(&fmt, embed_source_clipboard_format, TYMED_ISTORAGE);
     med.tymed = TYMED_ISTORAGE;
-    med.u.pstg = stg;
+    med.pstg = stg;
     med.pUnkForRelease = NULL;
     hr = IDataObject_GetDataHere(data, &fmt, &med);
     if(SUCCEEDED(hr))
