@@ -48,6 +48,7 @@ WINE_DEFAULT_DEBUG_CHANNEL(exec);
 
 typedef UINT_PTR (*SHELL_ExecuteW32)(const WCHAR *lpCmd, WCHAR *env, BOOL shWait,
 			    const SHELLEXECUTEINFOW *sei, LPSHELLEXECUTEINFOW sei_out);
+extern BOOL WINAPI PathResolveW(void *path, const WCHAR **paths, DWORD flags);
 
 static inline BOOL isSpace(WCHAR c)
 {
@@ -77,7 +78,7 @@ static BOOL SHELL_ArgifyW(WCHAR* out, int len, const WCHAR* fmt, const WCHAR* lp
     BOOL    found_p1 = FALSE;
     PWSTR   res = out;
     PCWSTR  cmd;
-    DWORD   used = 0;
+    DWORD   size, used = 0;
 
     TRACE("%p, %d, %s, %s, %p, %p\n", out, len, debugstr_w(fmt),
           debugstr_w(lpFile), pidl, args);
@@ -110,8 +111,6 @@ static BOOL SHELL_ArgifyW(WCHAR* out, int len, const WCHAR* fmt, const WCHAR* lp
                     if (*fmt == '*')
                     {
                         used++;
-                        if (used < len)
-                            *res++ = '"';
                         while(*args)
                         {
                             used++;
@@ -121,8 +120,6 @@ static BOOL SHELL_ArgifyW(WCHAR* out, int len, const WCHAR* fmt, const WCHAR* lp
                                 args++;
                         }
                         used++;
-                        if (used < len)
-                            *res++ = '"';
                     }
                     else
                     {
@@ -168,11 +165,16 @@ static BOOL SHELL_ArgifyW(WCHAR* out, int len, const WCHAR* fmt, const WCHAR* lp
             case 'l':
             case 'L':
 		if (lpFile) {
-		    used += lstrlenW(lpFile);
+		    size = SearchPathW(NULL, lpFile, L".exe", ARRAY_SIZE(xlpFile), xlpFile, NULL);
+		    if (size && size <= ARRAY_SIZE(xlpFile))
+		        cmd = xlpFile;
+		    else
+		        cmd = lpFile;
+		    used += lstrlenW(cmd);
 		    if (used < len)
 		    {
-			lstrcpyW(res, lpFile);
-			res += lstrlenW(lpFile);
+			lstrcpyW(res, cmd);
+			res += lstrlenW(cmd);
 		    }
 		}
 		found_p1 = TRUE;
@@ -284,7 +286,7 @@ static HRESULT SHELL_GetPathFromIDListForExecuteW(LPCITEMIDLIST pidl, LPWSTR psz
 	hr = IShellFolder_GetDisplayNameOf(desktop, pidl, SHGDN_FORPARSING, &strret);
 
 	if (SUCCEEDED(hr))
-	    StrRetToStrNW(pszPath, uOutSize, &strret, pidl);
+	    StrRetToBufW(&strret, pidl, pszPath, uOutSize);
 
 	IShellFolder_Release(desktop);
     }
@@ -292,6 +294,21 @@ static HRESULT SHELL_GetPathFromIDListForExecuteW(LPCITEMIDLIST pidl, LPWSTR psz
     return hr;
 }
 
+static HANDLE get_admin_token(void)
+{
+    TOKEN_ELEVATION_TYPE type;
+    TOKEN_LINKED_TOKEN linked;
+    DWORD size;
+
+    if (!GetTokenInformation(GetCurrentThreadEffectiveToken(), TokenElevationType, &type, sizeof(type), &size)
+            || type == TokenElevationTypeFull)
+        return NULL;
+
+    if (!GetTokenInformation(GetCurrentThreadEffectiveToken(), TokenLinkedToken, &linked, sizeof(linked), &size))
+        return NULL;
+    return linked.LinkedToken;
+}
+
 /*************************************************************************
  *	SHELL_ExecuteW [Internal]
  *
@@ -305,6 +322,7 @@ static UINT_PTR SHELL_ExecuteW(const WCHAR *lpCmd, WCHAR *env, BOOL shWait,
     UINT gcdret = 0;
     WCHAR curdir[MAX_PATH];
     DWORD dwCreationFlags;
+    HANDLE token = NULL;
 
     TRACE("Execute %s from directory %s\n", debugstr_w(lpCmd), debugstr_w(psei->lpDirectory));
 
@@ -326,8 +344,12 @@ static UINT_PTR SHELL_ExecuteW(const WCHAR *lpCmd, WCHAR *env, BOOL shWait,
     dwCreationFlags = CREATE_UNICODE_ENVIRONMENT;
     if (!(psei->fMask & SEE_MASK_NO_CONSOLE))
         dwCreationFlags |= CREATE_NEW_CONSOLE;
-    if (CreateProcessW(NULL, (LPWSTR)lpCmd, NULL, NULL, FALSE, dwCreationFlags, env,
-                       NULL, &startup, &info))
+
+    if (psei->lpVerb && !wcsicmp(psei->lpVerb, L"runas"))
+        token = get_admin_token();
+
+    if (CreateProcessAsUserW(token, NULL, (LPWSTR)lpCmd, NULL, NULL, FALSE,
+            dwCreationFlags, env, NULL, &startup, &info))
     {
         /* Give 30 seconds to the app to come up, if desired. Probably only needed
            when starting app immediately before making a DDE connection. */
@@ -347,6 +369,8 @@ static UINT_PTR SHELL_ExecuteW(const WCHAR *lpCmd, WCHAR *env, BOOL shWait,
         retval = ERROR_BAD_FORMAT;
     }
 
+    CloseHandle(token);
+
     TRACE("returning %Iu\n", retval);
 
     psei_out->hInstApp = (HINSTANCE)retval;
@@ -436,7 +460,7 @@ static void *SHELL_BuildEnvW( const WCHAR *path )
     if (!got_path) total += 5;  /* we need to create PATH */
     total++;  /* terminating null */
 
-    if (!(new_env = heap_alloc( total * sizeof(WCHAR) )))
+    if (!(new_env = malloc( total * sizeof(WCHAR) )))
     {
         FreeEnvironmentStringsW( strings );
         return NULL;
@@ -485,10 +509,22 @@ static BOOL SHELL_TryAppPathW( LPCWSTR szName, LPWSTR lpResult, WCHAR **env)
     BOOL found = FALSE;
 
     if (env) *env = NULL;
-    lstrcpyW(buffer, L"Software\\Microsoft\\Windows\\CurrentVersion\\App Paths\\");
-    lstrcatW(buffer, szName);
+    wcscpy(buffer, L"Software\\Microsoft\\Windows\\CurrentVersion\\App Paths\\");
+    if (wcslen(buffer) + wcslen(szName) + 1 > ARRAY_SIZE(buffer))
+    {
+        WARN("Name is too big.\n");
+        return FALSE;
+    }
+
+    wcscat(buffer, szName);
     res = RegOpenKeyExW(HKEY_LOCAL_MACHINE, buffer, 0, KEY_READ, &hkApp);
-    if (res) goto end;
+    if (res)
+    {
+        wcscat(buffer, L".exe");
+        res = RegOpenKeyExW(HKEY_LOCAL_MACHINE, buffer, 0, KEY_READ, &hkApp);
+    }
+    if (res)
+        goto end;
 
     len = MAX_PATH*sizeof(WCHAR);
     res = RegQueryValueW(hkApp, NULL, lpResult, &len);
@@ -530,6 +566,18 @@ static UINT SHELL_FindExecutableByVerb(LPCWSTR lpVerb, LPWSTR key, LPWSTR classn
     HKEY hkeyClass;
     WCHAR verb[MAX_PATH];
 
+    if (*classname == '.')
+    {
+        /* Extension, default key value holds class name. Extension may also have
+         * empty class name and shell\verb\command subkey. */
+        WCHAR class[MAX_PATH];
+        LONG len;
+
+        len = sizeof(class);
+        if (!RegQueryValueW(HKEY_CLASSES_ROOT, classname, class, &len) && *class)
+            wcscpy(classname, class);
+    }
+
     if (RegOpenKeyExW(HKEY_CLASSES_ROOT, classname, 0, 0x02000000, &hkeyClass))
         return SE_ERR_NOASSOC;
     if (!HCR_GetDefaultVerbW(hkeyClass, lpVerb, verb, ARRAY_SIZE(verb)))
@@ -584,6 +632,7 @@ static UINT SHELL_FindExecutableByVerb(LPCWSTR lpVerb, LPWSTR key, LPWSTR classn
  *
  * Utility for code sharing between FindExecutable and ShellExecute
  * in:
+ *      lpPath the path to search for the file
  *      lpFile the name of a file
  *      lpVerb the operation on it (open)
  * out:
@@ -603,8 +652,10 @@ static UINT SHELL_FindExecutable(LPCWSTR lpPath, LPCWSTR lpFile, LPCWSTR lpVerb,
     WCHAR wBuffer[256];      /* Used to GetProfileString */
     UINT  retval = SE_ERR_NOASSOC;
     WCHAR *tok;              /* token pointer */
-    WCHAR xlpFile[256];      /* result of SearchPath */
+    WCHAR xlpFile[MAX_PATH]; /* result of SearchPath */
     DWORD attribs;           /* file attributes */
+    WCHAR curdir[MAX_PATH];
+    const WCHAR *search_paths[3] = {0};
 
     TRACE("%s\n", debugstr_w(lpFile));
 
@@ -629,17 +680,64 @@ static UINT SHELL_FindExecutable(LPCWSTR lpPath, LPCWSTR lpFile, LPCWSTR lpVerb,
         return 33;
     }
 
-    if (SearchPathW(lpPath, lpFile, L".exe", ARRAY_SIZE(xlpFile), xlpFile, NULL))
+    GetCurrentDirectoryW(ARRAY_SIZE(curdir), curdir);
+    if (!PathIsFileSpecW(lpFile))
     {
-        TRACE("SearchPathW returned non-zero\n");
-        lpFile = xlpFile;
-        /* The file was found in the application-supplied default directory (or the system search path) */
+        BOOL found = FALSE;
+        if (lpPath && *lpPath)
+        {
+            TRACE("ASDF %s\n", debugstr_w(lpPath));
+            PathCombineW(xlpFile, lpPath, lpFile);
+            if (PathFileExistsDefExtW(xlpFile, 0xbf) || PathFileExistsW(xlpFile))
+            {
+                GetFullPathNameW(xlpFile, ARRAY_SIZE(xlpFile), xlpFile, NULL);
+                found = TRUE;
+            }
+        }
+        if (!found)
+        {
+            if (wcsncpy_s(xlpFile, ARRAY_SIZE(xlpFile), lpFile, _TRUNCATE))
+                return ERROR_FILE_NOT_FOUND;
+
+            if (PathFileExistsDefExtW(xlpFile, 0xbf) || PathFileExistsW(xlpFile))
+            {
+                GetFullPathNameW(xlpFile, ARRAY_SIZE(xlpFile), xlpFile, NULL);
+                found = TRUE;
+            }
+        }
+        if (found)
+        {
+            lpFile = xlpFile;
+            if (wcsncpy_s(lpResult, resultLen, xlpFile, _TRUNCATE))
+                return ERROR_FILE_NOT_FOUND;
+        }
+        else
+            xlpFile[0] = '\0';
     }
-    else if (lpPath && SearchPathW(NULL, lpFile, L".exe", ARRAY_SIZE(xlpFile), xlpFile, NULL))
+    else
     {
-        TRACE("SearchPathW returned non-zero\n");
-        lpFile = xlpFile;
-        /* The file was found in one of the directories in the system-wide search path */
+        if (lpPath && *lpPath)
+        {
+            search_paths[0] = lpPath;
+            search_paths[1] = curdir;
+        }
+        else
+            search_paths[0] = curdir;
+
+        if (wcsncpy_s(xlpFile, ARRAY_SIZE(xlpFile), lpFile, _TRUNCATE))
+            return ERROR_FILE_NOT_FOUND;
+
+        if (PathResolveW(xlpFile, search_paths, PRF_TRYPROGRAMEXTENSIONS | PRF_VERIFYEXISTS) ||
+            PathFindOnPathW(xlpFile, search_paths))
+        {
+            TRACE("PathResolveAW returned non-zero\n");
+            lpFile = xlpFile;
+            if (wcsncpy_s(lpResult, resultLen, xlpFile, _TRUNCATE))
+                return ERROR_FILE_NOT_FOUND;
+            /* The file was found in lpPath or one of the directories in the system-wide search path */
+        }
+        else
+            xlpFile[0] = '\0';
     }
 
     attribs = GetFileAttributesW(lpFile);
@@ -696,7 +794,6 @@ static UINT SHELL_FindExecutable(LPCWSTR lpPath, LPCWSTR lpFile, LPCWSTR lpVerb,
 
                 if (wcsicmp(tok, &extension[1]) == 0) /* have to skip the leading "." */
                 {
-                    lstrcpyW(lpResult, xlpFile);
                     /* Need to perhaps check that the file has a path
                      * attached */
                     TRACE("found %s\n", debugstr_w(lpResult));
@@ -764,16 +861,20 @@ static UINT SHELL_FindExecutable(LPCWSTR lpPath, LPCWSTR lpFile, LPCWSTR lpVerb,
         {
             if (*command)
             {
-                lstrcpyW(lpResult, command);
+                if (wcsncpy_s(lpResult, resultLen, command, _TRUNCATE))
+                    return ERROR_FILE_NOT_FOUND;
                 tok = wcschr(lpResult, '^'); /* should be ^.extension? */
                 if (tok != NULL)
                 {
                     tok[0] = '\0';
-                    lstrcatW(lpResult, xlpFile); /* what if no dir in xlpFile? */
+                    /* what if no dir in xlpFile? */
+                    if (wcscat_s(lpResult, resultLen, xlpFile))
+                        return ERROR_FILE_NOT_FOUND;
                     tok = wcschr(command, '^'); /* see above */
                     if ((tok != NULL) && (lstrlenW(tok)>5))
                     {
-                        lstrcatW(lpResult, &tok[5]);
+                        if (wcscat_s(lpResult, resultLen, &tok[5]))
+                            return ERROR_FILE_NOT_FOUND;
                     }
                 }
                 retval = 33; /* FIXME - see above */
@@ -781,7 +882,7 @@ static UINT SHELL_FindExecutable(LPCWSTR lpPath, LPCWSTR lpFile, LPCWSTR lpVerb,
         }
     }
 
-    TRACE("returning %s\n", debugstr_w(lpResult));
+    TRACE("returning path %s, retval %d\n", debugstr_w(lpResult), retval);
     return retval;
 }
 
@@ -960,7 +1061,7 @@ static unsigned dde_connect(const WCHAR* key, const WCHAR* start, WCHAR* ddeexec
     SHELL_ArgifyW(static_res, ARRAY_SIZE(static_res), exec, lpFile, pidl, szCommandline, &resultLen);
     if (resultLen > ARRAY_SIZE(static_res))
     {
-        res = dynamic_res = heap_alloc(resultLen * sizeof(WCHAR));
+        res = dynamic_res = malloc(resultLen * sizeof(WCHAR));
         SHELL_ArgifyW(dynamic_res, resultLen, exec, lpFile, pidl, szCommandline, NULL);
     }
     else
@@ -976,11 +1077,11 @@ static unsigned dde_connect(const WCHAR* key, const WCHAR* start, WCHAR* ddeexec
     else
     {
         DWORD lenA = WideCharToMultiByte(CP_ACP, 0, res, -1, NULL, 0, NULL, NULL);
-        char *resA = heap_alloc(lenA);
+        char *resA = malloc(lenA);
         WideCharToMultiByte(CP_ACP, 0, res, -1, resA, lenA, NULL, NULL);
         hDdeData = DdeClientTransaction( (LPBYTE)resA, lenA, hConv, 0L, 0,
                                          XTYP_EXECUTE, 10000, &tid );
-        heap_free(resA);
+        free(resA);
     }
     if (hDdeData)
         DdeFreeDataHandle(hDdeData);
@@ -988,7 +1089,7 @@ static unsigned dde_connect(const WCHAR* key, const WCHAR* start, WCHAR* ddeexec
         WARN("DdeClientTransaction failed with error %04x\n", DdeGetLastError(ddeInst));
     ret = 33;
 
-    heap_free(dynamic_res);
+    free(dynamic_res);
 
     DdeDisconnect(hConv);
 
@@ -1006,7 +1107,7 @@ static UINT_PTR execute_from_key(LPCWSTR key, LPCWSTR lpFile, WCHAR *env, LPCWST
 			     SHELL_ExecuteW32 execfunc,
                              LPSHELLEXECUTEINFOW psei, LPSHELLEXECUTEINFOW psei_out)
 {
-    WCHAR cmd[256], param[1024], ddeexec[256];
+    WCHAR cmd[256], parambuffer[1024], ddeexec[256], *param = parambuffer;
     LONG cmdlen = sizeof(cmd), ddeexeclen = sizeof(ddeexec);
     UINT_PTR retval = SE_ERR_NOASSOC;
     DWORD resultLen;
@@ -1028,9 +1129,12 @@ static UINT_PTR execute_from_key(LPCWSTR key, LPCWSTR lpFile, WCHAR *env, LPCWST
         if (cmdlen >= ARRAY_SIZE(cmd))
             cmdlen = ARRAY_SIZE(cmd) - 1;
         cmd[cmdlen] = '\0';
-        SHELL_ArgifyW(param, ARRAY_SIZE(param), cmd, lpFile, psei->lpIDList, szCommandline, &resultLen);
-        if (resultLen > ARRAY_SIZE(param))
-            ERR("Argify buffer not large enough, truncating\n");
+        SHELL_ArgifyW(param, ARRAY_SIZE(parambuffer), cmd, lpFile, psei->lpIDList, szCommandline, &resultLen);
+        if (resultLen > ARRAY_SIZE(parambuffer))
+        {
+            if ((param = malloc(resultLen * sizeof(*param))))
+                SHELL_ArgifyW(param, resultLen, cmd, lpFile, psei->lpIDList, szCommandline, &resultLen);
+        }
     }
 
     /* Get the parameters needed by the application
@@ -1053,6 +1157,7 @@ static UINT_PTR execute_from_key(LPCWSTR key, LPCWSTR lpFile, WCHAR *env, LPCWST
     else
         WARN("Nothing appropriate found for %s\n", debugstr_w(key));
 
+    if (param != parambuffer) free(param);
     return retval;
 }
 
@@ -1077,7 +1182,7 @@ HINSTANCE WINAPI FindExecutableA(LPCSTR lpFile, LPCSTR lpDirectory, LPSTR lpResu
     return retval;
 }
 
-static UINT_PTR SHELL_CrossOverFallback(const WCHAR* lpFile, const WCHAR* lpVerb, WCHAR* cmdFormat, int cmdFormatSize)
+static UINT_PTR SHELL_CrossOverFallback(const WCHAR* lpFile, const WCHAR* lpVerb, WCHAR* cmdFormat, int cmdFormatSize, UINT_PTR retval)
 {
     /* See if there is a native association for this file.
      * CrossOver Hack 2412.
@@ -1085,7 +1190,6 @@ static UINT_PTR SHELL_CrossOverFallback(const WCHAR* lpFile, const WCHAR* lpVerb
      * But before that, check that this is not the file we just opened
      * in order to avoid native -> wine -> native association loops.
      */
-    UINT_PTR retval = SE_ERR_NOASSOC;
     WCHAR* nofallback=NULL;
     int size=GetEnvironmentVariableW(L"NOCROSSOVERFALLBACK", NULL, 0);
     if (size)
@@ -1157,7 +1261,7 @@ HINSTANCE WINAPI FindExecutableW(LPCWSTR lpFile, LPCWSTR lpDirectory, LPWSTR lpR
 
     if (retval <= 32)
     {
-        retval=SHELL_CrossOverFallback(lpFile, L"open", lpResult, MAX_PATH*sizeof(*lpResult));
+        retval=SHELL_CrossOverFallback(lpFile, L"open", lpResult, MAX_PATH*sizeof(*lpResult), retval);
         if (retval > 32)
         {
             /* Remove double quotation marks and command line arguments */
@@ -1219,7 +1323,7 @@ static HKEY ShellExecute_GetClassKey( const SHELLEXECUTEINFOW *sei )
         if ( r == ERROR_SUCCESS && type == REG_SZ )
         {
             sz += sizeof (WCHAR);
-            cls = heap_alloc( sz );
+            cls = malloc( sz );
             cls[0] = 0;
             RegQueryValueExW( hkey, NULL, 0, &type, (LPBYTE) cls, &sz );
         }
@@ -1234,7 +1338,7 @@ static HKEY ShellExecute_GetClassKey( const SHELLEXECUTEINFOW *sei )
     if ( lpClass )
         RegOpenKeyW( HKEY_CLASSES_ROOT, lpClass, &hkey );
 
-    heap_free( cls );
+    free( cls );
 
     return hkey;
 }
@@ -1554,7 +1658,7 @@ static UINT_PTR SHELL_quote_and_execute( LPCWSTR wcmd, LPCWSTR wszParameters, LP
         /* Length of space plus length of parameters */
         len += 1 + lstrlenW(wszParameters);
     }
-    wszQuotedCmd = heap_alloc(len * sizeof(WCHAR));
+    wszQuotedCmd = malloc(len * sizeof(WCHAR));
     /* Must quote to handle case where cmd contains spaces,
      * else security hole if malicious user creates executable file "C:\\Program"
      */
@@ -1565,12 +1669,12 @@ static UINT_PTR SHELL_quote_and_execute( LPCWSTR wcmd, LPCWSTR wszParameters, LP
         lstrcatW(wszQuotedCmd, L" ");
         lstrcatW(wszQuotedCmd, wszParameters);
     }
-    TRACE("%s/%s => %s/%s\n", debugstr_w(wszApplicationName), debugstr_w(psei->lpVerb), debugstr_w(wszQuotedCmd), debugstr_w(wszKeyname));
+    TRACE("app %s verb %s => cmd %s key %s\n", debugstr_w(wszApplicationName), debugstr_w(psei->lpVerb), debugstr_w(wszQuotedCmd), debugstr_w(wszKeyname));
     if (*wszKeyname)
         retval = execute_from_key(wszKeyname, wszApplicationName, env, psei->lpParameters, wcmd, execfunc, psei, psei_out);
     else
         retval = execfunc(wszQuotedCmd, env, FALSE, psei, psei_out);
-    heap_free(wszQuotedCmd);
+    free(wszQuotedCmd);
     return retval;
 }
 
@@ -1595,7 +1699,7 @@ static UINT_PTR SHELL_execute_url( LPCWSTR lpFile, LPCWSTR wcmd, LPSHELLEXECUTEI
         len += lstrlenW(psei->lpVerb);
     else
         len += lstrlenW(L"open");
-    lpstrProtocol = heap_alloc(len * sizeof(WCHAR));
+    lpstrProtocol = malloc(len * sizeof(WCHAR));
     memcpy(lpstrProtocol, lpFile, iSize*sizeof(WCHAR));
     lpstrProtocol[iSize] = '\0';
     lstrcatW(lpstrProtocol, L"\\shell\\");
@@ -1604,7 +1708,7 @@ static UINT_PTR SHELL_execute_url( LPCWSTR lpFile, LPCWSTR wcmd, LPSHELLEXECUTEI
 
     retval = execute_from_key(lpstrProtocol, lpFile, NULL, psei->lpParameters,
                               wcmd, execfunc, psei, psei_out);
-    heap_free(lpstrProtocol);
+    free(lpstrProtocol);
     return retval;
 }
 
@@ -1629,13 +1733,13 @@ static WCHAR *expand_environment( const WCHAR *str )
     len = ExpandEnvironmentStringsW(str, NULL, 0);
     if (!len) return NULL;
 
-    buf = heap_alloc(len * sizeof(WCHAR));
+    buf = malloc(len * sizeof(WCHAR));
     if (!buf) return NULL;
 
     len = ExpandEnvironmentStringsW(str, buf, len);
     if (!len)
     {
-        heap_free(buf);
+        free(buf);
         return NULL;
     }
     return buf;
@@ -1678,22 +1782,25 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
     /* make copies of all path/command strings */
     if (!sei_tmp.lpFile)
     {
-        wszApplicationName = heap_alloc(dwApplicationNameLen*sizeof(WCHAR));
+        wszApplicationName = malloc(dwApplicationNameLen * sizeof(WCHAR));
         *wszApplicationName = '\0';
     }
     else if (*sei_tmp.lpFile == '\"' && sei_tmp.lpFile[(len = lstrlenW(sei_tmp.lpFile))-1] == '\"')
     {
         if(len-1 >= dwApplicationNameLen) dwApplicationNameLen = len;
-        wszApplicationName = heap_alloc(dwApplicationNameLen*sizeof(WCHAR));
+        wszApplicationName = malloc(dwApplicationNameLen * sizeof(WCHAR));
         memcpy(wszApplicationName, sei_tmp.lpFile+1, len*sizeof(WCHAR));
         if(len > 2)
             wszApplicationName[len-2] = '\0';
         TRACE("wszApplicationName=%s\n",debugstr_w(wszApplicationName));
     } else {
-        DWORD l = lstrlenW(sei_tmp.lpFile)+1;
-        if(l > dwApplicationNameLen) dwApplicationNameLen = l+1;
-        wszApplicationName = heap_alloc(dwApplicationNameLen*sizeof(WCHAR));
-        memcpy(wszApplicationName, sei_tmp.lpFile, l*sizeof(WCHAR));
+        /* remove trailing spaces */
+        WCHAR *buf = wcsdup(sei->lpFile), *end = buf + wcslen( buf ) - 1;
+        while (end >= buf && *end == ' ') *end-- = 0;
+
+        len = lstrlenW(buf)+1;
+        if(len > dwApplicationNameLen) dwApplicationNameLen = len+1;
+        wszApplicationName = buf;
     }
 
     wszParameters = parametersBuffer;
@@ -1702,7 +1809,7 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
         len = lstrlenW(sei_tmp.lpParameters) + 1;
         if (len > parametersLen)
         {
-            wszParameters = heap_alloc(len * sizeof(WCHAR));
+            wszParameters = malloc(len * sizeof(WCHAR));
             parametersLen = len;
         }
 	lstrcpyW(wszParameters, sei_tmp.lpParameters);
@@ -1715,7 +1822,7 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
     {
         len = lstrlenW(sei_tmp.lpDirectory) + 1;
         if (len > ARRAY_SIZE(dirBuffer))
-            wszDir = heap_alloc(len * sizeof(WCHAR));
+            wszDir = malloc(len * sizeof(WCHAR));
 	lstrcpyW(wszDir, sei_tmp.lpDirectory);
     }
     else
@@ -1745,11 +1852,11 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
 	    IShellExecuteHookW_Release(pSEH);
 
 	    if (hr == S_OK) {
-                heap_free(wszApplicationName);
+                free(wszApplicationName);
                 if (wszParameters != parametersBuffer)
-                    heap_free(wszParameters);
+                    free(wszParameters);
                 if (wszDir != dirBuffer)
-                    heap_free(wszDir);
+                    free(wszDir);
 		return TRUE;
             }
 	}
@@ -1768,7 +1875,7 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
             retval = SE_ERR_OOM;
             goto end;
         }
-        heap_free(wszApplicationName);
+        free(wszApplicationName);
         sei_tmp.lpFile = wszApplicationName = tmp;
 
         tmp = expand_environment(sei_tmp.lpDirectory);
@@ -1778,18 +1885,18 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
             goto end;
         }
         if (wszDir != dirBuffer)
-            heap_free(wszDir);
+            free(wszDir);
         sei_tmp.lpDirectory = wszDir = tmp;
     }
 
     if ( ERROR_SUCCESS == ShellExecute_FromContextMenu( &sei_tmp ) )
     {
         sei->hInstApp = (HINSTANCE) 33;
-        heap_free(wszApplicationName);
+        free(wszApplicationName);
         if (wszParameters != parametersBuffer)
-            heap_free(wszParameters);
+            free(wszParameters);
         if (wszDir != dirBuffer)
-            heap_free(wszDir);
+            free(wszDir);
         return TRUE;
     }
 
@@ -1799,11 +1906,11 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
                                       execfunc );
         if (retval <= 32 && !(sei_tmp.fMask & SEE_MASK_FLAG_NO_UI))
             do_error_dialog(retval, sei_tmp.hwnd);
-        heap_free(wszApplicationName);
+        free(wszApplicationName);
         if (wszParameters != parametersBuffer)
-            heap_free(wszParameters);
+            free(wszParameters);
         if (wszDir != dirBuffer)
-            heap_free(wszDir);
+            free(wszDir);
         return retval > 32;
     }
 
@@ -1823,14 +1930,14 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
         DWORD size;
 
         size = MAX_PATH;
-        buf = heap_alloc(size * sizeof(WCHAR));
+        buf = malloc(size * sizeof(WCHAR));
         if (!buf || FAILED(PathCreateFromUrlW(sei_tmp.lpFile, buf, &size, 0))) {
-            heap_free(buf);
+            free(buf);
             retval = SE_ERR_OOM;
             goto end;
         }
 
-        heap_free(wszApplicationName);
+        free(wszApplicationName);
         wszApplicationName = buf;
         sei_tmp.lpFile = wszApplicationName;
     }
@@ -1840,10 +1947,10 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
         if (len>0)
         {
             LPWSTR buf;
-            buf = heap_alloc((len + 1) * sizeof(WCHAR));
+            buf = malloc((len + 1) * sizeof(WCHAR));
 
             ExpandEnvironmentStringsW(sei_tmp.lpFile, buf, len + 1);
-            heap_free(wszApplicationName);
+            free(wszApplicationName);
             wszApplicationName = buf;
 
             sei_tmp.lpFile = wszApplicationName;
@@ -1852,15 +1959,27 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
 
     if (*sei_tmp.lpDirectory)
     {
+        LPWSTR buf;
         len = ExpandEnvironmentStringsW(sei_tmp.lpDirectory, NULL, 0);
         if (len > 0)
         {
-            LPWSTR buf;
             len++;
-            buf = heap_alloc(len * sizeof(WCHAR));
+            buf = malloc(len * sizeof(WCHAR));
             ExpandEnvironmentStringsW(sei_tmp.lpDirectory, buf, len);
             if (wszDir != dirBuffer)
-                heap_free(wszDir);
+                free(wszDir);
+            wszDir = buf;
+            sei_tmp.lpDirectory = wszDir;
+        }
+
+        len = GetFullPathNameW(sei_tmp.lpDirectory, 0, NULL, NULL);
+        if (len > 0)
+        {
+            len++;
+            buf = malloc(len * sizeof(WCHAR));
+            GetFullPathNameW(sei_tmp.lpDirectory, len, buf, NULL);
+            if (wszDir != dirBuffer)
+                free(wszDir);
             wszDir = buf;
             sei_tmp.lpDirectory = wszDir;
         }
@@ -1870,30 +1989,16 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
     TRACE("execute:%s,%s,%s\n", debugstr_w(wszApplicationName), debugstr_w(wszParameters), debugstr_w(wszDir));
     lpFile = sei_tmp.lpFile;
     wcmd = wcmdBuffer;
-    lstrcpyW(wcmd, wszApplicationName);
-    if (sei_tmp.lpDirectory)
+    len = lstrlenW(wszApplicationName) + 3;
+    if (sei_tmp.lpParameters[0])
+        len += 1 + lstrlenW(wszParameters);
+    if (len > wcmdLen)
     {
-        LPCWSTR searchPath[] = {
-            sei_tmp.lpDirectory,
-            NULL
-        };
-        PathFindOnPathW(wcmd, searchPath);
-    }
-    retval = SHELL_quote_and_execute( wcmd, wszParameters, L"",
-                                      wszApplicationName, NULL, &sei_tmp,
-                                      sei, execfunc );
-    if (retval > 32) {
-        heap_free(wszApplicationName);
-        if (wszParameters != parametersBuffer)
-            heap_free(wszParameters);
-        if (wszDir != dirBuffer)
-            heap_free(wszDir);
-        if (wcmd != wcmdBuffer)
-            heap_free(wcmd);
-        return TRUE;
+        wcmd = malloc(len * sizeof(WCHAR));
+        wcmdLen = len;
     }
 
-    /* Else, try to find the executable */
+    /* try to find the executable */
     wcmd[0] = '\0';
     retval = SHELL_FindExecutable(sei_tmp.lpDirectory, lpFile, sei_tmp.lpVerb, wcmd, wcmdLen, wszKeyname, &env, sei_tmp.lpIDList, sei_tmp.lpParameters);
     if (retval > 32)  /* Found */
@@ -1901,12 +2006,12 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
         retval = SHELL_quote_and_execute( wcmd, wszParameters, wszKeyname,
                                           wszApplicationName, env, &sei_tmp,
                                           sei, execfunc );
-        heap_free( env );
+        free( env );
     }
     else if (PathIsDirectoryW(lpFile))
     {
         WCHAR wExec[MAX_PATH];
-        WCHAR * lpQuotedFile = heap_alloc( sizeof(WCHAR) * (lstrlenW(lpFile) + 3) );
+        WCHAR * lpQuotedFile = malloc( sizeof(WCHAR) * (lstrlenW(lpFile) + 3) );
 
         if (lpQuotedFile)
         {
@@ -1922,9 +2027,9 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
                                                   wszKeyname,
                                                   wszApplicationName, env,
                                                   &sei_tmp, sei, execfunc );
-                heap_free( env );
+                free( env );
             }
-            heap_free( lpQuotedFile );
+            free( lpQuotedFile );
         }
         else
             retval = 0; /* Out of memory */
@@ -1945,7 +2050,7 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
     else
     {
         WCHAR wCmdFormat[1024];
-        retval=SHELL_CrossOverFallback(lpFile, sei_tmp.lpVerb, wCmdFormat, sizeof(wCmdFormat));
+        retval=SHELL_CrossOverFallback(lpFile, sei_tmp.lpVerb, wCmdFormat, sizeof(wCmdFormat), retval);
         if (retval > 32)
         {
 
@@ -1972,13 +2077,13 @@ static BOOL SHELL_execute( LPSHELLEXECUTEINFOW sei, SHELL_ExecuteW32 execfunc )
 end:
     TRACE("retval %Iu\n", retval);
 
-    heap_free(wszApplicationName);
+    free(wszApplicationName);
     if (wszParameters != parametersBuffer)
-        heap_free(wszParameters);
+        free(wszParameters);
     if (wszDir != dirBuffer)
-        heap_free(wszDir);
+        free(wszDir);
     if (wcmd != wcmdBuffer)
-        heap_free(wcmd);
+        free(wcmd);
 
     sei->hInstApp = (HINSTANCE)(retval > 32 ? 33 : retval);
 
