@@ -62,21 +62,11 @@
  *     WH_MOUSE_LL                  Implemented but should use SendMessage instead
  */
 
-#include <stdarg.h>
-#include <assert.h>
-
-#include "windef.h"
-#include "winbase.h"
-#include "winnls.h"
-#include "wingdi.h"
-#include "winuser.h"
-#include "winerror.h"
-#include "win.h"
+#include "ntstatus.h"
+#define WIN32_NO_STATUS
 #include "user_private.h"
-#include "wine/server.h"
 #include "wine/asm.h"
 #include "wine/debug.h"
-#include "winternl.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(hook);
 WINE_DECLARE_DEBUG_CHANNEL(relay);
@@ -84,6 +74,7 @@ WINE_DECLARE_DEBUG_CHANNEL(relay);
 static struct user_api_hook original_user_api =
 {
     USER_DefDlgProc,
+    USER_NonClientButtonDraw,
     USER_ScrollBarDraw,
     USER_ScrollBarProc,
 };
@@ -100,6 +91,16 @@ struct hook_info
     WCHAR module[MAX_PATH];
 };
 
+static CRITICAL_SECTION api_hook_cs;
+static CRITICAL_SECTION_DEBUG critsect_debug =
+{
+    0, 0, &api_hook_cs,
+    { &critsect_debug.ProcessLocksList, &critsect_debug.ProcessLocksList },
+    0, 0, { (DWORD_PTR)(__FILE__ ": api_hook_cs") }
+};
+static CRITICAL_SECTION api_hook_cs = { &critsect_debug, -1, 0, 0, 0, 0 };
+
+
 #define WH_WINEVENT (WH_MAXHOOK+1)
 
 static const char * const hook_names[WH_WINEVENT - WH_MINHOOK + 1] =
@@ -212,6 +213,7 @@ static LRESULT call_hook_AtoW( HOOKPROC proc, INT id, INT code, WPARAM wparam, L
         CREATESTRUCTW csW;
         LPWSTR nameW = NULL;
         LPWSTR classW = NULL;
+        WCHAR name_buf[3];
 
         cbtcwW.lpcs = &csW;
         cbtcwW.hwndInsertAfter = cbtcwA->hwndInsertAfter;
@@ -219,8 +221,18 @@ static LRESULT call_hook_AtoW( HOOKPROC proc, INT id, INT code, WPARAM wparam, L
 
         if (!IS_INTRESOURCE(cbtcwA->lpcs->lpszName))
         {
-            RtlCreateUnicodeStringFromAsciiz(&usBuffer,cbtcwA->lpcs->lpszName);
-            csW.lpszName = nameW = usBuffer.Buffer;
+            if (cbtcwA->lpcs->lpszName[0] != '\xff')
+            {
+                RtlCreateUnicodeStringFromAsciiz( &usBuffer, cbtcwA->lpcs->lpszName );
+                csW.lpszName = nameW = usBuffer.Buffer;
+            }
+            else
+            {
+                name_buf[0] = 0xffff;
+                name_buf[1] = MAKEWORD( cbtcwA->lpcs->lpszName[1], cbtcwA->lpcs->lpszName[2] );
+                name_buf[2] = 0;
+                csW.lpszName = name_buf;
+            }
         }
         if (!IS_INTRESOURCE(cbtcwA->lpcs->lpszClass))
         {
@@ -253,16 +265,29 @@ static LRESULT call_hook_WtoA( HOOKPROC proc, INT id, INT code, WPARAM wparam, L
         int len;
         LPSTR nameA = NULL;
         LPSTR classA = NULL;
+        char name_buf[4];
 
         cbtcwA.lpcs = &csA;
         cbtcwA.hwndInsertAfter = cbtcwW->hwndInsertAfter;
         csA = *(CREATESTRUCTA *)cbtcwW->lpcs;
 
-        if (!IS_INTRESOURCE(cbtcwW->lpcs->lpszName)) {
-            len = WideCharToMultiByte( CP_ACP, 0, cbtcwW->lpcs->lpszName, -1, NULL, 0, NULL, NULL );
-            nameA = HeapAlloc( GetProcessHeap(), 0, len*sizeof(CHAR) );
-            WideCharToMultiByte( CP_ACP, 0, cbtcwW->lpcs->lpszName, -1, nameA, len, NULL, NULL );
-            csA.lpszName = nameA;
+        if (!IS_INTRESOURCE(cbtcwW->lpcs->lpszName))
+        {
+            if (cbtcwW->lpcs->lpszName[0] != 0xffff)
+            {
+                len = WideCharToMultiByte( CP_ACP, 0, cbtcwW->lpcs->lpszName, -1, NULL, 0, NULL, NULL );
+                nameA = HeapAlloc( GetProcessHeap(), 0, len*sizeof(CHAR) );
+                WideCharToMultiByte( CP_ACP, 0, cbtcwW->lpcs->lpszName, -1, nameA, len, NULL, NULL );
+                csA.lpszName = nameA;
+            }
+            else
+            {
+                name_buf[0] = '\xff';
+                name_buf[1] = cbtcwW->lpcs->lpszName[1];
+                name_buf[2] = cbtcwW->lpcs->lpszName[1] >> 8;
+                name_buf[3] = 0;
+                csA.lpszName = name_buf;
+            }
         }
 
         if (!IS_INTRESOURCE(cbtcwW->lpcs->lpszClass)) {
@@ -289,14 +314,14 @@ static LRESULT call_hook_proc( HOOKPROC proc, INT id, INT code, WPARAM wparam, L
 {
     LRESULT ret;
 
-    TRACE_(relay)( "\1Call hook proc %p (id=%s,code=%x,wp=%08lx,lp=%08lx)\n",
+    TRACE_(relay)( "\1Call hook proc %p (id=%s,code=%x,wp=%08Ix,lp=%08Ix)\n",
                    proc, hook_names[id-WH_MINHOOK], code, wparam, lparam );
 
     if (!prev_unicode == !next_unicode) ret = proc( code, wparam, lparam );
     else if (prev_unicode) ret = call_hook_WtoA( proc, id, code, wparam, lparam );
     else ret = call_hook_AtoW( proc, id, code, wparam, lparam );
 
-    TRACE_(relay)( "\1Ret  hook proc %p (id=%s,code=%x,wp=%08lx,lp=%08lx) retval=%08lx\n",
+    TRACE_(relay)( "\1Ret  hook proc %p (id=%s,code=%x,wp=%08Ix,lp=%08Ix) retval=%08Ix\n",
                    proc, hook_names[id-WH_MINHOOK], code, wparam, lparam, ret );
 
     return ret;
@@ -356,21 +381,6 @@ void *get_hook_proc( void *proc, const WCHAR *module, HMODULE *free_module )
 }
 
 
-/***********************************************************************
- *		HOOK_CallHooks
- */
-LRESULT HOOK_CallHooks( INT id, INT code, WPARAM wparam, LPARAM lparam, BOOL unicode )
-{
-    struct win_hook_params params;
-    params.id = id;
-    params.code = code;
-    params.wparam = wparam;
-    params.lparam = lparam;
-    params.next_unicode = unicode;
-    return NtUserCallOneParam( (UINT_PTR)&params, NtUserCallHooks );
-}
-
-
 /***********************************************************************
  *		SetWindowsHookA (USER32.@)
  */
@@ -406,35 +416,6 @@ HHOOK WINAPI SetWindowsHookExW( INT id, HOOKPROC proc, HINSTANCE inst, DWORD tid
 }
 
 
-/***********************************************************************
- *		UnhookWindowsHook (USER32.@)
- */
-BOOL WINAPI UnhookWindowsHook( INT id, HOOKPROC proc )
-{
-    return NtUserUnhookWindowsHook( id, proc );
-}
-
-
-/***********************************************************************
- *		CallMsgFilterA (USER32.@)
- */
-BOOL WINAPI CallMsgFilterA( LPMSG msg, INT code )
-{
-    if (HOOK_CallHooks( WH_SYSMSGFILTER, code, 0, (LPARAM)msg, FALSE )) return TRUE;
-    return HOOK_CallHooks( WH_MSGFILTER, code, 0, (LPARAM)msg, FALSE );
-}
-
-
-/***********************************************************************
- *		CallMsgFilterW (USER32.@)
- */
-BOOL WINAPI CallMsgFilterW( LPMSG msg, INT code )
-{
-    if (HOOK_CallHooks( WH_SYSMSGFILTER, code, 0, (LPARAM)msg, TRUE )) return TRUE;
-    return HOOK_CallHooks( WH_MSGFILTER, code, 0, (LPARAM)msg, TRUE );
-}
-
-
 /***********************************************************************
  *           SetWinEventHook                            [USER32.@]
  *
@@ -461,7 +442,7 @@ HWINEVENTHOOK WINAPI SetWinEventHook(DWORD event_min, DWORD event_max,
     UNICODE_STRING str;
     DWORD len = 0;
 
-    TRACE("%d,%d,%p,%p,%08x,%04x,%08x\n", event_min, event_max, inst,
+    TRACE("%ld,%ld,%p,%p,%08lx,%04lx,%08lx\n", event_min, event_max, inst,
           proc, pid, tid, flags);
 
     if (inst && (!(len = GetModuleFileNameW( inst, module, MAX_PATH )) || len >= MAX_PATH))
@@ -474,41 +455,101 @@ HWINEVENTHOOK WINAPI SetWinEventHook(DWORD event_min, DWORD event_max,
     return NtUserSetWinEventHook( event_min, event_max, inst, &str, proc, pid, tid, flags );
 }
 
-BOOL WINAPI User32CallWinEventHook( const struct win_event_hook_params *params, ULONG size )
+NTSTATUS WINAPI User32CallWinEventHook( void *args, ULONG size )
 {
+    const struct win_event_hook_params *params = args;
     WINEVENTPROC proc = params->proc;
     HMODULE free_module = 0;
 
-    if (params->module[0] && !(proc = get_hook_proc( proc, params->module, &free_module ))) return FALSE;
+    if (params->module[0] && !(proc = get_hook_proc( proc, params->module, &free_module )))
+        return STATUS_INVALID_PARAMETER;
 
-    TRACE_(relay)( "\1Call winevent hook proc %p (hhook=%p,event=%x,hwnd=%p,object_id=%x,child_id=%x,tid=%04x,time=%x)\n",
+    TRACE_(relay)( "\1Call winevent hook proc %p (hhook=%p,event=%lx,hwnd=%p,object_id=%lx,child_id=%lx,tid=%04lx,time=%lx)\n",
                    proc, params->handle, params->event, params->hwnd, params->object_id,
                    params->child_id, params->tid, params->time );
 
     proc( params->handle, params->event, params->hwnd, params->object_id, params->child_id,
           params->tid, params->time );
 
-    TRACE_(relay)( "\1Ret  winevent hook proc %p (hhook=%p,event=%x,hwnd=%p,object_id=%x,child_id=%x,tid=%04x,time=%x)\n",
+    TRACE_(relay)( "\1Ret  winevent hook proc %p (hhook=%p,event=%lx,hwnd=%p,object_id=%lx,child_id=%lx,tid=%04lx,time=%lx)\n",
                    proc, params->handle, params->event, params->hwnd, params->object_id,
                    params->child_id, params->tid, params->time );
 
     if (free_module) FreeLibrary( free_module );
-    return TRUE;
+    return STATUS_SUCCESS;
 }
 
-BOOL WINAPI User32CallWindowsHook( const struct win_hook_params *params, ULONG size )
+NTSTATUS WINAPI User32CallWindowsHook( void *args, ULONG size )
 {
+    struct win_hook_params *params = args;
     HOOKPROC proc = params->proc;
     HMODULE free_module = 0;
+    void *ret_ptr = NULL;
+    CBT_CREATEWNDW cbtc;
+    UINT ret_size = 0;
+    size_t lparam_offset;
     LRESULT ret;
 
-    if (params->module[0] && !(proc = get_hook_proc( proc, params->module, &free_module ))) return FALSE;
+    lparam_offset = FIELD_OFFSET( struct win_hook_params, module[wcslen( params->module ) + 1]);
+
+    if (lparam_offset < size)
+    {
+        lparam_offset = (lparam_offset + 15) & ~15; /* align */
+        ret_size = size - lparam_offset;
+        ret_ptr = (char *)params + lparam_offset;
+        params->lparam = (LPARAM)ret_ptr;
+
+        switch (params->id)
+        {
+        case WH_CBT:
+            if (params->code == HCBT_CREATEWND)
+            {
+                cbtc.hwndInsertAfter = HWND_TOP;
+                unpack_message( (HWND)params->wparam, WM_CREATE, NULL, (LPARAM *)&cbtc.lpcs,
+                                ret_ptr, FALSE );
+                params->lparam = (LPARAM)&cbtc;
+                ret_size = sizeof(*cbtc.lpcs);
+            }
+            break;
+        case WH_CALLWNDPROC:
+            if (ret_size > sizeof(CWPSTRUCT))
+            {
+                CWPSTRUCT *cwp = (CWPSTRUCT *)params->lparam;
+                size_t offset = (lparam_offset + sizeof(*cwp) + 15) & ~15;
+
+                unpack_message( cwp->hwnd, cwp->message, &cwp->wParam, &cwp->lParam,
+                                (char *)params + offset, !params->prev_unicode );
+                ret_size = 0;
+                break;
+            }
+        case WH_CALLWNDPROCRET:
+            if (ret_size > sizeof(CWPRETSTRUCT))
+            {
+                CWPRETSTRUCT *cwpret = (CWPRETSTRUCT *)params->lparam;
+                size_t offset = (lparam_offset + sizeof(*cwpret) + 15) & ~15;
+
+                unpack_message( cwpret->hwnd, cwpret->message, &cwpret->wParam, &cwpret->lParam,
+                                (char *)params + offset, !params->prev_unicode );
+                ret_size = 0;
+                break;
+            }
+        }
+    }
+    if (params->module[0] && !(proc = get_hook_proc( proc, params->module, &free_module )))
+        return FALSE;
 
     ret = call_hook_proc( proc, params->id, params->code, params->wparam, params->lparam,
                           params->prev_unicode, params->next_unicode );
 
     if (free_module) FreeLibrary( free_module );
-    return ret;
+
+    if (ret_size)
+    {
+        LRESULT *result_ptr = (LRESULT *)ret_ptr - 1;
+        *result_ptr = ret;
+        return NtCallbackReturn( result_ptr, sizeof(*result_ptr) + ret_size, STATUS_SUCCESS );
+    }
+    return NtCallbackReturn( &ret, sizeof(ret), STATUS_SUCCESS );
 }
 
 /***********************************************************************
@@ -529,7 +570,7 @@ BOOL WINAPI User32CallWindowsHook( const struct win_hook_params *params, ULONG s
 BOOL WINAPI IsWinEventHookInstalled(DWORD dwEvent)
 {
     /* FIXME: Needed by Office 2007 installer */
-    WARN("(%d)-stub!\n", dwEvent);
+    WARN("(%ld)-stub!\n", dwEvent);
     return TRUE;
 }
 
@@ -539,12 +580,12 @@ BOOL WINAPI RegisterUserApiHook(const struct user_api_hook *new_hook, struct use
     if (!new_hook)
         return FALSE;
 
-    USER_Lock();
+    EnterCriticalSection( &api_hook_cs );
     hooked_user_api = *new_hook;
     user_api = &hooked_user_api;
     if (old_hook)
         *old_hook = original_user_api;
-    USER_Unlock();
+    LeaveCriticalSection( &api_hook_cs );
     return TRUE;
 }
 
