@@ -85,6 +85,7 @@
 #endif /* HAVE_IOKIT_USB_IOUSBLIB_H */
 
 #include <pthread.h>
+#include <sys/utsname.h>
 
 #include "ntstatus.h"
 #define WIN32_NO_STATUS
@@ -524,7 +525,7 @@ static void process_IOService_Device(io_object_t object)
 
     if ((kIOReturnSuccess != err) || (plugInInterface == nil) )
     {
-        ERR("Unable to create plug in interface for USB device");
+        ERR("Unable to create plug in interface for USB device\n");
         goto failed;
     }
 
@@ -599,6 +600,19 @@ static void handle_IOServiceTerminatedCallback(void *refcon, io_iterator_t iter)
     }
 }
 
+static BOOL sequoia_or_later(void)
+{
+    int result;
+    struct utsname name;
+    unsigned major, minor;
+
+    result = (uname( &name ) == 0 &&
+              sscanf( name.release, "%u.%u", &major, &minor ) == 2 &&
+              major >= 24 /* macOS 15 Sequoia */);
+
+    return (result == 1) ? TRUE : FALSE;
+}
+
 NTSTATUS xbox_bus_init(void *args)
 {
     IONotificationPortRef notificationObject;
@@ -609,6 +623,13 @@ NTSTATUS xbox_bus_init(void *args)
 
     TRACE("args %p\n", args);
 
+    /* In Sequoia and later, GameController.framework supports wired XBOX 360 controllers */
+    if (sequoia_or_later())
+    {
+        TRACE("disabled: running on macOS Sequoia or later\n");
+        return STATUS_SUCCESS;
+    }
+
     options = *(struct xbox_bus_options *)args;
 
     run_loop = CFRunLoopGetCurrent();
@@ -645,6 +666,13 @@ NTSTATUS xbox_bus_wait(void *args)
     /* cleanup previously returned event */
     bus_event_cleanup(result);
 
+    /* In Sequoia and later, GameController.framework supports wired XBOX 360 controllers */
+    if (sequoia_or_later())
+    {
+        TRACE("disabled: running on macOS Sequoia or later\n");
+        return STATUS_SUCCESS;
+    }
+
     do
     {
         if (bus_event_queue_pop(&event_queue, result)) return STATUS_PENDING;
