@@ -26,8 +26,11 @@
 #define VKD3D_NO_WIN32_TYPES
 #include "initguid.h"
 #include "wined3d_private.h"
+#include "wined3d_gl.h"
 #include "d3d12.h"
-#include <vkd3d.h>
+#define VK_NO_PROTOTYPES
+#include "wine/vulkan.h"
+#include <vkd3d_utils.h>
 
 WINE_DEFAULT_DEBUG_CHANNEL(d3d);
 WINE_DECLARE_DEBUG_CHANNEL(vkd3d);
@@ -114,7 +117,6 @@ struct wined3d_settings wined3d_settings =
 {
     .cs_multithreaded = WINED3D_CSMT_ENABLE,
     .max_gl_version = MAKEDWORD_VERSION(4, 4),
-    .offscreen_rendering_mode = ORM_FBO,
     .pci_vendor_id = PCI_VENDOR_NONE,
     .pci_device_id = PCI_DEVICE_NONE,
     .multisample_textures = TRUE,
@@ -134,16 +136,22 @@ struct wined3d_settings wined3d_settings =
 struct cxgames_hacks cxgames_hacks =
 {
     FALSE,                      /* safe_vs_consts */
-    WINED3D_MAPBUF_NEVER_NV,    /* Don't use glMapBuffer on dynamic buffers with NV threading */
-    FALSE,                      /* Don't force success when creating transform feedback buffers */
 };
 
-struct wined3d * CDECL wined3d_create(DWORD flags)
+enum wined3d_renderer CDECL wined3d_get_renderer(void)
+{
+    if (wined3d_settings.renderer == WINED3D_RENDERER_AUTO)
+        return WINED3D_RENDERER_OPENGL;
+
+    return wined3d_settings.renderer;
+}
+
+struct wined3d * CDECL wined3d_create(uint32_t flags)
 {
     struct wined3d *object;
     HRESULT hr;
 
-    if (!(object = heap_alloc_zero(FIELD_OFFSET(struct wined3d, adapters[1]))))
+    if (!(object = calloc(1, FIELD_OFFSET(struct wined3d, adapters[1]))))
     {
         ERR("Failed to allocate wined3d object memory.\n");
         return NULL;
@@ -154,8 +162,8 @@ struct wined3d * CDECL wined3d_create(DWORD flags)
 
     if (FAILED(hr = wined3d_init(object, flags)))
     {
-        WARN("Failed to initialize wined3d object, hr %#x.\n", hr);
-        heap_free(object);
+        WARN("Failed to initialize wined3d object, hr %#lx.\n", hr);
+        free(object);
         return NULL;
     }
 
@@ -212,7 +220,7 @@ static DWORD get_config_key(HKEY defkey, HKEY appkey, const char *env, const cha
     return ERROR_FILE_NOT_FOUND;
 }
 
-static DWORD get_config_key_dword(HKEY defkey, HKEY appkey, const char *env, const char *name, DWORD *value)
+static DWORD get_config_key_dword(HKEY defkey, HKEY appkey, const char *env, const char *name, unsigned int *value)
 {
     DWORD type, data, size;
     const char *env_value;
@@ -277,13 +285,13 @@ static BOOL wined3d_dll_init(HINSTANCE hInstDLL)
     const char *env;
     HKEY hkey = 0;
     HKEY appkey = 0;
-    DWORD tmpvalue;
+    unsigned int tmpvalue;
     WNDCLASSA wc;
 
     wined3d_context_tls_idx = TlsAlloc();
     if (wined3d_context_tls_idx == TLS_OUT_OF_INDEXES)
     {
-        DWORD err = GetLastError();
+        unsigned int err = GetLastError();
         ERR("Failed to allocate context TLS index, err %#x.\n", err);
         return FALSE;
     }
@@ -308,7 +316,7 @@ static BOOL wined3d_dll_init(HINSTANCE hInstDLL)
         ERR("Failed to register window class 'WineD3D_OpenGL'!\n");
         if (!TlsFree(wined3d_context_tls_idx))
         {
-            DWORD err = GetLastError();
+            unsigned int err = GetLastError();
             ERR("Failed to free context TLS index, err %#x.\n", err);
         }
         return FALSE;
@@ -356,52 +364,17 @@ static BOOL wined3d_dll_init(HINSTANCE hInstDLL)
         }
         if (!get_config_key(hkey, appkey, env, "shader_backend", buffer, size))
         {
+            if (!stricmp(buffer, "glsl-vkd3d"))
+            {
+                ERR_(winediag)("Using the vkd3d-shader GLSL shader backend.\n");
+                wined3d_settings.shader_backend = WINED3D_SHADER_BACKEND_GLSL_VKD3D;
+            }
             if (!stricmp(buffer, "glsl"))
             {
                 ERR_(winediag)("Using the GLSL shader backend.\n");
                 wined3d_settings.shader_backend = WINED3D_SHADER_BACKEND_GLSL;
             }
-            else if (!stricmp(buffer, "arb"))
-            {
-                ERR_(winediag)("Using the ARB shader backend.\n");
-                wined3d_settings.shader_backend = WINED3D_SHADER_BACKEND_ARB;
-            }
-            else if (!stricmp(buffer, "none"))
-            {
-                ERR_(winediag)("Disabling shader backends.\n");
-                wined3d_settings.shader_backend = WINED3D_SHADER_BACKEND_NONE;
-            }
         }
-        if (wined3d_settings.shader_backend == WINED3D_SHADER_BACKEND_ARB
-                || wined3d_settings.shader_backend == WINED3D_SHADER_BACKEND_NONE)
-        {
-            ERR_(winediag)("The GLSL shader backend has been disabled. You get to keep all the pieces if it breaks.\n");
-            TRACE("Use of GL Shading Language disabled.\n");
-        }
-        if ( !get_config_key( hkey, appkey, env, "hl2_disable_glsl", buffer, size) )
-        {
-            /* This allows disabling GLSL for single HL2 mods(they all use hl2.exe).
-             * It is important that this key is evaluated *after* the UseGLS one,
-             * otherwise the useGLSL key may overwrite decisions made here
-             */
-            char *token;
-            LPSTR cmdline;
-            cmdline = GetCommandLineA();
-            TRACE("Checking command line for disabling GLSL per HL2 mod\n");
-            token = strtok(buffer, ";");
-            while(token) {
-                TRACE("Looking for \"%s\"\n", token);
-                if(strstr(cmdline, token)) {
-                    TRACE("Disabling GLSL for this HL2 mod\n");
-                    wined3d_settings.shader_backend = WINED3D_SHADER_BACKEND_ARB;
-                    break;
-                }
-                token = strtok(NULL, ";");
-            }
-        }
-        if (!get_config_key(hkey, appkey, env, "OffscreenRenderingMode", buffer, size)
-                && !strcmp(buffer,"backbuffer"))
-            wined3d_settings.offscreen_rendering_mode = ORM_BACKBUFFER;
         if (!get_config_key_dword(hkey, appkey, env, "VideoPciDeviceID", &tmpvalue))
         {
             int pci_device_id = tmpvalue;
@@ -449,7 +422,7 @@ static BOOL wined3d_dll_init(HINSTANCE hInstDLL)
         {
             size_t len = strlen(buffer) + 1;
 
-            if (!(wined3d_settings.logo = heap_alloc(len)))
+            if (!(wined3d_settings.logo = malloc(len)))
                 ERR("Failed to allocate logo path memory.\n");
             else
                 memcpy(wined3d_settings.logo, buffer, len);
@@ -507,28 +480,15 @@ static BOOL wined3d_dll_init(HINSTANCE hInstDLL)
                 wined3d_settings.renderer = WINED3D_RENDERER_NO3D;
             }
         }
-        if (!get_config_key(hkey, appkey, env, "AllowGlMapBuffer", buffer, size))
-        {
-            if (!strcmp(buffer, "always"))
-            {
-                TRACE("Always using glMapBuffer if possible.\n");
-                cxgames_hacks.allow_glmapbuffer = WINED3D_MAPBUF_ALWAYS;
-            }
-            else if (!strcmp(buffer, "never") || !strcmp(buffer, "static"))
-            {
-                TRACE("Never using glMapBuffer.\n");
-                cxgames_hacks.allow_glmapbuffer = WINED3D_MAPBUF_NEVER;
-            }
-        }
         if (!get_config_key_dword(hkey, appkey, env, "cb_access_map_w", &tmpvalue) && tmpvalue)
         {
             TRACE("Forcing all constant buffers to be write-mappable.\n");
             wined3d_settings.cb_access_map_w = TRUE;
         }
-        if (!get_config_key_dword(hkey, appkey, env, "force_transform_feedback", &tmpvalue) && tmpvalue)
+        if (!get_config_key_dword(hkey, appkey, env, "ffp_hlsl", &tmpvalue))
         {
-            TRACE("Forcing success on transform feedback buffer creation.\n");
-            cxgames_hacks.force_transform_feedback = TRUE;
+            ERR_(winediag)("Using the HLSL-based FFP backend.\n");
+            wined3d_settings.ffp_hlsl = tmpvalue;
         }
     }
 
@@ -552,7 +512,7 @@ static BOOL wined3d_dll_init(HINSTANCE hInstDLL)
         else putenv( "VKD3D_SHADER_DEBUG=none" );
     }
 
-    vkd3d_set_log_callback(vkd3d_log_callback);
+    vkd3d_utils_set_log_callback(vkd3d_log_callback);
 
     return TRUE;
 }
@@ -564,7 +524,7 @@ static BOOL wined3d_dll_destroy(HINSTANCE hInstDLL)
 
     if (!TlsFree(wined3d_context_tls_idx))
     {
-        DWORD err = GetLastError();
+        unsigned int err = GetLastError();
         ERR("Failed to free context TLS index, err %#x.\n", err);
     }
 
@@ -578,17 +538,17 @@ static BOOL wined3d_dll_destroy(HINSTANCE hInstDLL)
          * these entries. */
         WARN("Leftover wndproc table entry %p.\n", &wndproc_table.entries[i]);
     }
-    heap_free(wndproc_table.entries);
+    free(wndproc_table.entries);
 
-    heap_free(swapchain_state_table.states);
+    free(swapchain_state_table.states);
     for (i = 0; i < swapchain_state_table.hook_count; ++i)
     {
         WARN("Leftover swapchain state hook %p.\n", &swapchain_state_table.hooks[i]);
         UnhookWindowsHookEx(swapchain_state_table.hooks[i].hook);
     }
-    heap_free(swapchain_state_table.hooks);
+    free(swapchain_state_table.hooks);
 
-    heap_free(wined3d_settings.logo);
+    free(wined3d_settings.logo);
     UnregisterClassA(WINED3D_OPENGL_WINDOW_CLASS_NAME, hInstDLL);
 
     DeleteCriticalSection(&wined3d_command_cs);
@@ -632,7 +592,7 @@ static struct wined3d_output * wined3d_get_output_from_window(const struct wined
     monitor_info.cbSize = sizeof(monitor_info);
     if (!GetMonitorInfoW(monitor, (MONITORINFO *)&monitor_info))
     {
-        ERR("GetMonitorInfoW failed, error %#x.\n", GetLastError());
+        ERR("GetMonitorInfoW failed, error %#lx.\n", GetLastError());
         return NULL;
     }
 
