@@ -22,43 +22,6 @@
 
 #include "macdrv_cocoa.h"
 
-#if !defined(MAC_OS_X_VERSION_10_12) || MAC_OS_X_VERSION_MAX_ALLOWED < MAC_OS_X_VERSION_10_12
-/* For older SDKs, #define the new names of constants deprecated/renamed in macOS 10.12. */
-#define NSCompositingOperationSourceOver    NSCompositeSourceOver
-#define NSEventMaskAny                      NSAnyEventMask
-#define NSEventMaskLeftMouseDown            NSLeftMouseDownMask
-#define NSEventMaskOtherMouseDown           NSOtherMouseDownMask
-#define NSEventMaskRightMouseDown           NSRightMouseDownMask
-#define NSEventModifierFlagCommand          NSCommandKeyMask
-#define NSEventModifierFlagControl          NSControlKeyMask
-#define NSEventModifierFlagOption           NSAlternateKeyMask
-#define NSEventModifierFlagShift            NSShiftKeyMask
-#define NSEventTypeAppKitDefined            NSAppKitDefined
-#define NSEventTypeApplicationDefined       NSApplicationDefined
-#define NSEventTypeFlagsChanged             NSFlagsChanged
-#define NSEventTypeKeyDown                  NSKeyDown
-#define NSEventTypeKeyUp                    NSKeyUp
-#define NSEventTypeLeftMouseDown            NSLeftMouseDown
-#define NSEventTypeLeftMouseDragged         NSLeftMouseDragged
-#define NSEventTypeLeftMouseUp              NSLeftMouseUp
-#define NSEventTypeMouseMoved               NSMouseMoved
-#define NSEventTypeOtherMouseDown           NSOtherMouseDown
-#define NSEventTypeOtherMouseDragged        NSOtherMouseDragged
-#define NSEventTypeOtherMouseUp             NSOtherMouseUp
-#define NSEventTypeRightMouseDown           NSRightMouseDown
-#define NSEventTypeRightMouseUp             NSRightMouseUp
-#define NSEventTypeRightMouseDragged        NSRightMouseDragged
-#define NSEventTypeScrollWheel              NSScrollWheel
-#define NSWindowStyleMaskBorderless         NSBorderlessWindowMask
-#define NSWindowStyleMaskClosable           NSClosableWindowMask
-#define NSWindowStyleMaskFullScreen         NSFullScreenWindowMask
-#define NSWindowStyleMaskMiniaturizable     NSMiniaturizableWindowMask
-#define NSWindowStyleMaskResizable          NSResizableWindowMask
-#define NSWindowStyleMaskTitled             NSTitledWindowMask
-#define NSWindowStyleMaskUtilityWindow      NSUtilityWindowMask
-#define NSWindowStyleMaskNonactivatingPanel NSNonactivatingPanelMask
-#endif
-
 #define ERR(...) do { if (macdrv_err_on) LogError(__func__, __VA_ARGS__); } while (false)
 
 
@@ -92,7 +55,6 @@ enum {
     CGEventSourceKeyboardType keyboardType;
     NSEvent* lastFlagsChanged;
     BOOL inputSourceIsInputMethod;
-    BOOL inputSourceIsInputMethodValid;
     uint32_t pressedKeyCodes[128 / 32];
 
     CGFloat primaryScreenHeight;
@@ -128,11 +90,14 @@ enum {
 
     NSImage* applicationIcon;
 
+    /* CW Hack 22310, 24199 */
+    NSString *explicitAppUserModelID;
+    BOOL terminatingDueToExternalRequest;
+
     BOOL beenActive;
 
     NSMutableSet* windowsBeingDragged;
 
-    BOOL useDragNotifications;
     // CrossOver Hack 10912: Mac Edit menu
     NSMutableArray* changedKeyEquivalents;
 }
@@ -142,6 +107,7 @@ enum {
 @property (readonly, nonatomic) BOOL areDisplaysCaptured;
 @property (readonly, nonatomic) BOOL displaysTemporarilyUncapturedForDialog;
 @property (readonly, nonatomic) BOOL temporarilyIgnoreResignEventsForDialog;
+@property (nonatomic, copy) NSString *explicitAppUserModelID;  /* CW Hack 22310 */
 
 @property (readonly) BOOL clippingCursor;
 @property (nonatomic) NSTimeInterval lastSetCursorPositionTime;
@@ -149,6 +115,7 @@ enum {
     + (WineApplicationController*) sharedController;
 
     - (void) transformProcessToForeground:(BOOL)activateIfTransformed;
+    - (void) tryToActivateIgnoringOtherApps:(BOOL)ignore;
 
     - (BOOL) registerEventQueue:(WineEventQueue*)queue;
     - (void) unregisterEventQueue:(WineEventQueue*)queue;
@@ -158,7 +125,7 @@ enum {
 
     - (void) windowGotFocus:(WineWindow*)window;
 
-    - (BOOL) waitUntilQueryDone:(int*)done timeout:(NSDate*)timeout processEvents:(BOOL)processEvents;
+    - (BOOL) waitUntilQueryDone:(bool*)done timeout:(NSDate*)timeout processEvents:(BOOL)processEvents;
 
     - (void) noteKey:(uint16_t)keyCode pressed:(BOOL)pressed;
 
