@@ -22,11 +22,11 @@
 #include <mach/mach.h>
 #include <mach/mach_time.h>
 
-#include "wine/hostaddrspace_enter.h"
-
 #include "macdrv_cocoa.h"
 #import "cocoa_app.h"
 
+#pragma GCC diagnostic ignored "-Wdeclaration-after-statement"
+
 
 /* Condition values for an NSConditionLock. Used to signal between run_cocoa_app
    and macdrv_start_cocoa_app so the latter knows when the former is running
@@ -66,44 +66,42 @@ static void run_cocoa_app(void* info)
     NSConditionLock* lock = startup_info->lock;
     BOOL created_app = FALSE;
 
-    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
-
-    if (!NSApp)
+    @autoreleasepool
     {
-        [WineApplication sharedApplication];
-        created_app = TRUE;
-
-        /* CrossOver hack 11196: Disable loading of input managers */
-        [[NSUserDefaults standardUserDefaults] registerDefaults:
-            [NSDictionary dictionaryWithObject:@"NO" forKey:@"NSUseCocoaInputServers"]];
+        if (!NSApp)
+        {
+            [WineApplication sharedApplication];
+            created_app = TRUE;
+        }
 
         /* CrossOver hack 12205: Prevent call to NSVersionOfRunTimeLibrary() during app startup.
                                  It can crash if a Wine thread unloads a dylib simultaneously. */
         [[NSUserDefaults standardUserDefaults] registerDefaults:
             [NSDictionary dictionaryWithObject:@"YES" forKey:@"NSUseActiveDisplayForMainScreen"]];
+
+        if ([NSApp respondsToSelector:@selector(setWineController:)])
+        {
+            WineApplicationController* controller = [WineApplicationController sharedController];
+            [NSApp setWineController:controller];
+            [controller computeEventTimeAdjustmentFromTicks:startup_info->tickcount uptime:startup_info->uptime_ns];
+            startup_info->success = TRUE;
+        }
+
+        /* Retain the lock while we're using it, so macdrv_start_cocoa_app()
+           doesn't deallocate it in the middle of us unlocking it. */
+        [lock retain];
+        [lock lock];
+        [lock unlockWithCondition:COCOA_APP_RUNNING];
+        [lock release];
     }
 
-    if ([NSApp respondsToSelector:@selector(setWineController:)])
-    {
-        WineApplicationController* controller = [WineApplicationController sharedController];
-        [NSApp setWineController:controller];
-        [controller computeEventTimeAdjustmentFromTicks:startup_info->tickcount uptime:startup_info->uptime_ns];
-        startup_info->success = TRUE;
-    }
-
-    /* Retain the lock while we're using it, so macdrv_start_cocoa_app()
-       doesn't deallocate it in the middle of us unlocking it. */
-    [lock retain];
-    [lock lock];
-    [lock unlockWithCondition:COCOA_APP_RUNNING];
-    [lock release];
-
-    [pool release];
-
     if (created_app && startup_info->success)
     {
-        /* Never returns */
-        [NSApp run];
+        @autoreleasepool
+        {
+            /* Never returns */
+            [NSApp run];
+        }
     }
 }
 
@@ -116,6 +114,8 @@ static void run_cocoa_app(void* info)
  * Returns 0 on success, non-zero on failure.
  */
 int macdrv_start_cocoa_app(unsigned long long tickcount)
+{
+@autoreleasepool
 {
     int ret = -1;
     CFRunLoopSourceRef source;
@@ -125,8 +125,6 @@ int macdrv_start_cocoa_app(unsigned long long tickcount)
     NSDate* timeLimit;
     CFRunLoopSourceContext source_context = { 0 };
 
-    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
-
     /* Make sure Cocoa is in multi-threading mode by detaching a
        do-nothing thread. */
     [NSThread detachNewThreadSelector:@selector(self)
@@ -162,6 +160,6 @@ int macdrv_start_cocoa_app(unsigned long long tickcount)
     if (source)
         CFRelease(source);
     [startup_info.lock release];
-    [pool release];
     return ret;
 }
+}
