@@ -20,13 +20,24 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
  */
 
+#if 0
+#pragma makedep unix
+#endif
+
 #include "config.h"
 
+#include <poll.h>
+
+#include "ntstatus.h"
+#define WIN32_NO_STATUS
 #include "macdrv.h"
-#include "winuser.h"
+#include "oleidl.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(event);
+WINE_DECLARE_DEBUG_CHANNEL(imm);
 
+pthread_mutex_t ime_composition_rect_mutex = PTHREAD_MUTEX_INITIALIZER;
+CGRect ime_composition_rect;
 
 /* return the name of an Mac event */
 static const char *dbgstr_event(int type)
@@ -35,6 +46,7 @@ static const char *dbgstr_event(int type)
         "APP_ACTIVATED",
         "APP_DEACTIVATED",
         "APP_QUIT_REQUESTED",
+        "CLIENT_SURFACE_PRESENTED", /* CW HACK 22435 */
         "DISPLAYS_CHANGED",
         "EDIT_MENU_COMMAND", /* CrossOver Hack 10912: Mac Edit menu */
         "HOTKEY_PRESS",
@@ -44,18 +56,18 @@ static const char *dbgstr_event(int type)
         "KEYBOARD_CHANGED",
         "LOST_PASTEBOARD_OWNERSHIP",
         "MOUSE_BUTTON",
-        "MOUSE_MOVED",
+        "MOUSE_MOVED_RELATIVE",
         "MOUSE_MOVED_ABSOLUTE",
         "MOUSE_SCROLL",
         "QUERY_EVENT",
         "QUERY_EVENT_NO_PREEMPT_WAIT",
         "REASSERT_WINDOW_POSITION",
         "RELEASE_CAPTURE",
-        "SENT_TEXT_INPUT",
         "STATUS_ITEM_MOUSE_BUTTON",
         "STATUS_ITEM_MOUSE_MOVE",
         "WINDOW_BROUGHT_FORWARD",
         "WINDOW_CLOSE_REQUESTED",
+        "WINDOW_DID_MINIMIZE",
         "WINDOW_DID_UNMINIMIZE",
         "WINDOW_DRAG_BEGIN",
         "WINDOW_DRAG_END",
@@ -67,6 +79,7 @@ static const char *dbgstr_event(int type)
         "WINDOW_RESIZE_ENDED",
         "WINDOW_RESTORE_REQUESTED",
     };
+    C_ASSERT(ARRAYSIZE(event_names) == NUM_EVENT_TYPES);
 
     if (0 <= type && type < NUM_EVENT_TYPES) return event_names[type];
     return wine_dbg_sprintf("Unknown event %d", type);
@@ -102,7 +115,7 @@ static macdrv_event_mask get_event_mask(DWORD mask)
 
     if (mask & QS_MOUSEMOVE)
     {
-        event_mask |= event_mask_for_type(MOUSE_MOVED);
+        event_mask |= event_mask_for_type(MOUSE_MOVED_RELATIVE);
         event_mask |= event_mask_for_type(MOUSE_MOVED_ABSOLUTE);
     }
 
@@ -111,6 +124,7 @@ static macdrv_event_mask get_event_mask(DWORD mask)
         event_mask |= event_mask_for_type(APP_ACTIVATED);
         event_mask |= event_mask_for_type(APP_DEACTIVATED);
         event_mask |= event_mask_for_type(APP_QUIT_REQUESTED);
+        event_mask |= event_mask_for_type(CLIENT_SURFACE_PRESENTED); /* CW HACK 22435 */
         event_mask |= event_mask_for_type(DISPLAYS_CHANGED);
         event_mask |= event_mask_for_type(IM_SET_TEXT);
         event_mask |= event_mask_for_type(LOST_PASTEBOARD_OWNERSHIP);
@@ -128,7 +142,6 @@ static macdrv_event_mask get_event_mask(DWORD mask)
         event_mask |= event_mask_for_type(QUERY_EVENT_NO_PREEMPT_WAIT);
         event_mask |= event_mask_for_type(REASSERT_WINDOW_POSITION);
         event_mask |= event_mask_for_type(RELEASE_CAPTURE);
-        event_mask |= event_mask_for_type(SENT_TEXT_INPUT);
         event_mask |= event_mask_for_type(WINDOW_BROUGHT_FORWARD);
         event_mask |= event_mask_for_type(WINDOW_CLOSE_REQUESTED);
         event_mask |= event_mask_for_type(WINDOW_DRAG_BEGIN);
@@ -142,6 +155,168 @@ static macdrv_event_mask get_event_mask(DWORD mask)
     return event_mask;
 }
 
+static void post_ime_update( HWND hwnd, UINT cursor_pos, WCHAR *comp_str, WCHAR *result_str )
+{
+    NtUserMessageCall( hwnd, WINE_IME_POST_UPDATE, cursor_pos, (LPARAM)comp_str,
+                       result_str, NtUserImeDriverCall, FALSE );
+}
+
+/***********************************************************************
+ *              macdrv_im_set_text
+ */
+static void macdrv_im_set_text(const macdrv_event *event)
+{
+    HWND hwnd = macdrv_get_window_hwnd(event->window);
+    WCHAR *text = NULL;
+
+    TRACE_(imm)("win %p/%p himc %p text %s complete %u\n", hwnd, event->window, event->im_set_text.himc,
+                debugstr_cf(event->im_set_text.text), event->im_set_text.complete);
+
+    if (event->im_set_text.text)
+    {
+        CFIndex length = CFStringGetLength(event->im_set_text.text);
+        if (!(text = malloc((length + 1) * sizeof(WCHAR)))) return;
+        if (length) CFStringGetCharacters(event->im_set_text.text, CFRangeMake(0, length), text);
+        text[length] = 0;
+    }
+
+    if (event->im_set_text.complete) post_ime_update(hwnd, -1, NULL, text);
+    else post_ime_update(hwnd,
+                         MAKELONG(event->im_set_text.cursor_begin, event->im_set_text.cursor_end),
+                         text, NULL);
+
+    free(text);
+}
+
+
+/**************************************************************************
+ *              drag_operations_to_dropeffects
+ */
+static DWORD drag_operations_to_dropeffects(uint32_t ops)
+{
+    DWORD effects = 0;
+    if (ops & (DRAG_OP_COPY | DRAG_OP_GENERIC))
+        effects |= DROPEFFECT_COPY;
+    if (ops & DRAG_OP_MOVE)
+        effects |= DROPEFFECT_MOVE;
+    if (ops & (DRAG_OP_LINK | DRAG_OP_GENERIC))
+        effects |= DROPEFFECT_LINK;
+    return effects;
+}
+
+
+/**************************************************************************
+ *              dropeffect_to_drag_operation
+ */
+static uint32_t dropeffect_to_drag_operation(DWORD effect, uint32_t ops)
+{
+    if (effect & DROPEFFECT_LINK && ops & DRAG_OP_LINK) return DRAG_OP_LINK;
+    if (effect & DROPEFFECT_COPY && ops & DRAG_OP_COPY) return DRAG_OP_COPY;
+    if (effect & DROPEFFECT_MOVE && ops & DRAG_OP_MOVE) return DRAG_OP_MOVE;
+    if (effect & DROPEFFECT_LINK && ops & DRAG_OP_GENERIC) return DRAG_OP_GENERIC;
+    if (effect & DROPEFFECT_COPY && ops & DRAG_OP_GENERIC) return DRAG_OP_GENERIC;
+
+    return DRAG_OP_NONE;
+}
+
+
+/**************************************************************************
+ *              query_drag_drop_drop
+ */
+static BOOL query_drag_drop_drop(macdrv_query *query)
+{
+    HWND hwnd = macdrv_get_window_hwnd(query->window);
+    struct macdrv_win_data *data;
+
+    if (!(data = get_win_data(hwnd)))
+    {
+        WARN("no win_data for win %p/%p\n", hwnd, query->window);
+        return FALSE;
+    }
+
+    release_win_data(data);
+
+    NtUserMessageCall(hwnd, WINE_DRAG_DROP_DROP, 0, 0, NULL, NtUserDragDropCall, FALSE);
+    return TRUE;
+}
+
+/**************************************************************************
+ *              query_drag_drop_enter
+ */
+static BOOL query_drag_drop_enter(macdrv_query *query)
+{
+    CFTypeRef pasteboard = query->drag_drop.pasteboard;
+    struct format_entry *entries;
+    UINT entries_size;
+
+    if (!(entries = get_format_entries(pasteboard, &entries_size))) return FALSE;
+    NtUserMessageCall(0, WINE_DRAG_DROP_ENTER, entries_size, (LPARAM)entries, NULL, NtUserDragDropCall, FALSE);
+    free(entries);
+
+    return TRUE;
+}
+
+/**************************************************************************
+ *              query_drag_drop_leave
+ */
+static BOOL query_drag_drop_leave(macdrv_query *query)
+{
+    NtUserMessageCall(0, WINE_DRAG_DROP_LEAVE, 0, 0, NULL, NtUserDragDropCall, FALSE);
+    return TRUE;
+}
+
+
+/**************************************************************************
+ *              query_drag_drop_drag
+ */
+static BOOL query_drag_drop_drag(macdrv_query *query)
+{
+    HWND hwnd = macdrv_get_window_hwnd(query->window);
+    struct macdrv_win_data *data = get_win_data(hwnd);
+    DWORD effect;
+    POINT point;
+
+    if (!data)
+    {
+        WARN("no win_data for win %p/%p\n", hwnd, query->window);
+        return FALSE;
+    }
+
+    effect = drag_operations_to_dropeffects(query->drag_drop.ops);
+    point.x = query->drag_drop.x + data->rects.visible.left;
+    point.y = query->drag_drop.y + data->rects.visible.top;
+    release_win_data(data);
+
+    effect = NtUserMessageCall(hwnd, WINE_DRAG_DROP_DRAG, MAKELONG(point.x, point.y), effect, NULL, NtUserDragDropCall, FALSE);
+    if (!effect) return FALSE;
+
+    query->drag_drop.ops = dropeffect_to_drag_operation(effect, query->drag_drop.ops);
+    return TRUE;
+}
+
+
+/***********************************************************************
+ *      SetIMECompositionRect (MACDRV.@)
+ */
+BOOL macdrv_SetIMECompositionRect(HWND hwnd, RECT rect)
+{
+    TRACE("hwnd %p, rect %s\n", hwnd, wine_dbgstr_rect(&rect));
+    pthread_mutex_lock(&ime_composition_rect_mutex);
+    ime_composition_rect = cgrect_from_rect(rect);
+    pthread_mutex_unlock(&ime_composition_rect_mutex);
+    return TRUE;
+}
+
+
+/***********************************************************************
+ *      NotifyIMEStatus (MACDRV.@)
+ */
+void macdrv_NotifyIMEStatus( HWND hwnd, UINT status )
+{
+    TRACE_(imm)( "hwnd %p, status %#x\n", hwnd, status );
+    if (!status) macdrv_clear_ime_text();
+}
+
 
 /***********************************************************************
  *              macdrv_query_event
@@ -155,21 +330,21 @@ static void macdrv_query_event(HWND hwnd, const macdrv_event *event)
 
     switch (query->type)
     {
-        case QUERY_DRAG_DROP:
-            TRACE("QUERY_DRAG_DROP\n");
-            success = query_drag_drop(query);
+        case QUERY_DRAG_DROP_ENTER:
+            TRACE("QUERY_DRAG_DROP_ENTER\n");
+            success = query_drag_drop_enter(query);
             break;
-        case QUERY_DRAG_EXITED:
-            TRACE("QUERY_DRAG_EXITED\n");
-            success = query_drag_exited(query);
+        case QUERY_DRAG_DROP_LEAVE:
+            TRACE("QUERY_DRAG_DROP_LEAVE\n");
+            success = query_drag_drop_leave(query);
             break;
-        case QUERY_DRAG_OPERATION:
-            TRACE("QUERY_DRAG_OPERATION\n");
-            success = query_drag_operation(query);
+        case QUERY_DRAG_DROP_DRAG:
+            TRACE("QUERY_DRAG_DROP_DRAG\n");
+            success = query_drag_drop_drag(query);
             break;
-        case QUERY_IME_CHAR_RECT:
-            TRACE("QUERY_IME_CHAR_RECT\n");
-            success = query_ime_char_rect(query);
+        case QUERY_DRAG_DROP_DROP:
+            TRACE("QUERY_DRAG_DROP_DROP\n");
+            success = query_drag_drop_drop(query);
             break;
         case QUERY_PASTEBOARD_DATA:
             TRACE("QUERY_PASTEBOARD_DATA\n");
@@ -224,6 +399,9 @@ void macdrv_handle_event(const macdrv_event *event)
     case APP_QUIT_REQUESTED:
         macdrv_app_quit_requested(event);
         break;
+    case CLIENT_SURFACE_PRESENTED:    /* CW HACK 22435 */
+	macdrv_client_surface_presented(event);
+	break;
     case DISPLAYS_CHANGED:
         macdrv_displays_changed(event);
         break;
@@ -250,7 +428,7 @@ void macdrv_handle_event(const macdrv_event *event)
     case MOUSE_BUTTON:
         macdrv_mouse_button(hwnd, event);
         break;
-    case MOUSE_MOVED:
+    case MOUSE_MOVED_RELATIVE:
     case MOUSE_MOVED_ABSOLUTE:
         macdrv_mouse_moved(hwnd, event);
         break;
@@ -267,9 +445,6 @@ void macdrv_handle_event(const macdrv_event *event)
     case RELEASE_CAPTURE:
         macdrv_release_capture(hwnd, event);
         break;
-    case SENT_TEXT_INPUT:
-        macdrv_sent_text_input(event);
-        break;
     case STATUS_ITEM_MOUSE_BUTTON:
         macdrv_status_item_mouse_button(event);
         break;
@@ -324,44 +499,26 @@ void macdrv_handle_event(const macdrv_event *event)
 }
 
 
-/***********************************************************************
- *              process_events
- */
-static int process_events(macdrv_event_queue queue, macdrv_event_mask mask)
+static int check_fd_events( int fd, int events )
 {
+    struct pollfd pfd = {.fd = fd, .events = events};
+    if (poll( &pfd, 1, 0 ) <= 0) return 0;
+    return pfd.revents;
+}
+
+/***********************************************************************
+ *              ProcessEvents   (MACDRV.@)
+ */
+BOOL macdrv_ProcessEvents(DWORD mask)
+{
+    struct macdrv_thread_data *data = macdrv_thread_data();
+    macdrv_event_mask event_mask = get_event_mask(mask);
     macdrv_event *event;
     int count = 0;
 
-    while (macdrv_copy_event_from_queue(queue, mask, &event))
-    {
-        count++;
-        macdrv_handle_event(event);
-        macdrv_release_event(event);
-    }
-    if (count) TRACE("processed %d events\n", count);
-    return count;
-}
+    TRACE("mask %x\n", mask);
 
-
-/***********************************************************************
- *              MsgWaitForMultipleObjectsEx   (MACDRV.@)
- */
-DWORD macdrv_MsgWaitForMultipleObjectsEx(DWORD count, const HANDLE *handles,
-                                         DWORD timeout, DWORD mask, DWORD flags)
-{
-    DWORD ret;
-    struct macdrv_thread_data *data = macdrv_thread_data();
-    macdrv_event_mask event_mask = get_event_mask(mask);
-
-    TRACE("count %d, handles %p, timeout %u, mask %x, flags %x\n", count,
-          handles, timeout, mask, flags);
-
-    if (!data)
-    {
-        if (!count && !timeout) return WAIT_TIMEOUT;
-        return WaitForMultipleObjectsEx(count, handles, flags & MWMO_WAITALL,
-                                        timeout, flags & MWMO_ALERTABLE);
-    }
+    if (!data) return FALSE;
 
     if (data->current_event && data->current_event->type != QUERY_EVENT &&
         data->current_event->type != QUERY_EVENT_NO_PREEMPT_WAIT &&
@@ -369,14 +526,13 @@ DWORD macdrv_MsgWaitForMultipleObjectsEx(DWORD count, const HANDLE *handles,
         data->current_event->type != WINDOW_DRAG_BEGIN)
         event_mask = 0;  /* don't process nested events */
 
-    if (process_events(data->queue, event_mask)) ret = count - 1;
-    else if (count || timeout)
+    while (macdrv_copy_event_from_queue(data->queue, event_mask, &event))
     {
-        ret = WaitForMultipleObjectsEx(count, handles, flags & MWMO_WAITALL,
-                                       timeout, flags & MWMO_ALERTABLE);
-        if (ret == count - 1) process_events(data->queue, event_mask);
+        count++;
+        macdrv_handle_event(event);
+        macdrv_release_event(event);
     }
-    else ret = WAIT_TIMEOUT;
 
-    return ret;
+    if (count) TRACE("processed %d events\n", count);
+    return mask == QS_ALLINPUT && !check_fd_events(macdrv_get_event_queue_fd(data->queue), POLLIN);
 }
