@@ -19,6 +19,10 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
  */
 
+#if 0
+#pragma makedep unix
+#endif
+
 #include "config.h"
 
 #include "macdrv.h"
@@ -42,20 +46,11 @@ static inline MACDRV_PDEVICE *get_macdrv_dev(PHYSDEV dev)
 static CGRect desktop_rect;     /* virtual desktop rectangle */
 static int horz_size;           /* horz. size of screen in millimeters */
 static int vert_size;           /* vert. size of screen in millimeters */
-static int bits_per_pixel;      /* pixel depth of screen */
-static int device_data_valid;   /* do the above variables have up-to-date values? */
+static bool device_data_valid;  /* do the above variables have up-to-date values? */
 
-int retina_on = FALSE;
-
-static CRITICAL_SECTION device_data_section;
-static CRITICAL_SECTION_DEBUG critsect_debug =
-{
-    0, 0, &device_data_section,
-    { &critsect_debug.ProcessLocksList, &critsect_debug.ProcessLocksList },
-      0, 0, { (DWORD_PTR)(__FILE__ ": device_data_section") }
-};
-static CRITICAL_SECTION device_data_section = { &critsect_debug, -1, 0, 0, 0, 0 };
+bool retina_on = false;
 
+static pthread_mutex_t device_data_mutex = PTHREAD_MUTEX_INITIALIZER;
 
 static const struct user_driver_funcs macdrv_funcs;
 
@@ -90,7 +85,7 @@ CGRect macdrv_get_desktop_rect(void)
 {
     CGRect ret;
 
-    EnterCriticalSection(&device_data_section);
+    pthread_mutex_lock(&device_data_mutex);
 
     if (!device_data_valid)
     {
@@ -99,7 +94,7 @@ CGRect macdrv_get_desktop_rect(void)
     }
     ret = desktop_rect;
 
-    LeaveCriticalSection(&device_data_section);
+    pthread_mutex_unlock(&device_data_mutex);
 
     TRACE("%s\n", wine_dbgstr_cgrect(ret));
 
@@ -116,7 +111,6 @@ static void device_init(void)
 {
     CGDirectDisplayID mainDisplay = CGMainDisplayID();
     CGSize size_mm = CGDisplayScreenSize(mainDisplay);
-    CGDisplayModeRef mode = CGDisplayCopyDisplayMode(mainDisplay);
 
     check_retina_status();
 
@@ -124,36 +118,17 @@ static void device_init(void)
     horz_size = size_mm.width;
     vert_size = size_mm.height;
 
-    bits_per_pixel = 32;
-    if (mode)
-    {
-        CFStringRef pixelEncoding = CGDisplayModeCopyPixelEncoding(mode);
-
-        if (pixelEncoding)
-        {
-            if (CFEqual(pixelEncoding, CFSTR(IO32BitDirectPixels)))
-                bits_per_pixel = 32;
-            else if (CFEqual(pixelEncoding, CFSTR(IO16BitDirectPixels)))
-                bits_per_pixel = 16;
-            else if (CFEqual(pixelEncoding, CFSTR(IO8BitIndexedPixels)))
-                bits_per_pixel = 8;
-            CFRelease(pixelEncoding);
-        }
-
-        CGDisplayModeRelease(mode);
-    }
-
     compute_desktop_rect();
 
-    device_data_valid = TRUE;
+    device_data_valid = true;
 }
 
 
 void macdrv_reset_device_metrics(void)
 {
-    EnterCriticalSection(&device_data_section);
-    device_data_valid = FALSE;
-    LeaveCriticalSection(&device_data_section);
+    pthread_mutex_lock(&device_data_mutex);
+    device_data_valid = false;
+    pthread_mutex_unlock(&device_data_mutex);
 }
 
 
@@ -161,11 +136,11 @@ static MACDRV_PDEVICE *create_mac_physdev(void)
 {
     MACDRV_PDEVICE *physDev;
 
-    EnterCriticalSection(&device_data_section);
+    pthread_mutex_lock(&device_data_mutex);
     if (!device_data_valid) device_init();
-    LeaveCriticalSection(&device_data_section);
+    pthread_mutex_unlock(&device_data_mutex);
 
-    if (!(physDev = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(*physDev)))) return NULL;
+    if (!(physDev = calloc(1, sizeof(*physDev)))) return NULL;
 
     return physDev;
 }
@@ -174,8 +149,7 @@ static MACDRV_PDEVICE *create_mac_physdev(void)
 /**********************************************************************
  *              CreateDC (MACDRV.@)
  */
-static BOOL CDECL macdrv_CreateDC(PHYSDEV *pdev, LPCWSTR device, LPCWSTR output,
-                                  const DEVMODEW* initData)
+static BOOL macdrv_CreateDC(PHYSDEV *pdev, LPCWSTR device, LPCWSTR output, const DEVMODEW* initData)
 {
     MACDRV_PDEVICE *physDev = create_mac_physdev();
 
@@ -192,7 +166,7 @@ static BOOL CDECL macdrv_CreateDC(PHYSDEV *pdev, LPCWSTR device, LPCWSTR output,
 /**********************************************************************
  *              CreateCompatibleDC (MACDRV.@)
  */
-static BOOL CDECL macdrv_CreateCompatibleDC(PHYSDEV orig, PHYSDEV *pdev)
+static BOOL macdrv_CreateCompatibleDC(PHYSDEV orig, PHYSDEV *pdev)
 {
     MACDRV_PDEVICE *physDev = create_mac_physdev();
 
@@ -209,13 +183,13 @@ static BOOL CDECL macdrv_CreateCompatibleDC(PHYSDEV orig, PHYSDEV *pdev)
 /**********************************************************************
  *              DeleteDC (MACDRV.@)
  */
-static BOOL CDECL macdrv_DeleteDC(PHYSDEV dev)
+static BOOL macdrv_DeleteDC(PHYSDEV dev)
 {
     MACDRV_PDEVICE *physDev = get_macdrv_dev(dev);
 
     TRACE("hdc %p\n", dev->hdc);
 
-    HeapFree(GetProcessHeap(), 0, physDev);
+    free(physDev);
     return TRUE;
 }
 
@@ -223,11 +197,11 @@ static BOOL CDECL macdrv_DeleteDC(PHYSDEV dev)
 /***********************************************************************
  *              GetDeviceCaps (MACDRV.@)
  */
-static INT CDECL macdrv_GetDeviceCaps(PHYSDEV dev, INT cap)
+static INT macdrv_GetDeviceCaps(PHYSDEV dev, INT cap)
 {
     INT ret;
 
-    EnterCriticalSection(&device_data_section);
+    pthread_mutex_lock(&device_data_mutex);
 
     if (!device_data_valid) device_init();
 
@@ -239,13 +213,10 @@ static INT CDECL macdrv_GetDeviceCaps(PHYSDEV dev, INT cap)
     case VERTSIZE:
         ret = vert_size;
         break;
-    case BITSPIXEL:
-        ret = bits_per_pixel;
-        break;
     case HORZRES:
     case VERTRES:
     default:
-        LeaveCriticalSection(&device_data_section);
+        pthread_mutex_unlock(&device_data_mutex);
         dev = GET_NEXT_PHYSDEV( dev, pGetDeviceCaps );
         ret = dev->funcs->pGetDeviceCaps( dev, cap );
         if ((cap == HORZRES || cap == VERTRES) && retina_on)
@@ -255,7 +226,7 @@ static INT CDECL macdrv_GetDeviceCaps(PHYSDEV dev, INT cap)
 
     TRACE("cap %d -> %d\n", cap, ret);
 
-    LeaveCriticalSection(&device_data_section);
+    pthread_mutex_unlock(&device_data_mutex);
     return ret;
 }
 
@@ -272,24 +243,28 @@ static const struct user_driver_funcs macdrv_funcs =
 
     .pActivateKeyboardLayout = macdrv_ActivateKeyboardLayout,
     .pBeep = macdrv_Beep,
-    .pChangeDisplaySettingsEx = macdrv_ChangeDisplaySettingsEx,
+    .pChangeDisplaySettings = macdrv_ChangeDisplaySettings,
     .pClipCursor = macdrv_ClipCursor,
-    .pCreateDesktopWindow = macdrv_CreateDesktopWindow,
-    .pCreateWindow = macdrv_CreateWindow,
+    .pNotifyIcon = macdrv_NotifyIcon,
+    .pCleanupIcons = macdrv_CleanupIcons,
+    .pClipboardWindowProc = macdrv_ClipboardWindowProc,
+    .pDesktopWindowProc = macdrv_DesktopWindowProc,
     .pDestroyCursorIcon = macdrv_DestroyCursorIcon,
     .pDestroyWindow = macdrv_DestroyWindow,
-    .pEnumDisplaySettingsEx = macdrv_EnumDisplaySettingsEx,
+    .pGetCurrentProcessExplicitAppUserModelID = macdrv_GetCurrentProcessExplicitAppUserModelID,  /* CW Hack 22310 */
     .pUpdateDisplayDevices = macdrv_UpdateDisplayDevices,
     .pGetCursorPos = macdrv_GetCursorPos,
     .pGetKeyboardLayoutList = macdrv_GetKeyboardLayoutList,
     .pGetKeyNameText = macdrv_GetKeyNameText,
     .pMapVirtualKeyEx = macdrv_MapVirtualKeyEx,
-    .pMsgWaitForMultipleObjectsEx = macdrv_MsgWaitForMultipleObjectsEx,
+    .pProcessEvents = macdrv_ProcessEvents,
     .pRegisterHotKey = macdrv_RegisterHotKey,
+    .pSetCurrentProcessExplicitAppUserModelID = macdrv_SetCurrentProcessExplicitAppUserModelID,  /* CW Hack 22310 */
     .pSetCapture = macdrv_SetCapture,
     .pSetCursor = macdrv_SetCursor,
     .pSetCursorPos = macdrv_SetCursorPos,
-    .pSetFocus = macdrv_SetFocus,
+    .pSetDesktopWindow = macdrv_SetDesktopWindow,
+    .pActivateWindow = macdrv_ActivateWindow,
     .pSetLayeredWindowAttributes = macdrv_SetLayeredWindowAttributes,
     .pSetParent = macdrv_SetParent,
     .pSetWindowRgn = macdrv_SetWindowRgn,
@@ -304,11 +279,16 @@ static const struct user_driver_funcs macdrv_funcs =
     .pUpdateClipboard = macdrv_UpdateClipboard,
     .pUpdateLayeredWindow = macdrv_UpdateLayeredWindow,
     .pVkKeyScanEx = macdrv_VkKeyScanEx,
+    .pImeProcessKey = macdrv_ImeProcessKey,
+    .pNotifyIMEStatus = macdrv_NotifyIMEStatus,
+    .pSetIMECompositionRect = macdrv_SetIMECompositionRect,
     .pWindowMessage = macdrv_WindowMessage,
     .pWindowPosChanged = macdrv_WindowPosChanged,
     .pWindowPosChanging = macdrv_WindowPosChanging,
-    .pwine_get_vulkan_driver = macdrv_wine_get_vulkan_driver,
-    .pwine_get_wgl_driver = macdrv_wine_get_wgl_driver,
+    .pGetWindowStyleMasks = macdrv_GetWindowStyleMasks,
+    .pCreateWindowSurface = macdrv_CreateWindowSurface,
+    .pVulkanInit = macdrv_VulkanInit,
+    .pOpenGLInit = macdrv_OpenGLInit,
 };
 
 
