@@ -24,162 +24,20 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
  */
 
+#if 0
+#pragma makedep unix
+#endif
+
 #include "config.h"
 
 #include "macdrv.h"
 #include "winuser.h"
-#include "wine/unicode.h"
 #include "wine/server.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(keyboard);
 WINE_DECLARE_DEBUG_CHANNEL(key);
 
 
-/* Carbon-style modifier mask definitions from <Carbon/HIToolbox/Events.h>. */
-enum {
-    cmdKeyBit       = 8,
-    shiftKeyBit     = 9,
-    alphaLockBit    = 10,
-    optionKeyBit    = 11,
-    controlKeyBit   = 12,
-};
-
-enum {
-    cmdKey      = 1 << cmdKeyBit,
-    shiftKey    = 1 << shiftKeyBit,
-    alphaLock   = 1 << alphaLockBit,
-    optionKey   = 1 << optionKeyBit,
-    controlKey  = 1 << controlKeyBit,
-};
-
-
-/* Mac virtual key code definitions from <Carbon/HIToolbox/Events.h>. */
-enum {
-    kVK_ANSI_A              = 0x00,
-    kVK_ANSI_S              = 0x01,
-    kVK_ANSI_D              = 0x02,
-    kVK_ANSI_F              = 0x03,
-    kVK_ANSI_H              = 0x04,
-    kVK_ANSI_G              = 0x05,
-    kVK_ANSI_Z              = 0x06,
-    kVK_ANSI_X              = 0x07,
-    kVK_ANSI_C              = 0x08,
-    kVK_ANSI_V              = 0x09,
-    kVK_ISO_Section         = 0x0A,
-    kVK_ANSI_B              = 0x0B,
-    kVK_ANSI_Q              = 0x0C,
-    kVK_ANSI_W              = 0x0D,
-    kVK_ANSI_E              = 0x0E,
-    kVK_ANSI_R              = 0x0F,
-    kVK_ANSI_Y              = 0x10,
-    kVK_ANSI_T              = 0x11,
-    kVK_ANSI_1              = 0x12,
-    kVK_ANSI_2              = 0x13,
-    kVK_ANSI_3              = 0x14,
-    kVK_ANSI_4              = 0x15,
-    kVK_ANSI_6              = 0x16,
-    kVK_ANSI_5              = 0x17,
-    kVK_ANSI_Equal          = 0x18,
-    kVK_ANSI_9              = 0x19,
-    kVK_ANSI_7              = 0x1A,
-    kVK_ANSI_Minus          = 0x1B,
-    kVK_ANSI_8              = 0x1C,
-    kVK_ANSI_0              = 0x1D,
-    kVK_ANSI_RightBracket   = 0x1E,
-    kVK_ANSI_O              = 0x1F,
-    kVK_ANSI_U              = 0x20,
-    kVK_ANSI_LeftBracket    = 0x21,
-    kVK_ANSI_I              = 0x22,
-    kVK_ANSI_P              = 0x23,
-    kVK_Return              = 0x24,
-    kVK_ANSI_L              = 0x25,
-    kVK_ANSI_J              = 0x26,
-    kVK_ANSI_Quote          = 0x27,
-    kVK_ANSI_K              = 0x28,
-    kVK_ANSI_Semicolon      = 0x29,
-    kVK_ANSI_Backslash      = 0x2A,
-    kVK_ANSI_Comma          = 0x2B,
-    kVK_ANSI_Slash          = 0x2C,
-    kVK_ANSI_N              = 0x2D,
-    kVK_ANSI_M              = 0x2E,
-    kVK_ANSI_Period         = 0x2F,
-    kVK_Tab                 = 0x30,
-    kVK_Space               = 0x31,
-    kVK_ANSI_Grave          = 0x32,
-    kVK_Delete              = 0x33,
-    kVK_Escape              = 0x35,
-    kVK_RightCommand        = 0x36, /* invented for Wine; co-opt unused key code */
-    kVK_Command             = 0x37,
-    kVK_Shift               = 0x38,
-    kVK_CapsLock            = 0x39,
-    kVK_Option              = 0x3A,
-    kVK_Control             = 0x3B,
-    kVK_RightShift          = 0x3C,
-    kVK_RightOption         = 0x3D,
-    kVK_RightControl        = 0x3E,
-    kVK_Function            = 0x3F,
-    kVK_F17                 = 0x40,
-    kVK_ANSI_KeypadDecimal  = 0x41,
-    kVK_ANSI_KeypadMultiply = 0x43,
-    kVK_ANSI_KeypadPlus     = 0x45,
-    kVK_ANSI_KeypadClear    = 0x47,
-    kVK_VolumeUp            = 0x48,
-    kVK_VolumeDown          = 0x49,
-    kVK_Mute                = 0x4A,
-    kVK_ANSI_KeypadDivide   = 0x4B,
-    kVK_ANSI_KeypadEnter    = 0x4C,
-    kVK_ANSI_KeypadMinus    = 0x4E,
-    kVK_F18                 = 0x4F,
-    kVK_F19                 = 0x50,
-    kVK_ANSI_KeypadEquals   = 0x51,
-    kVK_ANSI_Keypad0        = 0x52,
-    kVK_ANSI_Keypad1        = 0x53,
-    kVK_ANSI_Keypad2        = 0x54,
-    kVK_ANSI_Keypad3        = 0x55,
-    kVK_ANSI_Keypad4        = 0x56,
-    kVK_ANSI_Keypad5        = 0x57,
-    kVK_ANSI_Keypad6        = 0x58,
-    kVK_ANSI_Keypad7        = 0x59,
-    kVK_F20                 = 0x5A,
-    kVK_ANSI_Keypad8        = 0x5B,
-    kVK_ANSI_Keypad9        = 0x5C,
-    kVK_JIS_Yen             = 0x5D,
-    kVK_JIS_Underscore      = 0x5E,
-    kVK_JIS_KeypadComma     = 0x5F,
-    kVK_F5                  = 0x60,
-    kVK_F6                  = 0x61,
-    kVK_F7                  = 0x62,
-    kVK_F3                  = 0x63,
-    kVK_F8                  = 0x64,
-    kVK_F9                  = 0x65,
-    kVK_JIS_Eisu            = 0x66,
-    kVK_F11                 = 0x67,
-    kVK_JIS_Kana            = 0x68,
-    kVK_F13                 = 0x69,
-    kVK_F16                 = 0x6A,
-    kVK_F14                 = 0x6B,
-    kVK_F10                 = 0x6D,
-    kVK_F12                 = 0x6F,
-    kVK_F15                 = 0x71,
-    kVK_Help                = 0x72,
-    kVK_Home                = 0x73,
-    kVK_PageUp              = 0x74,
-    kVK_ForwardDelete       = 0x75,
-    kVK_F4                  = 0x76,
-    kVK_End                 = 0x77,
-    kVK_F2                  = 0x78,
-    kVK_PageDown            = 0x79,
-    kVK_F1                  = 0x7A,
-    kVK_LeftArrow           = 0x7B,
-    kVK_RightArrow          = 0x7C,
-    kVK_DownArrow           = 0x7D,
-    kVK_UpArrow             = 0x7E,
-};
-
-#include "wine/hostaddrspace_enter.h"
-extern const CFStringRef kTISTypeKeyboardLayout;
-#include "wine/hostaddrspace_exit.h"
-
 /* Indexed by Mac virtual keycode values defined above. */
 static const struct {
     WORD vkey;
@@ -397,18 +255,39 @@ static const struct {
     { VK_VOLUME_UP | 0x100,     "Volume Up" },
 };
 
-static BOOL char_matches_string(WCHAR wchar, UniChar *string, BOOL ignore_diacritics)
+
+static const struct {
+    WCHAR       wchar;
+    const char *name;
+} dead_key_names[] = {
+    { '^',                      "CIRCUMFLEX ACCENT" },
+    { '`',                      "GRAVE ACCENT" },
+    { 0x00B4,                   "ACUTE ACCENT" },
+    { '~',                      "TILDE" },
+    { 0x00A8,                   "DIAERESIS" },
+    { 0x00B8,                   "CEDILLA" },
+    { 0x02D8,                   "BREVE" },
+    { 0x02D9,                   "DOT ABOVE" },
+    { 0x00AF,                   "MACRON" },
+    { 0x02DA,                   "RING ABOVE" },
+    { 0x02DB,                   "OGONEK" },
+    { 0x02DC,                   "SMALL TILDE" },
+    { 0x02DD,                   "DOUBLE ACUTE ACCENT" },
+};
+
+
+static Boolean char_matches_string(WCHAR wchar, UniChar *string, CollatorRef collatorRef)
 {
-    BOOL ret;
-    CFStringRef s1 = CFStringCreateWithCharactersNoCopy(NULL, (UniChar*)&wchar, 1, kCFAllocatorNull);
-    CFStringRef s2 = CFStringCreateWithCharactersNoCopy(NULL, string, strlenW(string), kCFAllocatorNull);
-    CFStringCompareFlags flags = kCFCompareCaseInsensitive | kCFCompareNonliteral | kCFCompareWidthInsensitive;
-    if (ignore_diacritics)
-        flags |= kCFCompareDiacriticInsensitive;
-    ret = (CFStringCompare(s1, s2, flags) == kCFCompareEqualTo);
-    CFRelease(s1);
-    CFRelease(s2);
-    return ret;
+    Boolean equivalent;
+    OSStatus status;
+
+    status = UCCompareText(collatorRef, (UniChar*)&wchar, 1, string, wcslen(string), &equivalent, NULL);
+    if (status != noErr)
+    {
+        WARN("Failed to compare %s to %s\n", debugstr_wn(&wchar, 1), debugstr_w(string));
+        return FALSE;
+    }
+    return equivalent;
 }
 
 
@@ -439,27 +318,98 @@ struct layout
     BOOL enabled; /* is the input source enabled - ie displayed in the input source selector UI */
 };
 
-static CRITICAL_SECTION layout_list_section;
-static CRITICAL_SECTION_DEBUG critsect_debug =
-{
-    0, 0, &layout_list_section,
-    { &critsect_debug.ProcessLocksList, &critsect_debug.ProcessLocksList },
-    0, 0, { (DWORD_PTR)(__FILE__ ": layout_list_section") }
-};
-static CRITICAL_SECTION layout_list_section = { &critsect_debug, -1, 0, 0, 0, 0 };
+static pthread_mutex_t layout_list_mutex = PTHREAD_MUTEX_INITIALIZER;
 
 int macdrv_layout_list_needs_update = TRUE;
 
+static const NLS_LOCALE_HEADER *locale_table;
+
+static int compare_locale_names(const WCHAR *n1, const WCHAR *n2)
+{
+    for (;;)
+    {
+        WCHAR ch1 = *n1++;
+        WCHAR ch2 = *n2++;
+        if (ch1 >= 'a' && ch1 <= 'z') ch1 -= 'a' - 'A';
+        else if (ch1 == '_') ch1 = '-';
+        if (ch2 >= 'a' && ch2 <= 'z') ch2 -= 'a' - 'A';
+        else if (ch2 == '_') ch2 = '-';
+        if (!ch1 || ch1 != ch2) return ch1 - ch2;
+    }
+}
+
+
+static const NLS_LOCALE_LCNAME_INDEX *find_lcname_entry(const WCHAR *name)
+{
+    const NLS_LOCALE_LCNAME_INDEX *lcnames_index;
+    const WCHAR *locale_strings;
+    int min = 0, max = locale_table->nb_lcnames - 1;
+
+    locale_strings = (const WCHAR *)((char *)locale_table + locale_table->strings_offset);
+    lcnames_index = (const NLS_LOCALE_LCNAME_INDEX *)((char *)locale_table + locale_table->lcnames_offset);
+
+    while (min <= max)
+    {
+        int res, pos = (min + max) / 2;
+        const WCHAR *str = locale_strings + lcnames_index[pos].name;
+        res = compare_locale_names(name, str + 1);
+        if (res < 0) max = pos - 1;
+        else if (res > 0) min = pos + 1;
+        else return &lcnames_index[pos];
+    }
+    return NULL;
+}
+
+
 static DWORD get_lcid(CFStringRef lang)
 {
+    const NLS_LOCALE_LCNAME_INDEX *entry;
+    const NLS_LOCALE_DATA *locale;
     CFRange range;
     WCHAR str[10];
+    ULONG offset;
+
+    if (!locale_table)
+    {
+        struct
+        {
+            UINT ctypes;
+            UINT unknown1;
+            UINT unknown2;
+            UINT unknown3;
+            UINT locales;
+            UINT charmaps;
+            UINT geoids;
+            UINT scripts;
+        } *header;
+        LCID system_lcid;
+        LARGE_INTEGER size;
+
+        if (NtInitializeNlsFiles((void **)&header, &system_lcid, &size))
+        {
+            ERR("NtInitializeNlsFiles failed\n");
+            return 0;
+        }
+
+        if (InterlockedCompareExchangePointer((void **)&locale_table,
+                                              (char *)header + header->locales, NULL))
+            NtUnmapViewOfSection(GetCurrentProcess(), header);
+    }
 
     range.location = 0;
     range.length = min(CFStringGetLength(lang), ARRAY_SIZE(str) - 1);
     CFStringGetCharacters(lang, range, str);
     str[range.length] = 0;
-    return LocaleNameToLCID(str, 0);
+
+    if (!(entry = find_lcname_entry(str)))
+    {
+        ERR("%s not found\n", debugstr_w(str));
+        return 0;
+    }
+
+    offset = locale_table->locales_offset + entry->idx * locale_table->locale_size;
+    locale = (const NLS_LOCALE_DATA *)((const char *)locale_table + offset);
+    return locale->inotneutral ? entry->id : locale->idefaultlanguage;
 }
 
 static HKL get_hkl(CFStringRef lang, CFStringRef type)
@@ -488,7 +438,7 @@ static HKL get_hkl(CFStringRef lang, CFStringRef type)
 /******************************************************************
  *                get_layout_from_source
  *
- * Must be called while holding the layout_list_section.
+ * Must be called while holding the layout_list_mutex.
  * Note, returned layout may not currently be enabled.
  */
 static struct layout *get_layout_from_source(TISInputSourceRef input)
@@ -509,7 +459,7 @@ static struct layout *get_layout_from_source(TISInputSourceRef input)
 /***********************************************************************
  *            update_layout_list
  *
- * Must be called while holding the layout_list_section
+ * Must be called while holding the layout_list_mutex
  *
  * If an input source has been disabled (ie. removed from the UI) its
  * entry remains in the layout list but is marked as such and is not
@@ -522,7 +472,7 @@ static void update_layout_list(void)
     struct layout *layout;
     int i;
 
-    if (!InterlockedExchange(&macdrv_layout_list_needs_update, FALSE)) return;
+    if (!InterlockedExchange((LONG *)&macdrv_layout_list_needs_update, FALSE)) return;
 
     sources = macdrv_create_input_source_list();
 
@@ -539,7 +489,7 @@ static void update_layout_list(void)
             CFStringRef type = CFDictionaryGetValue(dict, macdrv_input_source_type_key);
             CFStringRef lang = CFDictionaryGetValue(dict, macdrv_input_source_lang_key);
 
-            layout = HeapAlloc(GetProcessHeap(), 0, sizeof(*layout));
+            layout = malloc(sizeof(*layout));
             layout->input_source = (TISInputSourceRef)CFRetain(input);
             layout->hkl = get_hkl(lang, type);
 
@@ -565,13 +515,13 @@ HKL macdrv_get_hkl_from_source(TISInputSourceRef input)
     struct layout *layout;
     HKL ret = 0;
 
-    EnterCriticalSection(&layout_list_section);
+    pthread_mutex_lock(&layout_list_mutex);
 
     update_layout_list();
     layout = get_layout_from_source(input);
     if (layout) ret = layout->hkl;
 
-    LeaveCriticalSection(&layout_list_section);
+    pthread_mutex_unlock(&layout_list_mutex);
 
     return ret;
 }
@@ -585,6 +535,9 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
     int keyc;
     WCHAR vkey;
     const UCKeyboardLayout *uchr;
+    LocaleRef localeRef;
+    CollatorRef collatorRef, caseInsensitiveCollatorRef, diacriticInsensitiveCollatorRef;
+    UCCollateOptions collateOptions = 0;
     const UInt32 modifier_combos[] = {
         0,
         shiftKey >> 8,
@@ -600,51 +553,72 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
     static const struct {
         WCHAR wchar;
         DWORD vkey;
+        /* Mac virtual key code that must match wchar under the current layout.
+         * A value of -1 means match-any.
+         * TODO: replace -1 with the actual mac virtual key codes for all mappings
+         * and their respective layouts to avoid false matches, when possible. */
+        int mac_keyc;
     } symbol_vkeys[] = {
-        { '-', VK_OEM_MINUS },
-        { '+', VK_OEM_PLUS },
-        { '_', VK_OEM_MINUS },
-        { ',', VK_OEM_COMMA },
-        { '.', VK_OEM_PERIOD },
-        { '=', VK_OEM_PLUS },
-        { '>', VK_OEM_PERIOD },
-        { '<', VK_OEM_COMMA },
-        { '|', VK_OEM_5 },
-        { '\\', VK_OEM_5 },
-        { '`', VK_OEM_3 },
-        { '[', VK_OEM_4 },
-        { '~', VK_OEM_3 },
-        { '?', VK_OEM_2 },
-        { ']', VK_OEM_6 },
-        { '/', VK_OEM_2 },
-        { ':', VK_OEM_1 },
-        { '}', VK_OEM_6 },
-        { '{', VK_OEM_4 },
-        { ';', VK_OEM_1 },
-        { '\'', VK_OEM_7 },
-        { ':', VK_OEM_PERIOD },
-        { ';', VK_OEM_COMMA },
-        { '"', VK_OEM_7 },
-        { 0x00B4, VK_OEM_4 }, /* 0x00B4 is ACUTE ACCENT */
-        { '\'', VK_OEM_2 },
-        { 0x00A7, VK_OEM_5 }, /* 0x00A7 is SECTION SIGN */
-        { '*', VK_OEM_PLUS },
-        { 0x00B4, VK_OEM_7 },
-        { '`', VK_OEM_4 },
-        { '[', VK_OEM_6 },
-        { '/', VK_OEM_5 },
-        { '^', VK_OEM_6 },
-        { '*', VK_OEM_2 },
-        { '{', VK_OEM_6 },
-        { '~', VK_OEM_1 },
-        { '?', VK_OEM_PLUS },
-        { '?', VK_OEM_4 },
-        { 0x00B4, VK_OEM_3 },
-        { '?', VK_OEM_COMMA },
-        { '~', VK_OEM_PLUS },
-        { ']', VK_OEM_4 },
-        { '\'', VK_OEM_3 },
-        { 0x00A7, VK_OEM_7 },
+        { '-', VK_OEM_PLUS, kVK_ANSI_Equal },
+        { '-', VK_OEM_MINUS, -1 },
+        { '+', VK_OEM_PLUS, -1 },
+        { '_', VK_OEM_MINUS, -1 },
+        { ',', VK_OEM_COMMA, -1 },
+        { '.', VK_OEM_PERIOD, -1 },
+        { '=', VK_OEM_8, kVK_ANSI_Slash },
+        { '=', VK_OEM_PLUS, -1 },
+        { '!', VK_OEM_8, kVK_ANSI_Slash },
+        { 0x00F9, VK_OEM_3, kVK_ANSI_Quote }, /* 0x00F9 is French U accent grave */
+        { '$', VK_OEM_1, kVK_ANSI_RightBracket },
+        { ':', VK_OEM_2, kVK_ANSI_Period },
+        { '*', VK_OEM_5, kVK_ANSI_Backslash },
+        { '`', VK_OEM_5, kVK_ANSI_Backslash },
+        { ';', VK_OEM_PERIOD, kVK_ANSI_Comma },
+        { ')', VK_OEM_4, kVK_ANSI_Minus },
+        { '>', VK_OEM_PERIOD, -1 },
+        { '<', VK_OEM_COMMA, -1 },
+        { '|', VK_OEM_5, -1 },
+        { '\\', VK_OEM_5, -1 },
+        { '`', VK_OEM_3, -1 },
+        { '[', VK_OEM_4, -1 },
+        { '~', VK_OEM_3, -1 },
+        { 0x00DF, VK_OEM_4, kVK_ANSI_Minus }, /* 0x00DF is ESZETT */
+        { 0x00FC, VK_OEM_1, kVK_ANSI_LeftBracket }, /* 0x00FC is German U Umlaut */
+        { 0x00F6, VK_OEM_3, kVK_ANSI_Semicolon }, /* 0x00F6 is German O Umlaut */
+        { 0x00E4, VK_OEM_7, kVK_ANSI_Quote }, /* 0x00B4 is German A Umlaut */
+        { '?', VK_OEM_2, -1 },
+        { ']', VK_OEM_6, -1 },
+        { '/', VK_OEM_2, -1 },
+        { ':', VK_OEM_1, -1 },
+        { '}', VK_OEM_6, -1 },
+        { '{', VK_OEM_4,  },
+        { ';', VK_OEM_1, -1 },
+        { '\'', VK_OEM_7, -1 },
+        { ':', VK_OEM_PERIOD, -1 },
+        { ';', VK_OEM_COMMA, -1 },
+        { '"', VK_OEM_7,  -1 },
+        { 0x00B4, VK_OEM_4, kVK_ANSI_Equal }, /* 0x00B4 is ACUTE ACCENT */
+        { '\'', VK_OEM_2, -1 },
+        { 0x00A7, VK_OEM_5, -1 }, /* 0x00A7 is SECTION SIGN */
+        { '*', VK_OEM_PLUS, -1 },
+        { 0x00B4, VK_OEM_7, -1 },
+        { '`', VK_OEM_4, -1 },
+        { '[', VK_OEM_6, -1 },
+        { '/', VK_OEM_5, -1 },
+        { '^', VK_OEM_6, -1 },
+        { '*', VK_OEM_2, -1 },
+        { '{', VK_OEM_6, -1 },
+        { 0x00B4, VK_OEM_6, -1 },
+        { '~', VK_OEM_1, -1 },
+        { '?', VK_OEM_PLUS, -1 },
+        { '?', VK_OEM_4, -1 },
+        { 0x00B4, VK_OEM_3, -1 },
+        { '?', VK_OEM_COMMA, -1 },
+        { '~', VK_OEM_PLUS, -1 },
+        { ']', VK_OEM_4, -1 },
+        { '\'', VK_OEM_3, -1 },
+        { 0x00A7, VK_OEM_7, -1 },
+        { '<', VK_OEM_102, -1 },
     };
     int i;
 
@@ -706,6 +680,13 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
 
     uchr = (const UCKeyboardLayout*)CFDataGetBytePtr(thread_data->keyboard_layout_uchr);
 
+    LocaleRefFromLocaleString("POSIX", &localeRef);
+    UCCreateCollator(localeRef, 0, collateOptions, &collatorRef);
+    collateOptions |= kUCCollateComposeInsensitiveMask | kUCCollateWidthInsensitiveMask | kUCCollateCaseInsensitiveMask;
+    UCCreateCollator(localeRef, 0, collateOptions, &caseInsensitiveCollatorRef);
+    collateOptions |= kUCCollateDiacritInsensitiveMask;
+    UCCreateCollator(localeRef, 0, collateOptions, &diacriticInsensitiveCollatorRef);
+
     /* Using the keyboard layout, build a map of key code + modifiers -> characters. */
     memset(map, 0, sizeof(map));
     for (keyc = 0; keyc < ARRAY_SIZE(map); keyc++)
@@ -752,7 +733,7 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
                     if (thread_data->keyc2vkey[keyc] || !map[keyc][combo][0])
                         continue;
 
-                    if (char_matches_string(vkey, map[keyc][combo], ignore_diacritics))
+                    if (char_matches_string(vkey, map[keyc][combo], ignore_diacritics ? diacriticInsensitiveCollatorRef : caseInsensitiveCollatorRef))
                     {
                         thread_data->keyc2vkey[keyc] = vkey;
                         vkey_used[vkey] = 1;
@@ -778,7 +759,7 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
                 if (thread_data->keyc2vkey[keyc] || !map[keyc][combo][0])
                     continue;
 
-                if (char_matches_string(vkey, map[keyc][combo], FALSE))
+                if (char_matches_string(vkey, map[keyc][combo], collatorRef))
                 {
                     thread_data->keyc2vkey[keyc] = vkey;
                     vkey_used[vkey] = 1;
@@ -792,22 +773,24 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
 
     /* Now try to match key codes for certain common punctuation characters to
        the most common OEM vkeys (e.g. '.' to VK_OEM_PERIOD). */
-    for (i = 0; i < ARRAY_SIZE(symbol_vkeys); i++)
+    for (combo = 0; combo < ARRAY_SIZE(modifier_combos); combo++)
     {
-        vkey = symbol_vkeys[i].vkey;
-
-        if (vkey_used[vkey])
-            continue;
-
-        for (combo = 0; combo < ARRAY_SIZE(modifier_combos); combo++)
+        for (i = 0; i < ARRAY_SIZE(symbol_vkeys); i++)
         {
+            vkey = symbol_vkeys[i].vkey;
+
+            if (vkey_used[vkey])
+                continue;
+
             for (keyc = 0; keyc < ARRAY_SIZE(map); keyc++)
             {
                 if (!thread_data->keyc2scan[keyc]) continue; /* not a known Mac key code */
                 if (thread_data->keyc2vkey[keyc] || !map[keyc][combo][0])
                     continue;
+                if (symbol_vkeys[i].mac_keyc != -1 && symbol_vkeys[i].mac_keyc != keyc)
+                    continue;
 
-                if (char_matches_string(symbol_vkeys[i].wchar, map[keyc][combo], FALSE))
+                if (char_matches_string(symbol_vkeys[i].wchar, map[keyc][combo], collatorRef))
                 {
                     thread_data->keyc2vkey[keyc] = vkey;
                     vkey_used[vkey] = 1;
@@ -816,9 +799,6 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
                     break;
                 }
             }
-
-            if (vkey_used[vkey])
-                break;
         }
     }
 
@@ -826,7 +806,7 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
        from the default map, if it's still available. */
     for (keyc = 0; keyc < ARRAY_SIZE(default_map); keyc++)
     {
-        DWORD vkey = default_map[keyc].vkey;
+        unsigned int vkey = default_map[keyc].vkey;
 
         if (!thread_data->keyc2scan[keyc]) continue; /* not a known Mac key code */
         if (thread_data->keyc2vkey[keyc]) continue; /* already assigned */
@@ -910,13 +890,17 @@ void macdrv_compute_keyboard_layout(struct macdrv_thread_data *thread_data)
         vkey_used[vkey] = 1;
         TRACE("keyc 0x%04x -> vkey 0x%04x (spare vkey)\n", keyc, vkey);
     }
+
+    UCDisposeCollator(&collatorRef);
+    UCDisposeCollator(&caseInsensitiveCollatorRef);
+    UCDisposeCollator(&diacriticInsensitiveCollatorRef);
 }
 
 
 /***********************************************************************
  *              macdrv_send_keyboard_input
  */
-static void macdrv_send_keyboard_input(HWND hwnd, WORD vkey, WORD scan, DWORD flags, DWORD time)
+static void macdrv_send_keyboard_input(HWND hwnd, WORD vkey, WORD scan, unsigned int flags, unsigned int time)
 {
     INPUT input;
 
@@ -929,26 +913,7 @@ static void macdrv_send_keyboard_input(HWND hwnd, WORD vkey, WORD scan, DWORD fl
     input.ki.time           = time;
     input.ki.dwExtraInfo    = 0;
 
-    __wine_send_input(hwnd, &input, NULL);
-}
-
-
-/***********************************************************************
- *           get_async_key_state
- */
-static BOOL get_async_key_state(BYTE state[256])
-{
-    BOOL ret;
-
-    SERVER_START_REQ(get_key_state)
-    {
-        req->async = 1;
-        req->key = -1;
-        wine_server_set_reply(req, state, 256);
-        ret = !wine_server_call(req);
-    }
-    SERVER_END_REQ;
-    return ret;
+    NtUserSendHardwareInput(hwnd, 0, &input, 0);
 }
 
 
@@ -1054,9 +1019,9 @@ void macdrv_keyboard_changed(const macdrv_event *event)
 
     macdrv_compute_keyboard_layout(thread_data);
 
-    ActivateKeyboardLayout(thread_data->active_keyboard_layout, 0);
+    NtUserActivateKeyboardLayout(thread_data->active_keyboard_layout, 0);
 
-    SendMessageW(GetActiveWindow(), WM_CANCELMODE, 0, 0);
+    send_message(get_active_window(), WM_CANCELMODE, 0, 0);
 }
 
 
@@ -1111,10 +1076,10 @@ void macdrv_edit_menu_command(const macdrv_event *event)
         }
         key_scan = vkey_to_scan(key_vkey);
 
-        left_alt_down = (GetAsyncKeyState(VK_LMENU) & 0x8000) != 0;
-        right_alt_down = (GetAsyncKeyState(VK_RMENU) & 0x8000) != 0;
+        left_alt_down = (NtUserGetAsyncKeyState(VK_LMENU) & 0x8000) != 0;
+        right_alt_down = (NtUserGetAsyncKeyState(VK_RMENU) & 0x8000) != 0;
         if (need_control)
-            control_down = (GetAsyncKeyState(VK_CONTROL) & 0x8000) != 0;
+            control_down = (NtUserGetAsyncKeyState(VK_CONTROL) & 0x8000) != 0;
 
         if (left_alt_down)
             left_alt_scan = vkey_to_scan(VK_LMENU);
@@ -1223,9 +1188,9 @@ void macdrv_edit_menu_command(const macdrv_event *event)
                 return;
         }
 
-        focus = GetFocus();
-        if (focus == hwnd || IsChild(hwnd, focus))
-            PostMessageW(focus, msg, 0, lparam);
+        focus = get_focus();
+        if (focus == hwnd || NtUserIsChild(hwnd, focus))
+            NtUserPostMessage(focus, msg, 0, lparam);
     }
 }
 
@@ -1250,7 +1215,7 @@ void macdrv_hotkey_press(const macdrv_event *event)
         BOOL got_keystate;
         DWORD flags;
 
-        if ((got_keystate = get_async_key_state(keystate)))
+        if ((got_keystate = NtUserGetAsyncKeyboardState(keystate)))
         {
             update_modifier_state(MOD_ALT, event->hotkey_press.mod_flags, keystate, VK_LMENU, VK_RMENU,
                                   0x38, 0x138, event->hotkey_press.time_ms, FALSE);
@@ -1286,15 +1251,38 @@ void macdrv_hotkey_press(const macdrv_event *event)
 
 
 /***********************************************************************
- *              macdrv_process_text_input
+ *              ImeProcessKey (MACDRV.@)
  */
-void macdrv_process_text_input(UINT vkey, UINT scan, UINT repeat, const BYTE *key_state, void *himc, int* done)
+UINT macdrv_ImeProcessKey(HIMC himc, UINT wparam, UINT lparam, const BYTE *key_state)
 {
     struct macdrv_thread_data *thread_data = macdrv_thread_data();
+    WORD scan = HIWORD(lparam) & 0x1ff, vkey = LOWORD(wparam);
+    BOOL repeat = !!(lparam >> 30), pressed = !(lparam >> 31);
     unsigned int flags;
     int keyc;
+    UINT ret;
 
-    TRACE("vkey 0x%04x scan 0x%04x repeat %u himc %p\n", vkey, scan, repeat, himc);
+    TRACE("himc %p, scan %#x, vkey %#x, repeat %u, pressed %u\n",
+          himc, scan, vkey, repeat, pressed);
+
+    if (!macdrv_using_input_method()) return 0;
+
+    if (!pressed)
+    {
+        /* Only key down events should be sent to the Cocoa input context. We do
+           not handle key ups, and instead let those go through as a normal
+           WM_KEYUP. */
+        return 0;
+    }
+
+    switch (vkey)
+    {
+        case VK_SHIFT:
+        case VK_CONTROL:
+        case VK_CAPITAL:
+        case VK_MENU:
+        return 0;
+    }
 
     flags = thread_data->last_modifiers;
     if (key_state[VK_SHIFT] & 0x80)
@@ -1318,15 +1306,13 @@ void macdrv_process_text_input(UINT vkey, UINT scan, UINT repeat, const BYTE *ke
     for (keyc = 0; keyc < ARRAY_SIZE(thread_data->keyc2vkey); keyc++)
         if (thread_data->keyc2vkey[keyc] == vkey) break;
 
-    if (keyc >= ARRAY_SIZE(thread_data->keyc2vkey))
-    {
-        *done = -1;
-        return;
-    }
+    if (keyc >= ARRAY_SIZE(thread_data->keyc2vkey)) return 0;
 
     TRACE("flags 0x%08x keyc 0x%04x\n", flags, keyc);
-
-    macdrv_send_text_input_event(((scan & 0x8000) == 0), flags, repeat, keyc, himc, done);
+    ret = (UINT)macdrv_ime_process_key(keyc, flags, repeat, himc);
+    NtUserMsgWaitForMultipleObjectsEx(0, NULL, 0, QS_POSTMESSAGE | QS_SENDMESSAGE, 0);
+    TRACE("returning %u\n", ret);
+    return ret;
 }
 
 
@@ -1344,7 +1330,7 @@ BOOL macdrv_ActivateKeyboardLayout(HKL hkl, UINT flags)
     if (hkl == thread_data->active_keyboard_layout)
         return TRUE;
 
-    EnterCriticalSection(&layout_list_section);
+    pthread_mutex_lock(&layout_list_mutex);
     update_layout_list();
 
     LIST_FOR_EACH_ENTRY(layout, &layout_list, struct layout, entry)
@@ -1367,7 +1353,7 @@ BOOL macdrv_ActivateKeyboardLayout(HKL hkl, UINT flags)
             break;
         }
     }
-    LeaveCriticalSection(&layout_list_section);
+    pthread_mutex_unlock(&layout_list_mutex);
 
     return ret;
 }
@@ -1395,12 +1381,11 @@ INT macdrv_GetKeyNameText(LONG lparam, LPWSTR buffer, INT size)
     {
         if (thread_data->keyc2scan[keyc] == scan)
         {
-            static const WCHAR dead[] = {' ','d','e','a','d',0};
             const UCKeyboardLayout *uchr;
             UInt32 deadKeyState = 0;
             UniCharCount len;
             OSStatus status;
-            DWORD vkey;
+            unsigned int vkey;
             int i;
 
             uchr = (const UCKeyboardLayout*)CFDataGetBytePtr(thread_data->keyboard_layout_uchr);
@@ -1436,17 +1421,20 @@ INT macdrv_GetKeyNameText(LONG lparam, LPWSTR buffer, INT size)
             {
                 if (vkey_names[i].vkey == vkey)
                 {
-                    len = MultiByteToWideChar(CP_UTF8, 0, vkey_names[i].name, -1, buffer, size);
-                    if (len) len--;
+                    len = min(strlen(vkey_names[i].name) + 1, size);
+                    ascii_to_unicode(buffer, vkey_names[i].name, len);
+                    if (len) buffer[--len] = 0;
                     break;
                 }
             }
 
             if (!len)
             {
-                static const WCHAR format[] = {'K','e','y',' ','0','x','%','0','2','x',0};
-                snprintfW(buffer, size, format, vkey);
-                len = strlenW(buffer);
+                char name[16];
+                len = snprintf(name, sizeof(name), "Key 0x%02x", vkey);
+                len = min(len + 1, size);
+                ascii_to_unicode(buffer, name, len);
+                if (len) buffer[--len] = 0;
             }
 
             if (!len)
@@ -1454,16 +1442,27 @@ INT macdrv_GetKeyNameText(LONG lparam, LPWSTR buffer, INT size)
 
             if (status == noErr && deadKeyState)
             {
-                lstrcpynW(buffer + len, dead, size - len);
-                len = strlenW(buffer);
+                for (i = 0; i < ARRAY_SIZE(dead_key_names); i++)
+                {
+                    if (dead_key_names[i].wchar == buffer[0])
+                    {
+                        len = min(strlen(dead_key_names[i].name) + 1, size);
+                        ascii_to_unicode(buffer, dead_key_names[i].name, len);
+                        if (len) buffer[--len] = 0;
+                        break;
+                    }
+                }
             }
 
-            TRACE("lparam 0x%08x -> %s\n", lparam, debugstr_w(buffer));
+            if (status == noErr && len == 1 && buffer[0] >= 'a' && buffer[0] <= 'z')
+                buffer[0] += 'A' - 'a';
+
+            TRACE("lparam 0x%08x -> %s\n", (unsigned int)lparam, debugstr_w(buffer));
             return len;
         }
     }
 
-    WARN("found no name for lparam 0x%08x\n", lparam);
+    WARN("found no name for lparam 0x%08x\n", (unsigned int)lparam);
     return 0;
 }
 
@@ -1478,7 +1477,7 @@ UINT macdrv_GetKeyboardLayoutList(INT size, HKL *list)
 
     TRACE("%d, %p\n", size, list);
 
-    EnterCriticalSection(&layout_list_section);
+    pthread_mutex_lock(&layout_list_mutex);
 
     update_layout_list();
 
@@ -1493,7 +1492,7 @@ UINT macdrv_GetKeyboardLayoutList(INT size, HKL *list)
         }
         count++;
     }
-    LeaveCriticalSection(&layout_list_section);
+    pthread_mutex_unlock(&layout_list_mutex);
 
     TRACE("returning %d\n", count);
     return count;
@@ -1615,7 +1614,7 @@ UINT macdrv_MapVirtualKeyEx(UINT wCode, UINT wMapType, HKL hkl)
             }
 
             if (status == noErr && len)
-                ret = toupperW(s[0]) | (deadKey ? 0x80000000 : 0);
+                ret = RtlUpcaseUnicodeChar(s[0]) | (deadKey ? 0x80000000 : 0);
 
             break;
         }
@@ -1663,9 +1662,9 @@ BOOL macdrv_RegisterHotKey(HWND hwnd, UINT mod_flags, UINT vkey)
     TRACE_(key)("keyc 0x%04x modifiers 0x%08x -> %d\n", keyc, modifiers, ret);
 
     if (ret == MACDRV_HOTKEY_ALREADY_REGISTERED)
-        SetLastError(ERROR_HOTKEY_ALREADY_REGISTERED);
+        RtlSetLastWin32Error(ERROR_HOTKEY_ALREADY_REGISTERED);
     else if (ret != MACDRV_HOTKEY_SUCCESS)
-        SetLastError(ERROR_GEN_FAILURE);
+        RtlSetLastWin32Error(ERROR_GEN_FAILURE);
 
     return ret == MACDRV_HOTKEY_SUCCESS;
 }
@@ -1704,7 +1703,7 @@ INT macdrv_ToUnicodeEx(UINT virtKey, UINT scanCode, const BYTE *lpKeyState,
     UInt32 modifierKeyState;
     OptionBits options;
     UInt32 deadKeyState, savedDeadKeyState;
-    UniCharCount len;
+    UniCharCount len = 0;
     BOOL dead = FALSE;
 
     TRACE_(key)("virtKey 0x%04x scanCode 0x%04x lpKeyState %p bufW %p bufW_size %d flags 0x%08x hkl %p\n",
@@ -1840,7 +1839,7 @@ INT macdrv_ToUnicodeEx(UINT virtKey, UINT scanCode, const BYTE *lpKeyState,
             &savedDeadKeyState, bufW_size, &len, bufW);
         if (status != noErr)
         {
-            ERR_(key)("Couldn't translate keycode 0x%04x, status %d\n", keyc, status);
+            ERR_(key)("Couldn't translate dead keycode 0x%04x, status %d\n", keyc, status);
             goto done;
         }
 
@@ -1850,8 +1849,7 @@ INT macdrv_ToUnicodeEx(UINT virtKey, UINT scanCode, const BYTE *lpKeyState,
     if (len > 0)
         len = strip_apple_private_chars(bufW, len);
 
-    if (dead && len > 0) ret = -1;
-    else ret = len;
+    ret = dead ? -len : len;
 
     /* Control-Return produces line feed instead of carriage return. */
     if (ret > 0 && (lpKeyState[VK_CONTROL] & 0x80) && virtKey == VK_RETURN)
@@ -1865,10 +1863,10 @@ INT macdrv_ToUnicodeEx(UINT virtKey, UINT scanCode, const BYTE *lpKeyState,
 done:
     /* Null-terminate the buffer, if there's room.  MSDN clearly states that the
        caller must not assume this is done, but some programs (e.g. Audiosurf) do. */
-    if (1 <= ret && ret < bufW_size)
-        bufW[ret] = 0;
+    if (len < bufW_size)
+        bufW[len] = 0;
 
-    TRACE_(key)("returning %d / %s\n", ret, debugstr_wn(bufW, abs(ret)));
+    TRACE_(key)("returning %d / %s\n", ret, debugstr_wn(bufW, len));
     return ret;
 }
 
