@@ -19,57 +19,53 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
  */
+
+#if 0
+#pragma makedep unix
+#endif
+
 #include "config.h"
 
 #include <Security/AuthSession.h>
 #include <IOKit/pwr_mgt/IOPMLib.h>
-#include <unistd.h> /* CrossOver Hack 11095 */
 
+#include "ntstatus.h"
+#define WIN32_NO_STATUS
 #include "macdrv.h"
-#include "winuser.h"
-#include "winreg.h"
+#include "shellapi.h"
 #include "wine/server.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(macdrv);
 
-#ifndef kIOPMAssertionTypePreventUserIdleDisplaySleep
-#define kIOPMAssertionTypePreventUserIdleDisplaySleep CFSTR("PreventUserIdleDisplaySleep")
-#endif
-#ifndef kCFCoreFoundationVersionNumber10_7
-#define kCFCoreFoundationVersionNumber10_7      635.00
-#endif
-
 #define IS_OPTION_TRUE(ch) \
     ((ch) == 'y' || (ch) == 'Y' || (ch) == 't' || (ch) == 'T' || (ch) == '1')
 
 C_ASSERT(NUM_EVENT_TYPES <= sizeof(macdrv_event_mask) * 8);
 
-DWORD thread_data_tls_index = TLS_OUT_OF_INDEXES;
-
 int topmost_float_inactive = TOPMOST_FLOAT_INACTIVE_NONFULLSCREEN;
-int capture_displays_for_fullscreen = 0;
-BOOL skip_single_buffer_flushes = FALSE;
+bool capture_displays_for_fullscreen = false;
 BOOL allow_vsync = TRUE;
 BOOL allow_set_gamma = TRUE;
 /* CrossOver Hack 10912: Mac Edit menu */
 int mac_edit_menu = MAC_EDIT_MENU_BY_KEY;
-int left_option_is_alt = 0;
-int right_option_is_alt = 0;
-int left_command_is_ctrl = 0;
-int right_command_is_ctrl = 0;
+bool left_option_is_alt = false;
+bool right_option_is_alt = false;
+bool left_command_is_ctrl = false;
+bool right_command_is_ctrl = false;
 BOOL allow_software_rendering = FALSE;
-BOOL disable_window_decorations = FALSE;
-int allow_immovable_windows = TRUE;
-int use_confinement_cursor_clipping = TRUE;
-int cursor_clipping_locks_windows = TRUE;
-int use_precise_scrolling = TRUE;
+bool allow_immovable_windows = true;
+bool use_confinement_cursor_clipping = true;
+bool cursor_clipping_locks_windows = true;
+bool use_precise_scrolling = true;
 int gl_surface_mode = GL_SURFACE_IN_FRONT_OPAQUE;
-int retina_enabled = FALSE;
-HMODULE macdrv_module = 0;
-int enable_app_nap = FALSE;
+bool retina_enabled = false;
+bool enable_app_nap = false;
 
-/* CrossOver Hack 14364 */
-BOOL force_backing_store = FALSE;
+UINT64 app_icon_callback = 0;
+UINT64 app_quit_request_callback = 0;
+UINT64 regcreateopenkeyexa_callback = 0;
+UINT64 regqueryvalueexa_callback = 0;
+UINT64 regsetvalueexa_callback = 0;
 
 CFDictionaryRef localized_strings;
 
@@ -80,8 +76,7 @@ CFDictionaryRef localized_strings;
 const char* debugstr_cf(CFTypeRef t)
 {
     CFStringRef s;
-    const char* HOSTPTR cstring = NULL;
-    const char* ret = NULL;
+    const char* ret;
 
     if (!t) return "(null)";
 
@@ -89,13 +84,13 @@ const char* debugstr_cf(CFTypeRef t)
         s = t;
     else
         s = CFCopyDescription(t);
-    cstring = CFStringGetCStringPtr(s, kCFStringEncodingUTF8);
-    if (cstring) ret = debugstr_a(cstring);
+    ret = CFStringGetCStringPtr(s, kCFStringEncodingUTF8);
+    if (ret) ret = debugstr_a(ret);
     if (!ret)
     {
-        const UniChar* HOSTPTR u = CFStringGetCharactersPtr(s);
+        const UniChar* u = CFStringGetCharactersPtr(s);
         if (u)
-            ret = debugstr_wn((const WCHAR* HOSTPTR)u, CFStringGetLength(s));
+            ret = debugstr_wn((const WCHAR*)u, CFStringGetLength(s));
     }
     if (!ret)
     {
@@ -109,16 +104,174 @@ const char* debugstr_cf(CFTypeRef t)
 }
 
 
+HKEY reg_open_key(HKEY root, const WCHAR *name, ULONG name_len)
+{
+    UNICODE_STRING nameW = { name_len, name_len, (WCHAR *)name };
+    OBJECT_ATTRIBUTES attr;
+    HANDLE ret;
+
+    attr.Length = sizeof(attr);
+    attr.RootDirectory = root;
+    attr.ObjectName = &nameW;
+    attr.Attributes = 0;
+    attr.SecurityDescriptor = NULL;
+    attr.SecurityQualityOfService = NULL;
+
+    return NtOpenKeyEx(&ret, MAXIMUM_ALLOWED, &attr, 0) ? 0 : ret;
+}
+
+
+HKEY open_hkcu_key(const char *name)
+{
+    WCHAR bufferW[256];
+    static HKEY hkcu;
+
+    if (!hkcu)
+    {
+        char buffer[256];
+        DWORD_PTR sid_data[(sizeof(TOKEN_USER) + SECURITY_MAX_SID_SIZE) / sizeof(DWORD_PTR)];
+        DWORD i, len = sizeof(sid_data);
+        SID *sid;
+
+        if (NtQueryInformationToken(GetCurrentThreadEffectiveToken(), TokenUser, sid_data, len, &len))
+            return 0;
+
+        sid = ((TOKEN_USER *)sid_data)->User.Sid;
+        len = snprintf(buffer, sizeof(buffer), "\\Registry\\User\\S-%u-%u", sid->Revision,
+                        MAKELONG(MAKEWORD(sid->IdentifierAuthority.Value[5],
+                                           sid->IdentifierAuthority.Value[4]),
+                                  MAKEWORD(sid->IdentifierAuthority.Value[3],
+                                            sid->IdentifierAuthority.Value[2])));
+        for (i = 0; i < sid->SubAuthorityCount; i++)
+            len += snprintf(buffer + len, sizeof(buffer) - len, "-%u", sid->SubAuthority[i]);
+
+        ascii_to_unicode(bufferW, buffer, len);
+        hkcu = reg_open_key(NULL, bufferW, len * sizeof(WCHAR));
+    }
+
+    return reg_open_key(hkcu, bufferW, asciiz_to_unicode(bufferW, name) - sizeof(WCHAR));
+}
+
+
+/* wrapper for NtCreateKey that creates the key recursively if necessary */
+HKEY reg_create_key(HKEY root, const WCHAR *name, ULONG name_len,
+                    DWORD options, DWORD *disposition)
+{
+    UNICODE_STRING nameW = { name_len, name_len, (WCHAR *)name };
+    OBJECT_ATTRIBUTES attr;
+    NTSTATUS status;
+    HANDLE ret;
+
+    attr.Length = sizeof(attr);
+    attr.RootDirectory = root;
+    attr.ObjectName = &nameW;
+    attr.Attributes = 0;
+    attr.SecurityDescriptor = NULL;
+    attr.SecurityQualityOfService = NULL;
+
+    status = NtCreateKey(&ret, MAXIMUM_ALLOWED, &attr, 0, NULL, options, disposition);
+    if (status == STATUS_OBJECT_NAME_NOT_FOUND)
+    {
+        static const WCHAR registry_rootW[] = { '\\','R','e','g','i','s','t','r','y','\\' };
+        DWORD pos = 0, i = 0, len = name_len / sizeof(WCHAR);
+
+        /* don't try to create registry root */
+        if (!root && len > ARRAY_SIZE(registry_rootW) &&
+            !memcmp(name, registry_rootW, sizeof(registry_rootW)))
+            i += ARRAY_SIZE(registry_rootW);
+
+        while (i < len && name[i] != '\\') i++;
+        if (i == len) return 0;
+        for (;;)
+        {
+            unsigned int subkey_options = options;
+            if (i < len) subkey_options &= ~(REG_OPTION_CREATE_LINK | REG_OPTION_OPEN_LINK);
+            nameW.Buffer = (WCHAR *)name + pos;
+            nameW.Length = (i - pos) * sizeof(WCHAR);
+            status = NtCreateKey(&ret, MAXIMUM_ALLOWED, &attr, 0, NULL, subkey_options, disposition);
+
+            if (attr.RootDirectory != root) NtClose(attr.RootDirectory);
+            if (!NT_SUCCESS(status)) return 0;
+            if (i == len) break;
+            attr.RootDirectory = ret;
+            while (i < len && name[i] == '\\') i++;
+            pos = i;
+            while (i < len && name[i] != '\\') i++;
+        }
+    }
+    return ret;
+}
+
+
+HKEY reg_create_ascii_key(HKEY root, const char *name, DWORD options, DWORD *disposition)
+{
+    WCHAR buf[256];
+    return reg_create_key(root, buf, asciiz_to_unicode(buf, name) - sizeof(WCHAR),
+                          options, disposition);
+}
+
+
+BOOL reg_delete_tree(HKEY parent, const WCHAR *name, ULONG name_len)
+{
+    char buffer[4096];
+    KEY_NODE_INFORMATION *key_info = (KEY_NODE_INFORMATION *)buffer;
+    DWORD size;
+    HKEY key;
+    BOOL ret = TRUE;
+
+    if (!(key = reg_open_key(parent, name, name_len))) return FALSE;
+
+    while (ret && !NtEnumerateKey(key, 0, KeyNodeInformation, key_info, sizeof(buffer), &size))
+        ret = reg_delete_tree(key, key_info->Name, key_info->NameLength);
+
+    if (ret) ret = !NtDeleteKey(key);
+    NtClose(key);
+    return ret;
+}
+
+
+ULONG query_reg_value(HKEY hkey, const WCHAR *name, KEY_VALUE_PARTIAL_INFORMATION *info, ULONG size)
+{
+    UNICODE_STRING str;
+
+    RtlInitUnicodeString(&str, name);
+    if (NtQueryValueKey(hkey, &str, KeyValuePartialInformation, info, size, &size))
+        return 0;
+
+    return size - FIELD_OFFSET(KEY_VALUE_PARTIAL_INFORMATION, Data);
+}
+
+
 /***********************************************************************
  *              get_config_key
  *
  * Get a config key from either the app-specific or the default config
  */
 static inline DWORD get_config_key(HKEY defkey, HKEY appkey, const char *name,
-                                   char *buffer, DWORD size)
+                                   WCHAR *buffer, DWORD size)
 {
-    if (appkey && !RegQueryValueExA(appkey, name, 0, NULL, (LPBYTE)buffer, &size)) return 0;
-    if (defkey && !RegQueryValueExA(defkey, name, 0, NULL, (LPBYTE)buffer, &size)) return 0;
+    WCHAR nameW[128];
+    char buf[2048];
+    KEY_VALUE_PARTIAL_INFORMATION *info = (void *)buf;
+
+    asciiz_to_unicode(nameW, name);
+
+    if (appkey && query_reg_value(appkey, nameW, info, sizeof(buf)))
+    {
+        size = min(info->DataLength, size - sizeof(WCHAR));
+        memcpy(buffer, info->Data, size);
+        buffer[size / sizeof(WCHAR)] = 0;
+        return 0;
+    }
+
+    if (defkey && query_reg_value(defkey, nameW, info, sizeof(buf)))
+    {
+        size = min(info->DataLength, size - sizeof(WCHAR));
+        memcpy(buffer, info->Data, size);
+        buffer[size / sizeof(WCHAR)] = 0;
+        return 0;
+    }
+
     return ERROR_FILE_NOT_FOUND;
 }
 
@@ -130,36 +283,41 @@ static inline DWORD get_config_key(HKEY defkey, HKEY appkey, const char *name,
  */
 static void setup_options(void)
 {
-    char buffer[MAX_PATH + 16];
+    static const WCHAR macdriverW[] = {'\\','M','a','c',' ','D','r','i','v','e','r',0};
+    WCHAR buffer[MAX_PATH + 16], *p, *appname;
     HKEY hkey, appkey = 0;
     DWORD len;
 
     /* @@ Wine registry key: HKCU\Software\Wine\Mac Driver */
-    if (RegOpenKeyA(HKEY_CURRENT_USER, "Software\\Wine\\Mac Driver", &hkey)) hkey = 0;
+    hkey = open_hkcu_key("Software\\Wine\\Mac Driver");
 
     /* open the app-specific key */
 
-    len = GetModuleFileNameA(0, buffer, MAX_PATH);
+    appname = NtCurrentTeb()->Peb->ProcessParameters->ImagePathName.Buffer;
+    if ((p = wcsrchr(appname, '/'))) appname = p + 1;
+    if ((p = wcsrchr(appname, '\\'))) appname = p + 1;
+    len = lstrlenW(appname);
+
     if (len && len < MAX_PATH)
     {
         HKEY tmpkey;
-        char *p, *appname = buffer;
-        if ((p = strrchr(appname, '/'))) appname = p + 1;
-        if ((p = strrchr(appname, '\\'))) appname = p + 1;
-        strcat(appname, "\\Mac Driver");
+        memcpy(buffer, appname, len * sizeof(WCHAR));
+        memcpy(buffer + len, macdriverW, sizeof(macdriverW));
         /* @@ Wine registry key: HKCU\Software\Wine\AppDefaults\app.exe\Mac Driver */
-        if (!RegOpenKeyA(HKEY_CURRENT_USER, "Software\\Wine\\AppDefaults", &tmpkey))
+        if ((tmpkey = open_hkcu_key("Software\\Wine\\AppDefaults")))
         {
-            if (RegOpenKeyA(tmpkey, appname, &appkey)) appkey = 0;
-            RegCloseKey(tmpkey);
+            appkey = reg_open_key(tmpkey, buffer, lstrlenW(buffer) * sizeof(WCHAR));
+            NtClose(tmpkey);
         }
     }
 
     if (!get_config_key(hkey, appkey, "WindowsFloatWhenInactive", buffer, sizeof(buffer)))
     {
-        if (!strcmp(buffer, "none"))
+        static const WCHAR noneW[] = {'n','o','n','e',0};
+        static const WCHAR allW[] = {'a','l','l',0};
+        if (!wcscmp(buffer, noneW))
             topmost_float_inactive = TOPMOST_FLOAT_INACTIVE_NONE;
-        else if (!strcmp(buffer, "all"))
+        else if (!wcscmp(buffer, allW))
             topmost_float_inactive = TOPMOST_FLOAT_INACTIVE_ALL;
         else
             topmost_float_inactive = TOPMOST_FLOAT_INACTIVE_NONFULLSCREEN;
@@ -168,9 +326,6 @@ static void setup_options(void)
     if (!get_config_key(hkey, appkey, "CaptureDisplaysForFullscreen", buffer, sizeof(buffer)))
         capture_displays_for_fullscreen = IS_OPTION_TRUE(buffer[0]);
 
-    if (!get_config_key(hkey, appkey, "SkipSingleBufferFlushes", buffer, sizeof(buffer)))
-        skip_single_buffer_flushes = IS_OPTION_TRUE(buffer[0]);
-
     if (!get_config_key(hkey, appkey, "AllowVerticalSync", buffer, sizeof(buffer)))
         allow_vsync = IS_OPTION_TRUE(buffer[0]);
 
@@ -180,9 +335,11 @@ static void setup_options(void)
     /* CrossOver Hack 10912: Mac Edit menu */
     if (!get_config_key(hkey, appkey, "EditMenu", buffer, sizeof(buffer)))
     {
-        if (!strcmp(buffer, "message"))
+        static const WCHAR messageW[] = {'m','e','s','s','a','g','e',0};
+        static const WCHAR keyW[] = {'k','e','y',0};
+        if (!wcscmp(buffer, messageW))
             mac_edit_menu = MAC_EDIT_MENU_BY_MESSAGE;
-        else if (!strcmp(buffer, "key"))
+        else if (!wcscmp(buffer, keyW))
             mac_edit_menu = MAC_EDIT_MENU_BY_KEY;
         else
             mac_edit_menu = MAC_EDIT_MENU_DISABLED;
@@ -205,10 +362,6 @@ static void setup_options(void)
     if (!get_config_key(hkey, appkey, "AllowSoftwareRendering", buffer, sizeof(buffer)))
         allow_software_rendering = IS_OPTION_TRUE(buffer[0]);
 
-    /* Value name chosen to match what's used in the X11 driver. */
-    if (!get_config_key(hkey, appkey, "Decorated", buffer, sizeof(buffer)))
-        disable_window_decorations = !IS_OPTION_TRUE(buffer[0]);
-
     if (!get_config_key(hkey, appkey, "AllowImmovableWindows", buffer, sizeof(buffer)))
         allow_immovable_windows = IS_OPTION_TRUE(buffer[0]);
 
@@ -223,9 +376,11 @@ static void setup_options(void)
 
     if (!get_config_key(hkey, appkey, "OpenGLSurfaceMode", buffer, sizeof(buffer)))
     {
-        if (!strcmp(buffer, "transparent"))
+        static const WCHAR transparentW[] = {'t','r','a','n','s','p','a','r','e','n','t',0};
+        static const WCHAR behindW[] = {'b','e','h','i','n','d',0};
+        if (!wcscmp(buffer, transparentW))
             gl_surface_mode = GL_SURFACE_IN_FRONT_TRANSPARENT;
-        else if (!strcmp(buffer, "behind"))
+        else if (!wcscmp(buffer, behindW))
             gl_surface_mode = GL_SURFACE_BEHIND;
         else
             gl_surface_mode = GL_SURFACE_IN_FRONT_OPAQUE;
@@ -239,37 +394,19 @@ static void setup_options(void)
     if (!get_config_key(hkey, NULL, "RetinaMode", buffer, sizeof(buffer)))
         retina_enabled = IS_OPTION_TRUE(buffer[0]);
 
-    /* CrossOver Hack 14364 */
-    if (!get_config_key(hkey, appkey, "ForceOpenGLBackingStore", buffer, sizeof(buffer)))
-        force_backing_store = IS_OPTION_TRUE(buffer[0]);
+    retina_on = retina_enabled;
 
-    if (appkey) RegCloseKey(appkey);
-    if (hkey) RegCloseKey(hkey);
+    if (appkey) NtClose(appkey);
+    if (hkey) NtClose(hkey);
 }
 
 
 /***********************************************************************
  *              load_strings
  */
-static void load_strings(HINSTANCE instance)
+static void load_strings(struct localized_string *str)
 {
-    static const unsigned int ids[] = {
-        STRING_MENU_WINE,
-        STRING_MENU_ITEM_HIDE_APPNAME,
-        STRING_MENU_ITEM_HIDE,
-        STRING_MENU_ITEM_HIDE_OTHERS,
-        STRING_MENU_ITEM_SHOW_ALL,
-        STRING_MENU_ITEM_QUIT_APPNAME,
-        STRING_MENU_ITEM_QUIT,
-
-        STRING_MENU_WINDOW,
-        STRING_MENU_ITEM_MINIMIZE,
-        STRING_MENU_ITEM_ZOOM,
-        STRING_MENU_ITEM_ENTER_FULL_SCREEN,
-        STRING_MENU_ITEM_BRING_ALL_TO_FRONT,
-    };
     CFMutableDictionaryRef dict;
-    int i;
 
     dict = CFDictionaryCreateMutable(NULL, 0, &kCFTypeDictionaryKeyCallBacks,
                                      &kCFTypeDictionaryValueCallBacks);
@@ -279,21 +416,21 @@ static void load_strings(HINSTANCE instance)
         return;
     }
 
-    for (i = 0; i < ARRAY_SIZE(ids); i++)
+    while (str->id)
     {
-        LPCWSTR str;
-        int len = LoadStringW(instance, ids[i], (LPWSTR)&str, 0);
-        if (str && len)
+        if (str->str && str->len)
         {
-            CFNumberRef key = CFNumberCreate(NULL, kCFNumberIntType, &ids[i]);
-            CFStringRef value = CFStringCreateWithCharacters(NULL, (UniChar*)str, len);
+            const UniChar *ptr = param_ptr(str->str);
+            CFNumberRef key = CFNumberCreate(NULL, kCFNumberIntType, &str->id);
+            CFStringRef value = CFStringCreateWithCharacters(NULL, ptr, str->len);
             if (key && value)
                 CFDictionarySetValue(dict, key, value);
             else
-                ERR("Failed to add string ID 0x%04x %s\n", ids[i], debugstr_wn(str, len));
+                ERR("Failed to add string ID 0x%04x %s\n", str->id, debugstr_wn(ptr, str->len));
         }
         else
-            ERR("Failed to load string ID 0x%04x\n", ids[i]);
+            ERR("Failed to load string ID 0x%04x\n", str->id);
+        str++;
     }
 
     localized_strings = dict;
@@ -301,44 +438,37 @@ static void load_strings(HINSTANCE instance)
 
 
 /***********************************************************************
- *              process_attach
+ *              macdrv_init
  */
-static BOOL process_attach(void)
+static NTSTATUS macdrv_init(void *arg)
 {
+    struct init_params *params = arg;
     SessionAttributeBits attributes;
     OSStatus status;
 
-    /* CrossOver Hack 11095.  Cocoa makes a similar call to confstr() during
-       its first pass through the event loop, which happens on the main thread.
-       However, if Wine is double-fork()-ing on a background thread simultaneously
-       with the first such call, the child process can become deadlocked.  It
-       appears to be a bug in the system library.
-
-       By calling this here, we greatly reduce the likelihood of such a race
-       and deadlock. */
-    char dummy[256];
-    confstr(_CS_DARWIN_USER_CACHE_DIR, dummy, sizeof(dummy));
+    app_icon_callback = params->app_icon_callback;
+    app_quit_request_callback = params->app_quit_request_callback;
+    regcreateopenkeyexa_callback = params->regcreateopenkeyexa_callback;
+    regqueryvalueexa_callback = params->regqueryvalueexa_callback;
+    regsetvalueexa_callback = params->regsetvalueexa_callback;
 
     status = SessionGetInfo(callerSecuritySession, NULL, &attributes);
     if (status != noErr || !(attributes & sessionHasGraphicAccess))
-        return FALSE;
+        return STATUS_UNSUCCESSFUL;
 
+    init_win_context();
     setup_options();
-    load_strings(macdrv_module);
-
-    if ((thread_data_tls_index = TlsAlloc()) == TLS_OUT_OF_INDEXES) return FALSE;
+    load_strings(params->strings);
 
     macdrv_err_on = ERR_ON(macdrv);
-    if (macdrv_start_cocoa_app(GetTickCount64()))
+    if (macdrv_start_cocoa_app(NtGetTickCount()))
     {
         ERR("Failed to start Cocoa app main loop\n");
-        return FALSE;
+        return STATUS_UNSUCCESSFUL;
     }
 
     init_user_driver();
-    macdrv_init_display_devices(FALSE);
-
-    return TRUE;
+    return STATUS_SUCCESS;
 }
 
 
@@ -354,9 +484,9 @@ void macdrv_ThreadDetach(void)
         macdrv_destroy_event_queue(data->queue);
         if (data->keyboard_layout_uchr)
             CFRelease(data->keyboard_layout_uchr);
-        HeapFree(GetProcessHeap(), 0, data);
+        free(data);
         /* clear data in case we get re-entered from user32 before the thread is truly dead */
-        TlsSetValue(thread_data_tls_index, NULL);
+        NtUserGetThreadInfo()->driver_data = 0;
     }
 }
 
@@ -374,7 +504,7 @@ static void set_queue_display_fd(int fd)
     if (wine_server_fd_to_handle(fd, GENERIC_READ | SYNCHRONIZE, 0, &handle))
     {
         MESSAGE("macdrv: Can't allocate handle for event queue fd\n");
-        ExitProcess(1);
+        NtTerminateProcess(0, 1);
     }
     SERVER_START_REQ(set_queue_fd)
     {
@@ -385,9 +515,9 @@ static void set_queue_display_fd(int fd)
     if (ret)
     {
         MESSAGE("macdrv: Can't store handle for event queue fd\n");
-        ExitProcess(1);
+        NtTerminateProcess(0, 1);
     }
-    CloseHandle(handle);
+    NtClose(handle);
 }
 
 
@@ -401,16 +531,16 @@ struct macdrv_thread_data *macdrv_init_thread_data(void)
 
     if (data) return data;
 
-    if (!(data = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(*data))))
+    if (!(data = calloc(1, sizeof(*data))))
     {
         ERR("could not create data\n");
-        ExitProcess(1);
+        NtTerminateProcess(0, 1);
     }
 
     if (!(data->queue = macdrv_create_event_queue(macdrv_handle_event)))
     {
         ERR("macdrv: Can't create event queue.\n");
-        ExitProcess(1);
+        NtTerminateProcess(0, 1);
     }
 
     macdrv_get_input_source_info(&data->keyboard_layout_uchr, &data->keyboard_type, &data->iso_keyboard, &input_source);
@@ -419,31 +549,13 @@ struct macdrv_thread_data *macdrv_init_thread_data(void)
     macdrv_compute_keyboard_layout(data);
 
     set_queue_display_fd(macdrv_get_event_queue_fd(data->queue));
-    TlsSetValue(thread_data_tls_index, data);
+    NtUserGetThreadInfo()->driver_data = (UINT_PTR)data;
 
-    ActivateKeyboardLayout(data->active_keyboard_layout, 0);
+    NtUserActivateKeyboardLayout(data->active_keyboard_layout, 0);
     return data;
 }
 
 
-/***********************************************************************
- *              DllMain
- */
-BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, LPVOID reserved)
-{
-    BOOL ret = TRUE;
-
-    switch(reason)
-    {
-    case DLL_PROCESS_ATTACH:
-        DisableThreadLibraryCalls( hinst );
-        macdrv_module = hinst;
-        ret = process_attach();
-        break;
-    }
-    return ret;
-}
-
 /***********************************************************************
  *              SystemParametersInfo (MACDRV.@)
  */
@@ -492,13 +604,7 @@ BOOL macdrv_SystemParametersInfo( UINT action, UINT int_param, void *ptr_param,
             }
             else if (powerAssertion == kIOPMNullAssertionID)
             {
-                CFStringRef assertName;
-                /*Are we running Lion or later?*/
-                if (kCFCoreFoundationVersionNumber >= kCFCoreFoundationVersionNumber10_7)
-                    assertName = kIOPMAssertionTypePreventUserIdleDisplaySleep;
-                else
-                    assertName = kIOPMAssertionTypeNoDisplaySleep;
-                IOPMAssertionCreateWithName( assertName, kIOPMAssertionLevelOn,
+                IOPMAssertionCreateWithName( kIOPMAssertionTypePreventUserIdleDisplaySleep, kIOPMAssertionLevelOn,
                                              CFSTR("Wine Process requesting no screen saver"),
                                              &powerAssertion);
             }
@@ -507,3 +613,73 @@ BOOL macdrv_SystemParametersInfo( UINT action, UINT int_param, void *ptr_param,
     }
     return FALSE;
 }
+
+/* CW Hack 22310 */
+NTSTATUS macdrv_SetCurrentProcessExplicitAppUserModelID(const WCHAR *aumid)
+{
+    if (!macdrv_set_current_process_explicit_app_user_model_id(aumid, lstrlenW(aumid)))
+        return STATUS_INVALID_PARAMETER;
+
+    return 0;
+}
+
+/* CW Hack 22310 */
+NTSTATUS macdrv_GetCurrentProcessExplicitAppUserModelID(WCHAR *buffer, INT size)
+{
+    if (!buffer) return STATUS_INVALID_PARAMETER;
+
+    if (!macdrv_get_current_process_explicit_app_user_model_id(buffer, size))
+        return STATUS_BUFFER_TOO_SMALL;
+
+    return 0;
+}
+
+static NTSTATUS macdrv_quit_result(void *arg)
+{
+    struct quit_result_params *params = arg;
+    macdrv_quit_reply(params->result);
+    return 0;
+}
+
+
+const unixlib_entry_t __wine_unix_call_funcs[] =
+{
+    macdrv_init,
+    macdrv_quit_result,
+};
+
+C_ASSERT( ARRAYSIZE(__wine_unix_call_funcs) == unix_funcs_count );
+
+#ifdef _WIN64
+
+static NTSTATUS wow64_init(void *arg)
+{
+    struct
+    {
+        ULONG strings;
+        UINT64 app_icon_callback;
+        UINT64 app_quit_request_callback;
+        UINT64 regcreateopenkeyexa_callback;
+        UINT64 regqueryvalueexa_callback;
+        UINT64 regsetvalueexa_callback;
+    } *params32 = arg;
+    struct init_params params;
+
+    params.strings = UlongToPtr(params32->strings);
+    params.app_icon_callback = params32->app_icon_callback;
+    params.app_quit_request_callback = params32->app_quit_request_callback;
+    params.regcreateopenkeyexa_callback = params32->regcreateopenkeyexa_callback;
+    params.regqueryvalueexa_callback = params32->regqueryvalueexa_callback;
+    params.regsetvalueexa_callback = params32->regsetvalueexa_callback;
+    return macdrv_init(&params);
+}
+
+const unixlib_entry_t __wine_unix_call_wow64_funcs[] =
+{
+    wow64_init,
+    macdrv_quit_result,
+};
+
+C_ASSERT( ARRAYSIZE(__wine_unix_call_wow64_funcs) == unix_funcs_count );
+
+#endif /* _WIN64 */
