@@ -38,12 +38,8 @@ static FILE_OBJECTID_BUFFER windir_id, sysdir_id;
 static inline NTSTATUS get_file_id( HANDLE handle, FILE_OBJECTID_BUFFER *id )
 {
     IO_STATUS_BLOCK32 io32;
-    IO_STATUS_BLOCK io;
+    IO_STATUS_BLOCK io = { .Pointer = &io32 };
 
-    /* HACK: this shouldn't be necessary since we open the file for synchronous
-     * I/O, but we currently ignore that in ntdll.so and always write the 32-bit
-     * IOSB */
-    io.Pointer = &io32;
     return NtFsControlFile( handle, 0, NULL, NULL, &io, FSCTL_GET_OBJECT_ID, NULL, 0, id, sizeof(*id) );
 }
 
@@ -64,12 +60,12 @@ static inline ULONG starts_with_path( const WCHAR *name, ULONG name_len, const W
 void init_file_redirects(void)
 {
     OBJECT_ATTRIBUTES attr;
-    UNICODE_STRING nameW;
+    UNICODE_STRING windows = RTL_CONSTANT_STRING( L"\\??\\C:\\windows" );
+    UNICODE_STRING system32 = RTL_CONSTANT_STRING( L"\\??\\C:\\windows\\system32" );
     IO_STATUS_BLOCK io;
     HANDLE handle;
 
-    InitializeObjectAttributes( &attr, &nameW, OBJ_CASE_INSENSITIVE, 0, NULL );
-    RtlInitUnicodeString( &nameW, L"\\??\\C:\\windows" );
+    InitializeObjectAttributes( &attr, &windows, OBJ_CASE_INSENSITIVE, 0, NULL );
     if (!NtOpenFile( &handle, SYNCHRONIZE | FILE_LIST_DIRECTORY, &attr, &io,
                      FILE_SHARE_READ, FILE_SYNCHRONOUS_IO_NONALERT |
                      FILE_OPEN_FOR_BACKUP_INTENT | FILE_DIRECTORY_FILE ))
@@ -77,7 +73,7 @@ void init_file_redirects(void)
         get_file_id( handle, &windir_id );
         NtClose( handle );
     }
-    RtlInitUnicodeString( &nameW, L"\\??\\C:\\windows\\system32" );
+    attr.ObjectName = &system32;
     if (!NtOpenFile( &handle, SYNCHRONIZE | FILE_LIST_DIRECTORY, &attr, &io,
                      FILE_SHARE_READ, FILE_SYNCHRONOUS_IO_NONALERT |
                      FILE_OPEN_FOR_BACKUP_INTENT | FILE_DIRECTORY_FILE ))
@@ -148,9 +144,8 @@ BOOL get_file_redirect( OBJECT_ATTRIBUTES *attr )
 
     /* CW HACK 20810: disable FS redirection when 32-bit-only bottle is being used */
     {
-        UNICODE_STRING name_str, val_str;
+        UNICODE_STRING val_str, name_str = RTL_CONSTANT_STRING( L"WINEWOW6432BPREFIXMODE" );
 
-        RtlInitUnicodeString( &name_str, L"WINEWOW6432BPREFIXMODE" );
         val_str.MaximumLength = 0;
         if (RtlQueryEnvironmentVariable_U( NULL, &name_str, &val_str ) != STATUS_VARIABLE_NOT_FOUND)
             return FALSE;
@@ -242,6 +237,24 @@ NTSTATUS WINAPI wow64_NtCancelIoFileEx( UINT *args )
 }
 
 
+/**********************************************************************
+ *           wow64_NtCancelSynchronousIoFile
+ */
+NTSTATUS WINAPI wow64_NtCancelSynchronousIoFile( UINT *args )
+{
+    HANDLE handle = get_handle( &args );
+    IO_STATUS_BLOCK32 *io_ptr = get_ptr( &args );
+    IO_STATUS_BLOCK32 *io32 = get_ptr( &args );
+
+    IO_STATUS_BLOCK io;
+    NTSTATUS status;
+
+    status = NtCancelSynchronousIoFile( handle, (IO_STATUS_BLOCK *)io_ptr, iosb_32to64( &io, io32 ));
+    put_iosb( io32, &io );
+    return status;
+}
+
+
 /**********************************************************************
  *           wow64_NtCreateFile
  */
@@ -410,6 +423,26 @@ NTSTATUS WINAPI wow64_NtFlushBuffersFile( UINT *args )
 }
 
 
+/**********************************************************************
+ *           wow64_NtFlushBuffersFileEx
+ */
+NTSTATUS WINAPI wow64_NtFlushBuffersFileEx( UINT *args )
+{
+    HANDLE handle = get_handle( &args );
+    ULONG flags = get_ulong( &args );
+    void *params = get_ptr( &args );
+    ULONG size = get_ulong( &args );
+    IO_STATUS_BLOCK32 *io32 = get_ptr( &args );
+
+    IO_STATUS_BLOCK io;
+    NTSTATUS status;
+
+    status = NtFlushBuffersFileEx( handle, flags, params, size, iosb_32to64( &io, io32 ) );
+    put_iosb( io32, &io );
+    return status;
+}
+
+
 /**********************************************************************
  *           wow64_NtFsControlFile
  */
@@ -654,8 +687,10 @@ NTSTATUS WINAPI wow64_NtReadFile( UINT *args )
     IO_STATUS_BLOCK io;
     NTSTATUS status;
 
+    if (pBTCpuNotifyReadFile) pBTCpuNotifyReadFile( handle, buffer, len, FALSE, 0 );
     status = NtReadFile( handle, event, apc_32to64( apc ), apc_param_32to64( apc, apc_param ),
                          iosb_32to64( &io, io32 ), buffer, len, offset, key );
+    if (pBTCpuNotifyReadFile) pBTCpuNotifyReadFile( handle, buffer, len, TRUE, status );
     put_iosb( io32, &io );
     return status;
 }
@@ -726,7 +761,11 @@ NTSTATUS WINAPI wow64_NtRemoveIoCompletionEx( UINT *args )
 
     NTSTATUS status;
     ULONG i;
-    FILE_IO_COMPLETION_INFORMATION *info = Wow64AllocateTemp( count * sizeof(*info) );
+    FILE_IO_COMPLETION_INFORMATION *info;
+
+    if (!count) return STATUS_INVALID_PARAMETER;
+
+    info = Wow64AllocateTemp( count * sizeof(*info) );
 
     status = NtRemoveIoCompletionEx( handle, info, count, written, timeout, alertable );
     for (i = 0; i < *written; i++)
@@ -784,11 +823,14 @@ NTSTATUS WINAPI wow64_NtSetInformationFile( UINT *args )
     case FileIoPriorityHintInformation:   /* FILE_IO_PRIORITY_HINT_INFO */
     case FileValidDataLengthInformation:   /* FILE_VALID_DATA_LENGTH_INFORMATION */
     case FileDispositionInformation:   /* FILE_DISPOSITION_INFORMATION */
+    case FileDispositionInformationEx:   /* FILE_DISPOSITION_INFORMATION_EX */
         status = NtSetInformationFile( handle, iosb_32to64( &io, io32 ), ptr, len, class );
         break;
 
     case FileRenameInformation:   /* FILE_RENAME_INFORMATION */
+    case FileRenameInformationEx:   /* FILE_RENAME_INFORMATION */
     case FileLinkInformation:   /* FILE_LINK_INFORMATION */
+    case FileLinkInformationEx:   /* FILE_LINK_INFORMATION */
         if (len >= sizeof(FILE_RENAME_INFORMATION32))
         {
             OBJECT_ATTRIBUTES attr;
@@ -803,7 +845,7 @@ NTSTATUS WINAPI wow64_NtSetInformationFile( UINT *args )
             get_file_redirect( &attr );
             size = offsetof( FILE_RENAME_INFORMATION, FileName[name.Length/sizeof(WCHAR)] );
             info = Wow64AllocateTemp( size );
-            info->ReplaceIfExists = info32->ReplaceIfExists;
+            info->Flags           = info32->Flags;
             info->RootDirectory   = attr.RootDirectory;
             info->FileNameLength  = name.Length;
             memcpy( info->FileName, name.Buffer, info->FileNameLength );
@@ -950,66 +992,3 @@ NTSTATUS WINAPI wow64___wine_rpc_NtReadFile( UINT *args )
     put_iosb( io32, &io );
     return status;
 }
-
-
-/**********************************************************************
- *           wow64_wine_nt_to_unix_file_name
- */
-NTSTATUS WINAPI wow64_wine_nt_to_unix_file_name( UINT *args )
-{
-    OBJECT_ATTRIBUTES32 *attr32 = get_ptr( &args );
-    char *nameA = get_ptr( &args );
-    ULONG *size = get_ptr( &args );
-    UINT disposition = get_ulong( &args );
-
-    struct object_attr64 attr;
-
-    return wine_nt_to_unix_file_name( objattr_32to64_redirect( &attr, attr32 ), nameA, size, disposition );
-}
-
-
-/**********************************************************************
- *           wow64_wine_server_fd_to_handle
- */
-NTSTATUS WINAPI wow64_wine_server_fd_to_handle( UINT *args )
-{
-    int fd = get_ulong( &args );
-    ACCESS_MASK access = get_ulong( &args );
-    ULONG attributes = get_ulong( &args );
-    ULONG *handle_ptr = get_ptr( &args );
-
-    HANDLE handle = 0;
-    NTSTATUS status;
-
-    *handle_ptr = 0;
-    status = wine_server_fd_to_handle( fd, access, attributes, &handle );
-    put_handle( handle_ptr, handle );
-    return status;
-}
-
-
-/**********************************************************************
- *           wow64_wine_server_handle_to_fd
- */
-NTSTATUS WINAPI wow64_wine_server_handle_to_fd( UINT *args )
-{
-    HANDLE handle = get_handle( &args );
-    ACCESS_MASK access = get_ulong( &args );
-    int *unix_fd = get_ptr( &args );
-    unsigned int *options = get_ptr( &args );
-
-    return wine_server_handle_to_fd( handle, access, unix_fd, options );
-}
-
-
-/**********************************************************************
- *           wow64_wine_unix_to_nt_file_name
- */
-NTSTATUS WINAPI wow64_wine_unix_to_nt_file_name( UINT *args )
-{
-    const char *name = get_ptr( &args );
-    WCHAR *buffer = get_ptr( &args );
-    ULONG *size = get_ptr( &args );
-
-    return wine_unix_to_nt_file_name( name, buffer, size );
-}
