@@ -25,6 +25,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/mman.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -33,20 +34,65 @@
 #ifdef HAVE_SYS_SYSCTL_H
 # include <sys/sysctl.h>
 #endif
+#ifdef __APPLE__
+# include <mach-o/dyld.h>
+#endif
 
 #ifdef __APPLE__ /* CrossOver Hack 13438 */
+#include <crt_externs.h>
 #include <mach/mach.h>
 #include <mach/vm_map.h>
 #include <mach-o/dyld.h>
 #include <mach-o/getsect.h>
+#include <mach-o/ldsyms.h>
 #endif
 
 #include "main.h"
 
-extern char **environ;
+#if defined(__APPLE__) && defined(__x86_64__) && !defined(HAVE_WINE_PRELOADER)
+
+/* Not using the preloader on x86_64:
+ * Reserve the same areas as the preloader does, but using zero-fill sections
+ * (the only way to prevent system frameworks from using them, including allocations
+ * before main() runs).
+ */
+__asm__(".zerofill WINE_RESERVE,WINE_RESERVE");
+static char __wine_reserve[0x1fffff000] __attribute__((section("WINE_RESERVE, WINE_RESERVE")));
+
+__asm__(".zerofill WINE_TOP_DOWN,WINE_TOP_DOWN");
+static char __wine_top_down[0x001ff0000] __attribute__((section("WINE_TOP_DOWN, WINE_TOP_DOWN")));
+
+static const struct wine_preload_info preload_info[] =
+{
+    { __wine_reserve,  sizeof(__wine_reserve)  }, /*         0x1000 -    0x200000000: low 8GB */
+    { __wine_top_down, sizeof(__wine_top_down) }, /* 0x7ff000000000 - 0x7ff001ff0000: top-down allocations + virtual heap */
+    { 0, 0 }                                      /* end of list */
+};
+
+const __attribute((visibility("default"))) struct wine_preload_info *wine_main_preload_info = preload_info;
+
+static void init_reserved_areas(void)
+{
+    int i;
+
+    for (i = 0; wine_main_preload_info[i].size != 0; i++)
+    {
+        /* Match how the preloader maps reserved areas: */
+        mmap(wine_main_preload_info[i].addr, wine_main_preload_info[i].size, PROT_NONE,
+             MAP_FIXED | MAP_NORESERVE | MAP_PRIVATE | MAP_ANON, -1, 0);
+    }
+}
+
+#else
 
 /* the preloader will set this variable */
-const struct wine_preload_info *wine_main_preload_info = NULL;
+const __attribute((visibility("default"))) struct wine_preload_info *wine_main_preload_info = NULL;
+
+static void init_reserved_areas(void)
+{
+}
+
+#endif
 
 /* canonicalize path and return its directory name */
 static char *realpath_dirname( const char *name )
@@ -89,7 +135,7 @@ static char *build_path( const char *dir, const char *name )
     return ret;
 }
 
-static const char *get_self_exe( char *argv0 )
+static const char *get_self_exe(void)
 {
 #if defined(__linux__) || defined(__FreeBSD_kernel__) || defined(__NetBSD__)
     return "/proc/self/exe";
@@ -100,79 +146,33 @@ static const char *get_self_exe( char *argv0 )
     if (path && !sysctl( pathname, sizeof(pathname)/sizeof(pathname[0]), path, &path_size, NULL, 0 ))
         return path;
     free( path );
-#endif
-
-    if (!strchr( argv0, '/' )) /* search in PATH */
-    {
-        char *p, *path = getenv( "PATH" );
-
-        if (!path || !(path = strdup(path))) return NULL;
-        for (p = strtok( path, ":" ); p; p = strtok( NULL, ":" ))
-        {
-            char *name = build_path( p, argv0 );
-            if (!access( name, X_OK ))
-            {
-                free( path );
-                return name;
-            }
-            free( name );
-        }
-        free( path );
-        return NULL;
-    }
-    return argv0;
-}
-
-static void *try_dlopen( const char *dir, const char *name )
-{
-    char *path = build_path( dir, name );
-    void *handle = dlopen( path, RTLD_NOW );
+#elif defined(__APPLE__)
+    uint32_t path_size = PATH_MAX;
+    char *path = malloc( path_size );
+    if (path && !_NSGetExecutablePath( path, &path_size ))
+        return path;
     free( path );
-    return handle;
+#endif
+    return NULL;
 }
 
-static void *load_ntdll( char *argv0 )
+static void *try_dlopen( const char *argv0 )
 {
-#ifdef __i386__
-#define SO_DIR "i386-unix/"
-#elif defined(__x86_64__)
-#define SO_DIR "x86_64-unix/"
-#elif defined(__arm__)
-#define SO_DIR "arm-unix/"
-#elif defined(__aarch64__)
-#define SO_DIR "aarch64-unix/"
-#else
-#define SO_DIR ""
-#endif
-    const char *self = get_self_exe( argv0 );
-    char *path, *p;
-    void *handle = NULL;
+    char *dir, *path, *p;
+    void *handle;
 
-    if (self && ((path = realpath_dirname( self ))))
-    {
-        if ((p = remove_tail( path, "/loader" )))
-        {
-            handle = try_dlopen( p, "dlls/ntdll/ntdll.so" );
-            free( p );
-        }
-        else handle = try_dlopen( path, BIN_TO_DLLDIR "/" SO_DIR "ntdll.so" );
-        free( path );
-    }
+    if (!argv0) return NULL;
+    if (!(dir = realpath_dirname( argv0 ))) return NULL;
 
-    if (!handle && (path = getenv( "WINEDLLPATH" )))
-    {
-        path = strdup( path );
-        for (p = strtok( path, ":" ); p; p = strtok( NULL, ":" ))
-        {
-            handle = try_dlopen( p, SO_DIR "ntdll.so" );
-            if (!handle) handle = try_dlopen( p, "ntdll.so" );
-            if (handle) break;
-        }
-        free( path );
-    }
-
-    if (!handle && !self) handle = try_dlopen( DLLDIR, SO_DIR "ntdll.so" );
+    if ((p = remove_tail( dir, "/loader" )))
+        path = build_path( p, "dlls/ntdll/ntdll.so" );
+    else
+        path = build_path( dir, "ntdll.so" );
 
+    handle = dlopen( path, RTLD_NOW );
+    free( p );
+    free( dir );
+    free( path );
     return handle;
 }
 
@@ -202,16 +202,17 @@ static void apple_override_bundle_name( int argc, char *argv[] )
     const char* p;
     size_t bundle_name_len, max_bundle_name_len;
     vm_address_t start, end;
-    const char* new_bundle_name;
+    const char* new_bundle_name = getenv("WINEPRELOADERAPPNAME");
     size_t new_bundle_name_len;
 
-    if (argc < 2)
+    unsetenv("WINEPRELOADERAPPNAME");
+
+    if (!new_bundle_name && argc < 2)
         return;
 
-    info_plist = getsectdata("__TEXT", "__info_plist", &remaining);
+    info_plist = (char *)getsectiondata(_NSGetMachExecuteHeader(), "__TEXT", "__info_plist", &remaining);
     if (!info_plist || !remaining)
         return;
-    info_plist += _dyld_get_image_vmaddr_slide(0);
 
     bundle_name = strnstr(info_plist, prefix, remaining);
     if (!bundle_name)
@@ -231,16 +232,22 @@ static void apple_override_bundle_name( int argc, char *argv[] )
         !memcmp(bundle_name + bundle_name_len + suffix_len, padding, padding_len))
         max_bundle_name_len += padding_len;
 
-    new_bundle_name = argv[1];
-    if ((p = strrchr(new_bundle_name, '\\'))) new_bundle_name = p + 1;
-    if ((p = strrchr(new_bundle_name, '/'))) new_bundle_name = p + 1;
-    if (strspn(new_bundle_name, "0123456789abcdefABCDEF") == 32 &&
-        new_bundle_name[32] == '.')
-        new_bundle_name += 33;
-    if ((p = strrchr(new_bundle_name, '.')) && p != new_bundle_name)
-        new_bundle_name_len = p - new_bundle_name;
+    if (!new_bundle_name)
+    {
+        new_bundle_name = argv[1];
+        if ((p = strrchr(new_bundle_name, '\\'))) new_bundle_name = p + 1;
+        if ((p = strrchr(new_bundle_name, '/'))) new_bundle_name = p + 1;
+        if (strspn(new_bundle_name, "0123456789abcdefABCDEF") == 32 &&
+            new_bundle_name[32] == '.')
+            new_bundle_name += 33;
+        if ((p = strrchr(new_bundle_name, '.')) && p != new_bundle_name)
+            new_bundle_name_len = p - new_bundle_name;
+        else
+            new_bundle_name_len = strlen(new_bundle_name);
+    }
     else
         new_bundle_name_len = strlen(new_bundle_name);
+
     if (!new_bundle_name_len)
         return;
 
@@ -268,14 +275,17 @@ int main( int argc, char *argv[] )
 {
     void *handle;
 
+    init_reserved_areas();
+
 #ifdef __APPLE__ /* CrossOver Hack 13438 */
     apple_override_bundle_name(argc, argv);
 #endif
 
-    if ((handle = load_ntdll( argv[0] )))
+    if ((handle = try_dlopen( get_self_exe() )) ||
+        (handle = try_dlopen( argv[0] )))
     {
-        void (*init_func)(int, char **, char **) = dlsym( handle, "__wine_main" );
-        if (init_func) init_func( argc, argv, environ );
+        void (*init_func)(int, char **) = dlsym( handle, "__wine_main" );
+        if (init_func) init_func( argc, argv );
         fprintf( stderr, "wine: __wine_main function not found in ntdll.so\n" );
         exit(1);
     }
