@@ -22,26 +22,27 @@
 #include <stdio.h>
 
 #define COBJMACROS
-#define NONAMELESSUNION
-#define NONAMELESSSTRUCT
 #define OEMRESOURCE
 #include <windows.h>
+#include <winternl.h>
 #include <rpc.h>
 #include <shlobj.h>
 #include <shellapi.h>
+#include <ntuser.h>
 #include "exdisp.h"
 
 #include "wine/debug.h"
 #include "explorer_private.h"
+#include "resource.h"
 
 WINE_DEFAULT_DEBUG_CHANNEL(explorer);
 
 #define DESKTOP_CLASS_ATOM ((LPCWSTR)MAKEINTATOM(32769))
 #define DESKTOP_ALL_ACCESS 0x01ff
 
-static const WCHAR default_driver[] = {'m','a','c',',','x','1','1',0};
+static const WCHAR default_driver[] = L"mac,x11,wayland";
 
-static BOOL using_root;
+static BOOL using_root = TRUE;
 
 struct launcher
 {
@@ -97,10 +98,11 @@ static HRESULT get_typeinfo(tid_t tid, ITypeInfo **typeinfo)
 {
     HRESULT hres;
 
-    if (!typelib)
+    if (!typelib) {
         hres = load_typelib();
-    if (!typelib)
-        return hres;
+        if (!typelib)
+            return hres;
+    }
 
     if (!typeinfos[tid]) {
         ITypeInfo *ti;
@@ -271,25 +273,19 @@ static void draw_launchers( HDC hdc, RECT update_rect )
                        DT_CENTER|DT_WORDBREAK|DT_EDITCONTROL|DT_END_ELLIPSIS );
     }
 
-    SelectObject( hdc, font );
+    font = SelectObject( hdc, font );
+    DeleteObject( font );
     SetTextColor( hdc, color );
     SetBkMode( hdc, mode );
 }
 
-static void do_launch( const struct launcher *launcher )
-{
-    static const WCHAR openW[] = {'o','p','e','n',0};
-    ShellExecuteW( NULL, openW, launcher->path, NULL, NULL, 0 );
-}
-
 static WCHAR *append_path( const WCHAR *path, const WCHAR *filename, int len_filename )
 {
     int len_path = lstrlenW( path );
     WCHAR *ret;
 
     if (len_filename == -1) len_filename = lstrlenW( filename );
-    if (!(ret = HeapAlloc( GetProcessHeap(), 0, (len_path + len_filename + 2) * sizeof(WCHAR) )))
-        return NULL;
+    if (!(ret = malloc( (len_path + len_filename + 2) * sizeof(WCHAR) ))) return NULL;
     memcpy( ret, path, len_path * sizeof(WCHAR) );
     ret[len_path] = '\\';
     memcpy( ret + len_path + 1, filename, len_filename * sizeof(WCHAR) );
@@ -358,7 +354,7 @@ static WCHAR *build_title( const WCHAR *filename, int len )
             break;
         }
     }
-    if (!(ret = HeapAlloc( GetProcessHeap(), 0, (len + 1) * sizeof(WCHAR) ))) return NULL;
+    if (!(ret = malloc( (len + 1) * sizeof(WCHAR) ))) return NULL;
     memcpy( ret, filename, len * sizeof(WCHAR) );
     ret[len] = 0;
     return ret;
@@ -372,13 +368,13 @@ static BOOL add_launcher( const WCHAR *folder, const WCHAR *filename, int len_fi
     if (nb_launchers == nb_allocated)
     {
         unsigned int count = nb_allocated * 2;
-        struct launcher **tmp = HeapReAlloc( GetProcessHeap(), 0, launchers, count * sizeof(*tmp) );
+        struct launcher **tmp = realloc( launchers, count * sizeof(*tmp) );
         if (!tmp) return FALSE;
         launchers = tmp;
         nb_allocated = count;
     }
 
-    if (!(launcher = HeapAlloc( GetProcessHeap(), 0, sizeof(*launcher) ))) return FALSE;
+    if (!(launcher = malloc( sizeof(*launcher) ))) return FALSE;
     if (!(launcher->path = append_path( folder, filename, len_filename ))) goto error;
     if (!(link = load_shelllink( launcher->path ))) goto error;
 
@@ -390,21 +386,21 @@ static BOOL add_launcher( const WCHAR *folder, const WCHAR *filename, int len_fi
         launchers[nb_launchers++] = launcher;
         return TRUE;
     }
-    HeapFree( GetProcessHeap(), 0, launcher->title );
+    free( launcher->title );
     DestroyIcon( launcher->icon );
 
 error:
-    HeapFree( GetProcessHeap(), 0, launcher->path );
-    HeapFree( GetProcessHeap(), 0, launcher );
+    free( launcher->path );
+    free( launcher );
     return FALSE;
 }
 
 static void free_launcher( struct launcher *launcher )
 {
     DestroyIcon( launcher->icon );
-    HeapFree( GetProcessHeap(), 0, launcher->path );
-    HeapFree( GetProcessHeap(), 0, launcher->title );
-    HeapFree( GetProcessHeap(), 0, launcher );
+    free( launcher->path );
+    free( launcher->title );
+    free( launcher );
 }
 
 static BOOL remove_launcher( const WCHAR *folder, const WCHAR *filename, int len_filename )
@@ -425,7 +421,7 @@ static BOOL remove_launcher( const WCHAR *folder, const WCHAR *filename, int len
             break;
         }
     }
-    HeapFree( GetProcessHeap(), 0, path );
+    free( path );
     return ret;
 }
 
@@ -440,7 +436,8 @@ static BOOL get_icon_text_metrics( HWND hwnd, TEXTMETRICW *tm )
     SystemParametersInfoW( SPI_GETICONTITLELOGFONT, sizeof(lf), &lf, 0 );
     hfont = SelectObject( hdc, CreateFontIndirectW( &lf ) );
     ret = GetTextMetricsW( hdc, tm );
-    SelectObject( hdc, hfont );
+    hfont = SelectObject( hdc, hfont );
+    DeleteObject( hfont );
     ReleaseDC( hwnd, hdc );
     return ret;
 }
@@ -498,8 +495,8 @@ static DWORD CALLBACK watch_desktop_folders( LPVOID param )
     }
     if (!(ovl0.hEvent = events[0] = CreateEventW( NULL, FALSE, FALSE, NULL ))) goto error;
     if (!(ovl1.hEvent = events[1] = CreateEventW( NULL, FALSE, FALSE, NULL ))) goto error;
-    if (!(buf0 = HeapAlloc( GetProcessHeap(), 0, size ))) goto error;
-    if (!(buf1 = HeapAlloc( GetProcessHeap(), 0, size ))) goto error;
+    if (!(buf0 = malloc( size ))) goto error;
+    if (!(buf1 = malloc( size ))) goto error;
 
     for (;;)
     {
@@ -540,21 +537,21 @@ error:
     CloseHandle( dir1 );
     CloseHandle( events[0] );
     CloseHandle( events[1] );
-    HeapFree( GetProcessHeap(), 0, buf0 );
-    HeapFree( GetProcessHeap(), 0, buf1 );
+    free( buf0 );
+    free( buf1 );
     if (SUCCEEDED( init )) CoUninitialize();
     return error;
 }
 
 static void add_folder( const WCHAR *folder )
 {
-    static const WCHAR lnkW[] = {'\\','*','.','l','n','k',0};
+    static const WCHAR lnkW[] = L"\\*.lnk";
     int len = lstrlenW( folder ) + lstrlenW( lnkW );
     WIN32_FIND_DATAW data;
     HANDLE handle;
     WCHAR *glob;
 
-    if (!(glob = HeapAlloc( GetProcessHeap(), 0, (len + 1) * sizeof(WCHAR) ))) return;
+    if (!(glob = malloc( (len + 1) * sizeof(WCHAR) ))) return;
     lstrcpyW( glob, folder );
     lstrcatW( glob, lnkW );
 
@@ -563,7 +560,7 @@ static void add_folder( const WCHAR *folder )
         do { add_launcher( folder, data.cFileName, -1 ); } while (FindNextFileW( handle, &data ));
         FindClose( handle );
     }
-    HeapFree( GetProcessHeap(), 0, glob );
+    free( glob );
 }
 
 #define BORDER_SIZE  4
@@ -595,17 +592,17 @@ static void initialize_launchers( HWND hwnd )
     hr = SHGetKnownFolderPath( &FOLDERID_Desktop, KF_FLAG_CREATE, NULL, &desktop_folder );
     if (FAILED( hr ))
     {
-        WINE_ERR("Could not get user desktop folder\n");
+        ERR( "Could not get user desktop folder\n" );
         return;
     }
     hr = SHGetKnownFolderPath( &FOLDERID_PublicDesktop, KF_FLAG_CREATE, NULL, &desktop_folder_public );
     if (FAILED( hr ))
     {
-        WINE_ERR("Could not get public desktop folder\n");
+        ERR( "Could not get public desktop folder\n" );
         CoTaskMemFree( desktop_folder );
         return;
     }
-    if ((launchers = HeapAlloc( GetProcessHeap(), 0, 2 * sizeof(launchers[0]) )))
+    if ((launchers = malloc( 2 * sizeof(launchers[0]) )))
     {
         nb_allocated = 2;
 
@@ -618,12 +615,184 @@ static void initialize_launchers( HWND hwnd )
     }
 }
 
+static void wait_named_mutex( const WCHAR *name )
+{
+    HANDLE mutex;
+
+    mutex = CreateMutexW( NULL, TRUE, name );
+    if (GetLastError() == ERROR_ALREADY_EXISTS)
+    {
+        TRACE( "waiting for mutex %s\n", debugstr_w( name ));
+        WaitForSingleObject( mutex, INFINITE );
+    }
+}
+
+/**************************************************************************
+ *		wait_clipboard_mutex
+ *
+ * Make sure that there's only one clipboard thread per window station.
+ */
+static BOOL wait_clipboard_mutex(void)
+{
+    static const WCHAR prefix[] = L"__wine_clipboard_";
+    WCHAR buffer[MAX_PATH + ARRAY_SIZE( prefix )];
+
+    memcpy( buffer, prefix, sizeof(prefix) );
+    if (!GetUserObjectInformationW( GetProcessWindowStation(), UOI_NAME,
+                                    buffer + ARRAY_SIZE( prefix ) - 1,
+                                    sizeof(buffer) - sizeof(prefix), NULL ))
+    {
+        ERR( "failed to get winstation name\n" );
+        return FALSE;
+    }
+    wait_named_mutex( buffer );
+    return TRUE;
+}
+
+
+/**************************************************************************
+ *		clipboard_wndproc
+ *
+ * Window procedure for the clipboard manager.
+ */
+static LRESULT CALLBACK clipboard_wndproc( HWND hwnd, UINT msg, WPARAM wp, LPARAM lp )
+{
+    switch (msg)
+    {
+    case WM_NCCREATE:
+    case WM_CLIPBOARDUPDATE:
+    case WM_RENDERFORMAT:
+    case WM_TIMER:
+    case WM_DESTROYCLIPBOARD:
+    case WM_USER:
+        return NtUserMessageCall( hwnd, msg, wp, lp, 0, NtUserClipboardWindowProc, FALSE );
+    }
+
+    return DefWindowProcW( hwnd, msg, wp, lp );
+}
+
+
+/**************************************************************************
+ *		clipboard_thread
+ *
+ * Thread running inside the desktop process to manage the clipboard
+ */
+static DWORD WINAPI clipboard_thread( void *arg )
+{
+    static const WCHAR clipboard_classname[] = L"__wine_clipboard_manager";
+    WNDCLASSW class;
+    ATOM atom;
+    MSG msg;
+
+    if (!wait_clipboard_mutex()) return 0;
+
+    memset( &class, 0, sizeof(class) );
+    class.lpfnWndProc   = clipboard_wndproc;
+    class.lpszClassName = clipboard_classname;
+
+    if (!(atom = RegisterClassW( &class )) && GetLastError() != ERROR_CLASS_ALREADY_EXISTS)
+    {
+        ERR( "could not register clipboard window class err %lu\n", GetLastError() );
+        return 0;
+    }
+    if (!CreateWindowW( clipboard_classname, NULL, 0, 0, 0, 0, 0, HWND_MESSAGE, 0, 0, NULL ))
+    {
+        TRACE( "failed to create clipboard window err %lu\n", GetLastError() );
+        UnregisterClassW( MAKEINTRESOURCEW(atom), NULL );
+        return 0;
+    }
+
+    while (GetMessageW( &msg, 0, 0, 0 )) DispatchMessageW( &msg );
+    return 0;
+}
+
+static HANDLE fullscreen_process;
+
+static LRESULT WINAPI display_settings_restorer_wndproc( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
+{
+    TRACE( "got msg %04x wp %Ix lp %Ix\n", message, wp, lp );
+
+    switch(message)
+    {
+    case WM_USER + 0:
+        TRACE( "fullscreen process id %Iu.\n", lp );
+
+        if (fullscreen_process)
+        {
+            CloseHandle( fullscreen_process );
+            fullscreen_process = NULL;
+        }
+
+        if (lp)
+            fullscreen_process = OpenProcess( SYNCHRONIZE, FALSE, lp );
+
+        return 0;
+    }
+
+    return DefWindowProcW( hwnd, message, wp, lp );
+}
+
+static DWORD WINAPI display_settings_restorer_thread( void *param )
+{
+    static const WCHAR *display_settings_restorer_classname = L"__wine_display_settings_restorer";
+    DWORD wait_result;
+    WNDCLASSW class;
+    MSG msg;
+
+    SetThreadDescription( GetCurrentThread(), L"wine_explorer_display_settings_restorer" );
+
+    wait_named_mutex( L"__wine_display_settings_restorer_mutex" );
+
+    memset( &class, 0, sizeof(class) );
+    class.lpfnWndProc   = display_settings_restorer_wndproc;
+    class.lpszClassName = display_settings_restorer_classname;
+
+    if (!RegisterClassW( &class ) && GetLastError() != ERROR_CLASS_ALREADY_EXISTS)
+    {
+        ERR( "could not register display settings restorer window class err %lu\n", GetLastError() );
+        return 0;
+    }
+    if (!CreateWindowW( display_settings_restorer_classname, NULL, 0, 0, 0, 0, 0, HWND_MESSAGE, 0, 0, NULL ))
+    {
+        WARN( "failed to create display settings restorer window err %lu\n", GetLastError() );
+        UnregisterClassW( display_settings_restorer_classname, NULL );
+        return 0;
+    }
+
+    for (;;)
+    {
+        if (PeekMessageW( &msg, NULL, 0, 0, PM_REMOVE ))
+        {
+            if (msg.message == WM_QUIT)
+                break;
+            DispatchMessageW( &msg );
+            continue;
+        }
+
+        wait_result = MsgWaitForMultipleObjects( fullscreen_process ? 1 : 0, &fullscreen_process,
+                                                 FALSE, INFINITE, QS_ALLINPUT );
+        if (wait_result == WAIT_FAILED)
+            break;
+        if (!fullscreen_process || wait_result != WAIT_OBJECT_0)
+            continue;
+
+        TRACE( "restoring display settings on process exit\n" );
+
+        ChangeDisplaySettingsExW( NULL, NULL, NULL, 0, NULL );
+
+        CloseHandle( fullscreen_process );
+        fullscreen_process = NULL;
+    }
+
+    return 0;
+}
+
 static WNDPROC desktop_orig_wndproc;
 
 /* window procedure for the desktop window */
 static LRESULT WINAPI desktop_wnd_proc( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
 {
-    WINE_TRACE( "got msg %04x wp %Ix lp %Ix\n", message, wp, lp );
+    TRACE( "got msg %04x wp %Ix lp %Ix\n", message, wp, lp );
 
     switch(message)
     {
@@ -663,7 +832,7 @@ static LRESULT WINAPI desktop_wnd_proc( HWND hwnd, UINT message, WPARAM wp, LPAR
         if (!using_root)
         {
             const struct launcher *launcher = launcher_from_point( (short)LOWORD(lp), (short)HIWORD(lp) );
-            if (launcher) do_launch( launcher );
+            if (launcher) ShellExecuteW( NULL, L"open", launcher->path, NULL, NULL, 0 );
         }
         return 0;
 
@@ -673,7 +842,7 @@ static LRESULT WINAPI desktop_wnd_proc( HWND hwnd, UINT message, WPARAM wp, LPAR
             BeginPaint( hwnd, &ps );
             if (!using_root)
             {
-                if (ps.fErase) PaintDesktop( ps.hdc );
+                PaintDesktop( ps.hdc );
                 draw_launchers( ps.hdc, ps.rcPaint );
             }
             EndPaint( hwnd, &ps );
@@ -684,20 +853,6 @@ static LRESULT WINAPI desktop_wnd_proc( HWND hwnd, UINT message, WPARAM wp, LPAR
     return desktop_orig_wndproc( hwnd, message, wp, lp );
 }
 
-/* create the desktop and the associated driver window, and make it the current desktop */
-static BOOL create_desktop( HMODULE driver, const WCHAR *name, unsigned int width, unsigned int height )
-{
-    BOOL ret = FALSE;
-    BOOL (CDECL *create_desktop_func)(unsigned int, unsigned int);
-
-    if (driver)
-    {
-        create_desktop_func = (void *)GetProcAddress( driver, "wine_create_desktop" );
-        if (create_desktop_func) ret = create_desktop_func( width, height );
-    }
-    return ret;
-}
-
 /* parse the desktop size specification */
 static BOOL parse_size( const WCHAR *size, unsigned int *width, unsigned int *height )
 {
@@ -714,10 +869,6 @@ static BOOL parse_size( const WCHAR *size, unsigned int *width, unsigned int *he
 /* retrieve the desktop name to use if not specified on the command line */
 static const WCHAR *get_default_desktop_name(void)
 {
-    static const WCHAR desktopW[] = {'D','e','s','k','t','o','p',0};
-    static const WCHAR defaultW[] = {'D','e','f','a','u','l','t',0};
-    static const WCHAR explorer_keyW[] = {'S','o','f','t','w','a','r','e','\\','W','i','n','e','\\',
-                                          'E','x','p','l','o','r','e','r',0};
     static WCHAR buffer[MAX_PATH];
     DWORD size = sizeof(buffer);
     HDESK desk = GetThreadDesktop( GetCurrentThreadId() );
@@ -726,13 +877,13 @@ static const WCHAR *get_default_desktop_name(void)
 
     if (desk && GetUserObjectInformationW( desk, UOI_NAME, buffer, ARRAY_SIZE( buffer ), NULL ))
     {
-        if (wcsicmp( buffer, defaultW )) return buffer;
+        if (wcsicmp( buffer, L"Default" )) return buffer;
     }
 
     /* @@ Wine registry key: HKCU\Software\Wine\Explorer */
-    if (!RegOpenKeyW( HKEY_CURRENT_USER, explorer_keyW, &hkey ))
+    if (!RegOpenKeyW( HKEY_CURRENT_USER, L"Software\\Wine\\Explorer", &hkey ))
     {
-        if (!RegQueryValueExW( hkey, desktopW, 0, NULL, (LPBYTE)buffer, &size ) && *buffer) ret = buffer;
+        if (!RegQueryValueExW( hkey, L"Desktop", 0, NULL, (LPBYTE)buffer, &size ) && *buffer) ret = buffer;
         RegCloseKey( hkey );
     }
     return ret;
@@ -741,9 +892,6 @@ static const WCHAR *get_default_desktop_name(void)
 /* retrieve the default desktop size from the registry */
 static BOOL get_default_desktop_size( const WCHAR *name, unsigned int *width, unsigned int *height )
 {
-    static const WCHAR desktop_keyW[] = {'S','o','f','t','w','a','r','e','\\','W','i','n','e','\\',
-                                         'E','x','p','l','o','r','e','r','\\',
-                                         'D','e','s','k','t','o','p','s',0};
     HKEY hkey;
     WCHAR buffer[64];
     DWORD size = sizeof(buffer);
@@ -753,7 +901,7 @@ static BOOL get_default_desktop_size( const WCHAR *name, unsigned int *width, un
     *height = 600;
 
     /* @@ Wine registry key: HKCU\Software\Wine\Explorer\Desktops */
-    if (!RegOpenKeyW( HKEY_CURRENT_USER, desktop_keyW, &hkey ))
+    if (!RegOpenKeyW( HKEY_CURRENT_USER, L"Software\\Wine\\Explorer\\Desktops", &hkey ))
     {
         if (!RegQueryValueExW( hkey, name, 0, NULL, (LPBYTE)buffer, &size ))
         {
@@ -767,44 +915,87 @@ static BOOL get_default_desktop_size( const WCHAR *name, unsigned int *width, un
 
 static BOOL get_default_enable_shell( const WCHAR *name )
 {
-    static const WCHAR desktop_keyW[] = {'S','o','f','t','w','a','r','e','\\','W','i','n','e','\\',
-                                         'E','x','p','l','o','r','e','r','\\',
-                                         'D','e','s','k','t','o','p','s',0};
-    static const WCHAR enable_shellW[] = {'E','n','a','b','l','e','S','h','e','l','l',0};
-    static const WCHAR shellW[] = {'s','h','e','l','l',0};
     HKEY hkey;
     BOOL found = FALSE;
     BOOL result;
     DWORD size = sizeof(result);
 
     /* @@ Wine registry key: HKCU\Software\Wine\Explorer\Desktops */
-    if (!RegOpenKeyW( HKEY_CURRENT_USER, desktop_keyW, &hkey ))
+    if (!RegOpenKeyW( HKEY_CURRENT_USER, L"Software\\Wine\\Explorer\\Desktops", &hkey ))
     {
-        if (!RegGetValueW( hkey, name, enable_shellW, RRF_RT_REG_DWORD, NULL, &result, &size ))
+        if (!RegGetValueW( hkey, name, L"EnableShell", RRF_RT_REG_DWORD, NULL, &result, &size ))
             found = TRUE;
         RegCloseKey( hkey );
     }
     /* Default off, except for the magic desktop name "shell" */
-    if (!found)
-        result = (lstrcmpiW( name, shellW ) == 0);
+    if (!found) result = (lstrcmpiW( name, L"shell" ) == 0);
     return result;
 }
 
-static HMODULE load_graphics_driver( const WCHAR *driver, const GUID *guid )
+static BOOL get_default_enable_launchers(void)
 {
-    static const WCHAR device_keyW[] = {
-        'S','y','s','t','e','m','\\',
-        'C','u','r','r','e','n','t','C','o','n','t','r','o','l','S','e','t','\\',
-        'C','o','n','t','r','o','l','\\',
-        'V','i','d','e','o','\\',
-        '{','%','0','8','x','-','%','0','4','x','-','%','0','4','x','-',
-        '%','0','2','x','%','0','2','x','-','%','0','2','x','%','0','2','x','%','0','2','x',
-        '%','0','2','x','%','0','2','x','%','0','2','x','}','\\','0','0','0','0',0};
-    static const WCHAR graphics_driverW[] = {'G','r','a','p','h','i','c','s','D','r','i','v','e','r',0};
-    static const WCHAR driversW[] = {'S','o','f','t','w','a','r','e','\\',
-                                     'W','i','n','e','\\','D','r','i','v','e','r','s',0};
-    static const WCHAR graphicsW[] = {'G','r','a','p','h','i','c','s',0};
-    static const WCHAR drv_formatW[] = {'w','i','n','e','%','s','.','d','r','v',0};
+    BOOL result;
+    DWORD size = sizeof(result);
+
+    if (!RegGetValueW( HKEY_LOCAL_MACHINE, L"Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer",
+                       L"NoDesktop", RRF_RT_REG_DWORD, NULL, &result, &size ))
+        return !result;
+
+    if (!RegGetValueW( HKEY_CURRENT_USER, L"Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer",
+                       L"NoDesktop", RRF_RT_REG_DWORD, NULL, &result, &size ))
+        return !result;
+
+    return TRUE;
+}
+
+static BOOL get_default_show_systray( const WCHAR *name )
+{
+    HKEY hkey;
+    BOOL found = FALSE;
+    BOOL result;
+    DWORD size = sizeof(result);
+
+    /* @@ Wine registry key: HKCU\Software\Wine\Explorer\Desktops */
+    if (name && !RegOpenKeyW( HKEY_CURRENT_USER, L"Software\\Wine\\Explorer\\Desktops", &hkey ))
+    {
+        if (!RegGetValueW( hkey, name, L"ShowSystray", RRF_RT_REG_DWORD, NULL, &result, &size ))
+            found = TRUE;
+        RegCloseKey( hkey );
+    }
+
+    /* Try again with a global Explorer setting */
+    /* @@ Wine registry key: HKCU\Software\Wine\Explorer */
+    if (!found && !RegOpenKeyW( HKEY_CURRENT_USER, L"Software\\Wine\\Explorer", &hkey ))
+    {
+        if (!RegGetValueW( hkey, NULL, L"ShowSystray", RRF_RT_REG_DWORD, NULL, &result, &size ))
+            found = TRUE;
+        RegCloseKey( hkey );
+    }
+
+    /* Default on */
+    if (!found) result = TRUE;
+    return result;
+}
+
+static BOOL get_no_tray_items_display(void)
+{
+    BOOL result;
+    DWORD size = sizeof(result);
+
+    if (!RegGetValueW( HKEY_CURRENT_USER, L"Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer",
+                       L"NoTrayItemsDisplay", RRF_RT_REG_DWORD, NULL, &result, &size ))
+        return result;
+
+    if (!RegGetValueW( HKEY_LOCAL_MACHINE, L"Software\\Microsoft\\Windows\\CurrentVersion\\Policies\\Explorer",
+                       L"NoTrayItemsDisplay", RRF_RT_REG_DWORD, NULL, &result, &size ))
+        return result;
+
+    return FALSE;
+}
+
+static void load_graphics_driver( const WCHAR *driver, GUID *guid )
+{
+    static const WCHAR device_keyW[] = L"System\\CurrentControlSet\\Control\\Video\\{%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x}\\0000";
 
     WCHAR buffer[MAX_PATH], libname[32], *name, *next;
     WCHAR key[ARRAY_SIZE( device_keyW ) + 39];
@@ -818,10 +1009,10 @@ static HMODULE load_graphics_driver( const WCHAR *driver, const GUID *guid )
         lstrcpyW( buffer, default_driver );
 
         /* @@ Wine registry key: HKCU\Software\Wine\Drivers */
-        if (!RegOpenKeyW( HKEY_CURRENT_USER, driversW, &hkey ))
+        if (!RegOpenKeyW( HKEY_CURRENT_USER, L"Software\\Wine\\Drivers", &hkey ))
         {
             DWORD count = sizeof(buffer);
-            RegQueryValueExW( hkey, graphicsW, 0, NULL, (LPBYTE)buffer, &count );
+            RegQueryValueExW( hkey, L"Graphics", 0, NULL, (LPBYTE)buffer, &count );
             RegCloseKey( hkey );
         }
     }
@@ -835,13 +1026,14 @@ static HMODULE load_graphics_driver( const WCHAR *driver, const GUID *guid )
 
         if (!wcscmp( name, L"null" ))
         {
+            memset( guid, 0, sizeof(*guid) );
             TRACE( "display %s using null driver\n", debugstr_guid(guid) );
             wcscpy( libname, L"null" );
             null_driver = TRUE;
             break;
         }
 
-        swprintf( libname, ARRAY_SIZE( libname ), drv_formatW, name );
+        swprintf( libname, ARRAY_SIZE( libname ), L"wine%s.drv", name );
         if ((module = LoadLibraryW( libname )) != 0) break;
         switch (GetLastError())
         {
@@ -849,7 +1041,7 @@ static HMODULE load_graphics_driver( const WCHAR *driver, const GUID *guid )
             strcpy( error, "The graphics driver is missing. Check your build!" );
             break;
         case ERROR_DLL_INIT_FAILED:
-            strcpy( error, "Make sure that your X server is running and that $DISPLAY is set correctly." );
+            strcpy( error, "Make sure that your display server is running and that its variables are set." );
             break;
         default:
             sprintf( error, "Unknown error (%lu).", GetLastError() );
@@ -868,76 +1060,105 @@ static HMODULE load_graphics_driver( const WCHAR *driver, const GUID *guid )
                           REG_OPTION_VOLATILE, KEY_SET_VALUE, NULL, &hkey, NULL  ))
     {
         if (module || null_driver)
-            RegSetValueExW( hkey, graphics_driverW, 0, REG_SZ,
+            RegSetValueExW( hkey, L"GraphicsDriver", 0, REG_SZ,
                             (BYTE *)libname, (lstrlenW(libname) + 1) * sizeof(WCHAR) );
         else
             RegSetValueExA( hkey, "DriverError", 0, REG_SZ, (BYTE *)error, strlen(error) + 1 );
         RegCloseKey( hkey );
     }
-
-    return module;
 }
 
-static void initialize_display_settings(void)
+static const char *debugstr_devmodew( const DEVMODEW *devmode )
 {
-    DISPLAY_DEVICEW ddW;
-    DEVMODEW dmW;
-    DWORD i = 0;
+    char position[32] = {0};
+
+    if (devmode->dmFields & DM_POSITION)
+    {
+        snprintf( position, sizeof(position), " at (%d,%d)",
+                 (int)devmode->dmPosition.x, (int)devmode->dmPosition.y );
+    }
+
+    return wine_dbg_sprintf( "%ux%u %ubits %uHz rotated %u degrees%s",
+                             (unsigned int)devmode->dmPelsWidth,
+                             (unsigned int)devmode->dmPelsHeight,
+                             (unsigned int)devmode->dmBitsPerPel,
+                             (unsigned int)devmode->dmDisplayFrequency,
+                             (unsigned int)devmode->dmDisplayOrientation * 90,
+                             position );
+}
+
+static void initialize_display_settings( unsigned int width, unsigned int height )
+{
+    DISPLAY_DEVICEW device = {.cb = sizeof(DISPLAY_DEVICEW)};
+    DWORD i = 0, flags = CDS_GLOBAL | CDS_UPDATEREGISTRY;
+    HANDLE thread;
 
     /* Store current display mode in the registry */
-    ddW.cb = sizeof(ddW);
-    memset(&dmW, 0, sizeof(dmW));
-    dmW.dmSize = sizeof(dmW);
-    while (EnumDisplayDevicesW( NULL, i++, &ddW, 0 ))
+    while (EnumDisplayDevicesW( NULL, i++, &device, 0 ))
     {
-        if (!EnumDisplaySettingsExW( ddW.DeviceName, ENUM_CURRENT_SETTINGS, &dmW, 0))
+        DEVMODEW devmode = {.dmSize = sizeof(DEVMODEW)};
+
+        if (!EnumDisplaySettingsExW( device.DeviceName, ENUM_CURRENT_SETTINGS, &devmode, 0))
         {
-            WINE_ERR( "Failed to query current display settings for %s.\n",
-                      wine_dbgstr_w( ddW.DeviceName ) );
+            ERR( "Failed to query current display settings for %s.\n", debugstr_w( device.DeviceName ) );
             continue;
         }
 
-        WINE_TRACE( "Device %s current display mode %lux%lu %luBits %luHz at %ld,%ld.\n",
-                    wine_dbgstr_w( ddW.DeviceName ), dmW.dmPelsWidth, dmW.dmPelsHeight,
-                    dmW.dmBitsPerPel, dmW.dmDisplayFrequency, dmW.u1.s2.dmPosition.x,
-                    dmW.u1.s2.dmPosition.y );
+        TRACE( "Device %s current display mode %s.\n", debugstr_w( device.DeviceName ), debugstr_devmodew( &devmode ) );
 
-        if (ChangeDisplaySettingsExW( ddW.DeviceName, &dmW, 0,
-                                      CDS_GLOBAL | CDS_NORESET | CDS_UPDATEREGISTRY, 0 ))
-            WINE_ERR( "Failed to initialize registry display settings for %s.\n",
-                       wine_dbgstr_w( ddW.DeviceName ) );
+        if (ChangeDisplaySettingsExW( device.DeviceName, &devmode, 0, flags | CDS_NORESET, 0 ))
+            ERR( "Failed to initialize registry display settings for %s.\n", debugstr_w( device.DeviceName ) );
     }
+
+    if (!using_root)
+    {
+        DEVMODEW devmode =
+        {
+            .dmSize = sizeof(DEVMODEW),
+            .dmFields = DM_PELSWIDTH | DM_PELSHEIGHT,
+            .dmPelsWidth = width,
+            .dmPelsHeight = height,
+        };
+
+        /* in virtual desktop mode, set the primary display settings to match desktop size */
+        if (ChangeDisplaySettingsExW( NULL, &devmode, 0, flags, NULL ))
+            ERR( "Failed to set primary display settings.\n" );
+    }
+
+    thread = CreateThread( NULL, 0, display_settings_restorer_thread, NULL, 0, NULL );
+    if (thread) CloseHandle( thread );
 }
 
 static void set_desktop_window_title( HWND hwnd, const WCHAR *name )
 {
-    static const WCHAR desktop_nameW[] = {'W','i','n','e',' ','d','e','s','k','t','o','p',0};
-    static const WCHAR desktop_name_separatorW[] = {' ', '-', ' ', 0};
+    static const WCHAR desktop_name_separatorW[] = L" - ";
+    WCHAR desktop_titleW[64];
     WCHAR *window_titleW = NULL;
     int window_title_len;
 
-    if (!name[0])
+    LoadStringW( NULL, IDS_DESKTOP_TITLE, desktop_titleW, ARRAY_SIZE(desktop_titleW) );
+
+    if (!name[0] || !wcscmp( name, L"Default" ))
     {
-        SetWindowTextW( hwnd, desktop_nameW );
+        SetWindowTextW( hwnd, desktop_titleW );
         return;
     }
 
-    window_title_len = lstrlenW(name) * sizeof(WCHAR)
-                     + sizeof(desktop_name_separatorW)
-                     + sizeof(desktop_nameW);
-    window_titleW = HeapAlloc( GetProcessHeap(), 0, window_title_len );
+    window_title_len = (wcslen(name) + wcslen(desktop_titleW)) * sizeof(WCHAR)
+                     + sizeof(desktop_name_separatorW);
+    window_titleW = malloc( window_title_len );
     if (!window_titleW)
     {
-        SetWindowTextW( hwnd, desktop_nameW );
+        SetWindowTextW( hwnd, desktop_titleW );
         return;
     }
 
     lstrcpyW( window_titleW, name );
     lstrcatW( window_titleW, desktop_name_separatorW );
-    lstrcatW( window_titleW, desktop_nameW );
+    lstrcatW( window_titleW, desktop_titleW );
 
     SetWindowTextW( hwnd, window_titleW );
-    HeapFree( GetProcessHeap(), 0, window_titleW );
+    free( window_titleW );
 }
 
 static inline BOOL is_whitespace(WCHAR c)
@@ -945,22 +1166,51 @@ static inline BOOL is_whitespace(WCHAR c)
     return c == ' ' || c == '\t';
 }
 
+/* Set the shell window if appropriate for the current desktop. We should set
+   the shell window on the "Default" desktop on a visible window station, but
+   not for other desktops. */
+static void set_shell_window( HWND hwnd )
+{
+    HWINSTA winsta;
+    USEROBJECTFLAGS flags;
+    HDESK desk;
+    WCHAR desk_name[MAX_PATH];
+
+    if (!(winsta = GetProcessWindowStation()) ||
+        !GetUserObjectInformationW( winsta, UOI_FLAGS, &flags, sizeof(flags), NULL ) ||
+        !(flags.dwFlags & WSF_VISIBLE))
+    {
+        return;
+    }
+
+    if (!(desk = GetThreadDesktop( GetCurrentThreadId() )) ||
+        !GetUserObjectInformationW( desk, UOI_NAME, desk_name, ARRAY_SIZE( desk_name ), NULL ) ||
+        (wcscmp( desk_name, L"Default" ) &&
+         wcscmp( desk_name, L"root" ) /* CW Hack 24270 */))
+    {
+        return;
+    }
+
+    SetShellWindow( hwnd );
+}
+
 /* main desktop management function */
 void manage_desktop( WCHAR *arg )
 {
-    static const WCHAR messageW[] = {'M','e','s','s','a','g','e',0};
     HDESK desktop = 0;
     GUID guid;
     MSG msg;
     HWND hwnd;
-    HMODULE graphics_driver;
     unsigned int width, height;
     WCHAR *cmdline = NULL, *driver = NULL;
     WCHAR *p = arg;
     const WCHAR *name = NULL;
-    BOOL enable_shell = FALSE;
+    BOOL enable_shell, enable_launchers, show_systray, no_tray_items;
     void (WINAPI *pShellDDEInit)( BOOL ) = NULL;
     HMODULE shell32;
+    HANDLE thread;
+    DWORD id;
+    NTSTATUS status;
 
     /* get the rest of the command line (if any) */
     while (*p && !is_whitespace(*p)) p++;
@@ -990,22 +1240,32 @@ void manage_desktop( WCHAR *arg )
         if (!get_default_desktop_size( name, &width, &height )) width = height = 0;
     }
 
-    if (name)
-        enable_shell = get_default_enable_shell( name );
+    enable_shell = name ? get_default_enable_shell( name ) : FALSE;
+    enable_launchers = get_default_enable_launchers();
+    show_systray = get_default_show_systray( name );
+    no_tray_items = get_no_tray_items_display();
+
+    UuidCreate( &guid );
+    TRACE( "display guid %s\n", debugstr_guid(&guid) );
+    load_graphics_driver( driver, &guid );
 
     if (name && width && height)
     {
-        if (!(desktop = CreateDesktopW( name, NULL, NULL, 0, DESKTOP_ALL_ACCESS, NULL )))
+        DEVMODEW devmode = {.dmPelsWidth = width, .dmPelsHeight = height};
+        /* magic: desktop "root" means use the root window */
+        if ((using_root = !wcsicmp( name, L"root" ))) desktop = CreateDesktopW( name, NULL, NULL, DF_WINE_ROOT_DESKTOP, DESKTOP_ALL_ACCESS, NULL );
+        else desktop = CreateDesktopW( name, NULL, &devmode, DF_WINE_VIRTUAL_DESKTOP, DESKTOP_ALL_ACCESS, NULL );
+        if (!desktop)
         {
-            WINE_ERR( "failed to create desktop %s error %ld\n", wine_dbgstr_w(name), GetLastError() );
+            ERR( "failed to create desktop %s error %ld\n", debugstr_w(name), GetLastError() );
             ExitProcess( 1 );
         }
         SetThreadDesktop( desktop );
     }
 
-    UuidCreate( &guid );
-    TRACE( "display guid %s\n", debugstr_guid(&guid) );
-    graphics_driver = load_graphics_driver( driver, &guid );
+    /* the desktop process should always have an admin token */
+    status = NtSetInformationProcess( GetCurrentProcess(), ProcessWineGrantAdminToken, NULL, 0 );
+    if (status) WARN( "couldn't set admin token for desktop, error %08lx\n", status );
 
     /* create the desktop window */
     hwnd = CreateWindowExW( 0, DESKTOP_CLASS_ATOM, NULL,
@@ -1014,26 +1274,25 @@ void manage_desktop( WCHAR *arg )
     if (hwnd)
     {
         /* create the HWND_MESSAGE parent */
-        CreateWindowExW( 0, messageW, NULL, WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
+        CreateWindowExW( 0, L"Message", NULL, WS_POPUP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
                          0, 0, 100, 100, 0, 0, 0, NULL );
 
         desktop_orig_wndproc = (WNDPROC)SetWindowLongPtrW( hwnd, GWLP_WNDPROC,
             (LONG_PTR)desktop_wnd_proc );
-        using_root = !desktop || !create_desktop( graphics_driver, name, width, height );
         SendMessageW( hwnd, WM_SETICON, ICON_BIG, (LPARAM)LoadIconW( 0, MAKEINTRESOURCEW(OIC_WINLOGO)));
         if (name) set_desktop_window_title( hwnd, name );
         SetWindowPos( hwnd, 0, GetSystemMetrics(SM_XVIRTUALSCREEN), GetSystemMetrics(SM_YVIRTUALSCREEN),
                       GetSystemMetrics(SM_CXVIRTUALSCREEN), GetSystemMetrics(SM_CYVIRTUALSCREEN),
                       SWP_SHOWWINDOW );
+        thread = CreateThread( NULL, 0, clipboard_thread, NULL, 0, &id );
+        if (thread) CloseHandle( thread );
         SystemParametersInfoW( SPI_SETDESKWALLPAPER, 0, NULL, FALSE );
         ClipCursor( NULL );
-        initialize_display_settings();
+        initialize_display_settings( width, height );
         initialize_appbar();
 
-        if (using_root) enable_shell = FALSE;
-
-        initialize_systray( graphics_driver, using_root, enable_shell );
-        if (!using_root) initialize_launchers( hwnd );
+        initialize_systray( using_root, enable_shell, show_systray, no_tray_items );
+        if (!using_root && enable_launchers) initialize_launchers( hwnd );
 
         if ((shell32 = LoadLibraryW( L"shell32.dll" )) &&
             (pShellDDEInit = (void *)GetProcAddress( shell32, (LPCSTR)188)))
@@ -1050,7 +1309,7 @@ void manage_desktop( WCHAR *arg )
 
         memset( &si, 0, sizeof(si) );
         si.cb = sizeof(si);
-        WINE_TRACE( "starting %s\n", wine_dbgstr_w(cmdline) );
+        TRACE( "starting %s\n", debugstr_w(cmdline) );
         if (CreateProcessW( NULL, cmdline, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi ))
         {
             CloseHandle( pi.hThread );
@@ -1059,15 +1318,18 @@ void manage_desktop( WCHAR *arg )
     }
 
     desktopshellbrowserwindow_init();
+    shellwindows_init();
+
+    /* Ideally we would set the window of an IShellView here, but we never
+       actually create one, so the desktop window itself will have to do. */
+    set_shell_window( hwnd );
 
     /* run the desktop message loop */
     if (hwnd)
     {
-        shellwindows_init();
-
-        WINE_TRACE( "desktop message loop starting on hwnd %p\n", hwnd );
+        TRACE( "desktop message loop starting on hwnd %p\n", hwnd );
         while (GetMessageW( &msg, 0, 0, 0 )) DispatchMessageW( &msg );
-        WINE_TRACE( "desktop message loop exiting for hwnd %p\n", hwnd );
+        TRACE( "desktop message loop exiting for hwnd %p\n", hwnd );
     }
 
     if (pShellDDEInit) pShellDDEInit( FALSE );
@@ -2305,7 +2567,7 @@ static void shellwindows_init(void)
     CoInitialize(NULL);
 
     shellwindows.IShellWindows_iface.lpVtbl = &shellwindowsvtbl;
-    InitializeCriticalSection(&shellwindows.cs);
+    InitializeCriticalSectionEx(&shellwindows.cs, 0, RTL_CRITICAL_SECTION_FLAG_FORCE_DEBUG_INFO);
     shellwindows.cs.DebugInfo->Spare[0] = (DWORD_PTR)(__FILE__ ": shellwindows.cs");
 
     hr = CoRegisterClassObject(&CLSID_ShellWindows,
