@@ -31,7 +31,6 @@
 #include <setupapi.h>
 
 #include "wine/debug.h"
-#include "wine/heap.h"
 #include "svcctl.h"
 
 #include "services.h"
@@ -43,46 +42,24 @@ WINE_DEFAULT_DEBUG_CHANNEL(service);
 struct scmdatabase *active_database;
 
 /* CW HACK 20218:
- * Double service_pipe_timeout, the Rockstar Launcher service takes 10+
+ * Quadruple service_pipe_timeout, the Rockstar Launcher service takes 20+
  * seconds to launch under Rosetta.
  */
-DWORD service_pipe_timeout = 20000;
+DWORD service_pipe_timeout = 40000;
 DWORD service_kill_timeout = 60000;
 static DWORD default_preshutdown_timeout = 180000;
 static DWORD autostart_delay = 120000;
 static void *environment = NULL;
 static HKEY service_current_key = NULL;
+static HANDLE job_object, job_completion_port;
 
 static const BOOL is_win64 = (sizeof(void *) > sizeof(int));
 
-static const WCHAR SZ_LOCAL_SYSTEM[] = {'L','o','c','a','l','S','y','s','t','e','m',0};
-
-/* Registry constants */
-static const WCHAR SZ_SERVICES_KEY[] = { 'S','y','s','t','e','m','\\',
-      'C','u','r','r','e','n','t','C','o','n','t','r','o','l','S','e','t','\\',
-      'S','e','r','v','i','c','e','s',0 };
-
-/* Service key values names */
-static const WCHAR SZ_DISPLAY_NAME[]      = {'D','i','s','p','l','a','y','N','a','m','e',0 };
-static const WCHAR SZ_TYPE[]              = {'T','y','p','e',0 };
-static const WCHAR SZ_START[]             = {'S','t','a','r','t',0 };
-static const WCHAR SZ_ERROR[]             = {'E','r','r','o','r','C','o','n','t','r','o','l',0 };
-static const WCHAR SZ_IMAGE_PATH[]        = {'I','m','a','g','e','P','a','t','h',0};
-static const WCHAR SZ_GROUP[]             = {'G','r','o','u','p',0};
-static const WCHAR SZ_DEPEND_ON_SERVICE[] = {'D','e','p','e','n','d','O','n','S','e','r','v','i','c','e',0};
-static const WCHAR SZ_DEPEND_ON_GROUP[]   = {'D','e','p','e','n','d','O','n','G','r','o','u','p',0};
-static const WCHAR SZ_OBJECT_NAME[]       = {'O','b','j','e','c','t','N','a','m','e',0};
-static const WCHAR SZ_TAG[]               = {'T','a','g',0};
-static const WCHAR SZ_DESCRIPTION[]       = {'D','e','s','c','r','i','p','t','i','o','n',0};
-static const WCHAR SZ_PRESHUTDOWN[]       = {'P','r','e','s','h','u','t','d','o','w','n','T','i','m','e','o','u','t',0};
-static const WCHAR SZ_WOW64[]             = {'W','O','W','6','4',0};
-static const WCHAR SZ_DELAYED_AUTOSTART[] = {'D','e','l','a','y','e','d','A','u','t','o','S','t','a','r','t',0};
-
 static DWORD process_create(const WCHAR *name, struct process_entry **entry)
 {
     DWORD err;
 
-    *entry = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(**entry));
+    *entry = calloc(1, sizeof(**entry));
     if (!*entry)
         return ERROR_NOT_ENOUGH_SERVER_MEMORY;
     (*entry)->ref_count = 1;
@@ -105,7 +82,7 @@ error:
         CloseHandle((*entry)->control_mutex);
     if ((*entry)->overlapped_event)
         CloseHandle((*entry)->overlapped_event);
-    HeapFree(GetProcessHeap(), 0, *entry);
+    free(*entry);
     return err;
 }
 
@@ -115,26 +92,26 @@ static void free_process_entry(struct process_entry *entry)
     CloseHandle(entry->control_mutex);
     CloseHandle(entry->control_pipe);
     CloseHandle(entry->overlapped_event);
-    HeapFree(GetProcessHeap(), 0, entry);
+    free(entry);
 }
 
 DWORD service_create(LPCWSTR name, struct service_entry **entry)
 {
-    *entry = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(**entry));
+    *entry = calloc(1, sizeof(**entry));
     if (!*entry)
         return ERROR_NOT_ENOUGH_SERVER_MEMORY;
-    (*entry)->name = strdupW(name);
+    (*entry)->name = wcsdup(name);
     list_init(&(*entry)->handles);
     if (!(*entry)->name)
     {
-        HeapFree(GetProcessHeap(), 0, *entry);
+        free(*entry);
         return ERROR_NOT_ENOUGH_SERVER_MEMORY;
     }
     (*entry)->status_changed_event = CreateEventW(NULL, TRUE, FALSE, NULL);
     if (!(*entry)->status_changed_event)
     {
-        HeapFree(GetProcessHeap(), 0, (*entry)->name);
-        HeapFree(GetProcessHeap(), 0, *entry);
+        free((*entry)->name);
+        free(*entry);
         return GetLastError();
     }
     (*entry)->ref_count = 1;
@@ -149,17 +126,17 @@ void free_service_entry(struct service_entry *entry)
 {
     assert(list_empty(&entry->handles));
     CloseHandle(entry->status_changed_event);
-    HeapFree(GetProcessHeap(), 0, entry->name);
-    HeapFree(GetProcessHeap(), 0, entry->config.lpBinaryPathName);
-    HeapFree(GetProcessHeap(), 0, entry->config.lpDependencies);
-    HeapFree(GetProcessHeap(), 0, entry->config.lpLoadOrderGroup);
-    HeapFree(GetProcessHeap(), 0, entry->config.lpServiceStartName);
-    HeapFree(GetProcessHeap(), 0, entry->config.lpDisplayName);
-    HeapFree(GetProcessHeap(), 0, entry->description);
-    HeapFree(GetProcessHeap(), 0, entry->dependOnServices);
-    HeapFree(GetProcessHeap(), 0, entry->dependOnGroups);
+    free(entry->name);
+    free(entry->config.lpBinaryPathName);
+    free(entry->config.lpDependencies);
+    free(entry->config.lpLoadOrderGroup);
+    free(entry->config.lpServiceStartName);
+    free(entry->config.lpDisplayName);
+    free(entry->description);
+    free(entry->dependOnServices);
+    free(entry->dependOnGroups);
     if (entry->process) release_process(entry->process);
-    HeapFree(GetProcessHeap(), 0, entry);
+    free(entry);
 }
 
 static DWORD load_service_config(HKEY hKey, struct service_entry *entry)
@@ -167,35 +144,22 @@ static DWORD load_service_config(HKEY hKey, struct service_entry *entry)
     DWORD err, value = 0;
     WCHAR *wptr;
 
-    if ((err = load_reg_string(hKey, SZ_IMAGE_PATH,   TRUE, &entry->config.lpBinaryPathName)) != 0)
-        return err;
-    if ((err = load_reg_string(hKey, SZ_GROUP,        0,    &entry->config.lpLoadOrderGroup)) != 0)
-        return err;
-    if ((err = load_reg_string(hKey, SZ_OBJECT_NAME,  TRUE, &entry->config.lpServiceStartName)) != 0)
-        return err;
-    if ((err = load_reg_string(hKey, SZ_DISPLAY_NAME, 0,    &entry->config.lpDisplayName)) != 0)
-        return err;
-    if ((err = load_reg_string(hKey, SZ_DESCRIPTION,  0,    &entry->description)) != 0)
-        return err;
-    if ((err = load_reg_multisz(hKey, SZ_DEPEND_ON_SERVICE, TRUE, &entry->dependOnServices)) != 0)
-        return err;
-    if ((err = load_reg_multisz(hKey, SZ_DEPEND_ON_GROUP, FALSE, &entry->dependOnGroups)) != 0)
-        return err;
+    if ((err = load_reg_string(hKey, L"ImagePath",   TRUE, &entry->config.lpBinaryPathName))) return err;
+    if ((err = load_reg_string(hKey, L"Group", FALSE, &entry->config.lpLoadOrderGroup))) return err;
+    if ((err = load_reg_string(hKey, L"ObjectName",  TRUE, &entry->config.lpServiceStartName))) return err;
+    if ((err = load_reg_string(hKey, L"DisplayName", FALSE, &entry->config.lpDisplayName))) return err;
+    if ((err = load_reg_string(hKey, L"Description",  FALSE, &entry->description))) return err;
+    if ((err = load_reg_multisz(hKey, L"DependOnService", TRUE, &entry->dependOnServices))) return err;
+    if ((err = load_reg_multisz(hKey, L"DependOnGroup", FALSE, &entry->dependOnGroups))) return err;
+    if ((err = load_reg_dword(hKey, L"Type",  &entry->config.dwServiceType))) return err;
+    if ((err = load_reg_dword(hKey, L"Start", &entry->config.dwStartType))) return err;
+    if ((err = load_reg_dword(hKey, L"ErrorControl", &entry->config.dwErrorControl))) return err;
+    if ((err = load_reg_dword(hKey, L"Tag", &entry->config.dwTagId))) return err;
+    if ((err = load_reg_dword(hKey, L"PreshutdownTimeout", &entry->preshutdown_timeout))) return err;
 
-    if ((err = load_reg_dword(hKey, SZ_TYPE,  &entry->config.dwServiceType)) != 0)
-        return err;
-    if ((err = load_reg_dword(hKey, SZ_START, &entry->config.dwStartType)) != 0)
-        return err;
-    if ((err = load_reg_dword(hKey, SZ_ERROR, &entry->config.dwErrorControl)) != 0)
-        return err;
-    if ((err = load_reg_dword(hKey, SZ_TAG,   &entry->config.dwTagId)) != 0)
-        return err;
-    if ((err = load_reg_dword(hKey, SZ_PRESHUTDOWN, &entry->preshutdown_timeout)) != 0)
-        return err;
-
-    if (load_reg_dword(hKey, SZ_WOW64, &value) == 0 && value == 1)
+    if (load_reg_dword(hKey, L"WOW64", &value) == 0 && value == 1)
         entry->is_wow64 = TRUE;
-    if (load_reg_dword(hKey, SZ_DELAYED_AUTOSTART, &value) == 0 && value == 1)
+    if (load_reg_dword(hKey, L"DelayedAutoStart", &value) == 0 && value == 1)
         entry->delayed_autostart = TRUE;
 
     WINE_TRACE("Image path           = %s\n", wine_dbgstr_w(entry->config.lpBinaryPathName) );
@@ -212,7 +176,7 @@ static DWORD load_service_config(HKEY hKey, struct service_entry *entry)
     return ERROR_SUCCESS;
 }
 
-static DWORD reg_set_string_value(HKEY hKey, LPCWSTR value_name, LPCWSTR string)
+static DWORD reg_set_string_value(HKEY hKey, LPCWSTR value_name, LPCWSTR string, BOOL expand)
 {
     if (!string)
     {
@@ -224,7 +188,8 @@ static DWORD reg_set_string_value(HKEY hKey, LPCWSTR value_name, LPCWSTR string)
         return ERROR_SUCCESS;
     }
 
-    return RegSetValueExW(hKey, value_name, 0, REG_SZ, (const BYTE*)string, sizeof(WCHAR)*(lstrlenW(string) + 1));
+    return RegSetValueExW(hKey, value_name, 0, expand ? REG_EXPAND_SZ : REG_SZ,
+            (const BYTE*)string, sizeof(WCHAR)*(lstrlenW(string) + 1));
 }
 
 static DWORD reg_set_multisz_value(HKEY hKey, LPCWSTR value_name, LPCWSTR string)
@@ -246,6 +211,11 @@ static DWORD reg_set_multisz_value(HKEY hKey, LPCWSTR value_name, LPCWSTR string
     return RegSetValueExW(hKey, value_name, 0, REG_MULTI_SZ, (const BYTE*)string, sizeof(WCHAR)*(ptr - string + 1));
 }
 
+static DWORD reg_set_dword_value(HKEY hKey, LPCWSTR value_name, DWORD val)
+{
+    return RegSetValueExW(hKey, value_name, 0, REG_DWORD, (const BYTE *)&val, sizeof(val));
+}
+
 DWORD save_service_config(struct service_entry *entry)
 {
     DWORD err;
@@ -255,41 +225,34 @@ DWORD save_service_config(struct service_entry *entry)
     if (err != ERROR_SUCCESS)
         goto cleanup;
 
-    if ((err = reg_set_string_value(hKey, SZ_DISPLAY_NAME, entry->config.lpDisplayName)) != 0)
-        goto cleanup;
-    if ((err = reg_set_string_value(hKey, SZ_IMAGE_PATH, entry->config.lpBinaryPathName)) != 0)
-        goto cleanup;
-    if ((err = reg_set_string_value(hKey, SZ_GROUP, entry->config.lpLoadOrderGroup)) != 0)
-        goto cleanup;
-    if ((err = reg_set_string_value(hKey, SZ_OBJECT_NAME, entry->config.lpServiceStartName)) != 0)
-        goto cleanup;
-    if ((err = reg_set_string_value(hKey, SZ_DESCRIPTION, entry->description)) != 0)
-        goto cleanup;
-    if ((err = reg_set_multisz_value(hKey, SZ_DEPEND_ON_SERVICE, entry->dependOnServices)) != 0)
-        goto cleanup;
-    if ((err = reg_set_multisz_value(hKey, SZ_DEPEND_ON_GROUP, entry->dependOnGroups)) != 0)
-        goto cleanup;
-    if ((err = RegSetValueExW(hKey, SZ_START, 0, REG_DWORD, (LPBYTE)&entry->config.dwStartType, sizeof(DWORD))) != 0)
-        goto cleanup;
-    if ((err = RegSetValueExW(hKey, SZ_ERROR, 0, REG_DWORD, (LPBYTE)&entry->config.dwErrorControl, sizeof(DWORD))) != 0)
-        goto cleanup;
-    if ((err = RegSetValueExW(hKey, SZ_TYPE, 0, REG_DWORD, (LPBYTE)&entry->config.dwServiceType, sizeof(DWORD))) != 0)
-        goto cleanup;
-    if ((err = RegSetValueExW(hKey, SZ_PRESHUTDOWN, 0, REG_DWORD, (LPBYTE)&entry->preshutdown_timeout, sizeof(DWORD))) != 0)
-        goto cleanup;
-    if ((err = RegSetValueExW(hKey, SZ_PRESHUTDOWN, 0, REG_DWORD, (LPBYTE)&entry->preshutdown_timeout, sizeof(DWORD))) != 0)
+    if ((err = reg_set_string_value(hKey, L"DisplayName", entry->config.lpDisplayName, FALSE))) goto cleanup;
+    if ((err = reg_set_string_value(hKey, L"ImagePath", entry->config.lpBinaryPathName, TRUE))) goto cleanup;
+    if ((err = reg_set_string_value(hKey, L"Group", entry->config.lpLoadOrderGroup, FALSE))) goto cleanup;
+    if ((err = reg_set_string_value(hKey, L"ObjectName", entry->config.lpServiceStartName, FALSE))) goto cleanup;
+    if ((err = reg_set_string_value(hKey, L"Description", entry->description, FALSE))) goto cleanup;
+    if ((err = reg_set_multisz_value(hKey, L"DependOnService", entry->dependOnServices))) goto cleanup;
+    if ((err = reg_set_multisz_value(hKey, L"DependOnGroup", entry->dependOnGroups))) goto cleanup;
+    if ((err = reg_set_dword_value(hKey, L"Start", entry->config.dwStartType))) goto cleanup;
+    if ((err = reg_set_dword_value(hKey, L"ErrorControl", entry->config.dwErrorControl))) goto cleanup;
+    if ((err = reg_set_dword_value(hKey, L"Type", entry->config.dwServiceType))) goto cleanup;
+    if ((err = reg_set_dword_value(hKey, L"PreshutdownTimeout", entry->preshutdown_timeout))) goto cleanup;
+
+    if (entry->delayed_autostart)
+        err = reg_set_dword_value(hKey, L"DelayedAutoStart", entry->delayed_autostart);
+    else
+        err = RegDeleteValueW(hKey, L"DelayedAutoStart");
+    if (err != 0 && err != ERROR_FILE_NOT_FOUND)
         goto cleanup;
+
     if (entry->is_wow64)
     {
-        const DWORD is_wow64 = 1;
-        if ((err = RegSetValueExW(hKey, SZ_WOW64, 0, REG_DWORD, (LPBYTE)&is_wow64, sizeof(DWORD))) != 0)
-            goto cleanup;
+        if ((err = reg_set_dword_value(hKey, L"WOW64", TRUE))) goto cleanup;
     }
 
     if (entry->config.dwTagId)
-        err = RegSetValueExW(hKey, SZ_TAG, 0, REG_DWORD, (LPBYTE)&entry->config.dwTagId, sizeof(DWORD));
+        err = reg_set_dword_value(hKey, L"Tag", entry->config.dwTagId);
     else
-        err = RegDeleteValueW(hKey, SZ_TAG);
+        err = RegDeleteValueW(hKey, L"Tag");
 
     if (err != 0 && err != ERROR_FILE_NOT_FOUND)
         goto cleanup;
@@ -333,10 +296,12 @@ static void scmdatabase_remove_service(struct scmdatabase *db, struct service_en
     service->entry.next = service->entry.prev = NULL;
 }
 
-static int __cdecl compare_tags(const void *a, const void *b)
+static int __cdecl compare_service(const void *a, const void *b)
 {
     struct service_entry *service_a = *(struct service_entry **)a;
     struct service_entry *service_b = *(struct service_entry **)b;
+    if (service_a->config.dwStartType != service_b->config.dwStartType)
+        return service_a->config.dwStartType - service_b->config.dwStartType;
     return service_a->config.dwTagId - service_b->config.dwTagId;
 }
 
@@ -353,8 +318,8 @@ static void CALLBACK delayed_autostart_cancel_callback(void *object, void *userd
     struct delayed_autostart_params *params = object;
     while(params->count--)
         release_service(params->services[params->count]);
-    heap_free(params->services);
-    heap_free(params);
+    free(params->services);
+    free(params);
 }
 
 static void CALLBACK delayed_autostart_callback(TP_CALLBACK_INSTANCE *instance, void *context,
@@ -383,8 +348,8 @@ static void CALLBACK delayed_autostart_callback(TP_CALLBACK_INSTANCE *instance,
 
     scmdatabase_unlock_startup(active_database);
 
-    heap_free(params->services);
-    heap_free(params);
+    free(params->services);
+    free(params);
     CloseThreadpoolTimer(timer);
 }
 
@@ -402,7 +367,7 @@ static BOOL schedule_delayed_autostart(struct service_entry **services, unsigned
         return FALSE;
     }
 
-    if (!(params = heap_alloc(sizeof(*params)))) return FALSE;
+    if (!(params = malloc(sizeof(*params)))) return FALSE;
     params->count = count;
     params->services = services;
 
@@ -418,7 +383,7 @@ static BOOL schedule_delayed_autostart(struct service_entry **services, unsigned
     if (!(timer = CreateThreadpoolTimer(delayed_autostart_callback, params, &environment)))
     {
         ERR("CreateThreadpoolWait failed: %lu\n", GetLastError());
-        heap_free(params);
+        free(params);
         return FALSE;
     }
 
@@ -447,7 +412,6 @@ static BOOL is_root_pnp_service(HDEVINFO set, const struct service_entry *servic
 
 static void scmdatabase_autostart_services(struct scmdatabase *db)
 {
-    static const WCHAR rootW[] = {'R','O','O','T',0};
     struct service_entry **services_list;
     unsigned int i = 0;
     unsigned int size = 32;
@@ -455,11 +419,11 @@ static void scmdatabase_autostart_services(struct scmdatabase *db)
     struct service_entry *service;
     HDEVINFO set;
 
-    services_list = HeapAlloc(GetProcessHeap(), 0, size * sizeof(services_list[0]));
+    services_list = malloc(size * sizeof(services_list[0]));
     if (!services_list)
         return;
 
-    if ((set = SetupDiGetClassDevsW( NULL, rootW, NULL, DIGCF_ALLCLASSES )) == INVALID_HANDLE_VALUE)
+    if ((set = SetupDiGetClassDevsW( NULL, L"ROOT", NULL, DIGCF_ALLCLASSES )) == INVALID_HANDLE_VALUE)
         WINE_ERR("Failed to enumerate devices, error %#lx.\n", GetLastError());
 
     scmdatabase_lock(db);
@@ -475,7 +439,7 @@ static void scmdatabase_autostart_services(struct scmdatabase *db)
             {
                 struct service_entry **slist_new;
                 size *= 2;
-                slist_new = HeapReAlloc(GetProcessHeap(), 0, services_list, size * sizeof(services_list[0]));
+                slist_new = realloc(services_list, size * sizeof(services_list[0]));
                 if (!slist_new)
                     break;
                 services_list = slist_new;
@@ -486,7 +450,7 @@ static void scmdatabase_autostart_services(struct scmdatabase *db)
     size = i;
 
     scmdatabase_unlock(db);
-    qsort(services_list, size, sizeof(services_list[0]), compare_tags);
+    qsort(services_list, size, sizeof(services_list[0]), compare_service);
     scmdatabase_lock_startup(db, INFINITE);
 
     for (i = 0; i < size; i++)
@@ -509,7 +473,7 @@ static void scmdatabase_autostart_services(struct scmdatabase *db)
     scmdatabase_unlock_startup(db);
 
     if (!delayed_cnt || !schedule_delayed_autostart(services_list, delayed_cnt))
-        heap_free(services_list);
+        free(services_list);
     SetupDiDestroyDeviceInfoList(set);
 }
 
@@ -560,7 +524,7 @@ BOOL validate_service_config(struct service_entry *entry)
     case SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS:
     case SERVICE_WIN32_SHARE_PROCESS | SERVICE_INTERACTIVE_PROCESS:
         /* These can be only run as LocalSystem */
-        if (entry->config.lpServiceStartName && wcsicmp(entry->config.lpServiceStartName, SZ_LOCAL_SYSTEM) != 0)
+        if (entry->config.lpServiceStartName && wcsicmp(entry->config.lpServiceStartName, L"LocalSystem") != 0)
         {
             WINE_ERR("Service %s is interactive but has a start name\n", wine_dbgstr_w(entry->name));
             return FALSE;
@@ -587,7 +551,7 @@ BOOL validate_service_config(struct service_entry *entry)
     }
 
     if (entry->config.lpServiceStartName == NULL)
-        entry->config.lpServiceStartName = strdupW(SZ_LOCAL_SYSTEM);
+        entry->config.lpServiceStartName = wcsdup(L"LocalSystem");
 
     return TRUE;
 }
@@ -663,7 +627,7 @@ static DWORD scmdatabase_create(struct scmdatabase **db)
 {
     DWORD err;
 
-    *db = HeapAlloc(GetProcessHeap(), 0, sizeof(**db));
+    *db = malloc(sizeof(**db));
     if (!*db)
         return ERROR_NOT_ENOUGH_SERVER_MEMORY;
 
@@ -671,14 +635,14 @@ static DWORD scmdatabase_create(struct scmdatabase **db)
     list_init(&(*db)->processes);
     list_init(&(*db)->services);
 
-    InitializeCriticalSection(&(*db)->cs);
+    InitializeCriticalSectionEx(&(*db)->cs, 0, RTL_CRITICAL_SECTION_FLAG_FORCE_DEBUG_INFO);
     (*db)->cs.DebugInfo->Spare[0] = (DWORD_PTR)(__FILE__ ": scmdatabase");
 
-    err = RegCreateKeyExW(HKEY_LOCAL_MACHINE, SZ_SERVICES_KEY, 0, NULL,
+    err = RegCreateKeyExW(HKEY_LOCAL_MACHINE, L"System\\CurrentControlSet\\Services", 0, NULL,
                           REG_OPTION_NON_VOLATILE, MAXIMUM_ALLOWED, NULL,
                           &(*db)->root_key, NULL);
     if (err != ERROR_SUCCESS)
-        HeapFree(GetProcessHeap(), 0, *db);
+        free(*db);
 
     return err;
 }
@@ -688,7 +652,7 @@ static void scmdatabase_destroy(struct scmdatabase *db)
     RegCloseKey(db->root_key);
     db->cs.DebugInfo->Spare[0] = 0;
     DeleteCriticalSection(&db->cs);
-    HeapFree(GetProcessHeap(), 0, db);
+    free(db);
 }
 
 static DWORD scmdatabase_load_services(struct scmdatabase *db)
@@ -798,9 +762,7 @@ void service_unlock(struct service_entry *service)
  * value here */
 static LPWSTR service_get_pipe_name(void)
 {
-    static const WCHAR format[] = { '\\','\\','.','\\','p','i','p','e','\\',
-        'n','e','t','\\','N','t','C','o','n','t','r','o','l','P','i','p','e','%','u',0};
-    static WCHAR name[ARRAY_SIZE(format) + 10]; /* lstrlenW("4294967295") */
+    static WCHAR name[ARRAY_SIZE(L"\\\\.\\pipe\\net\\NtControlPipe") + 10]; /* lstrlenW("4294967295") */
     static DWORD service_current = 0;
     DWORD len, value = -1;
     LONG ret;
@@ -813,7 +775,7 @@ static LPWSTR service_get_pipe_name(void)
         service_current = max(service_current, value + 1);
     RegSetValueExW(service_current_key, NULL, 0, REG_DWORD,
         (BYTE *)&service_current, sizeof(service_current));
-    swprintf(name, ARRAY_SIZE(name), format, service_current);
+    swprintf(name, ARRAY_SIZE(name), L"\\\\.\\pipe\\net\\NtControlPipe%u", service_current);
     service_current++;
     return name;
 }
@@ -822,7 +784,7 @@ static DWORD get_service_binary_path(const struct service_entry *service_entry,
 {
     DWORD size = ExpandEnvironmentStringsW(service_entry->config.lpBinaryPathName, NULL, 0);
 
-    *path = HeapAlloc(GetProcessHeap(), 0, size*sizeof(WCHAR));
+    *path = malloc(size * sizeof(WCHAR));
     if (!*path)
         return ERROR_NOT_ENOUGH_SERVER_MEMORY;
 
@@ -842,16 +804,16 @@ static DWORD get_service_binary_path(const struct service_entry *service_entry,
 
         GetSystemWow64DirectoryW( system_dir, MAX_PATH );
 
-        redirected = HeapAlloc( GetProcessHeap(), 0, (lstrlenW( *path ) + lstrlenW( system_dir ))*sizeof(WCHAR));
+        redirected = malloc( (wcslen( *path ) + wcslen( system_dir )) * sizeof(WCHAR) );
         if (!redirected)
         {
-            HeapFree( GetProcessHeap(), 0, *path );
+            free( *path );
             return ERROR_NOT_ENOUGH_SERVER_MEMORY;
         }
 
         lstrcpyW( redirected, system_dir );
         lstrcatW( redirected, &(*path)[len] );
-        HeapFree( GetProcessHeap(), 0, *path );
+        free( *path );
         *path = redirected;
         TRACE("redirected to %s\n", debugstr_w(redirected));
     }
@@ -861,7 +823,6 @@ static DWORD get_service_binary_path(const struct service_entry *service_entry,
 
 static DWORD get_winedevice_binary_path(struct service_entry *service_entry, WCHAR **path, BOOL *is_wow64)
 {
-    static const WCHAR winedeviceW[] = {'\\','w','i','n','e','d','e','v','i','c','e','.','e','x','e',0};
     WCHAR system_dir[MAX_PATH];
     DWORD type;
 
@@ -873,12 +834,12 @@ static DWORD get_winedevice_binary_path(struct service_entry *service_entry, WCH
         *is_wow64 = service_entry->is_wow64;
 
     GetSystemDirectoryW(system_dir, MAX_PATH);
-    HeapFree(GetProcessHeap(), 0, *path);
-    if (!(*path = HeapAlloc(GetProcessHeap(), 0, lstrlenW(system_dir) * sizeof(WCHAR) + sizeof(winedeviceW))))
+    free(*path);
+    if (!(*path = malloc(wcslen(system_dir) * sizeof(WCHAR) + sizeof(L"\\winedevice.exe"))))
        return ERROR_NOT_ENOUGH_SERVER_MEMORY;
 
     lstrcpyW(*path, system_dir);
-    lstrcatW(*path, winedeviceW);
+    lstrcatW(*path, L"\\winedevice.exe");
     return ERROR_SUCCESS;
 }
 
@@ -911,15 +872,14 @@ static struct process_entry *get_winedevice_process(struct service_entry *servic
 static DWORD add_winedevice_service(const struct service_entry *service, WCHAR *path, BOOL is_wow64,
                                     struct service_entry **entry)
 {
-    static const WCHAR format[] = {'W','i','n','e','d','e','v','i','c','e','%','u',0};
-    static WCHAR name[ARRAY_SIZE(format) + 10]; /* lstrlenW("4294967295") */
+    static WCHAR name[ARRAY_SIZE(L"Winedevice") + 10]; /* lstrlenW("4294967295") */
     static DWORD current = 0;
     struct scmdatabase *db = service->db;
     DWORD err;
 
     for (;;)
     {
-        swprintf(name, ARRAY_SIZE(name), format, ++current);
+        swprintf(name, ARRAY_SIZE(name), L"Winedevice%u", ++current);
         if (!scmdatabase_find_service(db, name)) break;
     }
 
@@ -932,14 +892,14 @@ static DWORD add_winedevice_service(const struct service_entry *service, WCHAR *
     (*entry)->config.dwStartType        = SERVICE_DEMAND_START;
     (*entry)->status.dwServiceType      = (*entry)->config.dwServiceType;
 
-    if (!((*entry)->config.lpBinaryPathName = strdupW(path)))
+    if (!((*entry)->config.lpBinaryPathName = wcsdup(path)))
         goto error;
-    if (!((*entry)->config.lpServiceStartName = strdupW(SZ_LOCAL_SYSTEM)))
+    if (!((*entry)->config.lpServiceStartName = wcsdup(L"LocalSystem")))
         goto error;
-    if (!((*entry)->config.lpDisplayName = strdupW(name)))
+    if (!((*entry)->config.lpDisplayName = wcsdup(name)))
         goto error;
     if (service->config.lpLoadOrderGroup &&
-        !((*entry)->config.lpLoadOrderGroup = strdupW(service->config.lpLoadOrderGroup)))
+        !((*entry)->config.lpLoadOrderGroup = wcsdup(service->config.lpLoadOrderGroup)))
         goto error;
 
     (*entry)->db = db;
@@ -996,29 +956,29 @@ static DWORD service_start_process(struct service_entry *service_entry, struct p
         if ((err = get_winedevice_binary_path(service_entry, &path, &is_wow64)))
         {
             service_unlock(service_entry);
-            HeapFree(GetProcessHeap(), 0, path);
+            free(path);
             return err;
         }
 
         if ((process = get_winedevice_process(service_entry, path, is_wow64)))
         {
-            HeapFree(GetProcessHeap(), 0, path);
+            free(path);
             goto found;
         }
 
         err = add_winedevice_service(service_entry, path, is_wow64, &winedevice_entry);
-        HeapFree(GetProcessHeap(), 0, path);
+        free(path);
         if (err != ERROR_SUCCESS)
         {
             service_unlock(service_entry);
             return err;
         }
 
-        group = strdupW(winedevice_entry->config.lpLoadOrderGroup);
+        group = wcsdup(winedevice_entry->config.lpLoadOrderGroup);
         service_unlock(service_entry);
 
         err = service_start(winedevice_entry, group != NULL, (const WCHAR **)&group);
-        HeapFree(GetProcessHeap(), 0, group);
+        free(group);
         if (err != ERROR_SUCCESS)
         {
             release_service(winedevice_entry);
@@ -1061,7 +1021,7 @@ found:
         WINE_ERR("failed to create process object for %s, error = %lu\n",
                  wine_dbgstr_w(service_entry->name), err);
         service_unlock(service_entry);
-        HeapFree(GetProcessHeap(), 0, path);
+        free(path);
         return err;
     }
 
@@ -1069,8 +1029,7 @@ found:
     si.cb = sizeof(STARTUPINFOW);
     if (!(service_entry->config.dwServiceType & SERVICE_INTERACTIVE_PROCESS))
     {
-        static WCHAR desktopW[] = {'_','_','w','i','n','e','s','e','r','v','i','c','e','_','w','i','n','s','t','a','t','i','o','n','\\','D','e','f','a','u','l','t',0};
-        si.lpDesktop = desktopW;
+        si.lpDesktop = (WCHAR *)L"__wineservice_winstation\\Default";
     }
 
     if (!environment && OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY | TOKEN_DUPLICATE, &token))
@@ -1079,9 +1038,9 @@ found:
         CreateEnvironmentBlock(&environment, token, FALSE);
         if (GetEnvironmentVariableW( L"WINEBOOTSTRAPMODE", val, ARRAY_SIZE(val) ))
         {
-            UNICODE_STRING name, value;
+            UNICODE_STRING name = RTL_CONSTANT_STRING(L"WINEBOOTSTRAPMODE");
+            UNICODE_STRING value;
 
-            RtlInitUnicodeString( &name, L"WINEBOOTSTRAPMODE" );
             RtlInitUnicodeString( &value, val );
             RtlSetEnvironmentVariable( (WCHAR **)&environment, &name, &value );
         }
@@ -1099,7 +1058,7 @@ found:
     service_unlock(service_entry);
 
     r = CreateProcessW(NULL, path, NULL, NULL, FALSE, CREATE_UNICODE_ENVIRONMENT | DETACHED_PROCESS, environment, NULL, &si, &pi);
-    HeapFree(GetProcessHeap(), 0, path);
+    free(path);
     if (!r)
     {
         err = GetLastError();
@@ -1107,6 +1066,8 @@ found:
         release_process(process);
         return err;
     }
+    if (!AssignProcessToJobObject(job_object, pi.hProcess))
+        WINE_ERR("Could not add object to job.\n");
 
     process->process_id = pi.dwProcessId;
     process->process = pi.hProcess;
@@ -1155,7 +1116,7 @@ static DWORD process_send_start_message(struct process_entry *process, BOOL shar
             handles[1] = process->process;
             if (WaitForMultipleObjects( 2, handles, FALSE, service_pipe_timeout ) != WAIT_OBJECT_0)
                 CancelIo(process->control_pipe);
-            if (!HasOverlappedIoCompleted( &overlapped ))
+            if (!GetOverlappedResult(process->control_pipe, &overlapped, &len, FALSE))
             {
                 WINE_ERR("service %s failed to start\n", wine_dbgstr_w(name));
                 return ERROR_SERVICE_REQUEST_TIMEOUT;
@@ -1173,7 +1134,7 @@ static DWORD process_send_start_message(struct process_entry *process, BOOL shar
         len += lstrlenW(argv[i])+1;
     len = (len + 1) * sizeof(WCHAR);
 
-    if (!(str = HeapAlloc(GetProcessHeap(), 0, len)))
+    if (!(str = malloc(len)))
         return ERROR_NOT_ENOUGH_SERVER_MEMORY;
 
     p = str;
@@ -1190,7 +1151,7 @@ static DWORD process_send_start_message(struct process_entry *process, BOOL shar
                               SERVICE_CONTROL_START, (const BYTE *)str, len, &result))
         result = ERROR_SERVICE_REQUEST_TIMEOUT;
 
-    HeapFree(GetProcessHeap(), 0, str);
+    free(str);
     return result;
 }
 
@@ -1249,52 +1210,101 @@ void process_terminate(struct process_entry *process)
 
 static void load_registry_parameters(void)
 {
-    static const WCHAR controlW[] =
-        { 'S','y','s','t','e','m','\\',
-          'C','u','r','r','e','n','t','C','o','n','t','r','o','l','S','e','t','\\',
-          'C','o','n','t','r','o','l',0 };
-    static const WCHAR pipetimeoutW[] =
-        {'S','e','r','v','i','c','e','s','P','i','p','e','T','i','m','e','o','u','t',0};
-    static const WCHAR killtimeoutW[] =
-        {'W','a','i','t','T','o','K','i','l','l','S','e','r','v','i','c','e','T','i','m','e','o','u','t',0};
-    static const WCHAR autostartdelayW[] =
-        {'A','u','t','o','S','t','a','r','t','D','e','l','a','y',0};
     HKEY key;
     WCHAR buffer[64];
     DWORD type, count, val;
 
-    if (RegOpenKeyW( HKEY_LOCAL_MACHINE, controlW, &key )) return;
+    if (RegOpenKeyW( HKEY_LOCAL_MACHINE, L"System\\CurrentControlSet\\Control", &key )) return;
 
     count = sizeof(buffer);
-    if (!RegQueryValueExW( key, pipetimeoutW, NULL, &type, (BYTE *)buffer, &count ) &&
+    if (!RegQueryValueExW( key, L"ServicesPipeTimeout", NULL, &type, (BYTE *)buffer, &count ) &&
         type == REG_SZ && (val = wcstol( buffer, NULL, 10 )))
         service_pipe_timeout = val;
 
     count = sizeof(buffer);
-    if (!RegQueryValueExW( key, killtimeoutW, NULL, &type, (BYTE *)buffer, &count ) &&
+    if (!RegQueryValueExW( key, L"WaitToKillServiceTimeout", NULL, &type, (BYTE *)buffer, &count ) &&
         type == REG_SZ && (val = wcstol( buffer, NULL, 10 )))
         service_kill_timeout = val;
 
     count = sizeof(val);
-    if (!RegQueryValueExW( key, autostartdelayW, NULL, &type, (BYTE *)&val, &count ) && type == REG_DWORD)
+    if (!RegQueryValueExW( key, L"AutoStartDelay", NULL, &type, (BYTE *)&val, &count ) && type == REG_DWORD)
         autostart_delay = val;
 
     RegCloseKey( key );
 }
 
+static DWORD WINAPI process_monitor_thread_proc( void *arg )
+{
+    struct scmdatabase *db = active_database;
+    struct service_entry *service;
+    struct process_entry *process;
+    OVERLAPPED *overlapped;
+    ULONG_PTR value;
+    DWORD key, pid;
+
+    while (GetQueuedCompletionStatus(job_completion_port, &key, &value, &overlapped, INFINITE))
+    {
+        if (!key)
+            break;
+        if (key != JOB_OBJECT_MSG_EXIT_PROCESS)
+            continue;
+        pid = (ULONG_PTR)overlapped;
+        WINE_TRACE("pid %04lx exited.\n", pid);
+        scmdatabase_lock(db);
+        LIST_FOR_EACH_ENTRY(service, &db->services, struct service_entry, entry)
+        {
+            if (service->status.dwCurrentState != SERVICE_RUNNING || !service->process
+                    || service->process->process_id != pid) continue;
+
+            WINE_TRACE("Stopping service %s.\n", debugstr_w(service->config.lpBinaryPathName));
+            service->status.dwCurrentState = SERVICE_STOPPED;
+            service->status.dwControlsAccepted = 0;
+            service->status.dwWin32ExitCode = ERROR_PROCESS_ABORTED;
+            service->status.dwServiceSpecificExitCode = 0;
+            service->status.dwCheckPoint = 0;
+            service->status.dwWaitHint = 0;
+            SetEvent(service->status_changed_event);
+
+            process = service->process;
+            service->process = NULL;
+            process->use_count--;
+            release_process(process);
+            notify_service_state(service);
+        }
+        scmdatabase_unlock(db);
+    }
+    WINE_TRACE("Terminating.\n");
+    return 0;
+}
+
 int __cdecl main(int argc, char *argv[])
 {
-    static const WCHAR service_current_key_str[] = { 'S','Y','S','T','E','M','\\',
-        'C','u','r','r','e','n','t','C','o','n','t','r','o','l','S','e','t','\\',
-        'C','o','n','t','r','o','l','\\',
-        'S','e','r','v','i','c','e','C','u','r','r','e','n','t',0};
     static const WCHAR svcctl_started_event[] = SVCCTL_STARTED_EVENT;
-    HANDLE started_event;
+    JOBOBJECT_EXTENDED_LIMIT_INFORMATION job_limit;
+    JOBOBJECT_ASSOCIATE_COMPLETION_PORT port_info;
+    HANDLE started_event, process_monitor_thread;
     DWORD err;
 
+    job_object = CreateJobObjectW(NULL, NULL);
+    job_limit.BasicLimitInformation.LimitFlags = JOB_OBJECT_LIMIT_BREAKAWAY_OK | JOB_OBJECT_LIMIT_SILENT_BREAKAWAY_OK;
+    if (!SetInformationJobObject(job_object, JobObjectExtendedLimitInformation, &job_limit, sizeof(job_limit)))
+    {
+        WINE_ERR("Failed to initialized job object, err %lu.\n", GetLastError());
+        return GetLastError();
+    }
+    job_completion_port = CreateIoCompletionPort(INVALID_HANDLE_VALUE, NULL, 0, 1);
+    port_info.CompletionPort = job_completion_port;
+    port_info.CompletionKey = job_object;
+    if (!SetInformationJobObject(job_object, JobObjectAssociateCompletionPortInformation,
+            &port_info, sizeof(port_info)))
+    {
+        WINE_ERR("Failed to set completion port for job, err %lu.\n", GetLastError());
+        return GetLastError();
+    }
+
     started_event = CreateEventW(NULL, TRUE, FALSE, svcctl_started_event);
 
-    err = RegCreateKeyExW(HKEY_LOCAL_MACHINE, service_current_key_str, 0,
+    err = RegCreateKeyExW(HKEY_LOCAL_MACHINE, L"SYSTEM\\CurrentControlSet\\Control\\ServiceCurrent", 0,
         NULL, REG_OPTION_VOLATILE, KEY_SET_VALUE | KEY_QUERY_VALUE, NULL,
         &service_current_key, NULL);
     if (err != ERROR_SUCCESS)
@@ -1309,8 +1319,11 @@ int __cdecl main(int argc, char *argv[])
     if ((err = RPC_Init()) == ERROR_SUCCESS)
     {
         scmdatabase_autostart_services(active_database);
+        process_monitor_thread = CreateThread(NULL, 0, process_monitor_thread_proc, NULL, 0, NULL);
         SetEvent(started_event);
         WaitForSingleObject(exit_event, INFINITE);
+        PostQueuedCompletionStatus(job_completion_port, 0, 0, NULL);
+        WaitForSingleObject(process_monitor_thread, INFINITE);
         scmdatabase_wait_terminate(active_database);
         if (delayed_autostart_cleanup)
         {
